/*
    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "bus.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/log.h>

#include "backend-vdcs-mock.h"
#include "businterface-mock.h"
#include "security/security-mock.h"

static ds::LogChannel dsLogChannel("busTest");

static const char* TAGS = "[dss][bus]";

using namespace dss;
using namespace trompeloeil;
using namespace std::literals;

namespace {

struct BusHooksMock {
  MAKE_MOCK0(linkPresent, void());
  MAKE_MOCK0(linkLost, void());
  MAKE_MOCK1(ds485MemberLost, void(const dsuid_t&));
  MAKE_MOCK1(ds485MemberJoin, void(const dsuid_t&));

  BusHooks busHooks() {
    return {
        .linkPresent = [&] { return linkPresent(); },
        .linkLost = [&] { return linkLost(); },
        .ds485MemberLost = [&](const dsuid_t& arg) { return ds485MemberLost(arg); },
        .ds485MemberJoin = [&](const dsuid_t& arg) { return ds485MemberJoin(arg); },
    };
  }
};

SCENARIO("dssBus", TAGS) {
  ds::asio::TestLoopResource loop; // RAII, ensure destructor runs last
  SecurityMock security{};
  BusInterfaceMock busInterface{};
  BackendVdcsMock backendVdcs{};

  Bus bus{security, busInterface};
  bus.setBackendVdcs(backendVdcs);

  dsuid_t dsuid{};

  SECTION("backendVdc") {
    WHEN("query structure of vdc") {
      StructureQueryBusInterfaceMock structureQuery{};
      REQUIRE_CALL(backendVdcs, isBackendVdc(_)).RETURN(true);
      REQUIRE_CALL(backendVdcs, structureQueryBusInterface()).LR_RETURN(structureQuery);
      CHECK(&bus.dsmQueryBus(dsuid) == &structureQuery);
    }
    WHEN("modify structure of vdc device") {
      StructureModifyingBusInterfaceMock structureModify{};
      REQUIRE_CALL(backendVdcs, isBackendVdc(_)).RETURN(true);
      REQUIRE_CALL(backendVdcs, structureModifyingBusInterface()).LR_RETURN(structureModify);
      CHECK(&bus.dsmModifyBus(dsuid) == &structureModify);
    }
  }

  GIVEN("PIMPL initialized") {
    BusHooksMock hooks{};

    loop.post([&] { return bus.makeImpl(loop, hooks.busHooks()); });
    loop.postStopRun();

    // ensure signals from bus components are forwarded

    WHEN("link present") {
      REQUIRE_CALL(hooks, linkPresent());
      busInterface.linkPresent.emit();
      loop.postStopRun();
    }
    WHEN("link lost") {
      REQUIRE_CALL(hooks, linkLost());
      busInterface.linkLost.emit();
      loop.postStopRun();
    }
    WHEN("bus member lost") {
      REQUIRE_CALL(hooks, ds485MemberLost(_));
      busInterface.ds485MemberLost.emit(dsuid_t{});
      loop.postStopRun();
    }
    WHEN("bus member join") {
      REQUIRE_CALL(hooks, ds485MemberJoin(_));
      busInterface.ds485MemberJoin.emit(dsuid_t{});
      loop.postStopRun();
    }
  }
}

} // namespace
