/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#include "cache/apartment-proxy.h"

#include <ds-model/model.h>
#include <ds/catch/catch.h>
#include <ds/conf-mock.h>
#include <ds/log.h>

#include <trompeloeil.hpp>

#include "businterface-mock.h"
#include "device-cache.h"
#include "ds485/dsactionrequest.h"
#include "dsm-api-mock.h"
#include "event-mock.h"
#include "hal/hardware-abstraction-mock.h"
#include "model/apartment-mock.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device-mock.h"
#include "model/group.h"
#include "model/set.h"
#include "model/zone.h"

static ds::LogChannel dsLogChannel("deviceCacheTest", ds::LogSeverity::DEBUG);
static const char* TAGS = "[dss][cache]";

using namespace dss;
using namespace trompeloeil;
using trompeloeil::_;

namespace {

static const Dsuid dsuid1{"0000000000000000000000000000000001"};
static const Dsuid dsuid2{"0000000000000000000000000000000002"};
static const Dsuid dsuid3{"0000000000000000000000000000000003"};
static const Dsuid dsuid4{"0000000000000000000000000000000004"};
static const Dsuid dsuid5{"0000000000000000000000000000000005"};

SCENARIO("SaveScene invocation", TAGS) {
  GIVEN("The cache contains 4 devices") {
    ds::sqlite3::Database d(":memory:");
    auto propertySystem = PropertySystem{};
    auto pApartment = makeApartment(nullptr, &propertySystem);
    auto& apt = *pApartment;

    auto& dev1 = *createDevice(apt, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(dev1);
    auto& dev2 = *createDevice(apt, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(dev2);
    auto& dev3 = *createDevice(apt, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(dev3);
    auto& dev4 = *createDevice(apt, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(dev4);

    auto halPtr = makeHardwareAbstraction(&apt, &d);
    auto& hal = *halPtr;
    auto& cacheDb = hal.getCache();
    ds::ConfMock conf;
    auto proxy = makeApartmentProxy(&apt, &hal, &d, nullptr, conf);
    DeviceCache A;
    A.scene.sceneTable[1].channels[ChannelType::brightness].value = 11;
    A.scene.sceneTable[2].channels[ChannelType::brightness].value = 21;
    DeviceCache B;
    B.scene.sceneTable[1].channels[ChannelType::shadePositionOutside].value = 12;
    B.scene.sceneTable[2].channels[ChannelType::shadePositionOutside].value = 22;
    DeviceCache C;
    C.scene.sceneTable[1].channels[ChannelType::brightness].value = 13;
    C.scene.sceneTable[1].channels[ChannelType::shadeOpeningAngleOutside].value = 14;
    C.scene.sceneTable[2].channels[ChannelType::brightness].value = 23;
    C.scene.sceneTable[2].channels[ChannelType::shadeOpeningAngleOutside].value = 24;
    DeviceCache D;
    D.scene.sceneTable[1].channels[ChannelType::brightness].value = 13;
    D.scene.sceneTable[1].flags = SceneFlags{.dontCare = false};
    D.scene.sceneTable[6].channels[ChannelType::brightness].value = 13;
    D.scene.sceneTable[6].flags = SceneFlags{.dontCare = false};
    cacheDb.update(dev1.getDSID(), A);
    cacheDb.update(dev2.getDSID(), B);
    cacheDb.update(dev3.getDSID(), C);
    cacheDb.update(dev4.getDSID(), D);

    WHEN("the scene 1 is invalidated for one device") {
      proxy->saveScene(dev1.getDSID(), 1);
      proxy->completeProcessing();

      THEN("values for scene 1 are deleted") {
        REQUIRE(!cacheDb.tryLookup(dev1.getDSID()).scene.sceneTable[1].channels[ChannelType::brightness].value);
      }
      AND_THEN("values for the other scenes are still intact") {
        REQUIRE(*cacheDb.tryLookup(dev1.getDSID()).scene.sceneTable[2].channels[ChannelType::brightness].value == 21);
      }
      AND_THEN("values for the other devices are still intact") {
        REQUIRE(
            *cacheDb.tryLookup(dev2.getDSID()).scene.sceneTable[1].channels[ChannelType::shadePositionOutside].value ==
            12);
        REQUIRE(
            *cacheDb.tryLookup(dev2.getDSID()).scene.sceneTable[2].channels[ChannelType::shadePositionOutside].value ==
            22);
        REQUIRE(*cacheDb.tryLookup(dev3.getDSID()).scene.sceneTable[1].channels[ChannelType::brightness].value == 13);
        REQUIRE(*cacheDb.tryLookup(dev3.getDSID())
                     .scene.sceneTable[1]
                     .channels[ChannelType::shadeOpeningAngleOutside]
                     .value == 14);
        REQUIRE(*cacheDb.tryLookup(dev3.getDSID()).scene.sceneTable[2].channels[ChannelType::brightness].value == 23);
        REQUIRE(*cacheDb.tryLookup(dev3.getDSID())
                     .scene.sceneTable[2]
                     .channels[ChannelType::shadeOpeningAngleOutside]
                     .value == 24);
      }
    }

    WHEN("the scene 6 (area) is invalidated") {
      proxy->saveScene(dev4.getDSID(), 6);
      proxy->completeProcessing();

      THEN("scene 1 and the flags for 1 and 6 are removed") {
        REQUIRE(!cacheDb.tryLookup(dev4.getDSID()).scene.sceneTable[1].channels[ChannelType::brightness].value);
        REQUIRE(!cacheDb.tryLookup(dev4.getDSID()).scene.sceneTable[6].channels[ChannelType::brightness].value);
        REQUIRE(!cacheDb.tryLookup(dev4.getDSID()).scene.sceneTable[1].flags.dontCare);
        REQUIRE(!cacheDb.tryLookup(dev4.getDSID()).scene.sceneTable[6].flags.dontCare);
      }
    }

    WHEN("the scene 1 is invalidated on a device with multiple channels") {
      proxy->saveScene(dev3.getDSID(), 1);
      proxy->completeProcessing();

      THEN("values for scene 1 are deleted") {
        REQUIRE(!cacheDb.tryLookup(dev3.getDSID()).scene.sceneTable[1].channels[ChannelType::brightness].value);
        REQUIRE(!cacheDb.tryLookup(dev3.getDSID())
                     .scene.sceneTable[1]
                     .channels[ChannelType::shadeOpeningAngleOutside]
                     .value);
      }
      AND_THEN("values for the other scenes are still intact") {
        REQUIRE(*cacheDb.tryLookup(dev3.getDSID()).scene.sceneTable[2].channels[ChannelType::brightness].value == 23);
        REQUIRE(*cacheDb.tryLookup(dev3.getDSID())
                     .scene.sceneTable[2]
                     .channels[ChannelType::shadeOpeningAngleOutside]
                     .value == 24);
      }
      AND_THEN("values for the other device are still intact") {
        REQUIRE(*cacheDb.tryLookup(dev1.getDSID()).scene.sceneTable[1].channels[ChannelType::brightness].value == 11);
        REQUIRE(*cacheDb.tryLookup(dev1.getDSID()).scene.sceneTable[2].channels[ChannelType::brightness].value == 21);
        REQUIRE(
            *cacheDb.tryLookup(dev2.getDSID()).scene.sceneTable[1].channels[ChannelType::shadePositionOutside].value ==
            12);
        REQUIRE(
            *cacheDb.tryLookup(dev2.getDSID()).scene.sceneTable[2].channels[ChannelType::shadePositionOutside].value ==
            22);
      }
    }

    WHEN("an yet unknown but valid scene is invalidated") {
      THEN("the call succeeds") { REQUIRE_NOTHROW(proxy->saveScene(dev3.getDSID(), 127)); }
    }

    WHEN("an invalid scene 129 is invalidated") {
      THEN("the call fails with an exception") { REQUIRE_THROWS(proxy->saveScene(dev3.getDSID(), 129)); }
    }
  }
}

SCENARIO("Scenario state", TAGS) {
  auto propertySystem = PropertySystem{};
  EventSinkMock eventSink;
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt = *pApartment;
  apt.setEventSink(eventSink);

  BusInterfaceMock busInterface;
  apt.setBusInterface(&busInterface);

  ds::DateTime now = std::chrono::system_clock::now();
  ds::ConfMock conf;
  ALLOW_CALL(dsm_api_mock, DsmApiGetDeviceDescription(_, _, _, _, _, _, _)).RETURN(0);

  DeviceSpec spec;
  spec.FunctionID = functionID100(DeviceClasses_t::DEVICE_CLASS_GE, FID100::output, FID100keys::_1);

  auto zone1 = apt.allocateZone(1);
  auto zone2 = apt.allocateZone(2);
  auto dev1 = apt.allocateDevice(dsuid1);
  dev1->setPartiallyFromSpec(spec);
  dev1->addToGroup(GroupIDYellow);
  dev1->setIsPresent(true);
  dev1->setIsConnected(true);
  auto dev2 = apt.allocateDevice(dsuid2);
  dev2->setPartiallyFromSpec(spec);
  dev2->addToGroup(GroupIDYellow);
  dev2->setIsPresent(true);
  dev2->setIsConnected(true);
  auto dev3 = apt.allocateDevice(dsuid3);
  dev3->setPartiallyFromSpec(spec);
  dev3->addToGroup(GroupIDYellow);
  dev3->setIsPresent(true);
  dev3->setIsConnected(true);
  auto dev4 = apt.allocateDevice(dsuid4);
  dev4->setPartiallyFromSpec(spec);
  dev4->addToGroup(GroupIDGray);
  dev4->setIsPresent(true);
  dev4->setIsConnected(true);
  auto dev5 = apt.allocateDevice(dsuid5);
  dev5->setPartiallyFromSpec(spec);
  dev5->addToGroup(GroupIDGray);
  dev5->setIsPresent(false);
  dev5->setIsConnected(false);

  DeviceReference d1(dev1, &apt);
  DeviceReference d2(dev2, &apt);
  DeviceReference d3(dev3, &apt);
  DeviceReference d4(dev4, &apt);
  DeviceReference d5(dev5, &apt);
  zone1->addDevice(d1);
  zone1->addDevice(d2);
  zone2->addDevice(d3);
  zone2->addDevice(d4);
  zone2->addDevice(d5);

  HardwareAbstractionMock hal{};
  ds::sqlite3::Database db(":memory:");

  CacheDb<ScenarioCache> sc{&db, false};
  auto dc = CacheDb<DeviceCache>(&db);
  {
    auto cache = dc.tryLookup(dsuid1);
    cache.scene.sceneTable[0].flags.dontCare = false;
    cache.scene.sceneTable[5].flags.dontCare = false;
    cache.scene.sceneTable[Scene2].flags.dontCare = true;
    cache.scene.sceneTable[18].flags.dontCare = false;
    dc.update(dsuid1, cache);
  }
  {
    auto cache = dc.tryLookup(dsuid2);
    cache.scene.sceneTable[0].flags.dontCare = false;
    cache.scene.sceneTable[5].flags.dontCare = true;
    cache.scene.sceneTable[Scene2].flags.dontCare = false;
    cache.scene.sceneTable[18].flags.dontCare = false;
    dc.update(dsuid2, cache);
  }
  {
    auto cache = dc.tryLookup(dsuid3);
    cache.scene.sceneTable[0].flags.dontCare = false;
    dc.update(dsuid3, cache);
  }
  {
    auto cache = dc.tryLookup(dsuid4);
    cache.scene.sceneTable[0].flags.dontCare = false;
    cache.scene.sceneTable[5].flags.dontCare = false;
    dc.update(dsuid4, cache);
  }
  REQUIRE_CALL(hal, getCache()).LR_RETURN(dc);

  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);

  GIVEN("A scene is called in a zone/group") {
    REQUIRE_CALL(hal, callScene(dsuid1, 5, false, _)).RETURN(true);
    REQUIRE_CALL(hal, callScene(dsuid2, 5, false, _)).RETURN(false);

    REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(AT_LEAST(1));
    REQUIRE_CALL(hal, getState(dsuid1))
        .RETURN(hal::DeviceState{
            .status = dsModel::OutputStatusStatus::ok,
            .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                   DeviceChannelValues{.value = 100.0}}},
            .localPriority = true,
        });
    REQUIRE_CALL(hal, getState(dsuid2))
        .RETURN(hal::DeviceState{
            .status = dsModel::OutputStatusStatus::ok,
            .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                   DeviceChannelValues{.value = 100.0}}},
            .localPriority = true,
        });
    REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1, 2).RETURN(std::vector<int>({SceneA11}));
    REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1, 2).RETURN(std::vector<int>({SceneA21}));

    // dev1/dev2 in their area state, other devices unknown
    REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

    REQUIRE_NOTHROW(proxy->callScene(1, 1, 5, false, now));
    proxy->completeProcessing();
    {
      auto scenario = sc.tryLookup(DSUID_NULL);
      REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
    }

    WHEN("A device changes disjunct state") {
      REQUIRE_CALL(hal, callScene(dsuid1, Scene2, false, _)).RETURN(false);
      REQUIRE_CALL(hal, callScene(dsuid2, Scene2, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));

      // dev1/dev2 in their area state, locally prioritized
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->callScene(1, 1, Scene2, false, now));
      proxy->completeProcessing();

      THEN("the scenario stays active") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s17"].status == dsModel::ScenarioStatusStatus::active);
      }
    }

    WHEN("A device changes shared state") {
      REQUIRE_CALL(hal, callScene(dsuid1, 18, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid2, 18, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));

      // dev1/dev2 in their area state, locally prioritized
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->callScene(1, 1, 18, false, now));
      proxy->completeProcessing();
      THEN("the scenario stays active") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s18"].status == dsModel::ScenarioStatusStatus::active);
      }
    }

    WHEN("callSceneMin is called") {
      REQUIRE_CALL(hal, callSceneMin(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(hal, callSceneMin(dsuid2, 5, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));

      // dev1/dev2 in their area state, locally prioritized
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->callSceneMin(1, 1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }

    WHEN("undoScene is called") {
      REQUIRE_CALL(hal, undoScene(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(hal, undoScene(dsuid2, 5, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));

      // dev1/dev2 in their area state, locally prioritized
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->undoScene(1, 1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }

    WHEN("setOutputValue is called") {
      REQUIRE_CALL(hal, setOutputValue(dsuid1, _, _)).RETURN(true);
      REQUIRE_CALL(hal, setOutputValue(dsuid2, _, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->setOutputValue(1, 1, 58, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }

    WHEN("A device callScene is called") {
      REQUIRE_CALL(hal, callScene(dsuid1, Scene2, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->callScene(dsuid1, Scene2, false, now));
      proxy->completeProcessing();
      THEN("the scenario becomes active") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s17"].status == dsModel::ScenarioStatusStatus::none);
      }
    }

    WHEN("callSceneMin is called on device") {
      REQUIRE_CALL(hal, callSceneMin(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->callSceneMin(dsuid1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("undoScene is called on device") {
      REQUIRE_CALL(hal, undoScene(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->undoScene(dsuid1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }

    WHEN("setOutputValue is called on device") {
      REQUIRE_CALL(hal, setOutputValue(dsuid1, 58, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->setOutputValue(dsuid1, 58, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }

    WHEN("setOutputValue is called on device") {
      REQUIRE_CALL(hal, setOutputValue(dsuid1, ChannelType::brightness, 58.25, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->setOutputValue(dsuid1, ChannelType::brightness, 58.25, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }

    WHEN("setSceneLocalPrio is called on device") {
      REQUIRE_CALL(hal, setSceneLocalPrio(dsuid1, 6)).RETURN(true);
      REQUIRE_CALL(hal, setSceneLocalPrio(dsuid2, 6)).RETURN(true);

      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(busInterface.action, setState(_, 255, 1, 1, 255, 255));
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->setSceneLocalPrio(1, 1, 6));
      proxy->completeProcessing();

      THEN("the scenario does not change") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("setOperationLock is called on device") {
      REQUIRE_CALL(hal, setOperationLock(dsuid1, true));

      REQUIRE_NOTHROW(proxy->setOperationLock(dsuid1, 5));
      proxy->completeProcessing();
      THEN("the scenario does not change") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("Querying scenarios work") {
      auto scenarioStatus = proxy->zoneScenarioStatus(dsModel::ScenarioId{"applicationZone-z1-g1-s5"});
      REQUIRE(scenarioStatus.id == dsModel::ScenarioId{"applicationZone-z1-g1-s5"});
      REQUIRE(scenarioStatus.status == dsModel::ScenarioStatusStatus::active);
      auto fakeScenarioStatus = proxy->zoneScenarioStatus(dsModel::ScenarioId{"applicationZone-z13-g4-s1"});
      REQUIRE(fakeScenarioStatus.id == dsModel::ScenarioId{"applicationZone-z13-g4-s1"});
      REQUIRE(fakeScenarioStatus.status == dsModel::ScenarioStatusStatus{});
    }
  }

  GIVEN("A scene is called in the whole apartment") {
    REQUIRE_CALL(hal, callScene(dsuid1, 5, false, _)).RETURN(true);
    REQUIRE_CALL(hal, callScene(dsuid2, 5, false, _)).RETURN(false);
    REQUIRE_CALL(hal, callScene(dsuid3, 5, false, _)).RETURN(true);
    REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(AT_LEAST(1));
    REQUIRE_CALL(hal, getState(dsuid1))
        .RETURN(hal::DeviceState{
            .status = dsModel::OutputStatusStatus::ok,
            .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                   DeviceChannelValues{.value = 100.0}}},
            .localPriority = true,
        });
    REQUIRE_CALL(hal, getState(dsuid2))
        .RETURN(hal::DeviceState{
            .status = dsModel::OutputStatusStatus::ok,
            .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                   DeviceChannelValues{.value = 100.0}}},
            .localPriority = true,
        });
    REQUIRE_CALL(hal, getState(dsuid3))
        .RETURN(hal::DeviceState{
            .status = dsModel::OutputStatusStatus::ok,
            .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                   DeviceChannelValues{.value = 100.0}}},
            .localPriority = true,
        });

    REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
    REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
    REQUIRE_CALL(hal, areaMember(dev3->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA31}));
    REQUIRE_CALL(busInterface.action, setState(zone1->getGroup(GroupIDYellow).get(), 255, 1, 1, 255, 255));
    REQUIRE_CALL(busInterface.action, setState(zone2->getGroup(GroupIDYellow).get(), 255, 255, 255, 1, 255));

    REQUIRE_NOTHROW(proxy->callScene(0, 1, 5, false, now));
    proxy->completeProcessing();
    {
      auto scenario = sc.tryLookup(DSUID_NULL);
      REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
    }

    WHEN("A device changes in a zone state") {
      REQUIRE_CALL(hal, callScene(dsuid1, 5, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid2, 5, false, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(zone1->getGroup(GroupIDYellow).get(), 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->callScene(1, 1, 5, false, now));
      proxy->completeProcessing();
      THEN("the apartment scenario becomes inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z1-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("A device changes in another group") {
      REQUIRE_CALL(hal, callScene(dsuid4, 5, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid4))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, areaMember(dsuid4)).RETURN(std::vector<int>({}));

      REQUIRE_NOTHROW(proxy->callScene(2, 2, 5, false, now));
      proxy->completeProcessing();
      THEN("the apartment scenario stays active") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::active);
        REQUIRE(scenario.scenarios["applicationZone-z2-g2-s5"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("An broadcast scene is called") {
      REQUIRE_CALL(hal, callScene(dsuid1, 0, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid2, 0, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid3, 0, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(dsuid4, 0, false, _)).RETURN(true);
      REQUIRE_CALL(eventSink, pushEvent(_));

      REQUIRE_NOTHROW(proxy->callScene(0, 0, 0, false, now));
      proxy->completeProcessing();
      THEN("all scenarios are inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
        REQUIRE(scenario.scenarios["applicationZone-z0-g0-s0"].status == dsModel::ScenarioStatusStatus::active);
      }
    }
    WHEN("callSceneMin is called") {
      REQUIRE_CALL(hal, callSceneMin(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(hal, callSceneMin(dsuid2, 5, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(zone1->getGroup(GroupIDYellow).get(), 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->callSceneMin(1, 1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("undoScene is called") {
      REQUIRE_CALL(hal, undoScene(dsuid1, 5, _)).RETURN(true);
      REQUIRE_CALL(hal, undoScene(dsuid2, 5, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(zone1->getGroup(GroupIDYellow).get(), 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->undoScene(1, 1, 5, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
    WHEN("setOutputValue is called") {
      REQUIRE_CALL(hal, setOutputValue(dsuid1, _, _)).RETURN(true);
      REQUIRE_CALL(hal, setOutputValue(dsuid2, _, _)).RETURN(false);
      REQUIRE_CALL(eventSink, pushEvent(_));
      REQUIRE_CALL(hal, getState(dsuid1))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, getState(dsuid2))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, areaMember(dev1->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
      REQUIRE_CALL(hal, areaMember(dev2->getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA21}));
      REQUIRE_CALL(busInterface.action, setState(zone1->getGroup(GroupIDYellow).get(), 255, 1, 1, 255, 255));

      REQUIRE_NOTHROW(proxy->setOutputValue(1, 1, 58, now));
      proxy->completeProcessing();
      THEN("the scenario becomes inactive ") {
        auto scenario = sc.tryLookup(DSUID_NULL);
        REQUIRE(scenario.scenarios["applicationZone-z0-g1-s5"].status == dsModel::ScenarioStatusStatus::inactive);
      }
    }
  }
  GIVEN("a certain scenario on a group") {
    THEN("get a list of participating devices for scene 5") {
      auto group = zone1->getGroup(1);
      auto result = proxy->scenarioDeviceList(*group.get(), 5);
      REQUIRE(result.size() == 1);
      REQUIRE(result[0] == ds::str(dsuid1));
    }
    THEN("get a list of participating devices for scene 18") {
      auto group = zone1->getGroup(1);
      auto result = proxy->scenarioDeviceList(*group.get(), 18);
      REQUIRE(result.size() == 2);
      REQUIRE(result[0] == ds::str(dsuid1));
      REQUIRE(result[1] == ds::str(dsuid2));
    }
    THEN("get an empty list of participating devices for scene 6") {
      auto group = zone1->getGroup(1);
      auto result = proxy->scenarioDeviceList(*group.get(), 6);
      REQUIRE(result.size() == 0);
    }
  }
}

SCENARIO("Read cache status") {
  auto propertySystem = PropertySystem{};
  EventSinkMock eventSink;
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt = *pApartment;
  apt.setEventSink(eventSink);

  BusInterfaceMock busInterface;
  apt.setBusInterface(&busInterface);

  DeviceSpec spec;
  spec.FunctionID = functionID100(DeviceClasses_t::DEVICE_CLASS_GE, FID100::output, FID100keys::_1);

  auto zone1 = apt.allocateZone(1);
  auto zone2 = apt.allocateZone(2);
  auto dev1 = apt.allocateDevice(dsuid1);
  dev1->setPartiallyFromSpec(spec);
  dev1->addToGroup(GroupIDYellow);
  dev1->setIsPresent(true);
  dev1->setIsConnected(true);
  auto dev2 = apt.allocateDevice(dsuid2);
  dev2->setPartiallyFromSpec(spec);
  dev2->addToGroup(GroupIDYellow);
  dev2->setIsPresent(true);
  dev2->setIsConnected(true);
  auto dev3 = apt.allocateDevice(dsuid3);
  dev3->setPartiallyFromSpec(spec);
  dev3->addToGroup(GroupIDYellow);
  dev3->setIsPresent(true);
  dev3->setIsConnected(true);
  auto dev4 = apt.allocateDevice(dsuid4);
  dev4->setPartiallyFromSpec(spec);
  dev4->addToGroup(GroupIDGray);
  dev4->setIsPresent(true);
  dev4->setIsConnected(true);
  auto dev5 = apt.allocateDevice(dsuid5);
  dev5->setPartiallyFromSpec(spec);
  dev5->addToGroup(GroupIDGray);
  dev5->setIsPresent(false);
  dev5->setIsConnected(false);

  ds::sqlite3::Database db(":memory:");
  auto dc = CacheDb<DeviceCache>(&db);

  ds::ConfMock conf;
  HardwareAbstractionMock hal{};
  REQUIRE_CALL(hal, getCache()).LR_RETURN(dc);
  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);

  using Id = dsModel::FunctionBlockId;
  WHEN("Reading complete cache") {
    REQUIRE_CALL(hal, deviceCacheable(_)).TIMES(4).RETURN(true);
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid1)}, .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid2)}, .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid3)}, .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid4)}, .total = 149, .completed = 149, .completion = 1});
    auto status = proxy->getParameterCacheStatus();
    REQUIRE(*status.cache.total == 4);
    REQUIRE(*status.cache.completed == 4);
    REQUIRE(*status.cache.completion == 100);
    REQUIRE(status.devices.size() == 4);
    REQUIRE(status.cache.status == dsModel::ParameterCacheStatus::complete);
  }

  WHEN("Reading partially complete cache") {
    REQUIRE_CALL(hal, deviceCacheable(_)).TIMES(4).RETURN(true);
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid1)}, .total = 149, .completed = 75, .completion = 0.5});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid2)}, .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid3)}, .total = 149, .completed = 149, .completion = 1});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid4)}, .total = 149, .completed = 149, .completion = 1});
    auto status = proxy->getParameterCacheStatus();
    REQUIRE(*status.cache.total == 4);
    REQUIRE(*status.cache.completed == 3);
    REQUIRE(roundDigits(*status.cache.completion, 1) == 87.6);
    REQUIRE(status.cache.status == dsModel::ParameterCacheStatus::incomplete);
    REQUIRE(status.devices.size() == 4);
  }

  WHEN("Reading empty cache") {
    REQUIRE_CALL(hal, deviceCacheable(_)).TIMES(4).RETURN(true);
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid1)}, .total = 149, .completed = 0, .completion = 0});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid2)}, .total = 149, .completed = 0, .completion = 0});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid3)}, .total = 149, .completed = 0, .completion = 0});
    REQUIRE_CALL(hal, cacheStatus(_))
        .RETURN(DeviceCacheState{.id = Id{ds::str(dsuid4)}, .total = 149, .completed = 0, .completion = 0});
    auto status = proxy->getParameterCacheStatus();
    REQUIRE(*status.cache.total == 4);
    REQUIRE(*status.cache.completed == 0);
    REQUIRE(roundDigits(*status.cache.completion, 3) == 0);
    REQUIRE(status.cache.status == dsModel::ParameterCacheStatus::incomplete);
    REQUIRE(status.devices.size() == 4);
  }
}

SCENARIO("A zone that contains SW terminal blocks in the audio group") {
  EventSinkMock eventSink;
  auto propertySystem = PropertySystem{};
  auto pApartment = makeApartment(nullptr, &propertySystem);
  auto& apt = *pApartment;
  apt.setEventSink(eventSink);

  BusInterfaceMock busInterface;
  apt.setBusInterface(&busInterface);

  ds::DateTime now = std::chrono::system_clock::now();

  DeviceSpec spec;
  spec.FunctionID = functionID100(DeviceClasses_t::DEVICE_CLASS_GE, FID100::output, FID100keys::_1);

  auto zone1 = apt.allocateZone(1);
  auto dev1 = apt.allocateDevice(dsuid1);
  dev1->setPartiallyFromSpec(spec);
  dev1->addToGroup(static_cast<int>(ApplicationType::audio));
  dev1->setIsPresent(true);
  dev1->setIsConnected(true);
  auto dev2 = apt.allocateDevice(dsuid2);
  dev2->setPartiallyFromSpec(spec);
  dev2->addToGroup(static_cast<int>(ApplicationType::audio));
  dev2->setIsPresent(true);
  dev2->setIsConnected(true);

  ds::sqlite3::Database db(":memory:");
  auto dc = CacheDb<DeviceCache>(&db);
  CacheDb<ScenarioCache> sc{&db, false};

  ds::ConfMock conf;
  HardwareAbstractionMock hal{};
  REQUIRE_CALL(hal, getCache()).LR_RETURN(dc);
  auto proxy = makeApartmentProxy(&apt, &hal, &db, nullptr, conf);

  DeviceReference d1(dev1, &apt);
  DeviceReference d2(dev2, &apt);
  zone1->addDevice(d1);
  zone1->addDevice(d2);

  WHEN("Preset1 is called in the zone") {
    REQUIRE_CALL(hal, callScene(dsuid1, Scene1, false, now)).RETURN(true);
    REQUIRE_CALL(hal, callScene(dsuid2, Scene1, false, now)).RETURN(true);
    REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(AT_LEAST(1));
    REQUIRE_CALL(hal, getState(dsuid1))
        .RETURN(hal::DeviceState{
            .status = dsModel::OutputStatusStatus::ok,
            .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::powerState,
                                                                   DeviceChannelValues{.value = 100.0}}},
            .localPriority = true,
        });
    REQUIRE_CALL(hal, getState(dsuid2))
        .RETURN(hal::DeviceState{
            .status = dsModel::OutputStatusStatus::ok,
            .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::powerState,
                                                                   DeviceChannelValues{.value = 100.0}}},
            .localPriority = true,
        });
    proxy->callScene(1, static_cast<int>(ApplicationType::audio), Scene1, false, now);
    proxy->completeProcessing();

    THEN("Zone is playing") {
      hal::DeviceState ds1{.lastChanged = {},
                           .terminatesAt = {},
                           .status = dsModel::OutputStatusStatus::ok,
                           .outputs = {{ChannelType::powerState, {.value = 0.0}}}};
      dsModel::ZoneApplicationsStatus zs1{.audio = {
                                            .status = dsModel::AudioStatusStatus::playing,
                                            .muted = true,
                                            .volume = 30.0
                                          }};
      REQUIRE_CALL(hal, getState(dsuid1)).TIMES(1).RETURN(ds1);
      REQUIRE_CALL(hal, getState(dsuid2)).TIMES(1).RETURN(ds1);
      REQUIRE_CALL(hal, getZoneState(1)).RETURN(zs1);
      REQUIRE(proxy->audioStatus(zone1->id()).status == dsModel::AudioStatusStatus::playing);
      auto scenario = sc.tryLookup(DSUID_NULL);
      REQUIRE(scenario.scenarios["applicationZone-z1-g4-s5"].status == dsModel::ScenarioStatusStatus::active);

      WHEN("Another scene is called in the zone") {
        REQUIRE_CALL(hal, callScene(dsuid1, SceneOff, false, now)).RETURN(true);
        REQUIRE_CALL(hal, callScene(dsuid2, SceneOff, false, now)).RETURN(true);
        REQUIRE_CALL(eventSink, pushEvent(_));
        REQUIRE_CALL(hal, getState(dsuid1))
            .RETURN(hal::DeviceState{
                .status = dsModel::OutputStatusStatus::ok,
                .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::powerState,
                                                                       DeviceChannelValues{.value = 100.0}}},
                .localPriority = true,
            });
        REQUIRE_CALL(hal, getState(dsuid2))
            .RETURN(hal::DeviceState{
                .status = dsModel::OutputStatusStatus::ok,
                .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::powerState,
                                                                       DeviceChannelValues{.value = 100.0}}},
                .localPriority = true,
            });
        proxy->callScene(1, static_cast<int>(ApplicationType::audio), SceneOff, false, now);
        proxy->completeProcessing();

        THEN("The first scene becomes inactive and the second one becomes active") {
          auto scenario = sc.tryLookup(DSUID_NULL);
          REQUIRE(scenario.scenarios["applicationZone-z1-g4-s0"].status == dsModel::ScenarioStatusStatus::active);
          REQUIRE(scenario.scenarios["applicationZone-z1-g4-s5"].status == dsModel::ScenarioStatusStatus::inactive);
        }
      }
    }

  }
}

SCENARIO("ventilation/recirculation", TAGS) {
  EventSinkMock eventSink;
  auto propertySystem = PropertySystem{};
  ApartmentMock apartment{nullptr, &propertySystem};
  apartment.setEventSink(eventSink);

  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);

  ds::sqlite3::Database db(":memory:");
  auto dc = CacheDb<DeviceCache>(&db);
  CacheDb<ScenarioCache> sc{&db, false};

  ds::ConfMock conf;
  HardwareAbstractionMock hal{};
  REQUIRE_CALL(hal, getCache()).LR_RETURN(dc);
  auto proxy = makeApartmentProxy(&apartment, &hal, &db, nullptr, conf);

  GIVEN("ventilation/recirculation devices in zone1") {
    auto& zone1 = *apartment.allocateZone(1);

    auto& ventilationDevice = *createDevice(apartment, {.hardwareType = HardwareType::SW_UMR200});
    auto& recirculationDevice = *createDevice(apartment, {.hardwareType = HardwareType::SW_UMR200});

    markOnline(ventilationDevice);
    markOnline(recirculationDevice);

    ventilationDevice.setActiveGroup(ApplicationType::ventilation);
    ventilationDevice.addToGroup(GroupIDVentilation);
    recirculationDevice.setActiveGroup(ApplicationType::recirculation);
    recirculationDevice.addToGroup(GroupIDRecirculation);

    zone1.addDevice({ventilationDevice.sharedFromThis(), &apartment});
    zone1.addDevice({recirculationDevice.sharedFromThis(), &apartment});

    WHEN("missing ventilation addon entries in property tree") {
      THEN("empty zone status") {
        REQUIRE(proxy->ventilationStatus(zone1.id()) == dsModel::VentilationStatus{});
        REQUIRE(proxy->recirculationStatus(zone1.id()) == dsModel::VentilationStatus{});
      }

      THEN("empty zone levels") {
        REQUIRE(proxy->ventilationLevels(zone1.id()) == dsModel::Ventilation{});
        REQUIRE(proxy->recirculationLevels(zone1.id()) == dsModel::Ventilation{});
      }
    }

    WHEN("ventilation addon entries in property tree exist") {
      auto zoneRecirculationNode =
          propertySystem.createProperty("/scripts/system-addon-ventilation-controller/zone-recirculation/1");

      auto zoneVentilationNode{
          propertySystem.createProperty("/scripts/system-addon-ventilation-controller/zone-ventilation/1")};

      WHEN("ventilationLevel == 1, off not available") {
        zoneRecirculationNode->setChildValue("ventilationLevel", 1);
        zoneVentilationNode->setChildValue("ventilationLevel", 1);

        THEN("then availableLevels: level1/boost") {
          auto availableLevels = std::vector<dsModel::VentilationLevel>{dsModel::VentilationLevel::level1,
                                                                        dsModel::VentilationLevel::boost};
          REQUIRE(proxy->recirculationLevels(zone1.id()) == dsModel::Ventilation{.availableLevels = availableLevels});
          REQUIRE(proxy->ventilationLevels(zone1.id()) == dsModel::Ventilation{.availableLevels = availableLevels});
        }
      }

      WHEN("ventilationLevel == 4, off available") {
        zoneRecirculationNode->setChildValue("ventilationLevel", 4);
        zoneVentilationNode->setChildValue("ventilationLevel", 4);
        zoneRecirculationNode->setChildValue("ventilationManual", true);
        zoneVentilationNode->setChildValue("ventilationManual", true);

        THEN("then availableLevels: off/level1/level2/level3/level4/boost") {
          auto availableLevels = std::vector<dsModel::VentilationLevel>{
              dsModel::VentilationLevel::off,    dsModel::VentilationLevel::level1, dsModel::VentilationLevel::level2,
              dsModel::VentilationLevel::level3, dsModel::VentilationLevel::level4, dsModel::VentilationLevel::boost};
          REQUIRE(proxy->recirculationLevels(zone1.id()) == dsModel::Ventilation{.availableLevels = availableLevels});
          REQUIRE(proxy->ventilationLevels(zone1.id()) == dsModel::Ventilation{.availableLevels = availableLevels});
        }
      }
    }

    WHEN("addon property nodes are populated") {
      auto zoneVentilationNode =
          propertySystem.createProperty("/scripts/system-addon-ventilation-controller/zone-ventilation/1");
      auto zoneRecirculationNode =
          propertySystem.createProperty("/scripts/system-addon-ventilation-controller/zone-recirculation/1");

#if 0
      // if checking available status in ventilationStatusImpl -> see TODO
      zoneVentilationNode->setChildValue("ventilationLevel", 2);
      zoneRecirculationNode->setChildValue("ventilationLevel", 2);
#endif

      auto statesNode = propertySystem.createProperty("/usr/states/");
      statesNode->setChildValue(
          ds::str("zone.", zone1.getID(), ".group.", GroupIDVentilation, ".status.malfunction/value"), 1);
      statesNode->setChildValue(ds::str("zone.", zone1.getID(), ".group.", GroupIDVentilation, ".status.service/value"),
                                0);
      statesNode->setChildValue(
          ds::str("zone.", zone1.getID(), ".group.", GroupIDRecirculation, ".status.malfunction/value"), 0);
      statesNode->setChildValue(
          ds::str("zone.", zone1.getID(), ".group.", GroupIDRecirculation, ".status.service/value"), 1);

      zone1.addDevice({ventilationDevice.sharedFromThis(), &apartment});
      zone1.addDevice({recirculationDevice.sharedFromThis(), &apartment});

      WHEN("call scene in zone/group") {
        REQUIRE_CALL(hal, callScene(ventilationDevice.getDSID(), SceneOff, false, _)).RETURN(true);
        REQUIRE_CALL(hal, callScene(recirculationDevice.getDSID(), Scene1, false, _)).RETURN(true);
        zone1.getGroup(GroupIDRecirculation)->setLastCalledScene(Scene1);
        zone1.getGroup(GroupIDVentilation)->setLastCalledScene(SceneOff);
        REQUIRE_CALL(eventSink, pushEvent(_)).TIMES(AT_LEAST(2));

        THEN("status is update upon callScene") {
          proxy->callScene(zone1.getID(), GroupIDVentilation, SceneOff, false, ds::DateTime());
          proxy->callScene(zone1.getID(), GroupIDRecirculation, Scene1, false, ds::DateTime());
          proxy->completeProcessing();

          CHECK(proxy->ventilationStatus(zone1.id()) ==
                dsModel::VentilationStatus{
                    .level = dsModel::VentilationLevel::off, .malfunction = true, .serviceRequired = false});
          CHECK(proxy->recirculationStatus(zone1.id()) ==
                dsModel::VentilationStatus{
                    .level = dsModel::VentilationLevel::level1, .malfunction = false, .serviceRequired = true});
        }
      }
    }
  }

  GIVEN("apartment ventilation/recirculation device status") {
    WHEN("no apartment ventilation devices in apartment") {
      THEN("global ventilation unit and status is empty") {
        CHECK(proxy->ventilationUnitLevels() == dsModel::Ventilation{});
        CHECK(proxy->ventilationUnitStatus() == dsModel::VentilationStatus{});
      }
    }

    WHEN("apartment ventilation device present and states present") {
      auto& ventilationDevice = *createDevice(apartment, {.hardwareType = HardwareType::SW_UMR200});
      ventilationDevice.setActiveGroup(ApplicationType::apartmentVentilation);
      ventilationDevice.addToGroup(GroupIDGlobalAppDsVentilation);
      markOnline(ventilationDevice);

      auto statesNode = propertySystem.createProperty("/usr/states/");
      statesNode->setChildValue("zone.0.group.64.status.malfunction/value", 2);
      statesNode->setChildValue("zone.0.group.64.status.service/value", 1);

      auto& zone0 = *apartment.allocateZone(0);
      zone0.getGroup(GroupIDGlobalAppDsVentilation)->setLastCalledScene(Scene1);

      THEN("return default setting") {
        CHECK(proxy->ventilationUnitLevels() ==
              dsModel::Ventilation{.availableLevels = std::vector<dsModel::VentilationLevel>{
                                       dsModel::VentilationLevel::off, dsModel::VentilationLevel::level1,
                                       dsModel::VentilationLevel::level2, dsModel::VentilationLevel::level3,
                                       dsModel::VentilationLevel::level4, dsModel::VentilationLevel::boost}});
      }
      THEN("status is taken from property tree entries and last called scene") {
        CHECK(proxy->ventilationUnitStatus() == dsModel::VentilationStatus{.level = dsModel::VentilationLevel::level1,
                                                                           .malfunction = false,
                                                                           .serviceRequired = true});
      }
    }

    WHEN("apartment ventilation device present but inactive") {
      auto& ventilationDevice = *createDevice(apartment, {.hardwareType = HardwareType::SW_UMR200});
      ventilationDevice.setActiveGroup(ApplicationType::apartmentVentilation);
      ventilationDevice.addToGroup(GroupIDGlobalAppDsVentilation);

      auto& zone0 = *apartment.allocateZone(0);
      zone0.getGroup(GroupIDGlobalAppDsVentilation)->setLastCalledScene(Scene1);

      THEN("return empty levels") {
        CHECK(proxy->ventilationUnitLevels() == dsModel::Ventilation{});
        CHECK(proxy->ventilationUnitStatus() == dsModel::VentilationStatus{});
      }
    }

    WHEN("zone ventilation device present but inactive") {
      auto& ventilationDevice = *createDevice(apartment, {.hardwareType = HardwareType::BL_UMV200});
      ventilationDevice.setActiveGroup(ApplicationType::recirculation);
      ventilationDevice.addToGroup(GroupIDRecirculation);

      auto& zone0 = *apartment.allocateZone(1);
      zone0.getGroup(GroupIDRecirculation)->setLastCalledScene(Scene1);

      THEN("return empty levels") {
        CHECK(proxy->ventilationUnitLevels() == dsModel::Ventilation{});
        CHECK(proxy->ventilationUnitStatus() == dsModel::VentilationStatus{});
      }
    }

  }
}

SCENARIO("stade status", TAGS) {
  EventSinkMock eventSink;
  auto propertySystem = PropertySystem{};
  ApartmentMock apartment{nullptr, &propertySystem};
  apartment.setEventSink(eventSink);

  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);

  ds::sqlite3::Database db(":memory:");
  auto dc = CacheDb<DeviceCache>(&db);
  CacheDb<ScenarioCache> sc{&db, false};

  ds::ConfMock conf;
  HardwareAbstractionMock hal{};
  REQUIRE_CALL(hal, getCache()).LR_RETURN(dc);
  auto proxy = makeApartmentProxy(&apartment, &hal, &db, nullptr, conf);

  GIVEN("zone with single device in all areas") {
    auto& shadeDevice = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL220});
    markOnline(shadeDevice);
    shadeDevice.addToGroup(GroupIDGray);

    auto& zone1 = *apartment.allocateZone(1);
    zone1.addDevice({shadeDevice.sharedFromThis(), &apartment});

    REQUIRE_CALL(hal, areaMember(shadeDevice.getDSID()))
        .TIMES(1)
        .RETURN(std::vector<int>({SceneA11, SceneA21, SceneA31, SceneA41}));

    WHEN("device on / non localPriority") {
      auto state = hal::DeviceState{
          .status = dsModel::OutputStatusStatus::ok,
          .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                 DeviceChannelValues{.value = 100.0}}},
          .localPriority = false,
      };
      REQUIRE_CALL(hal, getState(shadeDevice.getDSID())).TIMES(1).RETURN(state);

      THEN("all room/area states are on") {
        auto states = proxy->shadesStatus(dsModel::ZoneId{"1"});
        REQUIRE(states == dsModel::ShadesStatus{.status = dsModel::ShadesStatusStatus::open,
                                                .nonLocalPriority = dsModel::ShadesStatusStatus::open,
                                                .area1 = dsModel::ShadesStatusStatus::open,
                                                .area2 = dsModel::ShadesStatusStatus::open,
                                                .area3 = dsModel::ShadesStatusStatus::open,
                                                .area4 = dsModel::ShadesStatusStatus::open});
      }
    }
    WHEN("device on / localPriority") {
      // localPriority only when calculating DSM state, not accessible
      auto state = hal::DeviceState{
          .status = dsModel::OutputStatusStatus::ok,
          .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                 DeviceChannelValues{.value = 100.0}}},
          .localPriority = true,
      };
      REQUIRE_CALL(hal, getState(shadeDevice.getDSID())).TIMES(1).RETURN(state);

      THEN("all room/area states are on") {
        auto states = proxy->shadesStatus(dsModel::ZoneId{"1"});
        REQUIRE(states == dsModel::ShadesStatus{.status = dsModel::ShadesStatusStatus::open,
                                                .area1 = dsModel::ShadesStatusStatus::open,
                                                .area2 = dsModel::ShadesStatusStatus::open,
                                                .area3 = dsModel::ShadesStatusStatus::open,
                                                .area4 = dsModel::ShadesStatusStatus::open});
      }
    }
  }
  GIVEN("zone with two device, in disjunct areas") {
    auto& shadeDevice1 = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL220});
    markOnline(shadeDevice1);
    shadeDevice1.addToGroup(GroupIDGray);

    auto& shadeDevice2 = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL220});
    markOnline(shadeDevice2);
    shadeDevice2.addToGroup(GroupIDGray);

    auto& zone1 = *apartment.allocateZone(1);
    zone1.addDevice({shadeDevice1.sharedFromThis(), &apartment});
    zone1.addDevice({shadeDevice2.sharedFromThis(), &apartment});

    REQUIRE_CALL(hal, areaMember(shadeDevice1.getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
    REQUIRE_CALL(hal, areaMember(shadeDevice2.getDSID()))
        .TIMES(1)
        .RETURN(std::vector<int>({SceneA21, SceneA31, SceneA41}));

    WHEN("on / off -- none localPriority") {
      // localPriority only when calculating DSM state, not accessible
      REQUIRE_CALL(hal, getState(shadeDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(shadeDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = false,
          });
      THEN("room is mixed, area are on/off") {
        auto states = proxy->shadesStatus(dsModel::ZoneId{"1"});
        REQUIRE(states == dsModel::ShadesStatus{.status = dsModel::ShadesStatusStatus::mixed,
                                                .nonLocalPriority = dsModel::ShadesStatusStatus::mixed,
                                                .area1 = dsModel::ShadesStatusStatus::closed,
                                                .area2 = dsModel::ShadesStatusStatus::open,
                                                .area3 = dsModel::ShadesStatusStatus::open,
                                                .area4 = dsModel::ShadesStatusStatus::open});
      }
    }
    WHEN("on / off -- one localPriority") {
      // localPriority only when calculating DSM state, not accessible
      REQUIRE_CALL(hal, getState(shadeDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(shadeDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      THEN("room is mixed, area are on/off") {
        // localPriority not taken into account for union room state
        auto states = proxy->shadesStatus(dsModel::ZoneId{"1"});
        REQUIRE(states == dsModel::ShadesStatus{.status = dsModel::ShadesStatusStatus::mixed,
                                                .nonLocalPriority = dsModel::ShadesStatusStatus::closed,
                                                .area1 = dsModel::ShadesStatusStatus::closed,
                                                .area2 = dsModel::ShadesStatusStatus::open,
                                                .area3 = dsModel::ShadesStatusStatus::open,
                                                .area4 = dsModel::ShadesStatusStatus::open});
      }
    }
  }
}

SCENARIO("light status", TAGS) {
  EventSinkMock eventSink;
  auto propertySystem = PropertySystem{};
  ApartmentMock apartment{nullptr, &propertySystem};
  apartment.setEventSink(eventSink);

  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);

  ds::sqlite3::Database db(":memory:");
  auto dc = CacheDb<DeviceCache>(&db);
  CacheDb<ScenarioCache> sc{&db, false};

  ds::ConfMock conf;
  HardwareAbstractionMock hal{};
  REQUIRE_CALL(hal, getCache()).LR_RETURN(dc);
  auto proxy = makeApartmentProxy(&apartment, &hal, &db, nullptr, conf);

  GIVEN("zone with single device in all areas") {
    auto& lightDevice = *createDevice(apartment, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(lightDevice);
    lightDevice.addToGroup(GroupIDYellow);

    auto& zone1 = *apartment.allocateZone(1);
    zone1.addDevice({lightDevice.sharedFromThis(), &apartment});

    REQUIRE_CALL(hal, areaMember(lightDevice.getDSID()))
        .TIMES(1)
        .RETURN(std::vector<int>({SceneA11, SceneA21, SceneA31, SceneA41}));

    WHEN("device on / non localPriority") {
      REQUIRE_CALL(hal, getState(lightDevice.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = false,
          });

      THEN("all room/area states are on") {
        auto states = proxy->lightsStatus(dsModel::ZoneId{"1"});
        REQUIRE(states == dsModel::LightsStatus{.status = dsModel::LightsStatusStatus::on,
                                                .nonLocalPriority = dsModel::LightsStatusStatus::on,
                                                .area1 = dsModel::LightsStatusStatus::on,
                                                .area2 = dsModel::LightsStatusStatus::on,
                                                .area3 = dsModel::LightsStatusStatus::on,
                                                .area4 = dsModel::LightsStatusStatus::on});
      }
    }
    WHEN("device on / localPriority") {
      REQUIRE_CALL(hal, getState(lightDevice.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      THEN("all room/area states are on") {
        // nonLocalPriority not set if all devices are locally prioritized
        auto states = proxy->lightsStatus(dsModel::ZoneId{"1"});
        REQUIRE(states == dsModel::LightsStatus{.status = dsModel::LightsStatusStatus::on,
                                                .area1 = dsModel::LightsStatusStatus::on,
                                                .area2 = dsModel::LightsStatusStatus::on,
                                                .area3 = dsModel::LightsStatusStatus::on,
                                                .area4 = dsModel::LightsStatusStatus::on});
      }
    }
  }
  GIVEN("zone with two device, in disjunct areas") {
    auto& lightDevice1 = *createDevice(apartment, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(lightDevice1);
    lightDevice1.addToGroup(GroupIDYellow);

    auto& lightDevice2 = *createDevice(apartment, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(lightDevice2);
    lightDevice2.addToGroup(GroupIDYellow);

    auto& zone1 = *apartment.allocateZone(1);
    zone1.addDevice({lightDevice1.sharedFromThis(), &apartment});
    zone1.addDevice({lightDevice2.sharedFromThis(), &apartment});

    REQUIRE_CALL(hal, areaMember(lightDevice1.getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));
    REQUIRE_CALL(hal, areaMember(lightDevice2.getDSID()))
        .TIMES(1)
        .RETURN(std::vector<int>({SceneA21, SceneA31, SceneA41}));

    WHEN("on / off -- none localPriority") {
      REQUIRE_CALL(hal, getState(lightDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(lightDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = false,
          });
      THEN("room is mixed, area are on/off") {
        auto states = proxy->lightsStatus(dsModel::ZoneId{"1"});
        REQUIRE(states == dsModel::LightsStatus{.status = dsModel::LightsStatusStatus::on,
                                                .nonLocalPriority = dsModel::LightsStatusStatus::on,
                                                .area1 = dsModel::LightsStatusStatus::off,
                                                .area2 = dsModel::LightsStatusStatus::on,
                                                .area3 = dsModel::LightsStatusStatus::on,
                                                .area4 = dsModel::LightsStatusStatus::on});
      }
    }
    WHEN("on / off -- one localPriority") {
      REQUIRE_CALL(hal, getState(lightDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(lightDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      THEN("room is mixed, area are on/off") {
        auto states = proxy->lightsStatus(dsModel::ZoneId{"1"});
        REQUIRE(states == dsModel::LightsStatus{.status = dsModel::LightsStatusStatus::on,
                                                .nonLocalPriority = dsModel::LightsStatusStatus::off,
                                                .area1 = dsModel::LightsStatusStatus::off,
                                                .area2 = dsModel::LightsStatusStatus::on,
                                                .area3 = dsModel::LightsStatusStatus::on,
                                                .area4 = dsModel::LightsStatusStatus::on});
      }
    }
  }
}

// maybe sync with apartment-status-test which are more complex
SCENARIO("groupsStatus tests", TAGS) {
  EventSinkMock eventSink;
  auto propertySystem = PropertySystem{};
  ApartmentMock apartment{nullptr, &propertySystem};
  apartment.setEventSink(eventSink);

  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);

  ds::sqlite3::Database db(":memory:");
  auto dc = CacheDb<DeviceCache>(&db);
  CacheDb<ScenarioCache> sc{&db, false};

  ds::ConfMock conf;
  HardwareAbstractionMock hal{};
  REQUIRE_CALL(hal, getCache()).LR_RETURN(dc);
  auto proxy = makeApartmentProxy(&apartment, &hal, &db, nullptr, conf);

  GIVEN("2 lights devices in separate groups") {
    auto& lightDevice1 = *createDevice(apartment, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(lightDevice1);
    lightDevice1.addToGroup(GroupIDYellow);

    auto& lightDevice2 = *createDevice(apartment, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(lightDevice2);
    lightDevice2.addToGroup(GroupIDYellow);

    auto& zone1 = *apartment.allocateZone(1);
    zone1.addDevice({lightDevice1.sharedFromThis(), &apartment});
    auto& zone2 = *apartment.allocateZone(2);
    zone2.addDevice({lightDevice2.sharedFromThis(), &apartment});

    apartment.getZone(0)->addGroup(std::make_shared<Cluster>(16, apartment));
    auto cluster = apartment.getCluster(16);
    cluster->setApplicationType(ApplicationType::lights);

    lightDevice1.addToGroup(cluster->getID());
    lightDevice2.addToGroup(cluster->getID());

    // uses getLightApplicationStatus(devices) but ignores the area status
    ALLOW_CALL(hal, areaMember(_)).RETURN(std::vector<int>({}));

    // NOTE one device has localPriority set, which is don't care for union state over all device states
    WHEN("first off/second on, cluster is on") {
      REQUIRE_CALL(hal, getState(lightDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(lightDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      THEN("group status is on") {
        REQUIRE(proxy->groupsStatus() ==
                std::vector<dsModel::GroupStatus>{
                    {.id = "16", .application = {.lights = {.status = dsModel::LightsStatusStatus::on}}}});
      }
    }
    WHEN("first on/second off, cluster is on") {
      REQUIRE_CALL(hal, getState(lightDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(lightDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = true,
          });

      THEN("group status is on") {
        REQUIRE(proxy->groupsStatus() ==
                std::vector<dsModel::GroupStatus>{
                    {.id = "16", .application = {.lights = {.status = dsModel::LightsStatusStatus::on}}}});
      }
    }
    WHEN("both off, cluster is off") {
      REQUIRE_CALL(hal, getState(lightDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(lightDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = true,
          });

      THEN("group status is on") {
        REQUIRE(proxy->groupsStatus() ==
                std::vector<dsModel::GroupStatus>{
                    {.id = "16", .application = {.lights = {.status = dsModel::LightsStatusStatus::off}}}});
      }
    }
  }

  GIVEN("2 shade devices in separate groups") {
    auto& shadeDevice1 = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL220});
    markOnline(shadeDevice1);
    shadeDevice1.addToGroup(GroupIDYellow);

    auto& shadeDevice2 = *createDevice(apartment, {.hardwareType = HardwareType::GR_KL220});
    markOnline(shadeDevice2);
    shadeDevice2.addToGroup(GroupIDYellow);

    auto& zone1 = *apartment.allocateZone(1);
    zone1.addDevice({shadeDevice1.sharedFromThis(), &apartment});
    auto& zone2 = *apartment.allocateZone(2);
    zone2.addDevice({shadeDevice2.sharedFromThis(), &apartment});

    apartment.getZone(0)->addGroup(std::make_shared<Cluster>(16, apartment));
    auto cluster = apartment.getCluster(16);
    cluster->setApplicationType(ApplicationType::blinds);

    shadeDevice1.addToGroup(cluster->getID());
    shadeDevice2.addToGroup(cluster->getID());

    // uses getShadeApplicationStatus(devices) but ignores the area status
    ALLOW_CALL(hal, areaMember(_)).RETURN(std::vector<int>({}));

    WHEN("first open/second closed, cluster is mixed") {
      REQUIRE_CALL(hal, getState(shadeDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(shadeDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      THEN("group status is on") {
        REQUIRE(proxy->groupsStatus() ==
                std::vector<dsModel::GroupStatus>{
                    {.id = "16", .application = {.shades = {.status = dsModel::ShadesStatusStatus::mixed}}}});
      }
    }
    WHEN("first closed/second open, cluster is mixed") {
      REQUIRE_CALL(hal, getState(shadeDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(shadeDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = true,
          });

      THEN("group status is on") {
        REQUIRE(proxy->groupsStatus() ==
                std::vector<dsModel::GroupStatus>{
                    {.id = "16", .application = {.shades = {.status = dsModel::ShadesStatusStatus::mixed}}}});
      }
    }
    WHEN("both closed, cluster is closed") {
      REQUIRE_CALL(hal, getState(shadeDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(shadeDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 0.0}}},
              .localPriority = true,
          });

      THEN("group status is on") {
        REQUIRE(proxy->groupsStatus() ==
                std::vector<dsModel::GroupStatus>{
                    {.id = "16", .application = {.shades = {.status = dsModel::ShadesStatusStatus::closed}}}});
      }
    }
    WHEN("both open, cluster is open") {
      REQUIRE_CALL(hal, getState(shadeDevice1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = false,
          });
      REQUIRE_CALL(hal, getState(shadeDevice2.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::shadePositionOutside,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });

      THEN("group status is on") {
        REQUIRE(proxy->groupsStatus() ==
                std::vector<dsModel::GroupStatus>{
                    {.id = "16", .application = {.shades = {.status = dsModel::ShadesStatusStatus::open}}}});
      }
    }
  }
  GIVEN("1 audio devices in cluster") {
    auto& audioDevice1 = *createDevice(apartment, {.hardwareType = HardwareType::SW_TKM210});
    markOnline(audioDevice1);
    audioDevice1.addToGroup(GroupIDCyan);

    auto& zone1 = *apartment.allocateZone(1);
    zone1.addDevice({audioDevice1.sharedFromThis(), &apartment});

    apartment.getZone(0)->addGroup(std::make_shared<Cluster>(16, apartment));
    auto cluster = apartment.getCluster(16);
    cluster->setApplicationType(ApplicationType::audio);

    audioDevice1.addToGroup(cluster->getID());
    WHEN("requesting cluster status") {
      // audio cluster status not implemented
      THEN("empty status is returned") {
        REQUIRE(!cluster->getDevices().getByPresence(true).empty()); // device is there
        REQUIRE(proxy->groupsStatus() == std::vector<dsModel::GroupStatus>{});
      }
    }
  }
}

SCENARIO("Empty click for lights", TAGS) {
  auto propertySystem = PropertySystem{};
  ApartmentMock apartment{nullptr, &propertySystem};

  EventSinkMock eventSink;
  apartment.setEventSink(eventSink);
  ALLOW_CALL(eventSink, pushEvent(_));

  BusInterfaceMock busInterface;
  apartment.setBusInterface(&busInterface);

  ds::sqlite3::Database db(":memory:");
  auto dc = CacheDb<DeviceCache>(&db);
  CacheDb<ScenarioCache> sc{&db, false};

  ds::ConfMock conf;
  HardwareAbstractionMock hal{};
  REQUIRE_CALL(hal, getCache()).LR_RETURN(dc);
  auto proxy = makeApartmentProxy(&apartment, &hal, &db, nullptr, conf);

  GIVEN("2 lights devices same zone / both off") {
    auto& light1 = *createDevice(apartment, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(light1);
    light1.addToGroup(GroupIDYellow);

    auto& light2 = *createDevice(apartment, {.hardwareType = HardwareType::GE_TKM210});
    markOnline(light2);
    light2.addToGroup(GroupIDYellow);

    auto& zone1 = *apartment.allocateZone(1);
    zone1.addDevice({light1.sharedFromThis(), &apartment});
    zone1.addDevice({light2.sharedFromThis(), &apartment});

    WHEN("light1 turned on in area1 with local priority") {
      REQUIRE_CALL(hal, callScene(light1.getDSID(), SceneA11, false, _)).RETURN(true);
      REQUIRE_CALL(hal, callScene(light2.getDSID(), SceneA11, false, _)).RETURN(false);

      REQUIRE_CALL(hal, getState(light1.getDSID()))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
              .localPriority = true,
          });
      REQUIRE_CALL(hal, areaMember(light1.getDSID())).TIMES(1).RETURN(std::vector<int>({SceneA11}));

      // despite 2nd device does not act on call scene, syncDSMState requires it
      REQUIRE_CALL(hal, getState(light2.getDSID()))
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 0.0}}},
          });
      REQUIRE_CALL(hal, areaMember(light2.getDSID())).TIMES(1).RETURN(std::vector<int>({}));

      // nonLocalPriority room-state:off, area1: on
      REQUIRE_CALL(busInterface.action, setState(_, 0, 1, 255, 255, 255));

      proxy->callScene(zone1.getID(), GroupIDYellow, SceneA11, false, ds::DateTime());
      proxy->completeProcessing(); // fence

      THEN("dsm sees room as being off") {} // see setState above
    }

    GIVEN("light1 turned on in area1 but not with local priority") {
      REQUIRE_CALL(hal, getState(light1.getDSID()))
          .TIMES(1)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 100.0}}},
          });
      REQUIRE_CALL(hal, areaMember(light1.getDSID())).TIMES(3).RETURN(std::vector<int>({SceneA11}));

      REQUIRE_CALL(hal, getState(light2.getDSID()))
          .TIMES(3)
          .RETURN(hal::DeviceState{
              .status = dsModel::OutputStatusStatus::ok,
              .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                     DeviceChannelValues{.value = 0.0}}},
          });

      REQUIRE_CALL(hal, areaMember(light2.getDSID())).TIMES(3).RETURN(std::vector<int>({}));

      REQUIRE(proxy->lightsStatus(zone1.id()) ==
              dsModel::LightsStatus{.status = dsModel::LightsStatusStatus::on,
                                    .nonLocalPriority = dsModel::LightsStatusStatus::on,
                                    .area1 = dsModel::LightsStatusStatus::on});

      WHEN("local priority for light1 is enabled with an extra ds485 up-call") {
        REQUIRE_CALL(hal, setSceneLocalPrio(light1.getDSID(), SceneA11)).RETURN(true);
        REQUIRE_CALL(hal, setSceneLocalPrio(light2.getDSID(), SceneA11)).RETURN(false);

        REQUIRE_CALL(hal, getState(light1.getDSID()))
            .TIMES(2)
            .RETURN(hal::DeviceState{
                .status = dsModel::OutputStatusStatus::ok,
                .outputs = std::map<ChannelType, DeviceChannelValues>{{ChannelType::brightness,
                                                                       DeviceChannelValues{.value = 100.0}}},
                .localPriority = true,
            });

        REQUIRE_CALL(busInterface.action, setState(_, 0, 1, 255, 255, 255));

        proxy->setSceneLocalPrio(zone1.getID(), GroupIDYellow, SceneA11);
        proxy->completeProcessing(); // fence

        THEN("nonLocalPriority room-state is off") {
          REQUIRE(proxy->lightsStatus(zone1.id()) ==
                  dsModel::LightsStatus{.status = dsModel::LightsStatusStatus::on,
                                        .nonLocalPriority = dsModel::LightsStatusStatus::off,
                                        .area1 = dsModel::LightsStatusStatus::on});
        }
      }
    }
  }
}

} // namespace
