/*
    Copyright (c) 2018 digitalSTROM AG, Zurich, Switzerland

    This file is part of dSS.

    vdSM is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2 of the License.

    vdSM is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with dSS. If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

#include <chrono>
#include <functional>

#include "cache-db.h"
#include "device-cache.h"
#include "event.h"
#include "hal/hardware-abstraction.h"

DS_REFLECTED_ENUM((dss)(cache), CacheState, (hit)(miss))

namespace dss {

class Apartment;
class Device;
class HardwareAbstraction;

enum class ScanMode { FAST, NORMAL };

struct DeviceCacheMeta {
  int version{0};
};

class CacheUpdatePlugin : public EventInterpreterPlugin {
public:
  static constexpr const char* CacheUpdateEvent = "device_scene_cache_update";
  static constexpr const char* CacheUpdateEventFast = "device_scene_cache_update_fast";
  static constexpr const char* CacheUpdateEventResume = "device_scene_cache_update_resume";
  static constexpr const char* CacheUpdateEventSingleDevice = "device_scene_cache_update_single_device";
  static constexpr const int CacheUpdateDefaultBackoffInSec = 2;

  CacheUpdatePlugin(Apartment* apt, HardwareAbstraction* hal, ds::sqlite3::Database* db, EventInterpreter& interpreter,
                    std::shared_ptr<PropertySystem> propSystem, std::shared_ptr<EventQueue> eventQueue,
                    std::shared_ptr<EventRunner> eventRunner);
  virtual ~CacheUpdatePlugin() = default;
  virtual void subscribe();
  virtual void handleEvent(Event& _event, const EventSubscription& _subscription);

  bool isUpdateRunning() { return m_updateInProgress; }

protected:
  static constexpr int m_deviceCacheVersion = 1;
  void scanStart(ScanMode args);
  void scanDevice(const Dsuid& dsuid);
  void scanStep();
  void handleModelReady();

  DeviceCacheMeta getMetaInfo();
  void updateMetaInfo(const DeviceCacheMeta&);

private:
  void migrateFw387();
  void migrateDeviceCache();
  void reschedule();

  cache::CacheState readDevice(const Dsuid& dsuid, DeviceCache& cache);
  cache::CacheState readSceneTable(const Dsuid& dsuid, DeviceSceneCache& sceneCache);

  cache::CacheState readSceneFlags(const Dsuid& dsuid, const DeviceSceneId& adr, SceneFlags& flags);
  cache::CacheState readScene(const Dsuid& dsuid, const DeviceSceneId& adr, SceneEntry& entry);

  Apartment* m_apartment;
  HardwareAbstraction* m_hal;
  PropertyNodePtr m_scanActive;
  std::shared_ptr<PropertySystem> m_pPropSystem;
  std::shared_ptr<EventQueue> m_pEventQueue;
  std::shared_ptr<EventRunner> m_pEventRunner;
  mutable CacheDb<DeviceCacheMeta> m_metaInfoCache;

  bool m_updateInProgress{false};
  bool m_modelReady{false};
  int m_scanStepCount{};

  std::optional<Dsuid> m_currentDev{};
  std::deque<Dsuid> m_priorityQueue{};
  std::string m_currentPath{};
  DeviceCache m_currentEntry{};

  int m_hitBackoffInSec{};
};

} // namespace dss

DS_REFLECT_DECLARE_STRUCT((dss), DeviceCacheMeta, (version));
