#include "rest.h"

#include <ds/asio/promise.h>

#include <boost/algorithm/string/predicate.hpp>

DS_JSON_DEFINE_STRUCT((dsApi), HttpArgs);
DS_JSON_DEFINE_STRUCT((dsApi), HttpRequest);
DS_JSON_DEFINE_STRUCT((dsApi), HttpResult);
DS_JSON_DEFINE_STRUCT((dsApi), Error);
DS_JSON_DEFINE_STRUCT((dsApi), Document);
DS_JSON_DEFINE_STRUCT((dsApi), GetQuery);
DS_JSON_DEFINE_STRUCT((dsApi), PatchOperation);
DS_JSON_DEFINE_STRUCT((dsApi), PatchRequest);

namespace dsApi {

HttpResult AsyncCallee::blockingInvoke(ds::asio::ResourceContext& context, HttpRequest&& request) {
  auto promise = std::promise<HttpResult>{};
  auto path = request.path;
  asyncInvoke(std::move(request),
              ds::makePromise<dsApi::HttpResult>(context.io, [&](std::optional<HttpResult>&& result) {
                try {
                  DS_REQUIRE(result && result->isSuccess(), "Invoke failed.", path, result);
                  promise.set_value(*result);
                } catch (const std::exception& e) {
                  promise.set_exception(std::current_exception());
                }
              }));
  context.assertBlockingAllowed();
  return promise.get_future().get();
}

HttpResult noContentResult() {
  return {.status = 204};
}

HttpResult errorResult(int status, std::string&& message, HttpHeaders&& headers) {
  return {
      .status = status,
      .content = ds::toJsonFragment(dsApi::Document{
          .data = {}, .error = {.type = "error", .title = "error", .status = status, .detail = std::move(message)}}),
      .headers = std::move(headers)};
}

HttpResult errorResult(const std::exception& e) {
  return errorResult(500, e.what());
}

static std::string eraseTrailingSpace(std::string&& x) {
  // ds::context() tends to produce trailing space
  if (boost::ends_with(x, " ")) {
    x.erase(x.size() - 1, 1);
  }
  return std::move(x);
}

HttpResult resourceNotFoundResult() {
  return errorResult(404, eraseTrailingSpace(ds::str("Resource not found. ", ds::hideArg(ds::logContext()))));
}

HttpResult methodNotAllowedResult(HttpMethod method, const std::vector<HttpMethod>& allow) {
  auto allowHeader = [&] {
    std::string result;
    for (const auto& x : allow) {
      if (!result.empty()) {
        result += ", ";
      }
      result += ds::fromEnum<std::string>(x);
    }
    return result;
  };
  DS_CONTEXT(allow);
  return errorResult(
      405, eraseTrailingSpace(ds::str("Method not allowed. method:", method, " ", ds::hideArg(ds::logContext()))),
      HttpHeaders{{"Allow", {allowHeader()}}});
}

HttpResult notImplementedResult() {
  return errorResult(501, eraseTrailingSpace(ds::str("Not implemented. ", ds::hideArg(ds::logContext()))));
}

HttpResult cancelledResult() {
  return errorResult(500, eraseTrailingSpace(ds::str("Cancelled. ", ds::hideArg(ds::logContext()))));
}

Error error(const HttpResult& result) {
  const auto status = result.status;
  DS_REQUIRE(result.status / 100 != 2, "Failed to access error due to success http status.", status);
  return ds::fromJson<dsApi::Document>(result.content).error;
}

} // namespace dsApi
