/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "measurements-source.h"

#include <ds-apartment/ds-devices.h>
#include <ds-backend/connector.h>
#include <ds-model/model.h>
#include <ds/asio/ds-signal.h>
#include <ds/asio/promise.h>
#include <ds/asio/resource.h>
#include <event/event_create.h>
#include <model/group.h>

using namespace std::literals;
static ds::LogChannel dsLogChannel("dsBackendMeasurementsSource");

namespace dsBackend {

class MeasurementsSourceImpl : public MeasurementsSource,
                               public ds::asio::Resource<>,
                               public dss::RegisteredEventInterpreterPlugin {
public:
  MeasurementsSourceImpl(ds::asio::ResourceBase& parent, dss::EventInterpreter& eventInterpreter)
      : ds::asio::Resource<>(parent, "measurementsSource"),
        dss::RegisteredEventInterpreterPlugin("measurementsSource", eventInterpreter) {
    eventInterpreter.subscribe(*this, dss::EventName::ZoneSensorValue);
  }

private:
  ds::Signal<dsModel::Measurement> m_next;

  ds::Slots<dsModel::Measurement>& next() override { return m_next; }

  void handleEvent(dss::Event& event, const dss::EventSubscription&) override {
    DS_INFO_ENTER(event);
    handleEventImpl(event);
  }

  void handleEventImpl(const dss::Event& event) const {
    DS_DEBUG_ENTER(event);
    // Running from foreign thread(s) -> make sure to not modify this. signal emit is thread safe
    [& next = m_next, &event, &dsSystem = context().system] {
      const auto& name = event.getName();
      const auto& location = event.getRaiseLocation();
      if (name == dss::EventName::ZoneSensorValue && location == dss::erlGroup) {
        const auto sensorType = event.getPropertyByName<dss::SensorType>("sensorType");
        auto emit = [&](dsModel::MeasurementOrigin&& origin) {
          next.emit(dsModel::Measurement{
              .id = dsModel::MeasurementId{dsSystem.randomUuidString()},
              .type = "measurement",
              .lastChanged = static_cast<ds::DateTime>(event.getTimestamp()),
              .attributes = {.type = ds::toEnum<dsModel::MeasurementType>(sensorType),
                             .origin = std::move(origin),
                             .value = ds::requireDouble(event.getPropertyByName("sensorValueFloat"))}});
        };
        if (sensorType == dss::SensorType::TemperatureIndoors ||          //
            sensorType == dss::SensorType::HumidityIndoors ||             //
            sensorType == dss::SensorType::CO2Concentration ||            //
            sensorType == dss::SensorType::BrightnessIndoors ||           //
            sensorType == dss::SensorType::RoomTemperatureSetpoint ||     //
            sensorType == dss::SensorType::RoomTemperatureControlVariable //
        ) {
          emit(dsModel::MeasurementOriginZone{.type = "zone", .id = event.getRaisedAtGroup()->zoneId()});
        } else if (sensorType == dss::SensorType::TemperatureOutdoors || //
                   sensorType == dss::SensorType::HumidityOutdoors ||    //
                   sensorType == dss::SensorType::BrightnessOutdoors     //
        ) {
          emit(dsModel::MeasurementOriginOutdoor{.type = "outdoor"});
        }
      }
    }();
  }
};

std::unique_ptr<MeasurementsSource> makeMeasurementsSource(ds::asio::ResourceBase& parent,
                                                           dss::EventInterpreter& eventInterpreter) {
  return std::make_unique<MeasurementsSourceImpl>(parent, eventInterpreter);
}

} // namespace dsBackend
