/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-model.h"
#include <mutex>

#include <ds-apartment/api-revision.h>
#include <ds-apartment/ds-devices.h>
#include <ds-apartment/installation.h>
#include <ds-apartment/remote-connectivity.h>
#include <ds-api/array.h>
#include <ds-api/event.h>
#include <ds-api/service.h>
#include <ds-backend/connector.h>
#include <ds-model/api-revision.h>
#include <ds-model/model.h>
#include <ds/asio/loop.h>
#include <ds/asio/promise.h>
#include <ds/asio/state-poller.h>
#include <ds/asio/timer.h>
#include <ds/enum.h>
#include <ds/log.h>
#include <ds/system.h>
#include <ds/vector.h>
#include <model/cluster.h>
#include <model/device.h>
#include <model/dsmeter.h>
#include <model/floor.h>
#include <model/group.h>
#include <model/set.h>
#include <model/state.h>
#include <model/zone.h>

#include <boost/algorithm/string/predicate.hpp>
#include <boost/range/adaptor/transformed.hpp>
#include <boost/range/algorithm/copy.hpp>
#include <boost/range/algorithm/find.hpp>
#include <boost/range/algorithm/sort.hpp>

#include "cache/apartment-proxy.h"
#include "cache/cache-updater.h"
#include "ds-meterings.h"
#include "ds-scenarios.h"
#include "dss.h"
#include "event.h"
#include "model/shademanipulator.h"
#include "propertysystem_common_paths.h"
#include "security/security.h"
#include "security/user.h"
#include "sessionmanager.h"
#include "state-change-logger.h"
#include "structuremanipulator.h"
#include "util.h"

using namespace std::literals;
using namespace dss::model;

static ds::LogChannel dsLogChannel("dssDsModel");

std::optional<dsModel::ZoneApplicationType> DSToEnum<dsModel::ZoneApplicationType, dss::ApplicationType, void>::tryTo(
    dss::ApplicationType x) {
  switch (x) {
    case dss::ApplicationType::none: return {};
    case dss::ApplicationType::lights: return dsModel::ZoneApplicationType::lights;
    case dss::ApplicationType::blinds: return dsModel::ZoneApplicationType::shades;
    case dss::ApplicationType::heating: return dsModel::ZoneApplicationType::heating;
    case dss::ApplicationType::audio: return dsModel::ZoneApplicationType::audio;
    case dss::ApplicationType::video: return dsModel::ZoneApplicationType::video;
    case dss::ApplicationType::security: return dsModel::ZoneApplicationType::security;
    case dss::ApplicationType::access: return dsModel::ZoneApplicationType::access;
    case dss::ApplicationType::joker: return dsModel::ZoneApplicationType::joker;
    case dss::ApplicationType::cooling: return dsModel::ZoneApplicationType::cooling;
    case dss::ApplicationType::ventilation: return dsModel::ZoneApplicationType::ventilation;
    case dss::ApplicationType::window: return dsModel::ZoneApplicationType::window;
    case dss::ApplicationType::recirculation: return dsModel::ZoneApplicationType::recirculation;
    case dss::ApplicationType::temperature: return dsModel::ZoneApplicationType::temperature;
    case dss::ApplicationType::apartmentVentilation: return {};
    case dss::ApplicationType::awnings: return {};
    case dss::ApplicationType::apartmentRecirculation: return {};
  }
  return {};
}

std::optional<dsModel::ApartmentApplicationType> DSToEnum<dsModel::ApartmentApplicationType, dss::ApplicationType, void>::tryTo(
    dss::ApplicationType x) {
  switch (x) {
    case dss::ApplicationType::none: return {};
    case dss::ApplicationType::apartmentVentilation: return dsModel::ApartmentApplicationType::ventilationUnit;
    case dss::ApplicationType::lights: return {};
    case dss::ApplicationType::blinds: return {};
    case dss::ApplicationType::heating: return {};
    case dss::ApplicationType::audio: return {};
    case dss::ApplicationType::video: return {};
    case dss::ApplicationType::security: return {};
    case dss::ApplicationType::access: return {};
    case dss::ApplicationType::joker: return {};
    case dss::ApplicationType::cooling: return {};
    case dss::ApplicationType::ventilation: return {};
    case dss::ApplicationType::window: return {};
    case dss::ApplicationType::recirculation: return {};
    case dss::ApplicationType::temperature: return {};
    case dss::ApplicationType::awnings: return {};
    case dss::ApplicationType::apartmentRecirculation: return {};
  }
  return {};
}

namespace dss {

dsModel::ZoneApplicationType zoneApplication(const int& application) {
  switch (application) {
    case 0: break;  // usually the case for inactive devices. this is not an error.
    case 1: return dsModel::ZoneApplicationType::lights;
    case 2: return dsModel::ZoneApplicationType::shades;
    case 3: return dsModel::ZoneApplicationType::heating;
    case 4: return dsModel::ZoneApplicationType::audio;
    case 5: return dsModel::ZoneApplicationType::video;
    case 6: return dsModel::ZoneApplicationType::security;
    case 7: return dsModel::ZoneApplicationType::access;
    case 8: return dsModel::ZoneApplicationType::joker;
    case 9: return dsModel::ZoneApplicationType::cooling;
    case 10: return dsModel::ZoneApplicationType::ventilation;
    case 11: return dsModel::ZoneApplicationType::window;
    case 12: return dsModel::ZoneApplicationType::recirculation;
    case 48: return dsModel::ZoneApplicationType::temperature;
    default: DS_INFO("Incorrect zone application type.", application); break;
  }
  return {};
}

dsModel::ButtonType buttonType(const int& type, const int& activeGroup) {
  switch (type) {
    case 0: return dsModel::ButtonType::device;
    case 1: return dsModel::ButtonType::area1;
    case 2: return dsModel::ButtonType::area2;
    case 3: return dsModel::ButtonType::area3;
    case 4: return dsModel::ButtonType::area4;
    case 5: return dsModel::ButtonType::zone;
    case 6: return dsModel::ButtonType::zone1;
    case 7: return dsModel::ButtonType::zone2;
    case 8: return dsModel::ButtonType::zone3;
    case 9: return dsModel::ButtonType::zone4;
    case 10: return dsModel::ButtonType::zonex1;
    case 11: return dsModel::ButtonType::zonex2;
    case 12: return dsModel::ButtonType::zonex3;
    case 13: return dsModel::ButtonType::zonex4;
    case 14:
      if (activeGroup >= 64 && activeGroup <= 239) {
        return dsModel::ButtonType::application;
      }
      if (activeGroup >= 16 && activeGroup <= 39) {
        return dsModel::ButtonType::group;
      }
      break;
    case 15: return dsModel::ButtonType::appmode;
  }
  DS_INFO("Incorrect button type.", type, activeGroup);
  return {};
}

// https://intranet.aizo.net/_layouts/15/start.aspx#/Wiki/Parametersatz.aspx#BANK3 - the MODE register at offset "0"
// https://confluence.digitalstrom.org/display/DSD/Outputs
dsModel::OutputMode outputMode(const int& mode) {
  if ((mode >= 17 && mode <= 24) || (mode == 31) || (mode == 64) || (mode >= 48 && mode <= 51)) {
    return dsModel::OutputMode::gradual;
  } else if ((mode == 33) || (mode == 42)) {
    return dsModel::OutputMode::positional;
  } else if ((mode == 16) || (mode >= 25 && mode <= 27) || (mode == 32) || (mode >= 34 && mode <= 41) || (mode == 43) ||
             (mode >= 65 && mode <= 69)) {
    return dsModel::OutputMode::switched;
  } else if (mode == 0) {
    return dsModel::OutputMode::disabled;
  }
  DS_INFO("Unrecognized output mode.", mode);
  return {};
}

std::optional<int> outputLevels(const int& mode) {
  if ((mode == 16) || (mode >= 25 && mode <= 27) || (mode == 32) || (mode == 35) || (mode >= 39 && mode <= 41) ||
      (mode == 43) || (mode == 65)) {
    return 1;
  } else if ((mode == 34) || (mode == 36) || (mode == 37) || (mode == 43) || (mode == 67) || (mode == 69)) {
    return 2;
  } else if ((mode == 38) || (mode == 68)) {
    return 3;
  }
  DS_DEBUG("Output mode has no switched levels", mode);
  return {};
}

// Basic information about button input modes:
// https://intranet.aizo.net/_layouts/15/start.aspx#/Wiki/Parametersatz.aspx#BANK3 - address 0x1e
dsModel::ButtonMode buttonMode(const ButtonInputMode& mode, const ButtonUsage& usage) {
  if (usage != ButtonUsage::used) {
    return dsModel::ButtonMode::disabled;
  }
  int id = static_cast<int>(mode);
  if (id == 0 || id == 1 || id == 14 || id == 65) {
    return dsModel::ButtonMode::button1way;
  } else if (id >= 5 && id <= 13) {
    return dsModel::ButtonMode::button2way;
  } else if (id == 255) {
    return dsModel::ButtonMode::disabled;
  }
  DS_INFO("Unrecognized button mode.", mode);
  return {};
}

dsModel::ShadeAttributes shadeAttributes(Apartment& apartment, const Device& device) {
  auto shadeManipulator = makeShadeManipulator(apartment, *apartment.getDeviceBusInterface());
  return dsModel::ShadeAttributes{
      .shadeType = shadeManipulator->shadeType(device), .protectionClass = {}, .mountOrientation = {}};
}

dsModel::ZoneApplicationType submoduleApplication(const Device& device, const Zone& zone) {
  auto hardwareType = device.hardwareType();
  if (hardwareType == HardwareType::GR_KL210 || (device.isGRKL300Awning()) || (device.isBlindsButton() && zone.awningsOnly())) {
    return dsModel::ZoneApplicationType::awnings;
  } else {
    return zoneApplication(device.getActiveGroup());
  }
}

std::vector<dsModel::Sensor> getSensors(const Device& device) {
  std::vector<dsModel::Sensor> result = {};
  int sensorCount = 0;
  for (const auto& sensor : device.getSensors()) {
    const auto sensorId = dsModel::SensorId{ds::str(sensorCount)};
    DS_CONTEXT(sensorId);
    try {
      result.push_back({.id = sensorId,
                        .attributes = {
                            .name = "",
                            .technicalName = ds::str("Sensor#", sensorCount),
                            .type = ds::toEnum<dsModel::MeasurementType>(sensor->m_sensorType),
                            .usage = ds::toEnum<dsModel::MeasurementUsage>((int)sensor->m_sensorUsage),
                            .min = {},
                            .max = {},
                            .resolution = {}, // TODO: Find detailed information about those values
                        }});
    } catch (const std::exception& e) {
      DS_ERROR("Sensor failed.", e);
      result.push_back({.id = sensorId});
    }
    sensorCount++;
  };
  dsApi::normalizeIdArray(result);
  return result;
}

std::vector<dsModel::Button> getButtons(const Device& device) {
  std::vector<dsModel::Button> result = {};
  if (device.getButtonID() != 255 && device.getButtonInputCount() != 0) {
    const auto buttonId = dsModel::ButtonId{ds::str("button", device.getButtonInputIndex())};
    DS_CONTEXT(buttonId);
    try {
      result.push_back({.id = buttonId,
                        .attributes = {
                            .name = "",
                            .technicalName = ds::str("Pushbutton#0"),
                            .type = buttonType(device.getButtonID(), device.getActiveGroup()),
                            .mode = buttonMode(device.getButtonInputMode(), device.getButtonUsage()),
                            .channel = {},
                        }});
    } catch (const std::exception& e) {
      DS_ERROR("Button failed.", e);
      result.push_back({.id = buttonId});
    }
  }
  dsApi::normalizeIdArray(result);
  return result;
}

std::vector<dsModel::Output> getOutputs(const Device& device, std::vector<double>&& levelThresholds) {
  auto result = std::vector<dsModel::Output>{};
  auto channels = device.getOutputChannels();
  for (const auto& channel : channels) {
    const auto& outputId = dsModel::OutputId{channel->id};
    DS_CONTEXT(outputId);
    try {
      result.push_back({.id = outputId,
                        .attributes = {
                            .name = "",
                            .technicalName = channel->technicalName,
                            .type = ds::toEnum<dsModel::OutputType>(channel->type),
                            .function = {},
                            .min = [&]() -> double {
                              if (channel->type == ChannelType::colortemp) {
                                return 100;
                              } else {
                                return 0;
                              }
                            }(),
                            .max = [&]() -> double {
                              if (channel->type == ChannelType::hue) {
                                return 360;
                              } else if (channel->type == ChannelType::colortemp) {
                                return 1000;
                              } else if (channel->type == ChannelType::x || channel->type == ChannelType::y) {
                                return 1;
                              } else if (channel->type == ChannelType::waterTemperature) {
                                return 150;
                              } else {
                                return 100;
                              }
                            }(),
                            .resolution =
                                [&] {
                                  if (channel->type == ChannelType::colortemp) {
                                    return 1.0;
                                  } else if (channel->type == ChannelType::x || channel->type == ChannelType::y) {
                                    return 0.01;
                                  } else {
                                    return 0.1;
                                  }
                                }(),
                            .mode = outputMode(device.getOutputMode()),
                            .levels = outputLevels(device.getOutputMode()),
                            .levelThresholds = std::move(levelThresholds),
                        }});
    } catch (const std::exception& e) {
      DS_ERROR("Output failed.", e);
      result.push_back({.id = outputId});
    }
  };
  dsApi::normalizeIdArray(result);
  return result;
}

DecodedApplicationId decodeId(const dsModel::ApplicationId& applicationId) {
  auto result = DecodedApplicationId{};
  std::stringstream ss(applicationId);

  std::string applicationType;
  std::getline(ss, applicationType, '-');
  if (applicationType != "temperature") {
    DS_FAIL_REQUIRE("Unsupported application type.", applicationType);
  }

  std::string zoneId;
  std::getline(ss, zoneId);
  DS_REQUIRE(boost::starts_with(zoneId, "z"), "Wrong applicationId format.", applicationId);
  zoneId = zoneId.substr(1);
  result.zoneId = strToInt(zoneId);

  return result;
}

static std::string heatingLevelName(const dsModel::TemperatureLevel& level) {
  switch (level) {
    case dsModel::TemperatureLevel::heatingComfort: return "Comfort";
    case dsModel::TemperatureLevel::heatingEconomy: return "Economy";
    case dsModel::TemperatureLevel::heatingSetback: return "NotUsed";
    case dsModel::TemperatureLevel::heatingNight: return "Night";
    case dsModel::TemperatureLevel::heatingVacation: return "Holiday";
    case dsModel::TemperatureLevel::coolingComfort: return "Cooling";
    case dsModel::TemperatureLevel::coolingEconomy: return "CoolingEconomy";
    case dsModel::TemperatureLevel::coolingSetback: return "CoolingNotUsed";
    case dsModel::TemperatureLevel::coolingNight: return "CoolingNight";
    case dsModel::TemperatureLevel::coolingVacation: return "CoolingHoliday";
    case dsModel::TemperatureLevel::none:
    case dsModel::TemperatureLevel::heatingOff:
    case dsModel::TemperatureLevel::coolingOff:
    case dsModel::TemperatureLevel::passiveCoolingThreshold: break;
  }
  DS_FAIL_REQUIRE("Unrecognized temperature level", level);
}

static int heatingOperationMode(const dsModel::TemperatureLevel& level) {
  switch (level) {
    case dsModel::TemperatureLevel::none: break;
    case dsModel::TemperatureLevel::heatingOff: return HeatingOperationModeIDOff;
    case dsModel::TemperatureLevel::heatingComfort: return HeatingOperationModeIDComfort;
    case dsModel::TemperatureLevel::heatingEconomy: return HeatingOperationModeIDEconomy;
    case dsModel::TemperatureLevel::heatingSetback: return HeatingOperationModeIDNotUsed;
    case dsModel::TemperatureLevel::heatingNight: return HeatingOperationModeIDNight;
    case dsModel::TemperatureLevel::heatingVacation: return HeatingOperationModeIDHoliday;
    case dsModel::TemperatureLevel::coolingOff: return HeatingOperationModeIDCoolOff;
    case dsModel::TemperatureLevel::coolingComfort: return HeatingOperationModeIDCoolComfort;
    case dsModel::TemperatureLevel::coolingEconomy: return HeatingOperationModeIDCoolEconomy;
    case dsModel::TemperatureLevel::coolingSetback: return HeatingOperationModeIDCoolNotUsed;
    case dsModel::TemperatureLevel::coolingNight: return HeatingOperationModeIDCoolNight;
    case dsModel::TemperatureLevel::coolingVacation: return HeatingOperationModeIDCoolHoliday;
    case dsModel::TemperatureLevel::passiveCoolingThreshold: break;
  }
  DS_FAIL_REQUIRE("Unrecognized temperature level", level);
}

namespace {

class DsModelImpl final : public DsModel,
                          public RegisteredEventInterpreterPlugin,
                          public PropertyListener,
                          ds::asio::Resource<> {
  dsBackend::Connector& m_bus;
  Apartment& m_apartment;
  dsApi::EventHub& m_eventHub;
  ds::asio::ResourceContext& m_apartmentContext;
  dsApartment::ApiRevision& m_apiRevision;

  DsModelHooks m_hooks;
  ds::Slot<bool> DS_UNIQUE_NAME(slot){*this, m_bus.isConnectedChanged(),
                                      [this](bool x) { onBusIsConnectedChanged(x); }};
  ds::Slot<> DS_UNIQUE_NAME(slot){*this, m_apartment.structureChanged(), [this]() { pollApartmentStructureImpl(); }};
  ds::Slot<std::shared_ptr<ModelSnapshot>> DS_UNIQUE_NAME(slot){*this, m_apartment.modelStructureChanged(),
                                                                [this](auto arg) { pollApartmentStructureImpl(arg); }};
  ds::Slot<> DS_UNIQUE_NAME(slot){*this, m_apartment.statusChanged(), [this]() { pollApartmentStatusImpl(); }};
  PropertySystem& m_propertySystem = *m_apartment.getPropertySystem();
  PropertyNodePtr m_mshubActiveNode{m_propertySystem.getProperty(pp_websvc_mshub_active)};
  PropertyNodePtr m_apartmentNode{m_propertySystem.createProperty("/apartment")};
  PropertyNodePtr m_udasNode{m_propertySystem.createProperty("/scripts/system-addon-user-defined-actions/")};
  PropertyNodePtr m_udsNode{m_propertySystem.createProperty("/usr/addon-states/system-addon-user-defined-states/")};
  PropertyNodePtr m_statesNode{m_propertySystem.createProperty("/usr/states/")};
  PropertyNodePtr m_configNode{m_propertySystem.getProperty("/config")};
  PropertyNodePtr m_heatingGeneralSettingsNode{
      m_propertySystem.createProperty("/scripts/system-addon-heating-controller/generalSettings")};
  PropertyNodePtr m_heatingControllerConfigNode{
      m_propertySystem.createProperty("/scripts/system-addon-heating-controller/controllerConfig")};
  PropertyNodePtr m_heatingModesNode{m_propertySystem.createProperty("/scripts/system-addon-heating-controller/modes")};
  PropertyNodePtr m_heatingStatesNode{
      m_propertySystem.createProperty("/usr/addon-states/system-addon-heating-controller")};
  PropertyNodePtr m_systemNode{m_propertySystem.createProperty("/system")};
  mutable std::mutex m_structureMutex;
  ds::JsonFragment m_structure;
  mutable std::mutex m_statusMutex;
  ds::JsonFragment m_status;

  // event loop for with apartment level latency. io loop has higher latency requirements.
  StateChangeLogger m_structureChangeLogger{m_apartmentContext.io, "structure"};
  std::unique_ptr<ds::asio::StatePoller> m_apartmentPoller = ds::asio::invoke(m_apartmentContext.io, [&] {
    return std::make_unique<ds::asio::StatePoller>(
        m_apartmentContext.io, "apartmentStructurePoller",
        conf().readOr<ds::asio::StatePollerConfig>("apartmentStructurePoller", {.moderate = 10s, .loop = 1h}),
        ds::asio::StatePollerHooks{
            .poll = [this] { 
              return toJsonFragmentWithEmptyIncluded(apartment({.include = {}, .includeAll = true}));
            },
            .fire =
                [this](auto&& x) {
                  std::lock_guard<std::mutex> lock(m_structureMutex);
                  m_structure = std::move(x);
                  onStructurePollerFired();
                  pollApartmentStatusImpl();
                }});
  });
  std::unique_ptr<ds::asio::StatePoller> m_apartmentStatusPoller = ds::asio::invoke(m_apartmentContext.io, [&] {
    return std::make_unique<ds::asio::StatePoller>(
        m_apartmentContext.io, "apartmentStatusPoller",
        conf().readOr<ds::asio::StatePollerConfig>("apartmentStatusPoller", {.moderate = 1s, .loop = 1min}),
        ds::asio::StatePollerHooks{.poll =
                                       [this] {
                                         return toJsonFragmentWithEmptyIncluded(
                                            apartmentStatus({.include = {}, .includeAll = true}));
                                       },
                                   .fire =
                                       [this](auto&& x) {
                                         std::lock_guard<std::mutex> lock(m_statusMutex);
                                         m_status = std::move(x);
                                         onStatusPollerFired();
                                       }});
  });
  std::unique_ptr<dsModel::ScenariosApi> m_scenarios{makeDsScenarios(*this, m_apartment)};
  std::unique_ptr<dsModel::MeteringsApi> m_meterings{makeDsMeterings(*this, m_apartment, m_eventHub)};
  std::unique_ptr<dsApartment::Installation> m_installation{dsApartment::makeInstallation(*this, m_apartment, m_bus)};
  std::unique_ptr<SimpleHttpClient> m_httpClient = makeSimpleHttpClient();
  std::unique_ptr<dsModel::RemoteConnectivityApi> m_remoteConnectivity{
      dsApartment::makeRemoteConnectivity(*this, m_apartment, *m_httpClient, m_bus)};

public:
  DsModelImpl(ds::asio::ResourceBase& parent, dsBackend::Connector& bus, Apartment& apartment,
              EventInterpreter& eventInterpreter, dsApi::EventHub& eventHub,
              ds::asio::ResourceContext& apartmentContext, dsApartment::ApiRevision& apiRevision, DsModelHooks&& hooks)
      : RegisteredEventInterpreterPlugin("DSModel", eventInterpreter),
        ds::asio::Resource<>{parent, "dsModel"},
        m_bus{bus},
        m_apartment{apartment},
        m_eventHub{eventHub},
        m_apartmentContext{apartmentContext},
        m_apiRevision{apiRevision},
        m_hooks{std::move(hooks)} {
    DS_INFO_ENTER("");

    DS_REQUIRE(m_configNode, "Missing config property node");

    eventInterpreter.subscribe(*this, EventName::RetryRegistration);
    eventInterpreter.subscribe(*this, EventName::ApartmentProxyStateChanged);
    eventInterpreter.subscribe(*this, EventName::StateChange);
    eventInterpreter.subscribe(*this, EventName::AddonStateChange);
    eventInterpreter.subscribe(*this, EventName::StopMaintenance);

    if (m_mshubActiveNode) {
      m_mshubActiveNode->addListener(this);
    }
    m_udsNode->addListener(this);
    m_heatingGeneralSettingsNode->addListener(this);
    m_heatingControllerConfigNode->addListener(this);
    m_heatingModesNode->addListener(this);
    m_heatingStatesNode->addListener(this);
    updateBusActive();

    m_apartmentPoller->setActive(m_hooks.pollersActive);
    m_apartmentStatusPoller->setActive(m_hooks.pollersActive);
  }

  ~DsModelImpl() override { DS_INFO_ENTER(""); }

  [[nodiscard]] bool isApartmentIdMocked() const { return !conf().readOr<std::string>("apartmentId", {}).empty(); }

  void onStructurePollerFired() {
    onPollerFired(m_structure, "", dsModel::apartmentStructureChanged);
    m_structureChangeLogger.log(ds::JsonFragment(m_structure));
  }

  void onStatusPollerFired() { onPollerFired(m_status, "/status", dsModel::apartmentStatusChanged); }

  void onPollerFired(const ds::JsonFragment& value, std::string&& path, const dsApi::EventType& eventType) {
    DS_INFO_ENTER(eventType);

    auto emitEvent = [this, eventType]() -> void {
      // hint for signalR clients
      auto event = dsApi::Event{.type = eventType};
      DS_NOTICE("Publish event.", event);
      m_eventHub.emitEvent(std::move(event));
      DS_INFO("Publish event.", event, "done");
    };

    if (m_bus.isConnected()) { // payload for backend apartment cache
      auto request = dsApi::HttpRequest{.path = ds::str("/backend/v1/apartment", path),
                                        .httpArgs = {.method = dsApi::HttpMethod::PUT,
                                                     .content = ds::toJsonFragment(dsApi::Document{.data = value})}};
      DS_INFO("Publish", request); // too verbose for notice
      m_bus.asyncInvoke(std::move(request),
                        ds::makePromise<dsApi::HttpResult>(
                            m_apartmentContext.io, [path, emitEvent](std::optional<dsApi::HttpResult>&& result) {
                              DS_INFO_ENTER(path, result);
                              if (!result || !result->isSuccess()) {
                                DS_ERROR("Failed to put to backend cache", path, result);
                              } else {
                                emitEvent();
                              }
                            }));
    } else {
      DS_INFO("Not publishing apartment status. Not connected.");
      emitEvent();
    }

    DS_INFO_LEAVE(eventType);
    // we rely on poller to recover from exceptions
  }

  void pollApartmentStructureImpl(std::shared_ptr<ModelSnapshot> modelSnapshot) {
    DS_DEBUG_ENTER("");
    pollApartmentStructureImpl();
    io().dispatch([this, snapshot = std::move(modelSnapshot)] { m_apartmentPoller->poll(); });
  }

  void pollApartmentStructureImpl() {
    DS_DEBUG_ENTER("");
    m_hooks.apartmentStructurePolled();
    io().dispatch([this] { m_apartmentPoller->poll(); });
  }

  void pollApartmentStatusImpl() {
    DS_DEBUG_ENTER("");
    m_hooks.apartmentStatusPolled();
    io().dispatch([this] { m_apartmentStatusPoller->poll(); });
  }

  void onBusIsConnectedChanged(bool isBusConnected) {
    DS_INFO_ENTER(isBusConnected);
    onStructurePollerFired();
    onStatusPollerFired();
  }

  void updateBusActive() {
    m_bus.setActive(isApartmentIdMocked() || (m_mshubActiveNode && m_mshubActiveNode->getBoolValue()));
  }

  void handleEvent(Event& event, const EventSubscription&) override {
    DS_INFO_ENTER(event);
    handleEvent(event);
  }

  void handleEvent(const Event& event) override {
    DS_DEBUG_ENTER(event);
    const auto& eventName = event.getName();
    if (eventName == EventName::RetryRegistration) {
      pollApartmentStructureImpl();
    }
    if (eventName == EventName::ApartmentProxyStateChanged) {
      pollApartmentStatusImpl();
    }
    if (eventName == EventName::StateChange) { // many states lead to changes in status.
      pollApartmentStatusImpl();
    }

    if (eventName == EventName::AddonStateChange) {
      const auto& scriptId = event.getPropertyByName("scriptID");
      if (scriptId == "system-addon-user-defined-states") {
        pollApartmentStatusImpl();
      }
    }
    if (eventName == EventName::StopMaintenance) {
      dsModel::MaintenanceData data{};
      data.data.attributes.enabled = false;
      m_installation->toggleMaintenance(data);
    }
  }

  void propertyChanged(PropertyNodePtr node) {
    DS_DEBUG_ENTER(node->getName());
    if (node == m_mshubActiveNode) {
      updateBusActive();
    }
    if (node->isOnPathToRoot(*m_heatingControllerConfigNode) || node->isOnPathToRoot(*m_heatingModesNode) ||
        node->isOnPathToRoot(*m_udsNode)) {
      pollApartmentStructureImpl();
    }
    if (node->isOnPathToRoot(*m_heatingControllerConfigNode) || node->isOnPathToRoot(*m_heatingGeneralSettingsNode) ||
        node->isOnPathToRoot(*m_heatingStatesNode)) {
      pollApartmentStatusImpl();
    }
  }

  void propertyChanged(PropertyNodePtr parent, PropertyNodePtr node) override { propertyChanged(std::move(node)); }

  void propertyRemoved(PropertyNodePtr parent, PropertyNodePtr node) override { propertyChanged(std::move(parent)); }

  void propertyAdded(PropertyNodePtr parent, PropertyNodePtr node) override { propertyChanged(std::move(parent)); }

  dsModel::ScenariosApi& scenarios() override { return *m_scenarios; }
  [[nodiscard]] const dsModel::ScenariosApi& scenarios() const override { return *m_scenarios; }

  dsModel::MeteringsApi& meterings() override { return *m_meterings; }
  [[nodiscard]] const dsModel::MeteringsApi& meterings() const override { return *m_meterings; }

  [[nodiscard]] dsModel::Zone zone(const dsModel::ZoneId& zoneId) const override {
    DS_CONTEXT(zoneId);
    auto zonePtr = m_apartment.zone(zoneId);
    auto& zone = *zonePtr;
    try {
      return dsModel::Zone{
          .id = zoneId,
          .type = "zone",
          .lastChanged = {},
          .attributes = {
              .name = unescapeHTML(zone.getName()),
              .floor =
                  [&] {
                    if (zone.getFloorID() < 0) {
                      return dsModel::FloorId{};
                    }
                    return dsModel::FloorId{ds::str(zone.getFloorID())};
                  }(),
              .orderId = 0,
              .submodules = zone.submoduleIds(),
              .applications =
                  m_apiRevision.removals() < 1 ? zone.cachedApplications() : std::vector<dsModel::ZoneApplicationType>{},
              .applications2 = zoneApplications(zoneId),
              .applicationTypes = zone.cachedApplications(),
              .applicationDetails = zone.getZoneApplicationAttributes(),
              .scenarios = {}, // TODO(someday): next story
          }};
    } catch (const std::exception& e) {
      DS_ERROR("Zone failed.", e);
      return dsModel::Zone{.id = zoneId};
    }
  }

  [[nodiscard]] dsModel::Area zoneApplicationAreaAttributes(const dsModel::ZoneId& zoneId,
                                                            const std::string& zoneApplicationType,
                                                            const std::string& areaId) const override {
    DS_CONTEXT(zoneId, zoneApplicationType, areaId);
    auto zonePtr = m_apartment.zone(zoneId);
    auto& zone = *zonePtr;
    try {
      return zone.applicationAreaAttributes(ds::toEnum<dsModel::ZoneApplicationType>(zoneApplicationType), areaId);
    } catch (const std::exception& e) {
      DS_ERROR("Zone failed.", e);
      return dsModel::Area{.id = areaId};
    }
  }

  [[nodiscard]] std::vector<dsModel::Zone> zones() const override {
    return ds::makeVector(m_apartment.zoneIds() | boost::adaptors::transformed([&](const auto& x) { return zone(x); }));
  }

  void setZoneName(dsModel::ZoneArgs&& args, const std::string& value) override {
    DS_NOTICE("Set zone name", args, value);
    DS_REQUIRE(!value.empty(), "New name cannot be empty.");
    auto zone = m_apartment.zone(args.zoneId);
    DS_REQUIRE(zone->getID() != 0, "Zone not found.", args);
    DS_REQUIRE(zone->getID() != NEW_DEVICES_ZONE_ID, "New devices zone cannot be renamed.", args);
    zone->setName(escapeHTML(value));
  }

  void setAreaName(dsModel::ZoneArgs&& args, dsModel::ZoneApplicationType zoneApplicationType, const std::string areaId, const std::string& value) override {
    DS_NOTICE("Set area name", args, value);
    DS_REQUIRE(!value.empty(), "New name cannot be empty.");
    auto zone = m_apartment.zone(args.zoneId);
    DS_REQUIRE(zone->getID() != 0, "Zone not found.", args);
    DS_REQUIRE(zone->getID() != NEW_DEVICES_ZONE_ID, "New devices zone cannot be renamed.", args);
    zone->setAreaName(zoneApplicationType, areaId, value);
  }

  [[nodiscard]] dsModel::Cluster cluster(const dsModel::ClusterId& clusterId) const override {
    DS_CONTEXT(clusterId);
    auto clusterPtr = m_apartment.cluster(clusterId);
    auto& cluster = *clusterPtr;
    try {
      return dsModel::Cluster{.id = clusterId,
                              .type = "cluster",
                              .lastChanged = {},
                              .attributes = {.name = unescapeHTML(cluster.getName()),
                                             .orderId = {},
                                             .applicationType = cluster.applicationType(),
                                             .submodules = {}}};
    } catch (const std::exception& e) {
      DS_ERROR("Cluster failed.", e);
      return dsModel::Cluster{.id = clusterId};
    }
  }

  [[nodiscard]] dsModel::ClusterStatus clusterStatus(const dsModel::ClusterId& clusterId) const override {
    DS_CONTEXT(clusterId);
    try {
      auto wind = m_apartment.tryGetNonScriptState("wind.group" + ds::str(clusterId));
      auto hail = m_apartment.tryGetNonScriptState("hail.group" + ds::str(clusterId));
      std::optional<bool> windState;
      std::optional<bool> hailState;
      if (wind) {
        windState = wind->getState() == State_Active;
      }
      if (hail) {
        hailState = hail->getState() == State_Active;
      }
      return dsModel::ClusterStatus{
          .id = clusterId,
          .type = "clusterStatus",
          .lastChanged = {},
          .attributes = {.weather = {.wind = windState, .hail = hailState},
                         .operationsLocked = m_apartment.cluster(clusterId)->isOperationLock()}};
    } catch (const std::exception& e) {
      DS_ERROR("Cluster status failed.", e);
      return dsModel::ClusterStatus{.id = clusterId};
    }
  }

  [[nodiscard]] std::vector<dsModel::Cluster> clusters() const override {
    return ds::makeVector(m_apartment.clusterIds() |
                          boost::adaptors::transformed([&](const auto& x) { return cluster(x); }));
  }

  [[nodiscard]] std::vector<dsModel::ClusterStatus> clusterStatuses() const override {
    return ds::makeVector(m_apartment.clusterIds() |
                          boost::adaptors::transformed([&](const auto& x) { return clusterStatus(x); }));
  }

  void setClusterName(const dsModel::ClusterArgs& args, std::string&& name) override {
    DS_NOTICE("Set cluster name.", args, name);
    DS_REQUIRE(!name.empty(), "New name cannot be empty.");
    m_apartment.cluster(args.clusterId)->setName(escapeHTML(std::move(name)));
  }

  [[nodiscard]] dsModel::DSDevice device(const dsModel::DSDeviceId& deviceId) const override {
    DS_CONTEXT(deviceId);
    auto devicePtr = m_apartment.getDeviceByDSID(Dsuid{deviceId});
    auto& device = *devicePtr;
    try {
      return dsModel::DSDevice{.id = deviceId,
                               .type = "dsDevice",
                               .lastChanged = {},
                               .attributes = {
                                   .name = unescapeHTML(device.getName()),
                                   .model = {},
                                   .gtin =
                                       [&] {
                                         if (device.getOemInfoState() == DEVICE_OEM_VALID) {
                                           return device.getOemEanAsString();
                                         }
                                         return std::string();
                                       }(),
                                   .dsid =
                                       [&] {
                                         dsid_t dsid;
                                         std::string ret = "";
                                         if (dsuid_to_dsid(device.getDSID(), &dsid)) {
                                           ret = dsid2str(dsid);
                                         }
                                         return ret;
                                       }(),
                                   .displayId = device.getDisplayID(),
                                   .present = device.isPresent(),
                                   .firstSeen = {},
                                   .lastDiscovered = {},
                                   .inactiveSince = {},
                                   .shadeAttributes = shadeAttributes(m_apartment, device),
                                   .submodules = {device.submoduleId()},
                                   .zone = device.zoneId(),
                                   .scenarios = device.scenarioIds(),
                                   .states = device.states(),
                                   .controller = dsModel::ControllerId(dsuid2str(device.getDSMeterDSID()))
                               }};
    } catch (const std::exception& e) {
      DS_ERROR("Device failed.", e);
      return dsModel::DSDevice{.id = deviceId};
    }
  }

  [[nodiscard]] std::vector<dsModel::DSDevice> devices() const override {
    return ds::makeVector(m_apartment.deviceIds() |
                          boost::adaptors::transformed([&](const auto& x) { return device(x); }));
  }

  [[nodiscard]] dsModel::DSDeviceStatus deviceStatus(const dsModel::DSDeviceId& deviceId) const override {
    DS_CONTEXT(deviceId);
    try {
      auto result = m_apartment.proxy().dsDeviceStatus(deviceId);
      result.attributes.states = m_apartment.device(deviceId)->stateStatuses();
      return result;
    } catch (const std::exception& e) {
      DS_ERROR("Device status failed.", e);
      return dsModel::DSDeviceStatus{.id = deviceId};
    }
  }

  [[nodiscard]] std::vector<dsModel::DSDeviceStatus> deviceStatuses() const override {
    return ds::makeVector(m_apartment.deviceIds() |
                          boost::adaptors::transformed([&](auto&& x) { return deviceStatus(x); }));
  }

  void setOutputValue(dsModel::OutputStatusArgs&& args, double value) override {
    auto& device = *m_apartment.getDeviceByDSID(Dsuid{args.deviceId});
    auto& output = *device.getOutputChannelById(args.outputId);
    const auto size = getOutputChannelIdAndSize(output.id).second;
    const auto convertedValue = convertToOutputChannelValue(output.type, value);
    DS_NOTICE("Set output value", args, value, size, convertedValue);
    device.setDeviceOutputChannelValue(output.index, size, convertedValue);
  }

  void setShadeAttributes(const dsModel::DSDeviceId& deviceId, dsModel::ShadeAttributes&& shadeAttributes) override {
    DS_NOTICE("Set shade attributes", deviceId, shadeAttributes);

    if (shadeAttributes.shadeType != dsModel::ShadeType{}) {
      // it's not possible to set it back to ShadeType::none, problem?
      auto& device = *m_apartment.getDeviceByDSID(Dsuid{deviceId});
      auto shadeManipulator = makeShadeManipulator(m_apartment, *m_apartment.getDeviceBusInterface());
      shadeManipulator->setShadeType(device, shadeAttributes.shadeType);
    }
  }

  [[nodiscard]] dsModel::Submodule submodule(const dsModel::SubmoduleId& submoduleId) const override {
    DS_CONTEXT(submoduleId);
    auto devicePtr = m_apartment.getDeviceByDSID(Dsuid{submoduleId});
    auto& device = *devicePtr;
    auto& zone = *m_apartment.getZone(device.getZoneID());
    try {
      return dsModel::Submodule{.id = submoduleId,
                                .type = "submodule",
                                .lastChanged = {},
                                .attributes = {
                                    .name = unescapeHTML(device.getName()),
                                    .technicalName = device.getHWInfo(),
                                    .dsDevice = device.id(),
                                    .functionBlocks = {device.functionBlockId()},
                                    .zone = device.zoneId(),
                                    .application = submoduleApplication(device, zone),
                                }};
    } catch (const std::exception& e) {
      DS_ERROR("Submodule failed.", e);
      return dsModel::Submodule{.id = submoduleId};
    }
  }

  [[nodiscard]] std::vector<dsModel::Submodule> submodules() const override {
    return ds::makeVector(m_apartment.submoduleIds() |
                          boost::adaptors::transformed([&](const auto& x) { return submodule(x); }));
  }

  [[nodiscard]] dsModel::FunctionBlock functionBlock(const dsModel::FunctionBlockId& functionBlockId) const override {
    DS_CONTEXT(functionBlockId);
    auto devicePtr = m_apartment.getDeviceByDSID(Dsuid{functionBlockId});
    auto& device = *devicePtr;
    try {
      return dsModel::FunctionBlock{
          .id = functionBlockId,
          .type = "functionBlock",
          .lastChanged = {},
          .attributes = {
              .name = unescapeHTML(device.getName()),
              .technicalName = device.getHWInfo(),
              .active = true,
              .outputs = getOutputs(device, m_apartment.proxy().getDeviceThresholds(Dsuid{functionBlockId})),
              .buttonInputs = getButtons(device),
              .sensorInputs = getSensors(device),
              .submodule = device.submoduleId(),
              .deviceAdapter = dsModel::DeviceAdapterId{ds::str(device.getDSID())},
          }};
    } catch (const std::exception& e) {
      DS_ERROR("Function block failed.", e);
      return dsModel::FunctionBlock{.id = functionBlockId};
    }
  }

  [[nodiscard]] std::vector<dsModel::FunctionBlock> functionBlocks() const override {
    return ds::makeVector(m_apartment.functionBlockIds() |
                          boost::adaptors::transformed([&](const auto& x) { return functionBlock(x); }));
  }

  [[nodiscard]] dsModel::ApiRevision apiRevision() const override {
    DS_CONTEXT("apiRevision");
    try {
      return dsModel::ApiRevision{
          .id = dsModel::ApiRevisionId{"apiRevisionId"},
          .type = "apiRevision",
          .lastChanged = {},
          .attributes = {.additions = m_apiRevision.additions(), .removals = m_apiRevision.removals()}};
    } catch (const std::exception& e) {
      DS_ERROR("Api revision failed.", e);
      return dsModel::ApiRevision{.id = dsModel::ApiRevisionId{"apiRevisionId"}};
    }
  }

  [[nodiscard]] std::optional<bool> stateStatus(const State& state) const {
    if (state.getState() == State_Active) {
      return true;
    } else if (state.getState() == State_Inactive) {
      return false;
    }
    return {};
  }

  [[nodiscard]] std::optional<bool> systemStateStatus(std::string name) const {
    auto state = m_apartment.tryGetNonScriptState(name);
    if (state) {
      if (name == dss::StateName::Presence) {
        if (state->getEnumState<dss::PresenceStates>() == dss::PresenceStates::present) {
          return false;
        } else if (state->getEnumState<dss::PresenceStates>() == dss::PresenceStates::absent) {
          return true;
        }
      } else if (name == dss::StateName::Hibernation) {
        if (state->getEnumState<dss::HibernationStates>() == dss::HibernationStates::sleeping) {
          return true;
        } else if (state->getEnumState<dss::HibernationStates>() == dss::HibernationStates::awake) {
          return false;
        }
      } else {
        return stateStatus(*state);
      }
    }
    // Some states are populated lazily. We cannot log error uncoditionally.
    DS_DEBUG("Could not get state status", name);
    return {};
  }

  [[nodiscard]] dsModel::ApartmentAccessStatus apartmentAccessStatus() const override {
    dsModel::ApartmentAccessStatus accessStatus{};

    accessStatus.absent = systemStateStatus(dss::StateName::Presence);

    return accessStatus;
  }

  [[nodiscard]] dsModel::ApartmentUserStatus apartmentUserStatus() const override {
    dsModel::ApartmentUserStatus userStatus{};

    userStatus.sleeping = systemStateStatus(dss::StateName::Hibernation);

    return userStatus;
  }

  [[nodiscard]] dsModel::ApartmentSecurity apartmentSecurity() const override {
    auto group00 = m_apartment.getZone(0)->getGroup(0);
    return {
        .alarm1 = {.name = group00->getSceneName(SceneAlarm)},
        .alarm2 = {.name = group00->getSceneName(SceneAlarm2)},
        .alarm3 = {.name = group00->getSceneName(SceneAlarm3)},
        .alarm4 = {.name = group00->getSceneName(SceneAlarm4)},
    };
  }

  void setApartmentAlarmName(int scene, std::string&& value) {
    m_apartment.getZone(0)->getGroup(0)->setSceneName(scene, std::move(value));
  }

  void setApartmentAlarm1Name(std::string&& value) override {
    DS_NOTICE("Set apartment alarm1 name", value);
    setApartmentAlarmName(SceneAlarm, std::move(value));
  }

  void setApartmentAlarm2Name(std::string&& value) override {
    DS_NOTICE("Set apartment alarm2 name", value);
    setApartmentAlarmName(SceneAlarm2, std::move(value));
  }

  void setApartmentAlarm3Name(std::string&& value) override {
    DS_NOTICE("Set apartment alarm3 name", value);
    setApartmentAlarmName(SceneAlarm3, std::move(value));
  }

  void setApartmentAlarm4Name(std::string&& value) override {
    DS_NOTICE("Set apartment alarm4 name", value);
    setApartmentAlarmName(SceneAlarm4, std::move(value));
  }

  [[nodiscard]] dsModel::ApartmentSecurityStatus apartmentSecurityStatus() const override {
    dsModel::ApartmentSecurityStatus securityStatus{};

    securityStatus.panic = systemStateStatus(dss::StateName::Panic);
    securityStatus.fire = systemStateStatus(dss::StateName::Fire);
    securityStatus.alarm1 = systemStateStatus(dss::StateName::Alarm);
    securityStatus.alarm2 = systemStateStatus(dss::StateName::Alarm2);
    securityStatus.alarm3 = systemStateStatus(dss::StateName::Alarm3);
    securityStatus.alarm4 = systemStateStatus(dss::StateName::Alarm4);

    return securityStatus;
  }

  [[nodiscard]] dsModel::ApartmentTemperatureStatus apartmentTemperatureStatus() const override {
    dsModel::ApartmentTemperatureStatus temperatureStatus{};

    auto mode = [&]() -> dsModel::TemperatureMode {
      const auto& heatingEnabled = m_heatingGeneralSettingsNode->tryChildValue<bool>("heatingEnabled").value_or(false);
      if (!heatingEnabled) {
        return dsModel::TemperatureMode::off;
      }
      const auto& heatingForced =
          m_heatingGeneralSettingsNode->tryChildValue<bool>("overrideSystemModeAsHeating").value_or(false);
      const auto& coolingForced =
          m_heatingGeneralSettingsNode->tryChildValue<bool>("overrideSystemModeAsCooling").value_or(false);

      if(heatingForced) {
        return dsModel::TemperatureMode::heating;
      } else if (coolingForced) {
        return dsModel::TemperatureMode::cooling;
      } else {
        return dsModel::TemperatureMode::automatic;
      }
    };

    temperatureStatus.mode = mode();
    temperatureStatus.ecoPreferred =
        m_heatingStatesNode->tryChildValue<int>(ds::str("operationmode.0.Eco/value")) == 1 ? true : false;

    return temperatureStatus;
  }

  void setApartmentTemperatureStatusEcoPreferred(const dsModel::ApartmentArgs& args, bool value) override {
    DS_NOTICE("Set apartment temperature status eco preferred", args, value);
    m_apartment.pushEvent(
        std::make_shared<Event>("heating-controller.operation-mode", Properties{
                                                                           {"actions", "setOperationMode"},
                                                                           {"zoneID", "0"},
                                                                           {"operationMode", value ? "2" : "1"},
                                                                       }));
  }

  [[nodiscard]] dsModel::ApartmentWeatherStatus apartmentWeatherStatus() const override {
    const auto sensorStatus = m_apartment.getSensorStatus();
    const auto isSensorStatusValid = ds::DateTime{sensorStatus.m_WeatherTS} > system().now() - 3h;
    return {
        .wind = systemStateStatus(dss::StateName::Wind),
        .rain = systemStateStatus(dss::StateName::Rain),
        .hail = systemStateStatus(dss::StateName::Hail),
        .frost = systemStateStatus(dss::StateName::Frost),
        .dayNight =
            [&] { // day-night information is available only in icon
              if (isSensorStatusValid) {
                if (boost::ends_with(sensorStatus.m_WeatherIconId, "d")) {
                  return dsModel::DayNight::day;
                }
                if (boost::ends_with(sensorStatus.m_WeatherIconId, "n")) {
                  return dsModel::DayNight::night;
                }
              }
              return dsModel::DayNight::none;
            }(),
        .condition =
            dsModel::WeatherConditionId{isSensorStatusValid ? sensorStatus.m_WeatherConditionId : std::string{}},
        .icon = isSensorStatusValid ? sensorStatus.m_WeatherIconId : std::string{},
    };
  }

  [[nodiscard]] dsModel::ApartmentEnvironmentStatus apartmentEnvironmentStatus() const override {
    try {
      auto timeZoneOffset = system().timeZoneOffset();
      using Days = std::chrono::duration<signed, std::ratio<86400>>; // will be in <chrono> since c++20

      auto parseTime = [&](const std::string& path) -> ds::DateTime {
        DS_CONTEXT(path);
        try {
          auto timeStr = m_configNode->tryChildValue<std::string>(path).value();
          auto time = ds::toDateTime(ds::str("1970-01-01T", timeStr, "Z").c_str()).time_since_epoch();
          auto now = system().now();
          auto result = now - now.time_since_epoch() % Days{1} + time - timeZoneOffset;
          DS_DEBUG(timeStr, time, now, timeZoneOffset);
          return result;
        } catch (const std::exception& e) {
          DS_INFO("Failed to parse time in property node.", e);
          return {};
        }
      };
      return dsModel::ApartmentEnvironmentStatus{.sunrise = parseTime("geodata/sunrise"),
                                                 .sunset = parseTime("geodata/sunset"),
                                                 .dawn = parseTime("geodata/civil_dawn"),
                                                 .dusk = parseTime("geodata/civil_dusk")};
    } catch (const std::exception& e) {
      DS_INFO("Apartment environment status failed.", e);
      return {};
    }
  }

  [[nodiscard]] dsModel::ApartmentMeasurementStatus apartmentMeasurementStatus() const override {
    dsModel::ApartmentMeasurementStatus result{};

    ApartmentSensorStatus_t aSensorStatus = m_apartment.getSensorStatus();
    if (aSensorStatus.m_TemperatureTS != DateTime::NullDate) {
      result.temperature = aSensorStatus.m_TemperatureValue;
    }
    if (aSensorStatus.m_HumidityTS != DateTime::NullDate) {
      result.humidity = aSensorStatus.m_HumidityValue;
    }
    if (aSensorStatus.m_BrightnessTS != DateTime::NullDate) {
      result.brightness = aSensorStatus.m_BrightnessValue;
    }
    if (aSensorStatus.m_AirPressureTS != DateTime::NullDate) {
      result.airPressure = aSensorStatus.m_AirPressureValue;
    }
    if (aSensorStatus.m_PrecipitationTS != DateTime::NullDate) {
      result.precipitation = aSensorStatus.m_PrecipitationValue;
    }
    if (aSensorStatus.m_WindSpeedTS != DateTime::NullDate) {
      result.windSpeed = aSensorStatus.m_WindSpeedValue;
    }
    if (aSensorStatus.m_WindGustTS != DateTime::NullDate) {
      result.windGust = aSensorStatus.m_WindGustValue;
    }

    return result;
  }

  [[nodiscard]] dsModel::ZoneMeasurementsStatus zoneMeasurementsStatus(const dsModel::ZoneId& zoneId) const override {
    const auto& zone = *m_apartment.zone(zoneId);
    const auto sensorStatus = zone.getSensorStatus();
    return dsModel::ZoneMeasurementsStatus{
        .temperature = sensorStatus.temperatureValue,
        .humidity = sensorStatus.humidityValue,
        .brightness = sensorStatus.brightnessValue,
        .carbonDioxide = sensorStatus.co2ConcentrationValue,
        .motion = systemStateStatus(zoneStateName(StateName::Motion, zone.getID())),
        .presence = systemStateStatus(zoneStateName(StateName::Presence, zone.getID()))};
  }

  [[nodiscard]] dsModel::ZoneApplicationsStatus zoneApplicationsStatus(const dsModel::ZoneId& id) const override {
    const auto zone = m_apartment.zone(id);
    auto result = dsModel::ZoneApplicationsStatus{};
    for (const auto& application : zone->cachedApplications()) {
      switch (application) {
        case dsModel::ZoneApplicationType::none: break;
        case dsModel::ZoneApplicationType::lights: result.lights = m_apartment.proxy().lightsStatus(id); break;
        case dsModel::ZoneApplicationType::shades: result.shades = m_apartment.proxy().shadesStatus(id); break;
        case dsModel::ZoneApplicationType::awnings:
          result.awnings.status = [&] {
            switch (m_apartment.proxy().shadesStatus(id).status) {
              case dsModel::ShadesStatusStatus::none: return dsModel::AwningsStatusStatus::none;
              case dsModel::ShadesStatusStatus::closed: return dsModel::AwningsStatusStatus::in;
              case dsModel::ShadesStatusStatus::open: return dsModel::AwningsStatusStatus::out;
              case dsModel::ShadesStatusStatus::moving: return dsModel::AwningsStatusStatus::moving;
              case dsModel::ShadesStatusStatus::mixed: return dsModel::AwningsStatusStatus::mixed;
            }
            return dsModel::AwningsStatusStatus::none;
          }();
          break;
        case dsModel::ZoneApplicationType::audio: result.audio = m_apartment.proxy().audioStatus(id); break;
        case dsModel::ZoneApplicationType::video: result.video = m_apartment.proxy().videoStatus(id); break;
        case dsModel::ZoneApplicationType::security: break;
        case dsModel::ZoneApplicationType::access: break;
        case dsModel::ZoneApplicationType::heating: break;
        case dsModel::ZoneApplicationType::cooling: break;
        case dsModel::ZoneApplicationType::temperature: result.temperature = temperatureStatus(id); break;
        case dsModel::ZoneApplicationType::ventilation:
          result.ventilation = m_apartment.proxy().ventilationStatus(id);
          break;
        case dsModel::ZoneApplicationType::recirculation:
          result.recirculation = m_apartment.proxy().recirculationStatus(id);
          break;
        case dsModel::ZoneApplicationType::window: break;
        case dsModel::ZoneApplicationType::joker: break;
      }
    }
    return result;
  }

  [[nodiscard]] dsModel::ZoneStatus zoneStatus(const dsModel::ZoneId& zoneId) const override {
    DS_CONTEXT(zoneId);
    std::vector<dsModel::ScenarioStatus> scenarios{};

    try {
      for (const auto& scenarioId : zoneScenarioIds(*m_apartment.zone(zoneId))) {
        scenarios.push_back(m_apartment.proxy().zoneScenarioStatus(scenarioId));
      }

      return dsModel::ZoneStatus{.id = zoneId,
                                 .type = "zoneStatus",
                                 .lastChanged = {},
                                 .attributes = {
                                     .measurements = zoneMeasurementsStatus(zoneId),
                                     .applications = zoneApplicationsStatus(zoneId),
                                     .scenarios = std::move(scenarios),
                                 }};
    } catch (const std::exception& e) {
      DS_ERROR("Zone status failed.", e);
      return dsModel::ZoneStatus{.id = zoneId};
    }
  }

  [[nodiscard]] std::vector<dsModel::ZoneStatus> zoneStatuses() const override {
    return ds::makeVector(m_apartment.zoneIds() |
                          boost::adaptors::transformed([&](const auto& x) { return zoneStatus(x); }));
  }

  [[nodiscard]] dsModel::TemperatureScene temperatureLastScene(const int& scene) const {
    switch (scene) {
      case HeatingOperationModeIDOff: return dsModel::TemperatureScene::heatingOff;
      case HeatingOperationModeIDComfort: return dsModel::TemperatureScene::heatingComfort;
      case HeatingOperationModeIDEconomy: return dsModel::TemperatureScene::heatingEconomy;
      case HeatingOperationModeIDNotUsed: return dsModel::TemperatureScene::heatingSetback;
      case HeatingOperationModeIDNight: return dsModel::TemperatureScene::heatingNight;
      case HeatingOperationModeIDHoliday: return dsModel::TemperatureScene::heatingVacation;
      case HeatingOperationModeIDPassiveCoolingOn: return dsModel::TemperatureScene::passiveCoolingActive;
      case HeatingOperationModeIDPassiveCoolingOff: return dsModel::TemperatureScene::passiveCoolingInactive;
      case HeatingOperationModeIDCoolOff: return dsModel::TemperatureScene::coolingOff;
      case HeatingOperationModeIDCoolComfort: return dsModel::TemperatureScene::coolingComfort;
      case HeatingOperationModeIDCoolEconomy: return dsModel::TemperatureScene::coolingEconomy;
      case HeatingOperationModeIDCoolNotUsed: return dsModel::TemperatureScene::coolingSetback;
      case HeatingOperationModeIDCoolNight: return dsModel::TemperatureScene::coolingNight;
      case HeatingOperationModeIDCoolHoliday: return dsModel::TemperatureScene::coolingVacation;
    }
    return dsModel::TemperatureScene::none;
  }

  [[nodiscard]] dsModel::TemperatureStatusStatus temperatureStatusStatus(const std::optional<double>& controlValue,
                                                                         const dsModel::ZoneId& id) const {
    const auto heatingEnabled = m_heatingGeneralSettingsNode->tryChildValue<bool>("heatingEnabled").value_or(false);
    const auto heatingPossible =
        m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/heatingPossible")).value_or(false);
    const auto coolingPossible =
        m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/coolingPossible")).value_or(false);
    const auto passiveCoolingPossible =
        m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/passiveCoolingPossible")).value_or(false);

    if (!heatingEnabled) {
      return dsModel::TemperatureStatusStatus::off;
    } else {
      if (controlValue) {
        if (controlValue > 0) {
          if (heatingPossible) {
            return dsModel::TemperatureStatusStatus::heating;
          }
        } else if (controlValue < 0) {
          if (coolingPossible || passiveCoolingPossible) {
            return dsModel::TemperatureStatusStatus::cooling;
          }
        } else if (heatingPossible || coolingPossible || passiveCoolingPossible) {
          return dsModel::TemperatureStatusStatus::neutral;
        }
      }
    }
    return dsModel::TemperatureStatusStatus::off;
  }

  [[nodiscard]] dsModel::TemperatureStatus temperatureStatus(const dsModel::ZoneId& id) const override {
    DS_CONTEXT(id);
    const auto& zone = *m_apartment.zone(id);
    DS_REQUIRE(boost::find<boost::return_found_end>(zone.cachedApplications(), dsModel::ZoneApplicationType::temperature),
               "Temperature application does not exist.", id);
    try {
      const auto& config = zone.temperatureConfig();
      const auto& heatingStatus = zone.getHeatingStatus();
      const auto& controlValue = heatingStatus.controlValue;

      auto mode = [&]() -> dsModel::TemperatureMode {
        const auto& heatingEnabled =
            m_heatingGeneralSettingsNode->tryChildValue<bool>("heatingEnabled").value_or(false);
        if (!heatingEnabled) {
          return dsModel::TemperatureMode::off;
        }
        const auto& heatingForced =
            m_heatingGeneralSettingsNode->tryChildValue<bool>("overrideSystemModeAsHeating").value_or(false);
        const auto& coolingForced =
            m_heatingGeneralSettingsNode->tryChildValue<bool>("overrideSystemModeAsCooling").value_or(false);
        const auto heatingPossible =
            m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/heatingPossible")).value_or(false);
        const auto coolingPossible =
            m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/coolingPossible")).value_or(false);
        const auto passiveCoolingPossible =
            m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(id, "/passiveCoolingPossible")).value_or(false);
        if (!heatingPossible && !coolingPossible && !passiveCoolingPossible) {
          return dsModel::TemperatureMode::off;
        } else if (heatingForced) {
          if (heatingPossible) {
            return dsModel::TemperatureMode::heating;
          } else {
            return dsModel::TemperatureMode::off;
          }
        } else if (coolingForced) {
          if (coolingPossible || passiveCoolingPossible) {
            return dsModel::TemperatureMode::cooling;
          } else {
            return dsModel::TemperatureMode::off;
          }
        } else {
          return dsModel::TemperatureMode::automatic;
        }
      }();

      return dsModel::TemperatureStatus{
          .mode = mode,
          .status = temperatureStatusStatus(controlValue, id),
          .emergency = mode == dsModel::TemperatureMode::off
                           ? std::optional<bool>{}
                           : std::optional<bool>{config.m_HeatingControlState == HeatingControlStateIDEmergency},
          .setpoint = heatingStatus.nominalValue,
          .controlValue = controlValue,
          .lastCalledScene = temperatureLastScene(zone.getHeatingOperationMode()),
          .setbackRequested =
              m_heatingStatesNode->tryChildValue<int>(ds::str("operationmode.", zone.id(), ".NotUsed/value")) == 1
                  ? true
                  : false,
          .ecoPreferred =
              m_heatingStatesNode->tryChildValue<int>(ds::str("operationmode.", zone.id(), ".Eco/value")) == 1 ? true
                                                                                                               : false,
          .passiveCooling =
              m_heatingStatesNode->tryChildValue<int>(ds::str("passiveCooling.", zone.id(), "/value")) == 1 ? true
                                                                                                            : false,
      };
    } catch (const std::exception& e) {
      DS_ERROR("Failed to get temperature status", e);
      return {};
    }
  }

  [[nodiscard]] dsModel::Application temperatureApplication(const dsModel::ZoneId& zoneId) const override {
    DS_CONTEXT(zoneId);
    auto applicationId = dsModel::ApplicationId{ds::str("temperature-z", zoneId)};
    const auto zonePtr = m_apartment.zone(zoneId);
    const auto& zone = *zonePtr;
    DS_REQUIRE(boost::find<boost::return_found_end>(zone.cachedApplications(), dsModel::ZoneApplicationType::temperature),
               "Temperature application does not exist.", zoneId);
    try {
      const auto& config = zone.temperatureConfig();
      const auto& targetTemperatures = config.targetTemperatures;
      const auto heatingPossible =
          m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(zoneId, "/heatingPossible")).value_or(false);
      const auto coolingPossible =
          m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(zoneId, "/coolingPossible")).value_or(false);
      const auto passiveCoolingPossible =
          m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(zoneId, "/passiveCoolingPossible"))
              .value_or(false);

      std::vector<dsModel::TemperatureLevel> availableLevels{};
      std::vector<dsModel::TemperatureSetpoint> setpoints{};
      auto addAvailableLevel = [&](dsModel::TemperatureLevel level) {
        try {
          availableLevels.push_back(level);
          if (level == dsModel::TemperatureLevel::passiveCoolingThreshold) {
            setpoints.push_back(
                {.id = level,
                 .value = ds::tryToDouble(
                     m_heatingModesNode->tryChildValue<std::string>(ds::str(zoneId, "/minimalTemperatur"))
                         .value_or(std::string{}))});
          } else {
            setpoints.push_back({.id = level, .value = targetTemperatures[heatingOperationMode(level)]});
          }
        } catch (std::exception& e) {
          DS_ERROR("Failed to populate temperature level", level);
        }
      };

      if (heatingPossible) {
        addAvailableLevel(dsModel::TemperatureLevel::heatingComfort);
        addAvailableLevel(dsModel::TemperatureLevel::heatingEconomy);
        addAvailableLevel(dsModel::TemperatureLevel::heatingSetback);
        addAvailableLevel(dsModel::TemperatureLevel::heatingNight);
        addAvailableLevel(dsModel::TemperatureLevel::heatingVacation);
      }
      if (coolingPossible) {
        addAvailableLevel(dsModel::TemperatureLevel::coolingComfort);
        addAvailableLevel(dsModel::TemperatureLevel::coolingEconomy);
        addAvailableLevel(dsModel::TemperatureLevel::coolingSetback);
        addAvailableLevel(dsModel::TemperatureLevel::coolingNight);
        addAvailableLevel(dsModel::TemperatureLevel::coolingVacation);
      }
      if (passiveCoolingPossible) {
        addAvailableLevel(dsModel::TemperatureLevel::passiveCoolingThreshold);
      }

      return dsModel::Application{.id = applicationId,
                                  .attributes = dsModel::TemperatureApplicationAttributes{
                                      .applicationType = dsModel::ZoneApplicationType::temperature,
                                      .zone = zoneId,
                                      .availableLevels = std::move(availableLevels),
                                      .setpoints = std::move(setpoints),
                                  }};
    } catch (const std::exception& e) {
      DS_ERROR("Temperature application failed.", e);
      return dsModel::Application{.id = applicationId};
    }
  }

  [[nodiscard]] dsModel::Temperature temperature(const dsModel::ZoneId& zoneId) const override {
    DS_CONTEXT(zoneId);
    const auto zonePtr = m_apartment.zone(zoneId);
    const auto& zone = *zonePtr;
    DS_REQUIRE(boost::find<boost::return_found_end>(zone.cachedApplications(), dsModel::ZoneApplicationType::temperature),
               "Temperature application does not exist.", zoneId);

    const auto& config = zone.temperatureConfig();
    const auto& targetTemperatures = config.targetTemperatures;
    const auto heatingPossible =
        m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(zoneId, "/heatingPossible")).value_or(false);
    const auto coolingPossible =
        m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(zoneId, "/coolingPossible")).value_or(false);
    const auto passiveCoolingPossible =
        m_heatingControllerConfigNode->tryChildValue<bool>(ds::str(zoneId, "/passiveCoolingPossible")).value_or(false);

    std::vector<dsModel::TemperatureLevel> availableLevels{};
    std::vector<dsModel::TemperatureSetpoint> setpoints{};
    auto addAvailableLevel = [&](dsModel::TemperatureLevel level) {
      try {
        availableLevels.push_back(level);
        if (level == dsModel::TemperatureLevel::passiveCoolingThreshold) {
          setpoints.push_back({.id = level,
                               .value = ds::tryToDouble(
                                   m_heatingModesNode->tryChildValue<std::string>(ds::str(zoneId, "/minimalTemperatur"))
                                       .value_or(std::string{}))});
        } else {
          setpoints.push_back({.id = level, .value = targetTemperatures[heatingOperationMode(level)]});
        }
      } catch (std::exception& e) {
        DS_ERROR("Failed to populate temperature level", level);
      }
    };

    if (heatingPossible) {
      addAvailableLevel(dsModel::TemperatureLevel::heatingComfort);
      addAvailableLevel(dsModel::TemperatureLevel::heatingEconomy);
      addAvailableLevel(dsModel::TemperatureLevel::heatingSetback);
      addAvailableLevel(dsModel::TemperatureLevel::heatingNight);
      addAvailableLevel(dsModel::TemperatureLevel::heatingVacation);
    }
    if (coolingPossible) {
      addAvailableLevel(dsModel::TemperatureLevel::coolingComfort);
      addAvailableLevel(dsModel::TemperatureLevel::coolingEconomy);
      addAvailableLevel(dsModel::TemperatureLevel::coolingSetback);
      addAvailableLevel(dsModel::TemperatureLevel::coolingNight);
      addAvailableLevel(dsModel::TemperatureLevel::coolingVacation);
    }
    if (passiveCoolingPossible) {
      addAvailableLevel(dsModel::TemperatureLevel::passiveCoolingThreshold);
    }

    return dsModel::Temperature{
        .availableLevels = std::move(availableLevels),
        .setpoints = std::move(setpoints),
    };
  }

  [[nodiscard]] std::vector<dsModel::Application> applications() const override {
    auto result = std::vector<dsModel::Application>{};
    if (m_apiRevision.removals() < 1) {
      for (const auto& zoneId : m_apartment.zoneIds()) {
        auto zone = m_apartment.zone(zoneId);

        for (const auto& application : zone->cachedApplications()) {
          switch (application) {
            case dsModel::ZoneApplicationType::none: break;
            case dsModel::ZoneApplicationType::lights: break;
            case dsModel::ZoneApplicationType::shades: break;
            case dsModel::ZoneApplicationType::audio: break;
            case dsModel::ZoneApplicationType::video: break;
            case dsModel::ZoneApplicationType::security: break;
            case dsModel::ZoneApplicationType::access: break;
            case dsModel::ZoneApplicationType::heating: break;
            case dsModel::ZoneApplicationType::cooling: break;
            case dsModel::ZoneApplicationType::temperature: result.push_back(temperatureApplication(zone->id())); break;
            case dsModel::ZoneApplicationType::ventilation: break;
            case dsModel::ZoneApplicationType::recirculation: break;
            case dsModel::ZoneApplicationType::window: break;
            case dsModel::ZoneApplicationType::joker: break;
            case dsModel::ZoneApplicationType::awnings: break;
          }
        }
      }
    }
    return result;
  }

  [[nodiscard]] dsModel::ZoneApplications zoneApplications(const dsModel::ZoneId& zoneId) const override {
    auto result = dsModel::ZoneApplications{};
    auto zone = m_apartment.zone(zoneId);
    for (const auto& application : zone->cachedApplications()) {
      switch (application) {
        case dsModel::ZoneApplicationType::none: break;
        case dsModel::ZoneApplicationType::lights: break;
        case dsModel::ZoneApplicationType::shades: break;
        case dsModel::ZoneApplicationType::audio: break;
        case dsModel::ZoneApplicationType::video: break;
        case dsModel::ZoneApplicationType::security: break;
        case dsModel::ZoneApplicationType::access: break;
        case dsModel::ZoneApplicationType::heating: break;
        case dsModel::ZoneApplicationType::cooling: break;
        case dsModel::ZoneApplicationType::temperature: result.temperature = temperature(zone->id()); break;
        case dsModel::ZoneApplicationType::ventilation:
          result.ventilation = m_apartment.proxy().ventilationLevels(zone->id());
          break;
        case dsModel::ZoneApplicationType::recirculation:
          result.recirculation = m_apartment.proxy().recirculationLevels(zone->id());
          break;
        case dsModel::ZoneApplicationType::window: break;
        case dsModel::ZoneApplicationType::joker: break;
        case dsModel::ZoneApplicationType::awnings: break;
      }
    }
    return result;
  }

  [[nodiscard]] dsModel::ApartmentApplications apartmentApplications() const override {
    return dsModel::ApartmentApplications{.ventilationUnit = m_apartment.proxy().ventilationUnitLevels()};
  }

  [[nodiscard]] dsModel::ApartmentApplicationsStatus apartmentApplicationsStatus() const override {
    return dsModel::ApartmentApplicationsStatus{.ventilationUnit = m_apartment.proxy().ventilationUnitStatus()};
  }

  // include if requested in query
  template <class T>
  static auto included(const dsApi::GetQuery& query, std::string_view name, const T& f) -> decltype(f()) {
    if (query.includeAll || boost::find<boost::return_found_end>(query.include, name))
      return f();
    return {};
  };

  [[nodiscard]] dsModel::Apartment apartment(const dsApi::GetQuery& query) const override { //
    const auto apartmentId = m_apartment.id();
    try {
      //calculate cluster and zone cache
      auto zoneIds = m_apartment.zoneIds();
      for (auto zoneId : zoneIds) {
        auto zonePtr = m_apartment.zone(zoneId);
        zonePtr->calculateApplications();
      }
      auto clusterIds = m_apartment.clusterIds();
      for (auto clusterId : clusterIds) {
        auto clusterPtr = m_apartment.cluster(clusterId);
        clusterPtr->calculateCachedApplicationType();
      }
      return dsModel::Apartment{
          .id = apartmentId,
          .type = "apartment",
          .attributes = {.name = unescapeHTML(m_apartment.getName()),
                         .security = apartmentSecurity(),
                         .installation = {}, // TODO(someday): not in dss model
                         .zones = m_apartment.zoneIds(),
                         .dsDevices = m_apartment.deviceIds(),
                         .clusters = m_apartment.clusterIds(),
                         .applications2 = apartmentApplications()},
          .included = {
              .installation = included(query, "installation", [&] { return m_installation->installation(); }),
              .dsDevices = included(query, "dsDevices", [&] { return devices(); }),
              .submodules = included(query, "submodules", [&] { return submodules(); }),
              .functionBlocks = included(query, "functionBlocks", [&] { return functionBlocks(); }),
              .zones = included(query, "zones", [&] { return zones(); }),
              .scenarios = included(query, "scenarios", [&] { return scenarios().scenarios(); }),
              .userDefinedStates =
                  included(query, "userDefinedStates", [&] { return userDefinedStates().userDefinedStates; }),
              .floors = included(query, "floors", [&] { return floors().floors; }),
              .clusters = included(query, "clusters", [&] { return clusters(); }),
              .applications = included(query, "applications", [&] { return applications(); }),
              .dsServer = included(query, "dsServer", [&] { return dsServer(); }),
              .controllers = included(query, "controllers", [&] { return m_installation->controllers().controllers; }),
              .apiRevision = included(query, "apiRevision", [&] { return apiRevision(); }),
              .meterings = included(query, "meterings", [&] { return meterings().meterings().meterings; }),
          }};
    } catch (const std::exception& e) {
      DS_ERROR("Apartment failed.", e);
      return dsModel::Apartment{.id = apartmentId};
    }
  }

  [[nodiscard]] dsApi::HttpResult apartmentCached(const dsApi::HttpRequest& request) const override {
    std::lock_guard<std::mutex> lock(m_structureMutex);
    json_object *copy = nullptr;
    json_object_deep_copy(m_structure.get(), &copy, nullptr);
    return {.status = 200, .content = ds::toJsonFragment(dsApi::document(ds::makeJsonFragment(copy)))};
  }

  [[nodiscard]] dsApi::HttpResult apartmentStatusCached(const dsApi::HttpRequest& request) const override { 
    std::lock_guard<std::mutex> lock(m_statusMutex);
    json_object *copy = nullptr;
    json_object_deep_copy(m_status.get(), &copy, nullptr);
    return {.status = 200, .content = ds::toJsonFragment(dsApi::document(ds::makeJsonFragment(copy)))};
  }

  // backend api contract requires all fields in "included" to be serialized
  template <typename T>
  ds::JsonFragment toJsonFragmentWithEmptyIncluded(T&& value) {
    const auto included = std::move(value.included);
    value.included = decltype(included){};
    return ds::jsonAdd(ds::toJsonFragment(value), "included", ds::toJsonFragmentWithEmptyRootFields(included));
  }

  void pollApartment() override {
    DS_NOTICE("Poll apartment");
    pollApartmentStructureImpl();
  }

  [[nodiscard]] dsModel::ApartmentStatus apartmentStatus(const dsApi::GetQuery& query) const override {
    const auto apartmentId = m_apartment.id();
    /*DS_WARNING("step one");
    deviceStatuses();
    DS_WARNING("step two");
    zoneStatuses();
    DS_WARNING("step three");
    clusterStatuses();
    DS_WARNING("step four");
    userDefinedStateStatuses();
    DS_WARNING("step five");
    apartmentAccessStatus();
    apartmentUserStatus();
    apartmentSecurityStatus();
    apartmentTemperatureStatus();
    apartmentWeatherStatus();
    apartmentEnvironmentStatus();
    apartmentMeasurementStatus();
    apartmentApplicationsStatus();
    DS_WARNING("step six");*/
    try {
      return dsModel::ApartmentStatus{
          .id = apartmentId,
          .type = "apartmentStatus",
          .attributes = {.access = apartmentAccessStatus(),
                         .user = apartmentUserStatus(),
                         .security = apartmentSecurityStatus(),
                         .temperature = apartmentTemperatureStatus(),
                         .weather = apartmentWeatherStatus(),
                         .environment = apartmentEnvironmentStatus(),
                         .measurements = apartmentMeasurementStatus(),
                         .applications2 = apartmentApplicationsStatus()},
          .included = {
              .dsDevices = included(query, "dsDevices", [&] { return deviceStatuses(); }),
              .zones = included(query, "zones", [&] { return zoneStatuses(); }),
              .clusters = included(query, "clusters", [&] { return clusterStatuses(); }),
              .userDefinedStates = included(query, "userDefinedStates", [&] { return userDefinedStateStatuses(); })}};
    } catch (const std::exception& e) {
      DS_ERROR("Apartment status failed.", e);
      return dsModel::ApartmentStatus{.id = apartmentId};
    }
  }

  void pollApartmentStatus() override {
    DS_NOTICE("Poll apartment status");
    pollApartmentStatusImpl();
  }

  [[nodiscard]] dsModel::UserDefinedStates userDefinedStates() const override {
    auto result = dsModel::UserDefinedStates{};
    if (!m_udsNode) {
      return result;
    }
    for (auto&& udsNode : m_udsNode->getChildNodes()) {
      if (auto&& idNode = udsNode->getPropertyByName("name")) {
        if (auto&& nameNode = udsNode->getPropertyByName("displayName")) {
          if (auto&& setNameNode = udsNode->getPropertyByName("setName")) {
            if (auto&& resetNameNode = udsNode->getPropertyByName("resetName")) {
              if (auto&& phoneNode = udsNode->getPropertyByName("showOnPhone")) {
                if (auto&& typeNode = udsNode->getPropertyByName("type")) {
                  try {
                    result.userDefinedStates.push_back(dsModel::UserDefinedState{
                        .id = dsModel::UserDefinedStateId{unescapeHTML(idNode->getValue<std::string>())},
                        .type =
                            [&] {
                              auto name = typeNode->getValue<std::string>();
                              if (name == "custom-states")
                                return "manualState";
                              if ((name == "zone-sensor-states") || (name == "device-sensor-states"))
                                return "measurementState";
                              if (name == "consumption-states" || (name == "apartment-power-states") ||
                                  (name == "open-metering-system") || (name == "modbus"))
                                return "consumptionState";
                              if (name == "power-states")
                                return "circuitConsumptionState";
                              if (name == "triggered-states")
                                return "triggerState";
                              if (name == "combined-states")
                                return "combinedState";
                              if (name == "window-states")
                                return "windowState";
                              DS_FAIL_REQUIRE("Unkown UDS type", name);
                            }(),
                        .lastChanged = {},
                        .attributes = dsModel::UserDefinedStateAttributes{
                            .name = unescapeHTML(nameNode->getValue<std::string>()),
                            .setName = unescapeHTML(setNameNode->getValue<std::string>()),
                            .resetName = unescapeHTML(resetNameNode->getValue<std::string>()),
                            .visibleForUsers = phoneNode->getValue<bool>()}});
                  } catch (const std::exception& e) {
                    DS_ERROR("User defined state failed.", e);
                  }
                }
              }
            }
          }
        }
      }
    }
    boost::sort(result.userDefinedStates, [](const auto& a, const auto& b) { return a.id < b.id; });
    return result;
  }

  [[nodiscard]] std::vector<dsModel::UserDefinedStateStatus> userDefinedStateStatuses() const override {
    auto result = std::vector<dsModel::UserDefinedStateStatus>{};
    if (!m_udsNode) {
      return result;
    }
    for (auto&& udsNode : m_udsNode->getChildNodes()) {
      if (auto&& idNode = udsNode->getPropertyByName("name")) {
        if (auto&& stateNode = udsNode->getPropertyByName("state")) {
          const auto userDefinedStateId = dsModel::UserDefinedStateId{idNode->getValue<std::string>()};
          DS_CONTEXT(userDefinedStateId);
          try {
            result.push_back(dsModel::UserDefinedStateStatus{
                .id = userDefinedStateId,
                .type = "userDefinedStateStatus",
                .lastChanged = {},
                .attributes = {
                    .status = ds::toEnum<dsModel::UserDefinedStateStatusStatus>(stateNode->getValue<std::string>())}});
          } catch (const std::exception& e) {
            DS_ERROR("User defined state status failed.", e);
            result.push_back({.id = userDefinedStateId});
          }
        }
      }
    }
    boost::sort(result, [](const auto& a, const auto& b) { return a.id < b.id; });
    return result;
  }

  void setUserDefinedStateStatus(dsModel::UserDefinedStateStatusArgs&& args, const std::string& value) override {
    if (!m_udsNode) {
      return;
    }
    auto&& stateNode = m_udsNode->getPropertyByName(args.userDefinedStateId);
    DS_REQUIRE(stateNode, "State is unknown", args.userDefinedStateId);

    auto&& typeNode = stateNode->getPropertyByName("type");
    DS_REQUIRE(typeNode, "Type node unavailable");
    DS_REQUIRE(typeNode->getValue<std::string>() == "custom-states", "Only a manualState can be set");

    auto state = m_apartment.tryGetState(StateType_Script, args.userDefinedStateId);
    DS_REQUIRE(state, "State is unknown", args.userDefinedStateId);

    auto&& realValue = state->tryValueFromName(value);
    DS_REQUIRE(realValue, "Requested state value is illegal", value);
    state->setState(CallOrigin::json, *realValue);
    DS_NOTICE("Set UserDefinedState value", args, value);
  }

  [[nodiscard]] dsModel::ParameterCache parameterCacheStatus() const override {
    auto cacheStatus = m_apartment.proxy().getParameterCacheStatus();
    return dsModel::ParameterCache{
        .id = dsModel::ParameterCacheId{m_apartment.id()},
        .type = "parameterCache",
        .lastChanged = {},
        .attributes = std::move(cacheStatus.cache),
        .included = {.functionBlocks = ds::makeVector(
                         cacheStatus.devices | boost::adaptors::transformed([&](const auto& x) {
                           return dsModel::ParameterCacheFunctionBlock{.id = std::move(x.id),
                                                                       .type = "parameterCacheFunctionBlock",
                                                                       .lastChanged = {},
                                                                       .attributes = {.completion = x.completion}};
                         }))}};
  }

  void updateParameterCache(const dsModel::UpdateDeviceParameterCacheArgs& args) override {
    if (args.mode == dsModel::ParameterCacheUpdateMode::fast) {
      m_apartment.pushEvent(std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEventFast));
    } else {
      m_apartment.pushEvent(std::make_shared<Event>(CacheUpdatePlugin::CacheUpdateEvent));
    }
  }

  [[nodiscard]] dsModel::Floor floor(const dsModel::FloorArgs& args) const override {
    return m_installation->floor(args.floorId);
  }
  [[nodiscard]] dsModel::Floors floors() const override { return m_installation->floors(); }
  void setFloorName(const dsModel::FloorArgs& args, const std::string& value) override {
    m_installation->setFloorName(args.floorId, std::move(value));
  }

  void setTemperatureSetpoint(const dsModel::TemperatureSetpointArgs& args, double value) override {
    DS_NOTICE("Set temperature setpoint.", args, value);

    int zoneId = decodeId(args.applicationId).zoneId;
    DS_REQUIRE(m_apartment.zone(dsModel::ZoneId{ds::str(zoneId)}));

    if (args.temperatureLevel == dsModel::TemperatureLevel::passiveCoolingThreshold) {
      m_apartment.pushEvent(std::make_shared<Event>(
          "heating-controller.passive-cooling",
          Properties{{"actions", "setMinimalTemperature"}, {"zoneID", ds::str(zoneId)}, {"value", ds::str(value)}}));
    } else {
      m_apartment.pushEvent(std::make_shared<Event>(
          "heating-controller.operation-mode", Properties{{"actions", "setTemperatureValues"},
                                                          {"zoneID", ds::str(zoneId)},
                                                          {heatingLevelName(args.temperatureLevel), ds::str(value)}}));
    }
  }

  void setTemperatureStatusSetpoint(const dsModel::TemperatureStatusArgs& args, double value) override {
    DS_NOTICE("Set temperature status setpoint.", args, value);
    m_apartment.zone(args.zoneId)
        ->getGroup(GroupIDBroadcast)
        ->pushSensor(CallOrigin::json, SceneAccessCategory::manual, {}, SensorType::RoomTemperatureSetpoint, value, {});
  }

  void setTemperatureStatusSetbackRequested(const dsModel::TemperatureStatusArgs& args, bool value) override {
    DS_NOTICE("Set temperature status setback requested.", args, value);
    m_apartment.pushEvent(std::make_shared<Event>(
        "heating-controller.operation-mode", Properties{{"actions", value ? "setOperationMode" : "resetOperationMode"},
                                                        {"zoneID", args.zoneId},
                                                        {"operationMode", "3"}}));
  }

  void setTemperatureStatusEcoPreferred(const dsModel::TemperatureStatusArgs& args, bool value) override {
    DS_NOTICE("Set temperature status eco preferred.", args, value);
    m_apartment.pushEvent(std::make_shared<Event>(
        "heating-controller.operation-mode", Properties{{"actions", "setOperationMode"},
                                                        {"zoneID", args.zoneId},
                                                        {"operationMode", value ? "2" : "1"}}));
  }

  void setTemperatureStatusPassiveCooling(const dsModel::TemperatureStatusArgs& args, bool value) override {
    DS_NOTICE("Set temperature status passive cooling.", args, value);
    m_apartment.pushEvent(std::make_shared<Event>(
        "heating-controller.passive-cooling",
        Properties{{"actions", value ? "enable" : "disable"}, {"zoneID", args.zoneId}}));
  }

  [[nodiscard]] dsModel::Installation installation() const override { return m_installation->installation(); }

  [[nodiscard]] dsModel::DSServer dsServer() const override { return m_installation->dsServer(); }

  [[nodiscard]] dsModel::Controller controller(const dsModel::ControllerId& id) const override {
    return m_installation->controller(id);
  }

  dsModel::AuthorizeControllerResult authorizeController(const dsModel::ControllerId& id,
                                                         dsModel::AuthorizeControllerArgs&& args) override {
    return m_installation->authorizeController(id, std::move(args));
  }

  void toggleMaintenance(dsModel::MaintenanceData& maintenanceData) {
    DS_REQUIRE(maintenanceData.data.id == std::string(m_apartment.id()), "Apartment Id does not match.");
    DS_REQUIRE(maintenanceData.data.type == "apartment-maintenance", "Only type apartment-maintenance is supported.");
    m_installation->toggleMaintenance(maintenanceData);
    auto stopEvent = std::make_shared<Event>(EventName::StopMaintenance);
    stopEvent->setProperty(EventProperty::Time, "+86400");
    m_apartment.pushEvent(stopEvent);
  }

  void storeControllerAccessToken(const dsModel::ControllerId& id,
                                  dsModel::StoreControllerAccessTokenArgs&& args) override {
    return m_installation->storeControllerAccessToken(id, std::move(args));
  }
  [[nodiscard]] dsModel::Controllers controllers() const override { return m_installation->controllers(); }

  dsModel::ApplicationTokenId addApplicationToken(const dsModel::ApplicationToken& data) override {
    auto& applicationName = data.attributes.name;
    DS_REQUIRE(!applicationName.empty(), "Application name cannot be empty.");
    std::string applicationToken = dss::SessionTokenGenerator::generate();
    auto& security = DSS::getInstance()->getSecurity();
    User* dummyUser = NULL;

    security.createApplicationToken(applicationName, applicationToken);
    security.enableToken(applicationToken, dummyUser);

    DS_NOTICE("Added application token", applicationName, applicationToken);
    return dsModel::ApplicationTokenId{applicationToken};
  }

  void renameApplicationToken(const dsModel::ApplicationTokenId& id, const std::string& newName) override {
    DS_NOTICE("Renaming application token", id, newName);
    auto& security = DSS::getInstance()->getSecurity();
    security.setApplicationName(id, newName);
  }

  void revokeApplicationToken(const dsModel::ApplicationTokenId& id) override {
    DS_NOTICE("Revoking application token", id);
    auto& security = DSS::getInstance()->getSecurity();
    security.revokeToken(id);
  }

  dsModel::RemoteConnectivityApi& remoteConnectivity() override { return *m_remoteConnectivity; }
};
} // namespace

std::unique_ptr<DsModel> makeDsModel(ds::asio::ResourceBase& parent, dsBackend::Connector& bus, Apartment& apartment,
                                     EventInterpreter& eventInterpreter, dsApi::EventHub& eventHub,
                                     ds::asio::ResourceContext& apartmentContext, dsApartment::ApiRevision& apiRevision,
                                     DsModelHooks&& hooks) {
  return std::make_unique<DsModelImpl>(parent, bus, apartment, eventInterpreter, eventHub, apartmentContext,
                                       apiRevision, std::move(hooks));
}

} // namespace dss
