/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "session-base.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>
#include <ds/log.h>

using namespace trompeloeil;

namespace {

static const char *TAGS = "[dsNetSessionBase][dsNet]";

struct IntCallbackMock {
  MAKE_MOCK1(call, void(int));
};

struct SessionMock0 final : dsNet::Session<SessionMock0> {
  using dsNet::Session<SessionMock0>::Session;
  void close() override{};
};

struct SessionMock final : dsNet::Session<SessionMock>, dsNet::SessionsHook {
  SessionMock(dsNet::Sessions &sessions, std::string &&id)
      : dsNet::Session<SessionMock>{sessions, std::move(id)}, dsNet::SessionsHook{sessions} {}
  MAKE_MOCK0(close, void());
  MAKE_CONST_MOCK0(isClosed, bool());
};

struct SessionWithStateMock final : dsNet::Session<SessionWithStateMock>, dsNet::SessionsHook, dsNet::SessionStateBase {
  SessionWithStateMock(dsNet::Sessions &sessions, std::string &&id)
      : dsNet::Session<SessionWithStateMock>{sessions, std::move(id)}, dsNet::SessionsHook{sessions} {}
  MAKE_MOCK0(close, void());
};

SCENARIO("dsNetSession", TAGS) {
  auto loop = ds::asio::TestLoopResource{};
  GIVEN("sessions, session") {
    auto sessions = std::make_unique<dsNet::Sessions>(loop);
    auto x = ds::asio::Resource<>{*sessions, "x"};
    auto x2 = std::make_shared<ds::asio::Resource<>>(*sessions, "x2");
    auto x3 = std::make_shared<SessionMock0>(*sessions, "x3");
    auto session = std::make_shared<SessionMock>(*sessions, "session1");
    auto weakSession = std::weak_ptr{session};
    THEN("session has right path, id") {
      CHECK(session->id() == "session1");
      CHECK(session->path() == "/sessions/session1");
    }
    THEN("session is closed when sessions are destroyed") {
      REQUIRE_CALL(*session, close());
      sessions.reset();
    }
    THEN("session is removed from sessions in session destructor") {
      session.reset();
      // no close from sessions destructor
      sessions.reset();
      CHECK(!weakSession.lock());
    }
    SECTION("sharedCallback holds shared reference") {
      {
        auto sharedCallback = session->sharedCallback([&](SessionMock &) {});
        CHECK(weakSession.lock());
        session.reset();
        CHECK(weakSession.lock());
      }
      CHECK(!weakSession.lock());
    }
    GIVEN("sharedCallback") {
      auto intCallback = IntCallbackMock{};
      auto sharedCallback = session->sharedCallback([&, sessionPtr = session.get()](SessionMock &x, int arg) {
        CHECK(&x == sessionPtr);
        intCallback.call(arg);
      });
      THEN("!isClosed && session -> call") {
        ALLOW_CALL(*session, isClosed()).LR_RETURN(false);
        REQUIRE_CALL(intCallback, call(5));
        sharedCallback(5);
      }
      THEN("!isClosed && !session -> call") {
        ALLOW_CALL(*session, isClosed()).LR_RETURN(false);
        session.reset();
        REQUIRE_CALL(intCallback, call(5));
        sharedCallback(5);
      }
      THEN("isClosed && session -> !call") {
        ALLOW_CALL(*session, isClosed()).LR_RETURN(true);
        sharedCallback(5);
      }
    }
    SECTION("weakCallback does not hold shared reference") {
      auto sharedCallback = session->weakCallback([&](SessionMock &) {});
      CHECK(weakSession.lock());
      session.reset();
      CHECK(!weakSession.lock());
    }
    GIVEN("weakCallback") {
      auto intCallback = IntCallbackMock{};
      auto weakCallback = session->weakCallback([&](SessionMock &x, int arg) {
        CHECK(&x == session.get());
        intCallback.call(arg);
      });
      THEN("!isClosed && session -> call") {
        ALLOW_CALL(*session, isClosed()).LR_RETURN(false);
        REQUIRE_CALL(intCallback, call(5));
        weakCallback(5);
      }
      THEN("isClosed && session-> !call") {
        ALLOW_CALL(*session, isClosed()).LR_RETURN(true);
        weakCallback(5);
      }
      THEN("!session-> !isClosed && !call") {
        session.reset();
        weakCallback(5);
      }
    }
  }
}

SCENARIO("dsNetSessionStateBase", TAGS) {
  auto loop = ds::asio::TestLoopResource{};
  GIVEN("sessions, session") {
    auto sessions = std::make_unique<dsNet::Sessions>(loop);
    auto session = std::make_shared<SessionWithStateMock>(*sessions, "session1");
    auto checkClose = [&] {
      session->setClosed();
      CHECK(session->state() == dsNet::SessionStateType::closed);
      CHECK(session->isClosed());
    };

    THEN("can close") { checkClose(); }
    THEN("starts in init state") { CHECK(session->state() == dsNet::SessionStateType::init); }
    THEN("cannot setOpen") { CHECK_THROWS_FIND(session->setOpen(), "non-opening state"); }
    THEN("!requireOpen") { CHECK_THROWS_FIND(session->requireOpen(), "not open"); }
    WHEN("setOpening") {
      session->setOpening();
      CHECK(session->state() == dsNet::SessionStateType::opening);
      THEN("cannot setOpening") { CHECK_THROWS_FIND(session->setOpening(), "non-init state"); }
      THEN("can close") { checkClose(); }
      THEN("!requireOpen") { CHECK_THROWS_FIND(session->requireOpen(), "not open"); }

      WHEN("setOpen") {
        session->setOpen();
        CHECK(session->state() == dsNet::SessionStateType::open);
        THEN("requireOpen") { session->requireOpen(); }
        THEN("cannot setOpening") { CHECK_THROWS_FIND(session->setOpening(), "non-init state"); }
        THEN("cannot setOpen") { CHECK_THROWS_FIND(session->setOpen(), "non-opening state"); }
        THEN("can close") { checkClose(); }
      }
    }
  }
}

} // namespace
