/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <ds/asio/resource.h>

#include <boost/intrusive/list.hpp>

#include "session.h"

namespace dsNet {
namespace intrusive = boost::intrusive;

/// `Session` is base class for sessions providing common functionality
/// - enable_shared_from_this for shared_ptr ownership
/// - registration and unregistration in `Sessions`.
///     Sessions will call close on registered sessions when going out of scope
/// - Resource for visibility in inside.
/// - sharedCallback, weakCallback for capturing shared_this in callbacks
///
/// See `SessionStateBase` mixin for state management.
template <class T, class Config = ds::asio::EmptyResourceConfig>
struct Session : ds::asio::Resource<Config>, std::enable_shared_from_this<T>, virtual ISession {
  Session(ds::asio::ResourceBase& parent, std::string&& id, Config&& config = {})
      : ds::asio::Resource<Config>{parent, std::move(id), std::move(config)} {}

  using std::enable_shared_from_this<T>::shared_from_this;
  using std::enable_shared_from_this<T>::weak_from_this;

  std::shared_ptr<ISession> sharedSession() override { return shared_from_this(); }

  /// Functor calling member function with captured shared_from_this if !isClosed
  template <class F>
  auto sharedCallback(F&& f) {
    return [shared = shared_from_this(), f = DS_FWD(f)](auto&&... args) {
      if (!shared->isClosed()) {
        std::invoke(f, *shared, DS_FWD(args)...);
      }
    };
  }

  /// Functor calling member function with captured weak_from_this if !isClosed
  template <class F>
  auto weakCallback(F&& f) {
    return [weak = weak_from_this(), f = DS_FWD(f)](auto&&... args) {
      if (auto shared = weak.lock()) {
        if (!shared->isClosed()) {
          std::invoke(f, *shared, DS_FWD(args)...);
        }
      }
    };
  }
};

class Sessions;

/// `SessionsHook` mixin unregisters session from `Sessions` when session is destroyed.
struct SessionsHook : virtual ISession, intrusive::list_base_hook<intrusive::link_mode<intrusive::auto_unlink>> {
  SessionsHook(Sessions& sessions);
  /// Destroy unlinks from intrusive list
  virtual ~SessionsHook();
};

/// Tracks registered sessions and closes active ones in destructor.
///
/// Typically used in acceptor. When acceptor is destroyed,
/// all its session will be closed.
class Sessions final : public ds::asio::Resource<> {
  intrusive::list<SessionsHook, intrusive::constant_time_size<false>> m_sessions;

public:
  Sessions(ds::asio::ResourceBase& parent, std::string&& id = "sessions");
  ~Sessions();

private:
  void push_back(SessionsHook& session);
  friend struct SessionsHook;
};

/// `SessionBase` is base class for providing SessionState mixin.
class SessionStateBase : public virtual ds::asio::IResourceBase, public virtual SessionState {
  SessionStateType m_state = SessionStateType::init;

  void requireState(SessionStateType state) { DS_REQUIRE(m_state == state, "Invalid state.", m_state, *this); }

public:
  SessionStateType state() const final { return m_state; }
  void setOpening() {
    DS_REQUIRE(m_state == SessionStateType::init, "Opening session in non-init state.", m_state, *this);
    m_state = SessionStateType::opening;
  }
  void setOpen() {
    DS_REQUIRE(m_state == SessionStateType::opening, "Opened session in non-opening state.", m_state, *this);
    m_state = SessionStateType::open;
  }
  void requireOpen() { DS_REQUIRE(m_state == SessionStateType::open, "Session is not open.", m_state, *this); }
  void setClosed() { m_state = SessionStateType::closed; }
};

} // namespace dsNet
