/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "signalr.h"

#include <ds/asio/test.h>
#include <ds/catch/catch.h>

#include "message-stream-mock.h"
#include "session-base.h"

using namespace trompeloeil;
using namespace std::literals;

namespace {

static const char* TAGS = "[dsNetSignalr][dsNet]";

struct SignalrSessionHooksMock {
  MAKE_MOCK0(opened, void());
  MAKE_MOCK1(closed, void(dsNet::error_code));
  MAKE_MOCK2(invokeNoReply, void(std::string&& target, std::vector<ds::JsonFragment>&&));
  auto hooks() {
    return dsNet::SignalrSessionHooks{
        .opened = [this] { opened(); },
        .closed = [this](auto&& x) { closed(DS_FWD(x)); },
        .invokeNoReply = [this](auto&& x, auto&& y) { invokeNoReply(DS_FWD(x), DS_FWD(y)); },
        .bufferedMessageWriterConfig = {.countCapacity = 10, .sizeCapacity = 64000}};
  }
};

struct SignalrAcceptorHooksMock {
  MAKE_MOCK1(accept, void(dsNet::ClosePtr<dsNet::SignalrSession>));
  auto hooks() {
    return dsNet::SignalrAcceptorHooks{.accept = [this](auto&& x) { accept(DS_FWD(x)); }};
  }
};

SCENARIO("dsNetSignalr", TAGS) {
  auto loop = ds::asio::TestLoopResource{};
  auto messageStreamAcceptorPtr = std::make_unique<dsNet::MessageStreamAcceptorMock>();
  auto& messageStreamAcceptor = *messageStreamAcceptorPtr;
  auto signalrAcceptor = dsNet::makeSignalrAcceptor(loop, std::move(messageStreamAcceptorPtr));
  auto expectMessageStreamClose = std::unique_ptr<trompeloeil::expectation>{}; // close can come from Sessions
  auto signalrSessionHooks = SignalrSessionHooksMock{};
  auto expectSignalrSessionClosed = std::unique_ptr<trompeloeil::expectation>{}; // close can come from Sessions
  auto sessions = std::make_unique<dsNet::Sessions>(loop);

  SECTION("signalr acceptor asyncRun -> message stream acceptor asyncRun") {
    auto messageStreamAcceptorHooks = dsNet::MessageStreamAcceptorHooks{};
    REQUIRE_CALL(messageStreamAcceptor, asyncRun(_)).LR_SIDE_EFFECT(messageStreamAcceptorHooks = _1);
    auto signalrAcceptorHooks = SignalrAcceptorHooksMock{};
    signalrAcceptor->asyncRun(signalrAcceptorHooks.hooks());

    SECTION("nothing") {}
    SECTION("message stream accept -> signalr session accept") {
      auto signalrSession = dsNet::ClosePtr<dsNet::SignalrSession>{};
      REQUIRE_CALL(signalrAcceptorHooks, accept(_)).LR_SIDE_EFFECT(signalrSession = std::move(_1));
      auto messageStream = std::make_shared<dsNet::MessageStreamMock>(loop, "id1");
      expectMessageStreamClose = NAMED_REQUIRE_CALL(*messageStream, close());
      messageStreamAcceptorHooks.accept(std::shared_ptr{messageStream});
      SECTION("nothing") {}
      SECTION("signalr session asyncRun -> message stream asyncRun") {
        auto messageStreamHooks = dsNet::MessageStreamHooks{};
        REQUIRE_CALL(*messageStream, asyncRun(_)).LR_SIDE_EFFECT(messageStreamHooks = _1);
        signalrSession->asyncRun(signalrSessionHooks.hooks());
        expectSignalrSessionClosed = NAMED_REQUIRE_CALL(signalrSessionHooks, closed(_));

        auto checkMessageStreamClose = [&] {
          messageStreamHooks.closed(boost::asio::error::operation_aborted);
          expectMessageStreamClose.reset();
          expectSignalrSessionClosed.reset();
        };
        auto checkSignalrSessionClose = [&] {
          signalrSession->close();
          expectMessageStreamClose.reset();
          expectSignalrSessionClosed.reset();
        };
        SECTION("nothing") {}
        SECTION("message stream closed") { checkMessageStreamClose(); }
        SECTION("signalr session closed") { checkSignalrSessionClose(); }
        SECTION("message stream opened -> message stream async read") {
          auto readHandler = dsNet::MessageStream::ReadHandler{};
          REQUIRE_CALL(*messageStream, asyncRead(_)).LR_SIDE_EFFECT(readHandler = _1);
          messageStreamHooks.opened();

          SECTION("nothing") {}
          SECTION("message stream closed") { checkMessageStreamClose(); }
          SECTION("signalr session closed") { checkSignalrSessionClose(); }
          SECTION("invalid signalr handshake initiated from message stream -> close") {
            auto message = GENERATE(
                ""s, R"({"protocol":"json"})"s, R"({"version":1})"s, R"({"protocol":"json", "version":2})"s,
                R"({"protocol":"bson", "version":1})"s,
                "{\"protocol\":\"json\", \"version\":1}\x1e{}"s // we support only one record in handshake message
            );
            auto suffix = GENERATE(""s, "333\x1e"s);
            readHandler({}, ds::str(message, suffix));
            expectMessageStreamClose.reset();
          }
          SECTION("signalr handshake initiated from message stream -> signal opened") {
            auto writeHandler = dsNet::MessageStream::WriteHandler{};
            {
              REQUIRE_CALL(signalrSessionHooks, opened());
              auto nextReadHandler = dsNet::MessageStream::ReadHandler{};
              REQUIRE_CALL(*messageStream, asyncRead(_)).LR_SIDE_EFFECT(nextReadHandler = std::move(_1));
              REQUIRE_CALL(*messageStream, asyncWrite("{}\x1e", _)).LR_SIDE_EFFECT(writeHandler = std::move(_2));
              readHandler({}, R"({"protocol":"json", "version":1})"
                              "\x1e");
              readHandler = std::move(nextReadHandler);
            }

            auto checkWriteFails = [&] {
              writeHandler(boost::asio::error::operation_aborted);
              expectMessageStreamClose.reset();
            };

            SECTION("nothing") {}
            SECTION("message stream closed") { checkMessageStreamClose(); }
            SECTION("signalr session closed") { checkSignalrSessionClose(); }
            SECTION("write fails") { checkWriteFails(); }
            SECTION("hanshake write succeeds") {
              writeHandler({});

              SECTION("message stream invocation without id -> signalr session invokeNoReply") {
                REQUIRE_CALL(signalrSessionHooks, invokeNoReply("target1", std::vector{ds::toJsonFragment(1)}));
                REQUIRE_CALL(*messageStream, asyncRead(_));
                readHandler({}, R"({"type":1,"target":"target1","arguments":[1]})"
                                "\x1e");
              }

              SECTION("message stream invocation without id and argmuents-> signalr session invokeNoReply") {
                REQUIRE_CALL(signalrSessionHooks, invokeNoReply("target1", std::vector<ds::JsonFragment>{}));
                REQUIRE_CALL(*messageStream, asyncRead(_));
                readHandler({}, R"({"type":1,"target":"target1"})"
                                "\x1e");
              }

              SECTION("message stream invocation with id -> reply") {
                // no support for asynchronous && non-empty replies now
                {
                  auto nextReadHandler = dsNet::MessageStream::ReadHandler{};
                  REQUIRE_CALL(signalrSessionHooks, invokeNoReply("target1", std::vector{ds::toJsonFragment(1)}));
                  REQUIRE_CALL(*messageStream, asyncRead(_)).LR_SIDE_EFFECT(nextReadHandler = _1);
                  REQUIRE_CALL(*messageStream, asyncWrite(R"({"type":3,"invocationId":"1"})"
                                                          "\x1e",
                                                          _))
                      .LR_SIDE_EFFECT(writeHandler = _2);
                  readHandler({}, R"({"type":1,"invocationId":"1","target":"target1","arguments":[1]})"
                                  "\x1e");
                  readHandler = std::move(nextReadHandler);
                }

                SECTION("nothing") {}
                SECTION("message stream closed") { checkMessageStreamClose(); }
                SECTION("write fails") { checkWriteFails(); }
              }

              SECTION("batched message stream invocation with id -> batched signalr session invokeNoReply") {
                {
                  auto nextReadHandler = dsNet::MessageStream::ReadHandler{};
                  REQUIRE_CALL(signalrSessionHooks, invokeNoReply("target2", std::vector<ds::JsonFragment>{}));
                  REQUIRE_CALL(signalrSessionHooks, invokeNoReply("target1", std::vector<ds::JsonFragment>{}));
                  REQUIRE_CALL(*messageStream, asyncRead(_)).LR_SIDE_EFFECT(nextReadHandler = _1);
                  REQUIRE_CALL(*messageStream, asyncWrite(R"({"type":3,"invocationId":"1"})"
                                                          "\x1e",
                                                          _))
                      .LR_SIDE_EFFECT(writeHandler = _2);
                  readHandler({}, R"({"type":1,"invocationId":"1","target":"target1"})"
                                  "\x1e"
                                  R"({"type":1,"invocationId":"2","target":"target2"})");
                  readHandler = std::move(nextReadHandler);
                }

                SECTION("nothing") {}
                SECTION("message stream closed") { checkMessageStreamClose(); }
                SECTION("signalr session closed") { checkSignalrSessionClose(); }
                SECTION("write fails") { checkWriteFails(); }
                SECTION("buffered writes are writted") {
                  auto nextWriteHandler = dsNet::MessageStream::WriteHandler{};
                  REQUIRE_CALL(*messageStream, asyncWrite(R"({"type":3,"invocationId":"2"})"
                                                          "\x1e",
                                                          _))
                      .LR_SIDE_EFFECT(nextWriteHandler = std::move(_2));
                  writeHandler({});
                  nextWriteHandler({});
                }
              }

              SECTION("signalr session invocation without id -> message stream write") {
                REQUIRE_CALL(*messageStream, asyncWrite(R"({"type":1,"target":"target1","arguments":[1]})"
                                                        "\x1e",
                                                        _))
                    .LR_SIDE_EFFECT(writeHandler = _2);
                signalrSession->invokeNoReply("target1", std::vector{ds::toJsonFragment(1)});

                SECTION("nothing") {}
                SECTION("message stream closed") { checkMessageStreamClose(); }
                SECTION("write fails") { checkWriteFails(); }
              }
            }
          }
        }
      }
    }
  }
}

} // namespace
