/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "websocket.h"

#include <ds/asio/test.h>
#include <ds/log.h>

#include <boost/asio/connect.hpp>
#include <boost/beast/core.hpp>
#include <boost/beast/http.hpp>
#include <boost/beast/websocket.hpp>
#include <boost/beast/websocket/ssl.hpp>

#include "message-stream-mock.h"

using namespace std::literals;
using namespace trompeloeil;
using boost::system::error_code;
using tcp = boost::asio::ip::tcp;
namespace http = boost::beast::http;
namespace websocket = boost::beast::websocket;

namespace {

static const char *TAGS = "[dsNetWebsocket]";

struct AcceptorAccepted {
public:
  MAKE_MOCK1(call, void(dsNet::ClosePtr<dsNet::MessageStream>));
};

struct Completed {
public:
  MAKE_MOCK1(call, void(error_code));
};

struct WebsocketAcceptorHooksMock {
  MAKE_MOCK1(authenticate, void(const http::request<http::string_body> &));
  auto hooks() {
    return dsNet::WebsocketAcceptorHooks{.authenticate = [this](auto &&x) { authenticate(DS_FWD(x)); }};
  }
};

SCENARIO("dsNetWSAcceptor", TAGS) {
  auto expectStreamClosed = std::unique_ptr<trompeloeil::expectation>{}; // alive till io callback is destroyed
  auto streamHooks = dsNet::MessageStreamHooksMock{};
  auto websocketAcceptorHooks = WebsocketAcceptorHooksMock{};
  auto loop = ds::asio::TestLoopResource{};
  auto acceptorAccepted = AcceptorAccepted{};
  const auto host = "127.0.0.1"s;
  auto tcpAcceptorPtr = dsNet::makeTcpAcceptor(loop, {.endpoint = {.address = host}});
  auto &tcpAcceptor = *tcpAcceptorPtr;
  auto wsAcceptor = dsNet::makeWSAcceptor(loop, std::move(tcpAcceptorPtr), websocketAcceptorHooks.hooks());
  wsAcceptor->asyncRun({.accept = [&](auto &&x) { acceptorAccepted.call(DS_FWD(x)); }});
  auto resolver = tcp::resolver{loop.io};
  auto wsClient = websocket::stream<tcp::socket>{loop.io};
  const auto endpoints = resolver.resolve(tcpAcceptor.localAsioEndpoint());
  WHEN("tcp client connects") {
    auto stream = dsNet::ClosePtr<dsNet::MessageStream>{};
    REQUIRE_CALL(acceptorAccepted, call(_)).LR_SIDE_EFFECT(stream = std::move(_1); loop.stop());
    boost::asio::async_connect(wsClient.next_layer(), endpoints, [&](error_code ec, auto &&) {
      REQUIRE(!ec);
      loop.stop();
    });
    expectStreamClosed = NAMED_REQUIRE_CALL(streamHooks, closed(_));
    loop.run();
    loop.run();
    stream->asyncRun(streamHooks.hooks());
    WHEN("client starts connecting ws stream") {
      auto clientHandshaked = Completed{};
      wsClient.async_handshake(host, "/testPath", [&](error_code ec) { clientHandshaked.call(ec); });
      THEN("ws connection is established") {
        {
          REQUIRE_CALL(streamHooks, opened()).LR_SIDE_EFFECT(loop.stop(););
          REQUIRE_CALL(clientHandshaked, call(_)).LR_SIDE_EFFECT(CHECK(!_1); loop.stop(););
          REQUIRE_CALL(websocketAcceptorHooks, authenticate(_))
              .LR_SIDE_EFFECT(CHECK(_1.target() == "/testPath"); loop.stop(););
          loop.run();
          loop.run();
          loop.run();
        }

        const auto sessionId = ds::asio::ResourceBase::normalizeId(ds::str(wsClient.next_layer().local_endpoint()));
        THEN("stream id returns client endpoint") { CHECK(sessionId == stream->id()); }

        THEN("server reads what client writes") {
          stream->asyncRead([&](error_code ec, std::string &&s) {
            REQUIRE(!ec);
            REQUIRE(s == "clientWrites");
            loop.stop();
          });
          wsClient.async_write(boost::asio::buffer("clientWrites", 12), [&](error_code ec, std::size_t size) {
            CHECK(!ec);
            CHECK(size == 12);
            loop.stop();
          });
          loop.run();
          loop.run();
        }

        THEN("client reads what server writes") {
          boost::beast::multi_buffer readBuffer;
          wsClient.async_read(readBuffer, [&](error_code ec, std::size_t) {
            REQUIRE(!ec);
            loop.stop();
            auto message = boost::beast::buffers_to_string(readBuffer.data());
            readBuffer.consume(message.size());
            REQUIRE(message == "serverWrites");
          });
          stream->asyncWrite("serverWrites", [&](error_code ec) {
            REQUIRE(!ec);
            loop.stop();
          });
          loop.run();
          loop.run();
        }

        THEN("stream read fails when stream gets closed") {
          stream->asyncRead([&](error_code ec, std::string &&) {
            REQUIRE(ec);
            loop.stop();
          });
          loop.post([&] { stream->close(); });
          loop.run();
        }

        THEN("stream read fails when stream is closed") {
          stream->asyncRead([&](error_code ec, std::string &&) {
            REQUIRE(ec);
            loop.stop();
          });
          loop.post([&] { stream->close(); });
          loop.run();
        }

        THEN("stream write fails when stream is closed") {
          stream->close();
          loop.post([&] {
            stream->asyncWrite("aa", [&](error_code ec) {
              REQUIRE(ec);
              loop.stop();
            });
          });
          loop.run();
        }

        THEN("server read fails when client closes ws") {
          stream->asyncRead([&](error_code ec, std::string &&) {
            REQUIRE(ec);
            loop.stop();
          });
          wsClient.async_close(boost::beast::websocket::close_code::none, [&](error_code ec) {
            REQUIRE(!ec);
            loop.stop();
          });
          loop.run();
          loop.run();
        }

        THEN("server read fails when client closes socket") {
          stream->asyncRead([&](error_code ec, std::string &&) {
            REQUIRE(ec);
            loop.stop();
          });
          wsClient.next_layer().close();
          loop.run();
        }

        THEN("client read fails when server is destroyed") {
          boost::beast::multi_buffer readBuffer;
          wsClient.async_read(readBuffer, [&](error_code ec, std::size_t) {
            REQUIRE(ec);
            loop.stop();
          });
          wsAcceptor.reset();
          loop.run();
        }

        THEN("stream id is valid after client disconnects") {
          wsClient.next_layer().close();
          CHECK(sessionId == stream->id());
        }
      }

      THEN("ws connection is dropped when authenticate fails") {
        REQUIRE_CALL(clientHandshaked, call(_)).LR_SIDE_EFFECT(CHECK(_1); loop.stop());
        REQUIRE_CALL(websocketAcceptorHooks, authenticate(_))
            .LR_SIDE_EFFECT(loop.stop(); DS_FAIL_REQUIRE("test authentivate failure"));
        loop.run();
        loop.run();
      }

      THEN("connect fails if acceptor is destroyed") {
        REQUIRE_CALL(clientHandshaked, call(_)).LR_SIDE_EFFECT(CHECK(_1); loop.stop(););
        wsAcceptor.reset();
        loop.run();
      }
    }
  }
}

SCENARIO("dsNetParseWSSUrl", TAGS) {
  using dsNet::parseWSSUrl;
  using dsNet::WSSUrl;
  GIVEN("url without wss:// prefix throws") {
    REQUIRE_THROWS_FIND(parseWSSUrl(""), "missing wss://");
    REQUIRE_THROWS_FIND(parseWSSUrl("http://"), "missing wss://");
    REQUIRE_THROWS_FIND(parseWSSUrl("ws://"), "missing wss://");
  }

  GIVEN("url with host") {
    THEN("parses with default wss port") {
      REQUIRE(parseWSSUrl("wss://example.com/") == (WSSUrl{.host = "example.com", .port = "443", .target = "/"}));
    }
  }

  SECTION("url with host and port") {
    REQUIRE(parseWSSUrl("wss://example.com:443/") == (WSSUrl{.host = "example.com", .port = "443", .target = "/"}));
    REQUIRE(parseWSSUrl("wss://example.com:444/") == (WSSUrl{.host = "example.com", .port = "444", .target = "/"}));
  }

  SECTION("url with host, port, target") {
    REQUIRE(parseWSSUrl("wss://example.com:443/a") == (WSSUrl{.host = "example.com", .port = "443", .target = "/a"}));
    REQUIRE(parseWSSUrl("wss://example.com:444/a/b") ==
            (WSSUrl{.host = "example.com", .port = "444", .target = "/a/b"}));
  }

  SECTION("url with host, target") {
    REQUIRE(parseWSSUrl("wss://example.com/a") == (WSSUrl{.host = "example.com", .port = "443", .target = "/a"}));
    REQUIRE(parseWSSUrl("wss://example.com/a/b") == (WSSUrl{.host = "example.com", .port = "443", .target = "/a/b"}));
  }
}

// Certificate from boost::beast examples
//
// TODO(someday): move to shared place
struct SSLContext : public boost::asio::ssl::context {
  using Super = boost::asio::ssl::context;
  SSLContext() : Super{boost::asio::ssl::context::sslv23} {
    std::string const cert =
        "-----BEGIN CERTIFICATE-----\n"
        "MIIDaDCCAlCgAwIBAgIJAO8vBu8i8exWMA0GCSqGSIb3DQEBCwUAMEkxCzAJBgNV\n"
        "BAYTAlVTMQswCQYDVQQIDAJDQTEtMCsGA1UEBwwkTG9zIEFuZ2VsZXNPPUJlYXN0\n"
        "Q049d3d3LmV4YW1wbGUuY29tMB4XDTE3MDUwMzE4MzkxMloXDTQ0MDkxODE4Mzkx\n"
        "MlowSTELMAkGA1UEBhMCVVMxCzAJBgNVBAgMAkNBMS0wKwYDVQQHDCRMb3MgQW5n\n"
        "ZWxlc089QmVhc3RDTj13d3cuZXhhbXBsZS5jb20wggEiMA0GCSqGSIb3DQEBAQUA\n"
        "A4IBDwAwggEKAoIBAQDJ7BRKFO8fqmsEXw8v9YOVXyrQVsVbjSSGEs4Vzs4cJgcF\n"
        "xqGitbnLIrOgiJpRAPLy5MNcAXE1strVGfdEf7xMYSZ/4wOrxUyVw/Ltgsft8m7b\n"
        "Fu8TsCzO6XrxpnVtWk506YZ7ToTa5UjHfBi2+pWTxbpN12UhiZNUcrRsqTFW+6fO\n"
        "9d7xm5wlaZG8cMdg0cO1bhkz45JSl3wWKIES7t3EfKePZbNlQ5hPy7Pd5JTmdGBp\n"
        "yY8anC8u4LPbmgW0/U31PH0rRVfGcBbZsAoQw5Tc5dnb6N2GEIbq3ehSfdDHGnrv\n"
        "enu2tOK9Qx6GEzXh3sekZkxcgh+NlIxCNxu//Dk9AgMBAAGjUzBRMB0GA1UdDgQW\n"
        "BBTZh0N9Ne1OD7GBGJYz4PNESHuXezAfBgNVHSMEGDAWgBTZh0N9Ne1OD7GBGJYz\n"
        "4PNESHuXezAPBgNVHRMBAf8EBTADAQH/MA0GCSqGSIb3DQEBCwUAA4IBAQCmTJVT\n"
        "LH5Cru1vXtzb3N9dyolcVH82xFVwPewArchgq+CEkajOU9bnzCqvhM4CryBb4cUs\n"
        "gqXWp85hAh55uBOqXb2yyESEleMCJEiVTwm/m26FdONvEGptsiCmF5Gxi0YRtn8N\n"
        "V+KhrQaAyLrLdPYI7TrwAOisq2I1cD0mt+xgwuv/654Rl3IhOMx+fKWKJ9qLAiaE\n"
        "fQyshjlPP9mYVxWOxqctUdQ8UnsUKKGEUcVrA08i1OAnVKlPFjKBvk+r7jpsTPcr\n"
        "9pWXTO9JrYMML7d+XRSZA1n3856OqZDX4403+9FnXCvfcLZLLKTBvwwFgEFGpzjK\n"
        "UEVbkhd5qstF6qWK\n"
        "-----END CERTIFICATE-----\n";

    std::string const key =
        "-----BEGIN PRIVATE KEY-----\n"
        "MIIEvgIBADANBgkqhkiG9w0BAQEFAASCBKgwggSkAgEAAoIBAQDJ7BRKFO8fqmsE\n"
        "Xw8v9YOVXyrQVsVbjSSGEs4Vzs4cJgcFxqGitbnLIrOgiJpRAPLy5MNcAXE1strV\n"
        "GfdEf7xMYSZ/4wOrxUyVw/Ltgsft8m7bFu8TsCzO6XrxpnVtWk506YZ7ToTa5UjH\n"
        "fBi2+pWTxbpN12UhiZNUcrRsqTFW+6fO9d7xm5wlaZG8cMdg0cO1bhkz45JSl3wW\n"
        "KIES7t3EfKePZbNlQ5hPy7Pd5JTmdGBpyY8anC8u4LPbmgW0/U31PH0rRVfGcBbZ\n"
        "sAoQw5Tc5dnb6N2GEIbq3ehSfdDHGnrvenu2tOK9Qx6GEzXh3sekZkxcgh+NlIxC\n"
        "Nxu//Dk9AgMBAAECggEBAK1gV8uETg4SdfE67f9v/5uyK0DYQH1ro4C7hNiUycTB\n"
        "oiYDd6YOA4m4MiQVJuuGtRR5+IR3eI1zFRMFSJs4UqYChNwqQGys7CVsKpplQOW+\n"
        "1BCqkH2HN/Ix5662Dv3mHJemLCKUON77IJKoq0/xuZ04mc9csykox6grFWB3pjXY\n"
        "OEn9U8pt5KNldWfpfAZ7xu9WfyvthGXlhfwKEetOuHfAQv7FF6s25UIEU6Hmnwp9\n"
        "VmYp2twfMGdztz/gfFjKOGxf92RG+FMSkyAPq/vhyB7oQWxa+vdBn6BSdsfn27Qs\n"
        "bTvXrGe4FYcbuw4WkAKTljZX7TUegkXiwFoSps0jegECgYEA7o5AcRTZVUmmSs8W\n"
        "PUHn89UEuDAMFVk7grG1bg8exLQSpugCykcqXt1WNrqB7x6nB+dbVANWNhSmhgCg\n"
        "VrV941vbx8ketqZ9YInSbGPWIU/tss3r8Yx2Ct3mQpvpGC6iGHzEc/NHJP8Efvh/\n"
        "CcUWmLjLGJYYeP5oNu5cncC3fXUCgYEA2LANATm0A6sFVGe3sSLO9un1brA4zlZE\n"
        "Hjd3KOZnMPt73B426qUOcw5B2wIS8GJsUES0P94pKg83oyzmoUV9vJpJLjHA4qmL\n"
        "CDAd6CjAmE5ea4dFdZwDDS8F9FntJMdPQJA9vq+JaeS+k7ds3+7oiNe+RUIHR1Sz\n"
        "VEAKh3Xw66kCgYB7KO/2Mchesu5qku2tZJhHF4QfP5cNcos511uO3bmJ3ln+16uR\n"
        "GRqz7Vu0V6f7dvzPJM/O2QYqV5D9f9dHzN2YgvU9+QSlUeFK9PyxPv3vJt/WP1//\n"
        "zf+nbpaRbwLxnCnNsKSQJFpnrE166/pSZfFbmZQpNlyeIuJU8czZGQTifQKBgHXe\n"
        "/pQGEZhVNab+bHwdFTxXdDzr+1qyrodJYLaM7uFES9InVXQ6qSuJO+WosSi2QXlA\n"
        "hlSfwwCwGnHXAPYFWSp5Owm34tbpp0mi8wHQ+UNgjhgsE2qwnTBUvgZ3zHpPORtD\n"
        "23KZBkTmO40bIEyIJ1IZGdWO32q79nkEBTY+v/lRAoGBAI1rbouFYPBrTYQ9kcjt\n"
        "1yfu4JF5MvO9JrHQ9tOwkqDmNCWx9xWXbgydsn/eFtuUMULWsG3lNjfst/Esb8ch\n"
        "k5cZd6pdJZa4/vhEwrYYSuEjMCnRb0lUsm7TsHxQrUd6Fi/mUuFU/haC0o0chLq7\n"
        "pVOUFq5mW8p0zbtfHbjkgxyF\n"
        "-----END PRIVATE KEY-----\n";

    std::string const dh =
        "-----BEGIN DH PARAMETERS-----\n"
        "MIIBCAKCAQEArzQc5mpm0Fs8yahDeySj31JZlwEphUdZ9StM2D8+Fo7TMduGtSi+\n"
        "/HRWVwHcTFAgrxVdm+dl474mOUqqaz4MpzIb6+6OVfWHbQJmXPepZKyu4LgUPvY/\n"
        "4q3/iDMjIS0fLOu/bLuObwU5ccZmDgfhmz1GanRlTQOiYRty3FiOATWZBRh6uv4u\n"
        "tff4A9Bm3V9tLx9S6djq31w31Gl7OQhryodW28kc16t9TvO1BzcV3HjRPwpe701X\n"
        "oEEZdnZWANkkpR/m/pfgdmGPU66S2sXMHgsliViQWpDCYeehrvFRHEdR9NV+XJfC\n"
        "QMUk26jPTIVTLfXmmwU0u8vUkpR7LQKkwwIBAg==\n"
        "-----END DH PARAMETERS-----\n";

    set_password_callback([](std::size_t, boost::asio::ssl::context_base::password_purpose) { return "test"; });
    set_options(boost::asio::ssl::context::default_workarounds | boost::asio::ssl::context::no_sslv2 |
                boost::asio::ssl::context::single_dh_use);
    use_certificate_chain(boost::asio::buffer(cert.data(), cert.size()));
    use_private_key(boost::asio::buffer(key.data(), key.size()), boost::asio::ssl::context::file_format::pem);
    use_tmp_dh(boost::asio::buffer(dh.data(), dh.size()));
  }
};

SCENARIO("dsNetWSSClient", TAGS) {
  auto expectClosed = std::unique_ptr<trompeloeil::expectation>{}; // alive till io callback is destroyed
  auto hooks = dsNet::MessageStreamHooksMock{};
  auto loop = ds::asio::TestLoopResource{};
  SSLContext sslContext;
  boost::asio::ip::tcp::acceptor acceptor{loop.io};
  acceptor.open(boost::asio::ip::tcp::endpoint{}.protocol());
  acceptor.bind(boost::asio::ip::tcp::endpoint{});
  acceptor.listen(128);
  auto url = ds::str("wss://127.0.0.1:", acceptor.local_endpoint().port(), "/target");
  auto wssClient = dsNet::makeWSSClient(loop, {.url = url}, sslContext);
  boost::asio::ip::tcp::socket socket{loop.io};
  boost::beast::websocket::stream<boost::asio::ssl::stream<boost::asio::ip::tcp::socket &>> wss{socket, sslContext};

  auto stream = wssClient->asyncConnect();
  stream->asyncRun(hooks.hooks());
  expectClosed = NAMED_REQUIRE_CALL(hooks, closed(_)).LR_SIDE_EFFECT(loop.stop());

  THEN("stream id based on url") { CHECK(stream->id() == ds::asio::ResourceBase::normalizeId(url)); }
  WHEN("client starts connecting stream") {
    WHEN("server accepts") {
      acceptor.async_accept(socket, [&](error_code ec) {
        loop.stop();
        REQUIRE(!ec);
      });
      loop.run();

      THEN("client connect fails when server closes socket") {
        socket.close();
        loop.run();
        expectClosed.reset();
      }

      WHEN("server performs ssl handshake") {
        wss.next_layer().async_handshake(boost::asio::ssl::stream_base::server, [&](error_code ec) {
          REQUIRE(!ec);
          loop.stop();
        });
        loop.run();

        THEN("stream can be destroyed") {}

        THEN("client connect fails when server closes socket") {
          socket.close();
          loop.run();
          expectClosed.reset();
        }

        WHEN("server performs ws handshake") {
          wss.async_accept([&](error_code ec) {
            REQUIRE(!ec);
            loop.stop();
          });
          loop.run();

          THEN("stream can be destroyed") {}

          THEN("client connects") {
            REQUIRE_CALL(hooks, opened()).LR_SIDE_EFFECT(loop.stop(););
            loop.run();
            REQUIRE(stream);

            THEN("stream can be destroyed") {}

            THEN("stream can be destroyed when server closes socket") { socket.close(); }

            THEN("client reads what server writes") {
              stream->asyncRead([&](error_code ec, std::string &&s) {
                REQUIRE(!ec);
                loop.stop();
                REQUIRE(s == "clientReads");
              });
              wss.async_write(boost::asio::buffer("clientReads", 11),
                              [&](error_code ec, std::size_t) { REQUIRE(!ec); });
              loop.run();
            }

            THEN("server reads what client writes") {
              boost::beast::multi_buffer readBuffer;
              wss.async_read(readBuffer, [&](error_code ec, std::size_t) {
                REQUIRE(!ec);
                loop.stop();
                auto message = boost::beast::buffers_to_string(readBuffer.data());
                readBuffer.consume(message.size());
                REQUIRE(message == "clientWrites");
              });
              stream->asyncWrite("clientWrites", [&](error_code ec) {
                REQUIRE(!ec);
                loop.stop();
              });
              loop.run();
              loop.run();
            }

            THEN("stream read fails when stream is closed") {
              stream->asyncRead([&](error_code ec, std::string &&) {
                REQUIRE(ec);
                loop.stop();
              });
              loop.post([&] { stream->close(); });
              loop.run();
            }

            THEN("stream write fails when stream is closed") {
              stream->close();
              loop.post([&] {
                stream->asyncWrite("aa", [&](error_code ec) {
                  REQUIRE(ec);
                  loop.stop();
                });
              });
              loop.run();
            }

            THEN("client read fails when server closes wss") {
              stream->asyncRead([&](error_code ec, std::string &&) {
                REQUIRE(ec);
                loop.stop();
              });
              wss.async_close(boost::beast::websocket::close_code::none, [&](error_code) { loop.stop(); });
              loop.run();
              loop.run();
            }

            THEN("client read fails when server closes socket") {
              stream->asyncRead([&](error_code ec, std::string &&) {
                REQUIRE(ec);
                loop.stop();
              });
              socket.close();
              loop.run();
            }
          }
        }
      }
    }
  }
}

} // namespace
