/*
 *  Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <ds/log.h>
#include "timer.h"

namespace ds::asio {

using namespace std::literals;

/// Steady timer schedulling reconnects maintaining long lived connections.
///
/// The timer uses exponential backoff to schedule reconnects.
/// First reconnect is immediate.
///
/// Default reconnect intervals shall be suitable for internet use cases.
/// intervals: 0s, 2s, 5s, ... 20m, 20m, 20m, ...
/// resetIntervalsAfterConnected: 1min
///
class ReconnectTimer {
public:
  ReconnectTimer(boost::asio::io_service& io);
  ~ReconnectTimer();

  template <class F>
  void asyncDelayedReconnect(F&& f) {
    m_timer.randomlyExpiresFromNowPercentDown(raiseInterval(), 20, [f = std::forward<F>(f)] { f(); });
  }

  /// Connection is connected. Intervals will be reset if the connection is connected long enough
  void connected();

  /// Reset the reconnect interval to the first one. Next reconnect will happen fast.
  void resetInterval();

  /// Raise interval to next one unless the interval is at the maximum.
  Timer::duration raiseInterval();

  /// Raise interval up to `duration` or up to maximum interval.
  void raiseIntervalUpTo(Timer::duration duration);

  Timer::duration interval() { return *m_intervalsIt; }
  bool isLastInterval() { return std::next(m_intervalsIt) == m_intervals.cend(); }

  std::vector<Timer::duration> intervals() const { return m_intervals; }
  Timer::duration resetIntervalsAfterConnectedDuration() const { return m_resetIntervalsAfterConnectedDuration; }

private:
  Timer m_timer;
  const std::vector<Timer::duration> m_intervals{0s, 2s, 5s, 20s, 1min, 5min, 20min};
  const Timer::duration m_resetIntervalsAfterConnectedDuration{1min};
  decltype(m_intervals)::const_iterator m_intervalsIt{m_intervals.cbegin()};
  std::optional<Timer::time_point> m_connectedAt{};
};

} // namespace ds::asio
