/*
    Copyright (c) 2019 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <ds/conf.h>
#include <ds/system.h>

#include <boost/asio/io_service.hpp>
#include <future>
#include <string>
#include <string_view>

#include "resource-fwd.h"

namespace ds::asio {

extern ds::LogChannel& resourceLogChannel;

/// dispatches to io, blocks and returns result
template <class F>
auto invoke(boost::asio::io_service& io, F&& f) {
  using Result = std::decay_t<decltype(f())>;
  auto task = std::packaged_task<Result(void)>{[&] { return f(); }};
  io.dispatch([&, logContext = ds::logContext()] {
    DS_CONTEXT(ds::hideArg(logContext));
    task();
  });
  return task.get_future().get();
}

/// Allows to set / get named properties from base ResourceBase class.
struct ResourceProperty {
  gsl::czstring id;
  gsl::czstring description;
  ds::JsonFragment (*value)(const ResourceBase&);
  bool test = false; // is the resource available in test api?
  ds::JsonFragment (*defaultValue)(const ResourceBase&) = nullptr;
  void (*setValue)(ResourceBase&, ds::JsonFragment&&) = nullptr;
};

/// Allows to set / get named properties from base ResourceBase class.
struct ResourceMethod {
  gsl::czstring id;
  gsl::czstring description;
  void (*invoke)(ResourceBase&);
  bool test = false; // is the method available in test api?
};

/// Provides access inside the resource for investigattion.
struct ResourceInside {
  gsl::czstring description;
  std::initializer_list<ResourceProperty> properties;
  std::initializer_list<ResourceMethod> methods;
};

/// Instance must live longer than io passed to ResourceContext
///
/// TODO(someday): cleanup somehow
struct ResourceContextData {
  std::vector<std::reference_wrapper<ResourceBase>> resources;
};

/// Provides
/// - event loop io
/// - system
/// - register of resources for inside api
class ResourceContext {
  ResourceContextData& m_data;

public:
  boost::asio::io_service& io;
  ds::System& system;
  std::thread::id threadId;

  ResourceContext(ResourceContextData&, boost::asio::io_service&, ds::System&, std::thread::id);
  ResourceContext(ResourceContextData&, boost::asio::io_service&, ds::System&);
  ~ResourceContext();

  /// Is execution synchronized to resource thread?
  bool isSynchronized() const;
  void assertSynchronized() const;

  /// Is blocking allowed in current context?
  ///
  /// Blocking is allowed e.g. in thread pool thread.
  /// Blocking is not allowed e.g. in resource thread.
  bool isBlockingAllowed() const;
  void assertBlockingAllowed() const;

  template <class F>
  auto invoke(F&& f) const {
    if (isSynchronized()) {
      return f();
    } else {
      return ds::asio::invoke(io, std::forward<F>(f));
    }
  }

  ResourceBase& root() { return *m_root; }
  operator ResourceBase&() { return root(); }

  const std::vector<std::reference_wrapper<ResourceBase>>& resources();

private:
  const std::unique_ptr<ResourceBase> m_root;

  void addResource(ResourceBase&);
  void removeResource(ResourceBase&);

  friend class ResourceBase;
};

/// Global registry of resource context.
///
/// No synchronization, no removal. It is populated at startup and then constant.
std::vector<std::reference_wrapper<ResourceContext>>& resourceContexts();

/// resource context for resource path
ds::asio::ResourceContext& resourceContext(std::string_view path);

/// virtual interface for ResourceBase
struct IResourceBase {
  IResourceBase() = default;
  IResourceBase(const IResourceBase&) = delete;
  IResourceBase& operator=(const IResourceBase&) = delete;
  virtual ~IResourceBase() = default;

  virtual ResourceContext& context() const = 0;
  virtual const std::string& path() const = 0;
  virtual const std::string& id() const = 0;
  boost::asio::io_service& io() const { return context().io; }
  ds::System& system() const { return context().system; }
  ds::Conf& conf() const { return context().system.conf(); }
  virtual const ResourceInside& inside();
  virtual ds::JsonFragment jsonConfig() const;

  bool isSynchronized() const { return context().isSynchronized(); }
  void assertSynchronized() const;
  bool isBlockingAllowed() const { return context().isBlockingAllowed(); }
  void assertBlockingAllowed() const;

  template <class F>
  auto invoke(F&& f) const {
    return context().invoke(std::forward<F>(f));
  }

  // local ids of children
  std::vector<std::string> childIds() const;

  friend std::ostream& operator<<(std::ostream&, const IResourceBase&);
};

/// Resource is base for is Resource<Config> entity class that
///
/// * builds tree hierarchy
/// * identified by absolute path (e.g. '/model/zones')
/// * carries context with io for scheduling and synchronization
/// * carries context with mockable ds::System
/// * can be configured by type specific config in ds::Conf
/// * is registerd in context
///
class ResourceBase : public virtual IResourceBase {
public:
  // Child resource inheriting resource context from parent
  ResourceBase(ResourceBase& parent, std::string&& id);
  ~ResourceBase() override;

  ResourceContext& context() const final { return m_context; }
  const std::string& path() const final { return m_path; }
  const std::string& id() const final { return m_id; }

  static bool isIdNormalized(const std::string& id);
  static std::string normalizeId(const std::string& id);
  static std::string composePath(std::string_view parentPath, const std::string& id);

private:
  ResourceContext& m_context;
  const std::string m_path;
  const std::string m_id;

  struct Private {};
  // Root resource
  ResourceBase(Private, ResourceContext& context, std::string&& path, std::string&& id);
  friend class ResourceContext;
};

template <class Config>
struct IResource {
  IResource() = default;
  IResource(const IResource&) = delete;
  IResource& operator=(const IResource&) = delete;
  virtual ~IResource() = default;
  virtual const Config& config() const = 0;
};

struct EmptyResourceConfig {
  bool dummy;
};

template <class Config>
class Resource : public ResourceBase, virtual public IResource<Config> {
public:
  Resource(ResourceBase& parent, std::string&& id, Config&& defaultConfig = {})
      : ResourceBase(parent, std::move(id)), m_config{readConfigOr(std::move(defaultConfig))} {
    DS_CHANNEL_INFO_ENTER(ds::asio::resourceLogChannel, path(), config());
  }

  const Config& config() const override final { return m_config; }
  ds::JsonFragment jsonConfig() const override final { return ds::toJsonFragment(m_config); }

private:
  const Config m_config{};

  Config readConfigOr(Config&& defaultConfig) {
    return conf().template readOr<Config>(path(), std::move(defaultConfig));
  }
};

} // namespace ds::asio

DS_JSON_DECLARE_STRUCT((ds::asio), EmptyResourceConfig, (dummy));
