/*
    Copyright (c) 2020 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <memory>
#include <utility>
#include <vector>

#include "str.h"

namespace ds {

/// Immutable box for a single value of type T (never null).
/// Since a box is immutable, copying or moving is
/// * fast -> manipulates only reference counted pointers
/// * thread safe -> sharing const data is safe
///
/// The box is always copiable and movable.
/// The T copy or move operations are never called.
///
/// Implemented by std::shared_ptr<const T> under the hood.
/// Inspired by [immer](https://sinusoid.es/immer/index.html)
template <class T>
class Box : private std::shared_ptr<const T> {
  using Ptr = std::shared_ptr<const T>;

public:
  using Value = T;

  /// construct from T.
  /// E.g.
  /// - Box{7.7d}
  /// - Box{std::vector<int>{1,2,3}}
  Box(T&& x) : Ptr{std::make_shared<T>(std::move(x))} {}

  /// assign from T&&
  Box& operator=(T&& x) {
    *static_cast<Ptr*>(this) = std::make_shared<T>(std::move(x));
    return *this;
  }

  // copy/move constructors/assignments generated by compiler

  /// implicit cast to const T&
  operator const T&() const { return *static_cast<const Ptr&>(*this); }

  /// accessors from std::shared_ptr<const T>
  using Ptr::get;
  using Ptr::operator*;
  using Ptr::operator->;

  friend void swap(Box<T>& a, Box<T>& b) noexcept { std::swap(static_cast<Ptr&>(a), static_cast<Ptr&>(b)); }

  // comparison with other box<T>, T, anything comparable to T
  friend bool operator==(const Box<T>& a, const Box<T>& b) { return *a == *b; }
  friend bool operator!=(const Box<T>& a, const Box<T>& b) { return *a != *b; }
  template <typename U, typename std::enable_if_t<!std::is_base_of_v<Box<T>, U>, bool> = false>
  friend bool operator==(const Box<T>& a, const U& b) {
    return *a == b;
  }
  template <typename U, typename std::enable_if_t<!std::is_base_of_v<Box<T>, U>, bool> = false>
  friend bool operator!=(const Box<T>& a, const U& b) {
    return *a != b;
  }
  template <typename U, typename std::enable_if_t<!std::is_base_of_v<Box<T>, U>, bool> = false>
  friend bool operator==(const U& a, const Box<T>& b) {
    return a == *b;
  }
  template <typename U, typename std::enable_if_t<!std::is_base_of_v<Box<T>, U>, bool> = false>
  friend bool operator!=(const U& a, const Box<T>& b) {
    return a != *b;
  }
  // trompeloeil used to require this operator
  friend bool operator==(const Box<T>&, std::nullptr_t) { return false; }

  friend std::ostream& operator<<(std::ostream& s, const Box<T>& x) {
    ds::ostream(s, x);
    return s;
  }

  /// Mutable copy
  T unbox() const { return *this; }
};

namespace detail {
template <class T>
struct Str<ds::Box<T>> {
  static void f(std::ostream& ostream, const ds::Box<T>& x) { Str<std::decay_t<decltype(*x)>>::f(ostream, *x); }
};
} // namespace detail

} // namespace ds
