/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "common.h"

#include <ds/catch/catch.h>
#include <stdexcept>
#include "log.h"

static const char *TAGS = "[dsCommon][ds]";

namespace {
class Base {
public:
  virtual ~Base();
  virtual void virtualFoo();
};

TEST_CASE("DS_CONCAT", TAGS) {
  SECTION("concatenates two identifiers into one") {
    int ab = 7;
    CHECK(DS_CONCAT(a, b) == 7);
  }
}

TEST_CASE("DS_UNIQUE_NAME", TAGS) {
  SECTION("generates the same identifiers in the same line / macro") {
// Macro to force compilation into one line
#define MACRO                        \
  do {                               \
    auto DS_UNIQUE_NAME(foo) = 7;    \
    CHECK(DS_UNIQUE_NAME(foo) == 7); \
    DS_UNIQUE_NAME(foo) = 8;         \
    CHECK(DS_UNIQUE_NAME(foo) == 8); \
  } while (0)
    MACRO;
#undef MACRO
  }

  SECTION("generates unique identifiers on different lines") {
#define MACRO                   \
  auto DS_UNIQUE_NAME(foo) = 7; \
  CHECK(DS_UNIQUE_NAME(foo) == 7);

    MACRO;
    MACRO;
#undef MACRO
  }
}

namespace {
struct ExpectedException {};
}

TEST_CASE("DS_DEFER", TAGS) {
  SECTION("captures by reference and is called at the end of scope") {
    int calledCount = 0;
    try {
      DS_DEFER(calledCount++);
      throw ExpectedException();
    } catch (const ExpectedException &) {
    }
    CHECK(calledCount == 1);
  }
}

TEST_CASE("DS_IREQUIRE", TAGS) {
  SECTION("does nothing on success") { DS_IREQUIRE(true, ""); }

  SECTION("throws on failure") { CHECK_THROWS_FIND(DS_IREQUIRE(false, "Foo"), "Foo"); }
}

TEST_CASE("ds::araySize", TAGS) {
  int x0[] = {};
  int x5[] = {1, 2, 3, 4, 5};
  int x23[2][3];
  CHECK(ds::arraySize(x0) == 0);
  CHECK(ds::arraySize(x5) == 5);
  CHECK(ds::arraySize(x23) == 2);
  CHECK(ds::arraySize<0>(x23) == 2);
  CHECK(ds::arraySize<1>(x23) == 3);
}

TEST_CASE("ds::IsRange", TAGS) {
  static_assert(!ds::IsRange<int>::value, "Failed");
  static_assert(ds::IsRange<std::string>::value, "Failed");
  static_assert(ds::IsRange<std::vector<int>>::value, "Failed");
  static_assert(ds::IsRange<std::array<int, 4>>::value, "Failed");
}

} // namespace
