#include "http.h"

#include <boost/algorithm/string/predicate.hpp>
#include <boost/range/algorithm.hpp>
#include <iomanip>

#include "log.h"

namespace ds {

std::string urlDecode(gsl::cstring_span in) {
  DS_CONTEXT(in);
  std::string result{};
  result.reserve(in.size());

  auto hex2Int = [](char c) {
    if ((c >= '0') && (c <= '9'))
      return (c - '0');
    if ((c >= 'A') && (c <= 'F'))
      return (c - 'A' + 10);
    if ((c >= 'a') && (c <= 'f'))
      return (c - 'a' + 10);
    return 0;
  };

  for (auto it = in.begin(); it != in.end(); ++it) {
    const auto& c = *it;
    if (c == '+') {
      result.push_back(' ');
    } else if (c == '%') {
      if (++it == in.end()) {
        break;
      }
      auto x = hex2Int(*it);
      if (++it == in.end()) {
        break;
      }
      result.push_back(x * 16 + hex2Int(*it));
    } else {
      result.push_back(c);
    }
  }
  return result;
}

// from https://stackoverflow.com/questions/154536/encode-decode-urls-in-c/17708801
std::string urlEncode(std::string_view value) {
  std::ostringstream escaped;
  escaped.fill('0');
  escaped << std::hex;

  for (std::string_view::const_iterator i = value.begin(), n = value.end(); i != n; ++i) {
    std::string::value_type c = (*i);

    // Keep alphanumeric and other accepted characters intact
    if (isalnum(c) || c == '-' || c == '_' || c == '.' || c == '~') {
      escaped << c;
      continue;
    }

    // Any other characters are percent-encoded
    escaped << std::uppercase;
    escaped << '%' << std::setw(2) << int((unsigned char)c);
    escaped << std::nouppercase;
  }

  return escaped.str();
}

std::optional<std::string> tryHttpQueryArg(gsl::cstring_span query, gsl::cstring_span nameIn) {
  auto name = gsl::to_string(nameIn);
  gsl::cstring_span nameMatch{};
  name.append("=");
  if (boost::starts_with(query, name)) {
    nameMatch = query;
  } else {
    name.insert(0, "&");
    if (auto foundEnd = boost::search<boost::return_found_end>(query, name)) {
      nameMatch = {foundEnd.begin(), foundEnd.end()};
    }
  }

  if (nameMatch.empty()) {
    return {};
  }

  gsl::cstring_span valueMatch = nameMatch.subspan(name.length());
  auto value = gsl::cstring_span{valueMatch.begin(), boost::find<boost::return_found>(valueMatch, '&')};

  return urlDecode(value);
}

std::string httpQueryArg(gsl::cstring_span query, gsl::cstring_span name) {
  if (auto result = tryHttpQueryArg(query, name)) {
    return std::move(*result);
  }
  DS_FAIL_REQUIRE("Http query argument not found.", name, query);
}

} // namespace ds
