/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "ds-locale.h"
#include <ds/catch/catch.h>
#include "backend.h"

static const char* TAGS = "[dsLocale][ds]";

namespace {

TEST_CASE("dsLocaleLanguage", TAGS) {
  SECTION("with poPo locale") {
    boost::locale::generator generator(ds::locale::backend());
    auto locale = generator("po_PO");
    SECTION("ds::locale::language returns langugage passed to generator") {
      CHECK(ds::locale::language(locale) == "po_PO");
    }
    SECTION("ds::locale::httpLanguage returns langugage passed to generator in http format") {
      CHECK(ds::locale::httpLanguage(locale) == "po-PO");
    }
    SECTION("ds::str returns meaningful description") { CHECK(ds::str(locale) == "po_PO"); }
  }
  SECTION("with default constructed locale") {
    auto locale = std::locale();
    SECTION("ds::locale::language returns en_US") { CHECK(ds::locale::language(locale) == "en_US"); }
    SECTION("ds::locale::httpLanguage returns langugage passed to generator in http format") {
      CHECK(ds::locale::httpLanguage(locale) == "en-US");
    }
    SECTION("ds::str returns meaningful description") { CHECK(ds::str(locale) == "en_US"); }
  }
}

TEST_CASE("dsLocaleTr", TAGS) {
  boost::locale::generator generator(ds::locale::backend());
  // generator.set_default_messages_domain(...) is buggy and crashes on boost_1.5.8
  generator.add_messages_domain("ds-locale-messages-test");
  // DS_LOCALE_SRC_DIR is passed as -D... by build system
  generator.add_messages_path(ds::str(DS_LOCALE_SRC_DIR, "/test"));
  auto locale = generator("de_DE");

  CHECK(dsTr(locale, "welcomeId") == "Willkommen");
  CHECK(dsTr(locale, "context1", "welcomeId") == "Willkommen1");
  CHECK(dsTr(locale, "context2", "welcomeId") == "Willkommen2");
  CHECK(dsTr(locale, "itemId", "itemIds", 1) == "Stück");
  CHECK(dsTr(locale, "itemId", "itemIds", 2) == "Stücke");
  CHECK(dsTr(locale, "context1", "itemId", "itemIds", 1) == "Stück1");
  CHECK(dsTr(locale, "context1", "itemId", "itemIds", 2) == "Stücke1");
  CHECK(dsTr(locale, "context2", "itemId", "itemIds", 1) == "Stück2");
  CHECK(dsTr(locale, "context2", "itemId", "itemIds", 2) == "Stücke2");

  std::string domain = "ds-locale-messages-test1-en,ds-locale-messages-test2-de";
  CHECK(dsDtr(locale, domain, "welcomeId") == "welcomeId");
  CHECK(dsDtr(locale, domain, "context1", "welcomeId") == "Welcome1-1");
  CHECK(dsDtr(locale, domain, "context2", "welcomeId") == "Willkommen2-2");
  CHECK(dsDtr(locale, domain, "itemId", "itemIds", 1) == "itemId");
  CHECK(dsDtr(locale, domain, "itemId", "itemIds", 2) == "itemIds");
  CHECK(dsDtr(locale, domain, "context1", "itemId", "itemIds", 1) == "Item1-1");
  CHECK(dsDtr(locale, domain, "context1", "itemId", "itemIds", 2) == "Items1-1");
  CHECK(dsDtr(locale, domain, "context2", "itemId", "itemIds", 1) == "Stück2-2");
  CHECK(dsDtr(locale, domain, "context2", "itemId", "itemIds", 2) == "Stücke2-2");

  SECTION("dsApiTranslatable without domain is raw message") {
    CHECK(dsTr(locale, dsApi::Translatable{.domain = {}, .message = "welcome"}) == "welcome");
  }
  SECTION("dsApiTranslatable with domain") {
    CHECK(dsTr(locale, dsApi::Translatable{.domain = domain, .message = "welcomeId"}) == "welcomeId");
  }
  SECTION("dsApiTranslatable with domain and context") {
    CHECK(dsTr(locale, dsApi::Translatable{.domain = domain, .message = "welcomeId", .context = "context1"}) ==
          "Welcome1-1");
  }
}

TEST_CASE("dsLocaleSanitize", TAGS) {
  CHECK(ds::locale::sanitizeLanguage("") == "en_US");
  CHECK(ds::locale::sanitizeLanguage("en_US") == "en_US");
  CHECK(ds::locale::sanitizeLanguage("de_DE") == "de_DE");
  CHECK(ds::locale::sanitizeLanguage("de_CH") == "de_DE");
  CHECK(ds::locale::sanitizeLanguage("cn_CN") == "en_US");
}

} // namespace
