/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#include "messages.h"
#include <ds/catch/catch.h>
#include <ds/log.h>
#include <boost/locale/gnu_gettext.hpp>
#include <boost/locale/message.hpp>

static const char* TAGS = "[dsLocale][ds]";

namespace {

TEST_CASE("dsLocaleMessages", TAGS) {
  ds::locale::MessagesInfo info;
  info.language = "de_DE";
  // DS_LOCALE_SRC_DIR is passed as -D... by build system
  info.paths = {ds::str(DS_LOCALE_SRC_DIR, "/test")};
  info.domains = {std::string("ds-locale-messages-test")};

  auto locale = std::locale(std::locale::classic(), ds::locale::makeMessages(info));
  auto& messages = std::use_facet<ds::locale::Messages>(locale);

  SECTION("Domain in info structure is default domain") { CHECK(messages.domain("ds-locale-messages-test") == 0); }

  SECTION("Domains not in info structure are added") { CHECK(messages.domain("ds-locale-messages-test1") == 1); }

  SECTION("Get with de and en translation returns de") {
    auto&& domainId = messages.domain("ds-locale-messages-test");
    CHECK(ds::str(messages.get(domainId, nullptr, "welcomeId")) == "Willkommen");
    CHECK(ds::str(messages.get(domainId, "context1", "welcomeId")) == "Willkommen1");
    CHECK(ds::str(messages.get(domainId, "context2", "welcomeId")) == "Willkommen2");
    CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 1)) == "Stück");
    CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 2)) == "Stücke");
    CHECK(ds::str(messages.get(domainId, "context1", "itemId", 1)) == "Stück1");
    CHECK(ds::str(messages.get(domainId, "context1", "itemId", 2)) == "Stücke1");
    CHECK(ds::str(messages.get(domainId, "context2", "itemId", 1)) == "Stück2");
    CHECK(ds::str(messages.get(domainId, "context2", "itemId", 2)) == "Stücke2");
  }

  SECTION("Get with de translation returns de") {
    auto&& domainId = messages.domain("ds-locale-messages-test-de");
    CHECK(ds::str(messages.get(domainId, nullptr, "welcomeId")) == "Willkommen");
    CHECK(ds::str(messages.get(domainId, "context1", "welcomeId")) == "Willkommen1");
    CHECK(ds::str(messages.get(domainId, "context2", "welcomeId")) == "Willkommen2");
    CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 1)) == "Stück");
    CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 2)) == "Stücke");
    CHECK(ds::str(messages.get(domainId, "context1", "itemId", 1)) == "Stück1");
    CHECK(ds::str(messages.get(domainId, "context1", "itemId", 2)) == "Stücke1");
    CHECK(ds::str(messages.get(domainId, "context2", "itemId", 1)) == "Stück2");
    CHECK(ds::str(messages.get(domainId, "context2", "itemId", 2)) == "Stücke2");
  }

  SECTION("Get with en translation returns en") {
    auto&& domainId = messages.domain("ds-locale-messages-test-en");
    CHECK(ds::str(messages.get(domainId, nullptr, "welcomeId")) == "Welcome");
    CHECK(ds::str(messages.get(domainId, "context1", "welcomeId")) == "Welcome1");
    CHECK(ds::str(messages.get(domainId, "context2", "welcomeId")) == "Welcome2");
    CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 1)) == "Item");
    CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 2)) == "Items");
    CHECK(ds::str(messages.get(domainId, "context1", "itemId", 1)) == "Item1");
    CHECK(ds::str(messages.get(domainId, "context1", "itemId", 2)) == "Items1");
    CHECK(ds::str(messages.get(domainId, "context2", "itemId", 1)) == "Item2");
    CHECK(ds::str(messages.get(domainId, "context2", "itemId", 2)) == "Items2");
  }

  SECTION("Get without translation returns nullptr") {
    auto&& domainId = messages.domain("ds-locale-messages-test9");
    CHECK(ds::str(messages.get(domainId, nullptr, "welcomeId")) == "null");
    CHECK(ds::str(messages.get(domainId, "context1", "welcomeId")) == "null");
    CHECK(ds::str(messages.get(domainId, "context2", "welcomeId")) == "null");
    CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 1)) == "null");
    CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 2)) == "null");
    CHECK(ds::str(messages.get(domainId, "context1", "itemId", 1)) == "null");
    CHECK(ds::str(messages.get(domainId, "context1", "itemId", 2)) == "null");
    CHECK(ds::str(messages.get(domainId, "context2", "itemId", 1)) == "null");
    CHECK(ds::str(messages.get(domainId, "context2", "itemId", 2)) == "null");
  }

  SECTION("Get with sequence of domains searches whole sequence in de and then in en") {
    SECTION("0 1-de 2-en") {
      auto&& domainId =
          messages.domain("ds-locale-messages-test0,ds-locale-messages-test1-de,ds-locale-messages-test2-en");
      CHECK(ds::str(messages.get(domainId, nullptr, "welcomeId")) == "Willkommen-0");
      CHECK(ds::str(messages.get(domainId, "context1", "welcomeId")) == "Willkommen1-1");
      CHECK(ds::str(messages.get(domainId, "context2", "welcomeId")) == "Welcome2-2");
      CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 1)) == "Stück-0");
      CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 2)) == "Stücke-0");
      CHECK(ds::str(messages.get(domainId, "context1", "itemId", 1)) == "Stück1-1");
      CHECK(ds::str(messages.get(domainId, "context1", "itemId", 2)) == "Stücke1-1");
      CHECK(ds::str(messages.get(domainId, "context2", "itemId", 1)) == "Item2-2");
      CHECK(ds::str(messages.get(domainId, "context2", "itemId", 2)) == "Items2-2");
    }

    SECTION("- 1-en 2-de") {
      auto&& domainId =
          messages.domain("ds-locale-messages-test9,ds-locale-messages-test1-en,ds-locale-messages-test2-de");
      CHECK(ds::str(messages.get(domainId, nullptr, "welcomeId")) == "null");
      CHECK(ds::str(messages.get(domainId, "context1", "welcomeId")) == "Welcome1-1");
      CHECK(ds::str(messages.get(domainId, "context2", "welcomeId")) == "Willkommen2-2");
      CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 1)) == "null");
      CHECK(ds::str(messages.get(domainId, nullptr, "itemId", 2)) == "null");
      CHECK(ds::str(messages.get(domainId, "context1", "itemId", 1)) == "Item1-1");
      CHECK(ds::str(messages.get(domainId, "context1", "itemId", 2)) == "Items1-1");
      CHECK(ds::str(messages.get(domainId, "context2", "itemId", 1)) == "Stück2-2");
      CHECK(ds::str(messages.get(domainId, "context2", "itemId", 2)) == "Stücke2-2");
    }
  }
}

} // namespace
