#pragma once

#include <boost/locale/gnu_gettext.hpp>
#include <boost/locale/localization_backend.hpp>
#include <boost/locale/message.hpp>

namespace ds {
namespace locale {

typedef boost::locale::gnu_gettext::messages_info MessagesInfo;

/// Makes message facet compatible with `boost::locale`.
///
/// Example assuming translation files in /usr/share/locale/de/LC_MESSAGES/default-domain.po:
///
///     ds::locale::MessageInfo info;
///     info.language = "de";
///     info.domains = { std::string("default-domain") };
///     auto locale = std::locale(std::locale::classic(), ds::locale::makeMessage(info));
///
///     boost::locale::translate("Hello").str(locale); // default domain
///     boost::locale::translate("Hello2").str(locale, "domain2"); // specific domain
///     boost::locale::translate("Hello2").str(locale, "default-domain,domain2"); // domains hierarchy
///
/// Domains are loaded on demand.
///
/// Messages are translated into target language with fail back to `en` language.
///
/// Domain names with `,` inside the name are hierarchical domains.
/// Hierarchical domains are split to sub-domains.
/// Messages are translated into the first sub-domain with existing translation.
///

class Messages : public boost::locale::message_format<char> {
public:
  virtual const std::string& language() const = 0;

protected:
  // std::locale is the only one who can own and delete facets
  ~Messages();
};

/// Intended to be passed to std::locale constructor that takes its ownership.
Messages* makeMessages(const MessagesInfo& info);

// TODO(someday): Consider to provide own `boost::locale::message`.
//
// `boost::locale::message` does not provide `operator ==`. Missed in unit tests.
// `ds::locale::test::makeLocale` is a workaround.
//
// `boost::locale::message` provides implicit cast to `std::string`. Not wanted.
//
// `boost::locale::message` could provide `empty()` or `explicit operator bool()`.
// Useful on some places.

} // namespace locale
} // namespace ds
