/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/
#pragma once

#include <gsl.h>

#include <array>
#include <boost/system/error_code.hpp>
#include <chrono>
#include <filesystem>
#include <memory>
#include <optional>
#include <sstream>
#include <variant>

#include "common.h"
#include "str-fwd.h"

namespace google {
namespace protobuf {
class Message; // Avoid linking to protobuf
} // namespace protobuf
} // namespace google

namespace ds {

/// Concatenate all arguments into string using std::ostream operator <<
/// Example: ds:str("int:", 7, " double:", 7.7, " char:", '7'); // == "int:7 double:7.7 char:7"
template <typename... Args>
std::string str(Args&&... args);

/// Convert string in decimal (10 base) representation to int.
///
/// - 0 prefix is not recognized as 8 base
/// - 0x, 0X prefix is not recognized as 16 base
std::optional<int> tryToInt(const std::string&);
int requireInt(const std::string&);

std::optional<double> tryToDouble(const std::string&);
double requireDouble(const std::string&);

std::string intToString(const long long integer, int base = 10);

// implementation details

namespace detail {

// uint8_t and int8_t shall be serialized as numbers, not characters
template <>
struct Str<unsigned char> {
  static void f(std::ostream& ostream, unsigned char x);
};

template <>
struct Str<signed char> {
  static void f(std::ostream& ostream, signed char x);
};

template <typename T>
struct StrPtr {
  static void f(std::ostream& ostream, const T* x) {
    if (x) {
      Str<std::decay_t<decltype(*x)>>::f(ostream, *x);
    } else {
      ostream << "null";
    }
  }
};

template <typename T>
struct Str<std::unique_ptr<T>> {
  static void f(std::ostream& ostream, const std::unique_ptr<T>& x) {
    StrPtr<std::decay_t<decltype(*x)>>::f(ostream, x.get());
  }
};

template <typename T>
struct Str<std::shared_ptr<T>> {
  static void f(std::ostream& ostream, const std::shared_ptr<T>& x) {
    StrPtr<std::decay_t<decltype(*x)>>::f(ostream, x.get());
  }
};

template <typename T>
struct Str<std::optional<T>> {
  static void f(std::ostream& ostream, const std::optional<T>& x) {
    StrPtr<std::decay_t<decltype(*x)>>::f(ostream, x ? &(*x) : nullptr);
  }
};

template <>
struct Str<std::exception> {
  static void f(std::ostream& ostream, const std::exception& x);
};

template <>
struct Str<boost::system::error_code> {
  static void f(std::ostream& ostream, const boost::system::error_code& x) { ostream << x.message(); }
};

template <class Rep, class Period>
struct Str<std::chrono::duration<Rep, Period>> {
  static void f(std::ostream& ostream, std::chrono::duration<Rep, Period> x) {
    Str<double>::f(ostream, std::chrono::duration_cast<std::chrono::duration<double>>(x).count());
  }
};

template <class Clock, class Duration>
struct Str<std::chrono::time_point<Clock, Duration>> {
  static void f(std::ostream& ostream, std::chrono::time_point<Clock, Duration> x) {
    Str<Duration>::f(ostream, x.time_since_epoch());
  }
};

template <typename Range>
struct Str<Range,
    typename std::enable_if<IsRange<Range>::value && !std::is_convertible<Range, gsl::cstring_span>::value>::type> {
  static inline void f(std::ostream& ostream, const Range& xs) {
    ostream << '[';
    auto first = true;
    for (const auto& x : xs) {
      if (!first) {
        ostream << ',';
      } else {
        first = false;
      }
      Str<typename std::decay<decltype(x)>::type>::f(ostream, x);
    }
    ostream << ']';
  }
};

template <typename T1, typename T2>
struct Str<std::pair<T1, T2>> {
  static inline void f(std::ostream& ostream, const std::pair<T1, T2>& x) {
    ostream << '(';
    Str<typename std::decay<T1>::type>::f(ostream, x.first);
    ostream << ',';
    Str<typename std::decay<T2>::type>::f(ostream, x.second);
    ostream << ')';
  }
};

// Serialize protobuf messages, avoid linking to protobuf library.
template <typename T>
struct Str<T, typename std::enable_if<std::is_base_of<google::protobuf::Message, T>::value>::type> {
  static inline void f(std::ostream& ostream, const T& x) { ostream << x.DebugString(); }
};

// (const char *) nullptr is serialized as null.
// Default behavior is to terminate the stream.
template <>
struct Str<const char*> {
  static inline void f(std::ostream& ostream, const char* x) {
    if (x) {
      ostream << x;
    } else {
      ostream << "null";
    }
  }
};

template <>
struct Str<std::nullptr_t> {
  static void f(std::ostream& ostream, std::nullptr_t) { ostream << "null"; }
};

template <>
struct Str<std::monostate> {
  static void f(std::ostream& ostream, std::monostate x) { Str<std::nullptr_t>::f(ostream, nullptr); }
};

template <class... Types>
struct Str<std::variant<Types...>> {
  static void f(std::ostream& ostream, const std::variant<Types...>& variant) {
    ostream << '{' << variant.index() << ":";
    std::visit([&](const auto& x) { Str<std::decay_t<decltype(x)>>::f(ostream, x); }, variant);
    ostream << "}";
  }
};

// support for BOOST_STRONG_TYPEDEF(std::string, T)
template <typename T>
struct Str<T, typename std::enable_if_t<std::is_convertible_v<T, std::string>>> {
  static inline void f(std::ostream& ostream, const T& x) { ostream << static_cast<std::string>(x); }
};

template <>
struct Str<std::filesystem::path> {
  static void f(std::ostream& ostream, const std::filesystem::path& path) { ostream << path.string(); }
};

} // namespace detail

template <typename Arg, typename... Args>
void ostream(std::ostream& ostream, Arg&& arg, Args&&... args) {
  (detail::Str<std::decay_t<Args>>::f(ostream, std::forward<Args>(args)), ...);
}

template <typename... Args>
std::string str(Args&&... args) {
  std::ostringstream ostream;
  (detail::Str<std::decay_t<Args>>::f(ostream, std::forward<Args>(args)), ...);
  return ostream.str();
}

} // namespace ds
