/*
 *  Copyright (c) 2018 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once
#include <boost/range/algorithm_ext/push_back.hpp>
#include <boost/range/irange.hpp>
#include <vector>

#include "cache/cache-db.h"
#include "cache/device-cache.h"
#include "ds-model/devices-fwd.h"
#include "ds-model/groups-fwd.h"
#include "ds-model/model-fwd.h"
#include "ds/date-time.h"
#include "dssfwd.h"
#include "model/modelconst.h"
#include "propertysystem.h"

DS_REFLECTED_ENUM((dss), DeviceDimmingStatus, (none)(dimming)(stepping));
DS_REFLECTED_ENUM((dss), SceneProperty, (blink)(dimTime)(dontCare)(ignoreLocalPrio)(ledConfig)(specialHandling));
DS_REFLECTED_ENUM((dss), ChannelProperty, (dontCare));

namespace dss::hal {

class HalDevice;
class HalLightDevice;
class HalShadeDevice;
class HalJokerDevice;
class HalAudioDevice;
class HalClimateDevice;

struct DeviceChannelValues {
  double value;
  double targetValue;
  std::optional<int> level;
};

struct DeviceState {
  ds::DateTime lastChanged;
  std::optional<ds::DateTime> terminatesAt;
  dsModel::OutputStatusStatus status;
  std::map<ChannelType, DeviceChannelValues> outputs;
  bool localPriority;
};

struct DeviceCacheState {
  dsModel::FunctionBlockId id{};
  int total{};
  int completed{};
  double completion{};
};

// TODO hide the internal stuff
struct DeviceInternalOutputState {
  std::optional<double> current;
  std::optional<double> target;
  std::optional<int> level;
};

struct DeviceInternalState {
  ds::DateTime lastChanged;
  std::optional<ds::DateTime> terminatesAt;
  std::map<ChannelType, DeviceInternalOutputState> channels;
  bool localPriority;
  std::optional<int> lastSceneCalled;
  std::optional<int> beforeLastSceneCalled;
  std::optional<int> sceneInRAMBank; // values in RAM bank match non-volatile scene bank, only shades yet
  bool operationLocked;
  std::optional<int> lastDimmCommand;
  int fineDimmSteps;
  DeviceDimmingStatus dimmStatus;
  std::optional<ds::DateTime> dimmingTimeout;
  std::optional<bool> standby;
};

} // namespace dss::hal

namespace dss {

struct DevicePropertyId {
  Dsuid dsuid;
  DeviceProperty property;
};

struct DeviceSceneId {
  Dsuid dsuid;
  int sceneId;
};

struct DeviceScenePropertyId {
  DeviceSceneId scene;
  SceneProperty property;
};

struct DeviceSceneChannelId {
  DeviceSceneId scene;
  ChannelType type;
};

struct DeviceSceneChannelPropertyId {
  DeviceSceneChannelId channel;
  ChannelProperty property;
};

// start legacy route functions
std::string toPath(const DevicePropertyId&);
std::string toPath(const DeviceSceneId&);
std::string toPath(const DeviceScenePropertyId&);
std::string toPath(const DeviceSceneChannelId&);
std::string toPath(const DeviceSceneChannelPropertyId&);
// end legacy route functions

class HardwareAbstraction {
public:
  virtual ~HardwareAbstraction() = default;
  virtual void write(const Dsuid& dsuid, const std::string& parameter, double value) const = 0;
  virtual void write(const Dsuid& dsuid, const std::string& parameter, bool value) const = 0;
  virtual void write(const Dsuid& dsuid, const std::string& parameter, int value) const = 0;
  virtual void write(const Dsuid& dsuid, const std::string& parameter, const std::string& value) const = 0;
  virtual void updateDeviceConfig(const Dsuid& dsuid, int bank, int offset, int value) const = 0;

  virtual double readDouble(const Dsuid& dsuid, const std::string& parameter) const = 0;
  virtual bool readBool(const DevicePropertyId& id) const = 0;
  virtual bool readBool(const DeviceScenePropertyId& id) const = 0;
  virtual bool readBool(const DeviceSceneChannelPropertyId& id) const = 0;
  virtual std::string readString(const Dsuid& dsuid, const std::string& parameter) const = 0;

  virtual bool callScene(const Dsuid& dsuid, int sceneId, bool forced, ds::DateTime now) const = 0;
  virtual bool undoScene(const Dsuid& dsuid, int sceneId, ds::DateTime now) const = 0;
  virtual void saveScene(const Dsuid& dsuid, int sceneId, ds::DateTime now) const = 0;
  virtual bool setOutputValue(const Dsuid& dsuid, uint8_t value, ds::DateTime now) const = 0;
  virtual bool setOutputValue(const Dsuid& dsuid, ChannelType channel, double value, ds::DateTime now) const = 0;
  virtual bool setSceneLocalPrio(const Dsuid& dsuid, int sceneId) const = 0;
  virtual bool callSceneMin(const Dsuid& dsuid, int sceneId, ds::DateTime now) const = 0;
  virtual void setOperationLock(const Dsuid& device, bool lock) const = 0;
  virtual bool stopDimming(const Dsuid& dsuid, int sceneId, ds::DateTime now) const = 0;
  virtual bool completeMoving(const Dsuid& dsuid, ds::DateTime now) const = 0;

  virtual hal::DeviceState getState(const Dsuid& dsuid) const = 0;
  virtual dsModel::ZoneApplicationsStatus getZoneState(int zoneId) const = 0;

  virtual hal::DeviceCacheState cacheStatus(const Dsuid& dsuid) const = 0;
  virtual void populateCache(const Dsuid& dsuid) const = 0; // unit tests

  virtual bool deviceCacheable(const Dsuid& dsuid) const = 0;
  virtual CacheDb<DeviceCache>& getCache() const = 0;
  virtual CacheDb<hal::DeviceInternalState>& getStateCache() const = 0;

  virtual std::vector<int> areaMember(const Dsuid& dsuid) const = 0;

  virtual std::vector<double> getDeviceThresholds(const Dsuid& dsuid) const = 0;
  virtual std::vector<int> getPlayerScenarios(int zoneId) const = 0;

  virtual void publishToPropertyTree(const Dsuid& dsuid) const = 0; // ugly hack in dss::Device
};

std::unique_ptr<HardwareAbstraction> makeHardwareAbstraction(Apartment* apt, ds::sqlite3::Database* db);

} /* namespace dss */


DS_REFLECT_DECLARE_STRUCT((dss), DevicePropertyId, (dsuid)(property));
DS_REFLECT_DECLARE_STRUCT((dss), DeviceSceneId, (dsuid)(sceneId));
DS_REFLECT_DECLARE_STRUCT((dss), DeviceScenePropertyId, (scene)(property));
DS_REFLECT_DECLARE_STRUCT((dss), DeviceSceneChannelId, (scene)(type));
DS_REFLECT_DECLARE_STRUCT((dss), DeviceSceneChannelPropertyId, (channel)(property));

DS_JSON_DECLARE_STRUCT((dss)(hal), DeviceChannelValues, (value)(targetValue)(level));
DS_JSON_DECLARE_STRUCT((dss)(hal), DeviceState, (lastChanged)(terminatesAt)(status)(outputs)(localPriority));
DS_JSON_DECLARE_STRUCT((dss)(hal), DeviceCacheState, (id)(total)(completed)(completion));

// TODO internal stuff should be hidden
DS_JSON_DECLARE_STRUCT(
    (dss)(hal), DeviceInternalState,
    (lastChanged)(channels)(terminatesAt)(localPriority)(lastSceneCalled)(beforeLastSceneCalled)(sceneInRAMBank)(
        operationLocked)(lastDimmCommand)(fineDimmSteps)(dimmStatus)(dimmingTimeout)(standby));
DS_JSON_DECLARE_STRUCT((dss)(hal), DeviceInternalOutputState, (current)(target)(level));
