/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "archives-sanity.h"

#include <ds/conf.h>
#include <ds/log.h>
#include <ds/system.h>

#include <boost/algorithm/string.hpp>

#include "metering/archives-connection.h"
#include "metering/archives-storage.h"
#include "metering/config.h"

static ds::LogChannel dsLogChannel("dssMeteringArchivesSanity");

namespace dss::metering {

using namespace std::literals;
using namespace dss;

namespace {
struct ArchivesSanityImpl final : public ArchivesSanity {
  ds::System& m_system;
  ArchivesConnection& m_archivesConnection;
  ArchivesStorage& m_archivesStorage;
  ElectricityMeteringConfig m_electricityMeteringConfig;

  ~ArchivesSanityImpl() override = default;
  ArchivesSanityImpl(ArchivesConnection& archivesConnection, ArchivesStorage& archivesStorage, ds::System& system)
      : m_system(system),
        m_archivesConnection(archivesConnection),
        m_archivesStorage(archivesStorage),
        m_electricityMeteringConfig(system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {})) {}

  bool withoutUpdates(const std::string& filename) const override {
    auto lastAccess = m_archivesConnection.lastTimestamp(filename);
    if (!lastAccess) {
      // probably invalid file format, rely on later checks to fail as well
      return false;
    }

    // If new sample time is more than MAX_INTERPOLATION_INTERVAL newer than last
    // db entry, it would take too long to roll forward. (more than 15 minutes!)
    // If the the new sample ist older than latest sample in the database,
    // rrd will silently ignore it (ntp update), and no data will be stored
    // until time is going forward again.
    // If the jump into the past is too big discard all data and start over.

    auto now = std::chrono::system_clock::now();
    if (std::chrono::abs(now - *lastAccess) > m_electricityMeteringConfig.maxInterpolationInterval) {
      DS_NOTICE("withoutUpdates", (now - *lastAccess), m_electricityMeteringConfig.maxInterpolationInterval);
      return true;
    }

    return false;
  }

  bool removeDeadArchives() const override {
    bool restartCached{false};
    for (const auto& [compoundId, filename] : m_archivesStorage.searchArchives()) {
      if (withoutUpdates(filename)) {
        DS_WARNING("Metering DB ", filename, " is out of date. removing ");
        m_system.remove(filename);
        restartCached = true;
      }
    }
    return restartCached;
  }

  bool restoreBackups() const override {
    // see https://git.digitalstrom.org/dss11-websetup/dss11-websetup/-/blob/master/pages/cgi-bin/restorebackup.hs
    // MOTIVATION: rrd database are very perishable, if the user uploads an old backup, the last update might be older
    // then Config::maxInterpolationInterval, hence we might just throw it away upon dss startup
    // For that reason dss11-websetup created backups, if files in the backup collided with already present archives.
    // In this case we just rename those and make them the active archives again...

    auto rrdFiles = m_archivesStorage.searchRrdFiles();

    decltype(rrdFiles) backups{};
    std::copy_if(std::cbegin(rrdFiles), std::cend(rrdFiles), std::back_inserter(backups),
                 [](const auto& file) { return boost::ends_with(file.stem().string(), "-backup"s); });

    auto stripBackupPostfix = [](const auto& file) {
      auto stem = file.stem().string();
      auto filename = ds::str(stem.substr(0, stem.size() - "-backup"s.size()), file.extension());
      return file.parent_path() / filename;
    };

    decltype(backups) restoreBackups{};
    std::copy_if(std::cbegin(backups), std::cend(backups), std::back_inserter(restoreBackups),
                 [&rrdFiles, &stripBackupPostfix](const auto& path) {
                   // non-backup version no longer exists
                   return std::none_of(
                       std::cbegin(rrdFiles), std::cend(rrdFiles),
                       [activeRrd = stripBackupPostfix(path)](const auto& elt) { return activeRrd == elt; });
                 });

    std::for_each(
        std::cbegin(restoreBackups), std::cend(restoreBackups),
        [this, &stripBackupPostfix](const auto backup) { m_system.rename(backup, stripBackupPostfix(backup)); });

    return !restoreBackups.empty();
  }

  void migrateFilenames() const override {
    for (auto element : m_archivesStorage.searchArchives()) {
      auto file = std::get<std::string>(element);
      const auto& compoundId = std::get<CompoundId>(element);
      DS_NOTICE(compoundId, "->", file);

      std::visit(
          ds::Visitor{[](const std::monostate&) { DS_FAIL_REQUIRE("monostate"); },
                      [&](const dsid_t& dsid) {
                        auto dsuid = dsuid_from_dsid(&dsid);
                        auto oldFile = std::exchange(file, m_archivesStorage.filepath({dsuid, Category::consumption}));

                        if (m_system.existsFile(file)) {
                          DS_WARNING("Failed to rename old dsid based archive, new file already exists");
                          m_system.remove(oldFile);
                        } else {
                          m_system.rename(oldFile, file);
                          DS_NOTICE("Migrated metering data from ", oldFile, " to ", file);
                        }
                      },
                      [&](const ArchiveId& archiveId) {},
                      [&](const ModbusMeteringId& modbusMeteringId) {
                        DS_NOTICE("not implemented", modbusMeteringId);
                      },
                      [&](const OpenMeteringSystemId& openMeteringSystemId) {
                        DS_NOTICE("not implemented", openMeteringSystemId);
                      }},
          compoundId);
    }
  }
};

} // namespace

std::unique_ptr<ArchivesSanity> makeArchivesSanity(ArchivesConnection& archivesConnection,
                                                   ArchivesStorage& archivesStorage, ds::System& system) {
  return std::make_unique<ArchivesSanityImpl>(archivesConnection, archivesStorage, system);
}

} // namespace dss::metering
