/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "open-metering-system.h"

#include <ds/conf.h>
#include <ds/log.h>
#include <ds/system.h>

#include <regex>

#include "metering/archives-connection.h"
#include "metering/archives-storage.h"
#include "metering/config.h"
#include "metering/readings/open-metering-system.h"
#include "metering/rrd-interface.h"

static ds::LogChannel dsLogChannel("dssMeteringOmsElectricityArchive");

using namespace std::literals;

using namespace dss;
using namespace dss::metering;

namespace dss::metering::archives {

namespace {

using Reading = readings::OpenMeteringSystem;

using std::chrono::system_clock;

class OpenMeteringSystemArchiveImpl final : public OpenMeteringSystemArchive {
  ds::System& m_system;
  ArchivesConnection& m_archivesConnection;
  ArchivesStorage& m_archivesStorage;
  const ElectricityMeteringConfig m_electricityMeteringConfig;
  mutable std::unordered_map<int, int> m_nonFiniteValueHistogram{}; // rate -> count

public:
  OpenMeteringSystemArchiveImpl(ArchivesConnection& archivesConnection, ArchivesStorage& archivesStorage,
                                ds::System& system)
      : m_system(system),
        m_archivesConnection(archivesConnection),
        m_archivesStorage(archivesStorage),
        m_electricityMeteringConfig(m_system.conf().readOr<ElectricityMeteringConfig>("electricityMetering", {})) {}
  virtual ~OpenMeteringSystemArchiveImpl() override = default;

  void create(const OpenMeteringSystemId& openMeteringSystemArchiveId, ds::DateTime startTime) override;
  void addValue(const OpenMeteringSystemId& openMeteringSystemArchiveId, Reading&& value) override;

  std::optional<Reading> lastValue(const OpenMeteringSystemId& openMeteringSystemArchiveId) const override;

  ValueRange getSeries(const OpenMeteringSystemId& openMeteringSystemId, Category category, SeriesTypes seriesType,
                       time_t start, time_t end, int step, EnergyUnit energyUnit) const override;

  ValidationResult validate(const OpenMeteringSystemId& openMeteringSystemId) const override;
  void migrate(const OpenMeteringSystemId& openMeteringSystemArchiveId) override{};

  std::optional<ds::DateTime> lastTimestamp(OpenMeteringSystemId openMeteringSystemId) const override {
    return m_archivesConnection.lastTimestamp(m_archivesStorage.filepath(openMeteringSystemId));
  }

  void flush(std::vector<OpenMeteringSystemId> ids) const override {
    std::vector<std::string> files;
    std::transform(std::cbegin(ids), std::cend(ids), std::back_inserter(files),
                   [&](const auto elt) { return m_archivesStorage.filepath(elt); });
    m_archivesConnection.flush(files);
  }
};

void OpenMeteringSystemArchiveImpl::create(const OpenMeteringSystemId& openMeteringSystemArchiveId,
                                           ds::DateTime startTime) {
  DS_DEBUG("Creating new RRD database", openMeteringSystemArchiveId);

  // counter never decreases, only wrap around, unlike dSM which has jumps/resets upon restart
  std::vector<std::string> lines;
  lines.push_back("DS:powerConsumed:GAUGE:20:0:U");
  lines.push_back("DS:powerProduced:GAUGE:20:0:U");
  // A ds-name must be 1 to 19 characters long in the characters [a-zA-Z0-9_].
  lines.push_back("DS:energyConsumedTari1:DCOUNTER:30:U:U");
  lines.push_back("DS:energyProducedTari1:DCOUNTER:30:U:U");
  lines.push_back("DS:energyConsumedTari2:DCOUNTER:30:U:U");
  lines.push_back("DS:energyProducedTari2:DCOUNTER:30:U:U");

  for (const auto& archiveConfig : m_electricityMeteringConfig.timeSeries) {
    std::stringstream sstream;
    sstream << "RRA:AVERAGE:0.5:" << std::chrono::seconds(archiveConfig.resolution).count() << ":"
            << archiveConfig.numberOfValues;
    lines.push_back(sstream.str());
  }

  m_archivesConnection.rrdInterface().create(m_archivesStorage.filepath(openMeteringSystemArchiveId), 1,
                                             std::chrono::system_clock::to_time_t(startTime), lines);
}

void OpenMeteringSystemArchiveImpl::addValue(const OpenMeteringSystemId& openMeteringSystemArchiveId, Reading&& value) {
  DS_CONTEXT(ds::hideArg(openMeteringSystemArchiveId));
  DS_REQUIRE(openMeteringSystemArchiveId != OpenMeteringSystemId{}, "addValue invalid archive descriptor",
             openMeteringSystemArchiveId);

  std::vector<std::string> lines;
  lines.push_back("update");
  if (auto rrdCachedPath = m_archivesConnection.rrdCachedPath(); !rrdCachedPath.empty()) {
    lines.push_back("--daemon");
    lines.push_back(rrdCachedPath);
  }
  lines.push_back(m_archivesStorage.filepath(openMeteringSystemArchiveId));

  lines.push_back(ds::str(std::chrono::system_clock::to_time_t(value.lastChanged), ":", //
                          value.powerConsumed, ":", value.powerProduced, ":",           //
                          value.t1EnergyConsumed, ":", value.t1EnergyProduced, ":",     //
                          value.t2EnergyConsumed, ":", value.t2EnergyProduced));

  DS_DEBUG(lines);
  m_archivesConnection.rrdInterface().update(lines);
}

std::optional<Reading> OpenMeteringSystemArchiveImpl::lastValue(
    const OpenMeteringSystemId& openMeteringSystemArchiveId) const {
  try {
    auto [lastChanged, data] =
        m_archivesConnection.rrdInterface().lastValue(m_archivesStorage.filepath(openMeteringSystemArchiveId));

    auto ret = Reading{.lastChanged = ds::DateTime{std::chrono::seconds{lastChanged}}};
    for (const auto& [name, value] : data) {
      if (name == "powerConsumed"s) {
        ret.powerConsumed = value;
      } else if (name == "powerProduced") {
        ret.powerProduced = value;
      } else if (name == "energyConsumedTari1") {
        ret.t1EnergyConsumed = value;
      } else if (name == "energyProducedTari1") {
        ret.t1EnergyProduced = value;
      } else if (name == "energyConsumedTari2") {
        ret.t2EnergyConsumed = value;
      } else if (name == "energyProducedTari2") {
        ret.t2EnergyProduced = value;
      } else {
        DS_WARNING("unknown field", name);
        // TODO rely on validateDBSeries to verify that all fields are present
      }
    }
    DS_DEBUG(ret);
    return ret;

  } catch (const std::exception& e) {
    // probably unknown value in db or missing file, conversion to double failed
    DS_NOTICE("lastValue", e);
  }

  return std::nullopt;
}

ValueRange OpenMeteringSystemArchiveImpl::getSeries(const OpenMeteringSystemId& openMeteringSystemId, Category category,
                                                    SeriesTypes seriesType, time_t start, time_t end, int step,
                                                    EnergyUnit energyUnit) const {
  // DS_REQUIRE(m_archivesStorage.hasArchive(openMeteringSystemId), "no archive file present", openMeteringSystemId);

  std::vector<std::string> lines;
  lines.push_back("xport");
  if (auto rrdCachedPath = m_archivesConnection.rrdCachedPath(); !rrdCachedPath.empty()) {
    lines.push_back("--daemon");
    lines.push_back(rrdCachedPath);
  }
  lines.push_back("--start");
  lines.push_back(ds::str(start));
  lines.push_back("--end");
  lines.push_back(ds::str(end));
  lines.push_back("--step");
  lines.push_back(ds::str(step));
  lines.push_back("--maxrows");
  lines.push_back("3000"); // TODO
  {
    auto mapUndefined = [](const auto& raw, const auto& data) {
      // return 0 if undefined else pass value
      return ds::str("CDEF:", data, "=", raw, ",UN,0,", raw, ",IF");
    };

    auto convert = [](Category category) {
      if (category == Category::consumption) {
        return "Consumed"s;
      } else {
        return "Produced"s;
      }
    };

    auto archiveId = openMeteringSystemId;
    if (seriesType == SeriesTypes::power) {
      lines.push_back(
          ds::str("DEF:raw0=", m_archivesStorage.filepath(archiveId), ":power", convert(category), ":AVERAGE"));
      lines.push_back(mapUndefined("raw0"sv, "data0"sv));
      lines.push_back("CDEF:sum=data0");
    } else {
      lines.push_back(
          ds::str("DEF:raw0=", m_archivesStorage.filepath(archiveId), ":energy", convert(category), "Tari1:AVERAGE"));
      lines.push_back(mapUndefined("raw0"sv, "data0"sv));
      lines.push_back(
          ds::str("DEF:raw1=", m_archivesStorage.filepath(archiveId), ":energy", convert(category), "Tari2:AVERAGE"));
      lines.push_back(mapUndefined("raw1"sv, "data1"sv));
      lines.push_back("CDEF:sum=data0,data1,+");
    }
  }

  switch (seriesType) {
    case SeriesTypes::energy:
    case SeriesTypes::energyDelta: {
      lines.push_back(ds::str("CDEF:ratePerStep=sum,", step, ",*"));
      if (energyUnit == EnergyUnit::Wh) {
        lines.push_back("CDEF:adj=ratePerStep,3600,/");
        lines.push_back("XPORT:adj");
      } else {
        lines.push_back("XPORT:ratePerStep");
      }
      break;
    }
    case SeriesTypes::power:
    default: {
      lines.push_back("XPORT:sum");
      break;
    }
  }

  auto results = m_archivesConnection.rrdInterface().xport(lines);
  // DS_REQUIRE(start, end, step) -> move to rrd-interface

  for (auto& [value, timestamp] : results.values) {
    if (std::isfinite(value)) {
      continue;
    }
    auto total = m_nonFiniteValueHistogram[step]++;
    DS_NOTICE("NaN/Infinity value detected", timestamp, value, step, total);
    value = 0.0;
  }

  DS_REQUIRE(std::none_of(std::cbegin(results.values), std::cend(results.values),
                          [](const auto& data) { return !std::isfinite(data.value); }),
             "values contain non-finite values");

  return results;
}

ValidationResult OpenMeteringSystemArchiveImpl::validate(
    const OpenMeteringSystemId& openMeteringSystemArchiveId) const {
#if 0
  static const auto rraIndexRegex = std::regex(R"(rra\[(\d+)\].rows)");
  unsigned long rrdMatchCount = 0;

  auto seriesConfig = m_electricityMeteringConfig.timeSeries;

  for (auto& rrdInfo :
       m_archivesConnection.rrdInterface().info(m_archivesStorage.filepath(openMeteringSystemArchiveId))) {
    DS_DEBUG(rrdInfo);
    if ((rrdInfo.key == "ds[power].index" && (rrdInfo.type == RrdInfoType::counter) &&
         (std::get<counter_t>(rrdInfo.value) == 0)) ||
        (rrdInfo.key == "ds[energy].index" && (rrdInfo.type == RrdInfoType::counter) &&
         (std::get<counter_t>(rrdInfo.value) == 1))) {
      rrdMatchCount++;
    } else {
      std::smatch match;
      if (std::regex_search(rrdInfo.key, match, rraIndexRegex)) {
        auto index = std::stoul(match[1]);
        DS_REQUIRE(index < seriesConfig.size(), "RRA index exceeded");
        if (std::get<counter_t>(rrdInfo.value) == seriesConfig[index].numberOfValues) {
          rrdMatchCount++;
        }
      }
    }

    if (rrdInfo.key == "ds[power].max") {
      DS_DEBUG(rrdInfo);
      if ((rrdInfo.type == RrdInfoType::value) && (std::get<rrd_value_t>(rrdInfo.value) != limitPower(archiveId))) {
        DS_NOTICE("validateDBSeries: invalid upper power limit", rrdInfo, archiveId);
        _tunePowerMaxSetting = true;
      }
    }
  }

  DS_DEBUG("RRD MatchCount", rrdMatchCount, _tunePowerMaxSetting);
  return (rrdMatchCount == (2 + seriesConfig.size()));
#endif
  return ValidationResult::ok;
}

} // namespace

std::unique_ptr<OpenMeteringSystemArchive> makeOpenMeteringSystemArchive(ArchivesConnection& archivesConnection,
                                                                         ArchivesStorage& archivesStorage,
                                                                         ds::System& system) {
  return std::make_unique<OpenMeteringSystemArchiveImpl>(archivesConnection, archivesStorage, system);
};

} // namespace dss::metering::archives
