/*
    Copyright (c) 2016 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "metering.h"

#include <ds/catch/catch.h>
#include <ds/system-mock.h>
#include <model/apartment-mock.h>

#include <cmath>
#include <trompeloeil.hpp>

#include "archives-connection-mock.h"
#include "archives-sanity-mock.h"
#include "archives-storage-mock.h"
#include "controller-direction-mock.h"
#include "model/state.h"
#include "open-metering-system-mock.h"
#include "modbusmeteringsystem-mock.h"
#include "power-states.h"
#include "rrd-interface.h"

using namespace trompeloeil;
using namespace std::literals;

using dss::Dsuid;
using namespace dss::metering;
using namespace dss::metering::archives;

using system_clock = std::chrono::system_clock;

static ds::LogChannel dsLogChannel("dssMeteringMeteringTest");

namespace {

static const char* TAGS = "[dSS][metering][core]";


SCENARIO("add/retrieve data", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto createController = [&apartment](const dsModel::ControllerId& controllerId) {
    static int ct{};
    auto controller = apartment.allocateDSMeter(dss::Dsuid{controllerId});
    controller->setName(ds::str("controller", ct++));
    controller->setIsPresent(true);
    controller->setIsConnected(true); // isPresent checks isConnected internally
    controller->setCapability_HasMetering(true);
    return controller;
  };

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = createController(controllerId1);

  auto system = ds::SystemMock();
  auto archivesConnection = std::make_shared<ArchivesConnectionMock>();
  auto archivesSanity_ = std::make_unique<ArchivesSanityMock>();
  [[maybe_unused]] auto& archivesSanity = *archivesSanity_;

  auto openMeteringSystem_ = std::make_unique<OpenMeteringSystemMock>();
  auto& openMeteringSystem = *openMeteringSystem_;

  auto modbusMeteringSystem_ = std::make_unique<ModbusMeteringSystemMock>();

  ArchivesStorageMock archivesStorage{};
  ALLOW_CALL(*archivesConnection, getArchivesStorage()).LR_RETURN(archivesStorage);

  auto consumption_ = std::make_unique<ControllerDirectionMock>();
  auto& consumption = *consumption_;
  auto production_ = std::make_unique<ControllerDirectionMock>();
  auto& production = *production_;

  auto metering = makeMockedMetering(apartment, archivesConnection, std::move(archivesSanity_), std::move(consumption_),
                                     std::move(production_), std::move(openMeteringSystem_), std::move(modbusMeteringSystem_), system);

  SECTION("getSeries") {
    WHEN("single valid controller") {
      REQUIRE_CALL(consumption, getSeries(std::vector{Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                          TimeDuration{ds::DateTime{1s}, ds::DateTime{600s}}, 0))
          .RETURN([]() -> ValueRange {
            std::vector<Value> ret;
            while (ret.size() < 600) {
              ret.push_back(Value{1, ds::DateTime{1s}}); // all valid data, incl. last entry
            }
            return ValueRange{.step = 1, .values = std::move(ret)};
          }());

      auto range = metering->getSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                       {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
      CHECK(range.values.size() == 600);
    }
    WHEN("empty list of dsmid given") {
      THEN("no data returned") {
        REQUIRE_CALL(consumption, getSeries(std::vector<Dsuid>{}, 1s, SeriesTypes::power, false,
                                            TimeDuration{ds::DateTime{1s}, ds::DateTime{600s}}, 0))
            .RETURN(ValueRange{});
        auto range = metering->getSeries({}, 1s, SeriesTypes::power, false, {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
        CHECK(range == ValueRange{});
      }
    }
  }

  GIVEN("addConsumedenergy") {
    REQUIRE_CALL(consumption, addEnergy(Dsuid{controllerId1}, ElectricityMeasurement{75, 100, ds::DateTime(100s)}));
    REQUIRE_CALL(consumption, getEnergy(Dsuid{controllerId1})).RETURN(ArchiveValue{75, 100, ds::DateTime(100s)});
    metering->addConsumedEnergy(*controller1, {75, 100, ds::DateTime(100s)});
  }

  GIVEN("getEnergy") {
    WHEN("value is known") {
      REQUIRE_CALL(consumption, getEnergy(dss::Dsuid{controllerId1}))
          .RETURN(ArchiveValue{.power = 75.0, .energy = 100.0, .lastChanged = ds::DateTime{100s}});
      CHECK(metering->getEnergy(Dsuid{controllerId1}) == ArchiveValue{75, 100, ds::DateTime(100s)});
    }
    WHEN("when value is unknow") {
      REQUIRE_CALL(consumption, getEnergy(dss::Dsuid{controllerId1})).RETURN(std::nullopt);
      THEN("return is nullopt") { CHECK(!metering->getEnergy(dss::Dsuid{controllerId1})); }
    }
  }

  GIVEN("getLastReading") {
    using dsModel::ApartmentMeteringOrigin;
    using dsModel::ControllerMeteringOrigin;
    using dsModel::MeteringType;
    using dsModel::OpenMeteringSystemOrigin;

    SECTION("ControllerOrigin") {
      REQUIRE_CALL(consumption, getLastReading(controller1->getDSID(), MeteringType::energyMetering, EnergyUnit::Ws))
          .TIMES(2)
          .LR_RETURN(LastReading{MeteringType::energyMetering, 3600.0, ds::DateTime{10s}});
      REQUIRE_CALL(consumption, getLastReading(controller1->getDSID(), MeteringType::energyMetering, EnergyUnit::Wh))
          .LR_RETURN(LastReading{MeteringType::energyMetering, 1.0, ds::DateTime{10s}});

      CHECK(metering->getLastReading(ControllerMeteringOrigin{.id = controllerId1}, MeteringType::energyMetering) ==
            LastReading{MeteringType::energyMetering, 3600.0, ds::DateTime{10s}}); // Ws is default
      CHECK(metering->getLastReading(ControllerMeteringOrigin{.id = controllerId1}, MeteringType::energyMetering,
                                     EnergyUnit::Ws) ==
            LastReading{MeteringType::energyMetering, 3600.0, ds::DateTime{10s}});
      CHECK(metering->getLastReading(ControllerMeteringOrigin{.id = controllerId1}, MeteringType::energyMetering,
                                     EnergyUnit::Wh) ==
            LastReading{MeteringType::energyMetering, 1.0, ds::DateTime{10s}});
    }
    SECTION("ApartmentOrigin") {
      auto controllerId2 = dsModel::ControllerId{"1111111111111111111111111111111200"};
      auto controller2 = createController(controllerId2);

      REQUIRE_CALL(consumption, getLastReadingApartment(MeteringType::energyMetering, EnergyUnit::Ws))
          .LR_RETURN(LastReading{MeteringType::energyMetering, 5400.0, ds::DateTime{10s}});
      REQUIRE_CALL(consumption, getLastReadingApartment(MeteringType::powerMetering, _))
          .LR_RETURN(LastReading{MeteringType::powerMetering, 30.0, ds::DateTime{10s}});

      CHECK(metering->getLastReading(ApartmentMeteringOrigin{}, MeteringType::energyMetering) ==
            LastReading{MeteringType::energyMetering, 5400.0, ds::DateTime{10s}});
      CHECK(metering->getLastReading(ApartmentMeteringOrigin{}, MeteringType::powerMetering) ==
            LastReading{MeteringType::powerMetering, 30.0, ds::DateTime{10s}});
    }
    SECTION("open-metering-system") {
      OpenMeteringSystemId id{"oms-3000"};
      REQUIRE_CALL(openMeteringSystem, getLastReading(id, MeteringType::energyMetering, EnergyUnit::Wh))
          .RETURN(LastReading{MeteringType::energyMetering, 50.0, ds::DateTime{17s}});
      CHECK(metering->getLastReading(OpenMeteringSystemOrigin{.serialNumber = id}, MeteringType::energyMetering,
                                     EnergyUnit::Wh) ==
            LastReading{MeteringType::energyMetering, 50.0, ds::DateTime{17s}});
    }
  }

  SECTION("15min average uploads") {
    REQUIRE_CALL(consumption,
                 get15MinuteAverage(std::vector<Dsuid>{controller1->getDSID()}, SeriesTypes::energyDelta, Unit::Wh))
        .RETURN(10.0);
    CHECK(metering->get15minConsumed({dss::Dsuid{controllerId1}}, SeriesTypes::energyDelta) == 10.0);
  }

  SECTION("meterings listing") {
    using dsModel::ApartmentMeteringOrigin;
    using dsModel::ControllerMeteringOrigin;
    using dsModel::MeteringOrigin;
    using dsModel::MeteringType;
    using dsModel::OpenMeteringSystemOrigin;

    GIVEN("consumption meter") {
      ALLOW_CALL(openMeteringSystem, meterings()).RETURN(std::vector<std::pair<OpenMeteringSystemId, MeteringType>>{});
      REQUIRE_CALL(consumption, meterings())
          .RETURN(std::vector{std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::powerMetering},
                              std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::energyMetering}});
      REQUIRE_CALL(production, meterings()).RETURN(std::vector<std::pair<Dsuid, MeteringType>>{});

      THEN("consumption meterings exist") {
        CHECK(metering->meterings() ==
              std::vector{
                  std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                          MeteringType::powerMetering},
                  std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                          MeteringType::energyMetering},
                  std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                  std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering}});
      }
    }

    GIVEN("production meter") {
      ALLOW_CALL(openMeteringSystem, meterings()).RETURN(std::vector<std::pair<OpenMeteringSystemId, MeteringType>>{});

      REQUIRE_CALL(consumption, meterings())
          .RETURN(std::vector{std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::powerMetering},
                              std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::energyMetering}});
      REQUIRE_CALL(production, meterings())
          .RETURN(
              std::vector{std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::powerProducedMetering},
                          std::pair<Dsuid, MeteringType>{Dsuid{controllerId1}, MeteringType::energyProducedMetering}});

      THEN("production meterings exist") {
        CHECK(
            metering->meterings() ==
            std::vector{
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ControllerMeteringOrigin{.id = controllerId1},
                                                        MeteringType::energyProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::energyMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{}, MeteringType::powerProducedMetering},
                std::pair<MeteringOrigin, MeteringType>{ApartmentMeteringOrigin{},
                                                        MeteringType::energyProducedMetering}});
      }
    }
    GIVEN("open-metering-system") {
      OpenMeteringSystemId id{"oms-3000"};

      REQUIRE_CALL(production, meterings()).RETURN(std::vector<std::pair<Dsuid, MeteringType>>{});
      REQUIRE_CALL(consumption, meterings()).RETURN(std::vector<std::pair<Dsuid, MeteringType>>{});

      REQUIRE_CALL(openMeteringSystem, meterings())
          .RETURN(std::vector<std::pair<OpenMeteringSystemId, MeteringType>>{
              {id, MeteringType::powerMetering},
          });
      CHECK(metering->meterings() == std::vector{std::pair<MeteringOrigin, MeteringType>{
                                         OpenMeteringSystemOrigin{.serialNumber = id}, MeteringType::powerMetering}});

      THEN("serial number exists") {
        REQUIRE_CALL(openMeteringSystem, activeMeterId()).RETURN(id);
        CHECK(metering->openMeteringSystemOrigin() == dsModel::OpenMeteringSystemOrigin{.serialNumber = id});
      }
    }
    GIVEN("open-metering-system missing") {
      THEN("no serial number exists") {
        REQUIRE_CALL(openMeteringSystem, activeMeterId()).RETURN(std::nullopt);
        CHECK(metering->openMeteringSystemOrigin() == std::nullopt);
      }
    }
  }

  SECTION("flushDb") {
    auto controllerId2 = dsModel::ControllerId{"1111111111111111111111111111111200"};
    auto controller2 = createController(controllerId2);

    auto mockPath = [](const auto& archiveId) { return ds::str("/storage/", archiveId, ".rrd"s); };

    WHEN("3 archives needs flushing") {
      REQUIRE_CALL(consumption, needsFlushing(_))
          .RETURN(std::vector{mockPath(ArchiveId{controller1->id(), Category::consumption}),
                              mockPath(ArchiveId{controller2->id(), Category::consumption})});
      REQUIRE_CALL(production, needsFlushing(_))
          .LR_RETURN(std::vector{mockPath(ArchiveId{controller1->id(), Category::production})});

      REQUIRE_CALL(*archivesConnection,
                   flush(std::vector{"/storage/(1111111111111111111111111111111100,consumption).rrd"s,
                                     "/storage/(1111111111111111111111111111111200,consumption).rrd"s,
                                     "/storage/(1111111111111111111111111111111100,production).rrd"s}));
      metering->flushDb(ds::DateTime{42s});
    }
  }

  SECTION("checkInitialState upon createState") {
    using dsModel::MeteringType;
    WHEN("apartment metering state is allocataed") {
      const auto x = GENERATE(table<PowerStateConfig, TotalPower, int>(
          {{{.category = Category::consumption, .setThreshold = 12}, {.consumption = 13}, dss::State_Active},
           {{.category = Category::consumption, .setThreshold = 12}, {.consumption = 12}, dss::State_Inactive},
           {{.category = Category::production, .setThreshold = 12}, {.production = 13}, dss::State_Active},
           {{.category = Category::production, .setThreshold = 12}, {.production = 12}, dss::State_Inactive}}));

      // REQUIRE_CALL(powerStates);

      auto totalPower = std::get<TotalPower>(x);
      REQUIRE_CALL(consumption, getLastReadingApartment(MeteringType::powerMetering, _))
          .RETURN(LastReading{.value = double(totalPower.consumption)});
      REQUIRE_CALL(production, getLastReadingApartment(MeteringType::powerProducedMetering, _))
          .LR_RETURN(LastReading{.value = double(totalPower.production)});

      auto state =
          metering->createState(MeteringSystem::apartmentSummation, PowerStateConfig(std::get<PowerStateConfig>(x)));
      CHECK(state->getState() == std::get<int>(x));
    }

    WHEN("open metering system state is allocataed") {
      const auto x = GENERATE(table<PowerStateConfig, TotalPower, int>(
          {{{.category = Category::consumption, .setThreshold = 12}, {.consumption = 13}, dss::State_Active},
           {{.category = Category::consumption, .setThreshold = 12}, {.consumption = 12}, dss::State_Inactive},
           {{.category = Category::production, .setThreshold = 12}, {.production = 13}, dss::State_Active},
           {{.category = Category::production, .setThreshold = 12}, {.production = 12}, dss::State_Inactive}}));

      WHEN("open-metering-system is present") {
        REQUIRE_CALL(openMeteringSystem, activeMeterId()).RETURN(OpenMeteringSystemId{});
        auto totalPower = std::get<TotalPower>(x);
        REQUIRE_CALL(openMeteringSystem, getLastReading(_, MeteringType::powerMetering, _))
            .RETURN(LastReading{.value = double(totalPower.consumption)});
        REQUIRE_CALL(openMeteringSystem, getLastReading(_, MeteringType::powerProducedMetering, _))
            .LR_RETURN(LastReading{.value = double(totalPower.production)});

        auto state =
            metering->createState(MeteringSystem::openMeteringSystem, PowerStateConfig(std::get<PowerStateConfig>(x)));
        CHECK(state->getState() == std::get<int>(x));
      }
      WHEN("open-metering-system is absent") {
        REQUIRE_CALL(openMeteringSystem, activeMeterId()).RETURN(std::nullopt);
        auto state =
            metering->createState(MeteringSystem::openMeteringSystem, PowerStateConfig(std::get<PowerStateConfig>(x)));
        CHECK(state->getState() == dss::State_Inactive);
      }
    }
  }
}


SCENARIO("energy production", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});
  controller1->setName("controller1");
  controller1->setCapability_HasMetering(true);

  auto system = ds::SystemMock();
  auto archivesConnection = std::make_shared<ArchivesConnectionMock>();
  auto archivesSanity_ = std::make_unique<ArchivesSanityMock>();
  [[maybe_unused]] auto& archivesSanity = *archivesSanity_;
  auto openMeteringSystem = std::make_unique<OpenMeteringSystemMock>();
  auto modbusMeteringSystem = std::make_unique<ModbusMeteringSystemMock>();

  ArchivesStorageMock archivesStorage{};
  ALLOW_CALL(*archivesConnection, getArchivesStorage()).LR_RETURN(archivesStorage);

  auto consumption_ = std::make_unique<ControllerDirectionMock>();
  auto production_ = std::make_unique<ControllerDirectionMock>();
  auto& production = *production_;

  auto metering = makeMockedMetering(apartment, archivesConnection, std::move(archivesSanity_), std::move(consumption_),
                                     std::move(production_), std::move(openMeteringSystem), 
                                     std::move(modbusMeteringSystem), system);

  SECTION("addProducedEnergy") {
    REQUIRE_CALL(production, addEnergy(_, _));
    metering->addProducedEnergy(*controller1, {75, 100, ds::DateTime(100s)}); // baseline
  }

  WHEN("getProducedSeries") {
    WHEN("single valid controller") {
      REQUIRE_CALL(production, getSeries(std::vector{Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                         TimeDuration{ds::DateTime{1s}, ds::DateTime{600s}}, 0))
          .RETURN([]() -> ValueRange {
            std::vector<Value> ret;
            while (ret.size() < 600) {
              ret.push_back(Value{1, ds::DateTime{1s}}); // all valid data, incl. last entry
            }
            return ValueRange{.step = 1, .values = std::move(ret)};
          }());

      auto range = metering->getProducedSeries({Dsuid{controllerId1}}, 1s, SeriesTypes::power, false,
                                               {ds::DateTime{1s}, ds::DateTime{600s}}, 0);
      CHECK(range.values.size() == 600);
    }
  }
}

SCENARIO("total power", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  auto system = ds::SystemMock();
  auto archivesConnection = std::make_shared<ArchivesConnectionMock>();
  auto archivesSanity_ = std::make_unique<ArchivesSanityMock>();
  [[maybe_unused]] auto& archivesSanity = *archivesSanity_;

  auto openMeteringSystem = std::make_unique<OpenMeteringSystemMock>();
  auto modbusMeteringSystem = std::make_unique<ModbusMeteringSystemMock>();

  ArchivesStorageMock archivesStorage{};
  ALLOW_CALL(*archivesConnection, getArchivesStorage()).LR_RETURN(archivesStorage);

  auto consumption_ = std::make_unique<ControllerDirectionMock>();
  auto& consumption = *consumption_;
  auto production_ = std::make_unique<ControllerDirectionMock>();
  auto& production = *production_;

  auto metering = makeMockedMetering(apartment, archivesConnection, std::move(archivesSanity_), std::move(consumption_),
                                     std::move(production_), std::move(openMeteringSystem), 
                                     std::move(modbusMeteringSystem), system);

  GIVEN("consumption/production power") {
    THEN("getApartmentPower") {
      REQUIRE_CALL(consumption, getApartmentPower()).RETURN(25);
      REQUIRE_CALL(production, getApartmentPower()).RETURN(75);
      CHECK(metering->getConsumedPowerApartment() == 25);
      CHECK(metering->getProducedPowerApartment() == 75);
    }
  }
}

SCENARIO("get instant power", TAGS) {
  auto propertySystem = dss::PropertySystem{};
  auto configNode = propertySystem.createProperty("/config");
  dss::ApartmentMock apartment{nullptr, &propertySystem};

  auto controllerId1 = dsModel::ControllerId{"1111111111111111111111111111111100"};
  auto controller1 = apartment.allocateDSMeter(dss::Dsuid{controllerId1});

  auto system = ds::SystemMock();
  auto archivesConnection = std::make_shared<ArchivesConnectionMock>();
  auto archivesSanity_ = std::make_unique<ArchivesSanityMock>();
  [[maybe_unused]] auto& archivesSanity = *archivesSanity_;
  auto openMeteringSystem = std::make_unique<OpenMeteringSystemMock>();
  auto modbusMeteringSystem = std::make_unique<ModbusMeteringSystemMock>();

  ArchivesStorageMock archivesStorage{};
  ALLOW_CALL(*archivesConnection, getArchivesStorage()).LR_RETURN(archivesStorage);

  auto consumption_ = std::make_unique<ControllerDirectionMock>();
  auto& consumption = *consumption_;
  auto production_ = std::make_unique<ControllerDirectionMock>();
  auto& production = *production_;

  auto metering = makeMockedMetering(apartment, archivesConnection, std::move(archivesSanity_), std::move(consumption_),
                                     std::move(production_), std::move(openMeteringSystem),
                                     std::move(modbusMeteringSystem), system);

  GIVEN("current data") {

    THEN("instant power is set") {
      REQUIRE_CALL(consumption, getInstantPower(dss::Dsuid{controllerId1})).RETURN(25);
      REQUIRE_CALL(production, getInstantPower(dss::Dsuid{controllerId1})).RETURN(75);
      CHECK(75 == metering->getProducedPower(dss::Dsuid{controllerId1}));
      CHECK(25 == metering->getConsumedPower(dss::Dsuid{controllerId1}));
    }
    THEN("instant energy is set") {
      REQUIRE_CALL(consumption, getLastEnergy(dss::Dsuid{controllerId1})).RETURN(1000);
      REQUIRE_CALL(production, getLastEnergy(dss::Dsuid{controllerId1})).RETURN(100);
      CHECK(metering->getLastEnergy(dss::Dsuid{controllerId1}) == TotalEnergy{.production = 100, .consumption = 1000});
    }
  }
}

} // namespace
