#include "serial-toolkit.h"
#include <errno.h>
#include <string.h>

#include <ds/json.h>
#include <ds/log.h>

#include <filesystem>
#include <modbus/modbus.h>
#include <http_client.h>
#include <linux/serial.h>
#include "config-file.h"

static ds::LogChannel dsLogChannel("modbusSerial");

modbus_t * ctx;



namespace dss::modbusreader {

struct SerialPortConfigReply {
  dsModel::SerialPortConfig data;
};

struct AttachedMetersReply {
  std::vector<dsModel::ModbusMeter> data;
};

struct MeterConfigReply {
  std::vector<dsModel::ElectricityMeterConfig> data;
};

void SenderQueue::send(std::string meterReading) {
  // we don't re-use socket client yet
  std::make_shared<dss::modbusreader::SocketClient>(m_io, m_socketPath)->send(meterReading.substr(0));
}

void handler(const boost::system::error_code& error)
{
  if (!error)
  {
    // Timer expired.
  }
}

void SerialConnection::start() {
  //doReconnect();

  //while (true) {
    // Construct a timer with an absolute expiry time.
    //boost::asio::deadline_timer timer(m_io,boost::posix_time::seconds(5));

    // Start an asynchronous wait.
    
    m_busy = true;
    DS_DEBUG("Main loop is alive");
    //timer.async_wait(handler);
    querySerialConnectionConfig();
    doRead();
    sleep(1);
    m_busy = false;
  //}
}

void SerialConnection::doRead() {
  std::string result;
  auto code = m_client.get("http://127.0.0.1:8088/api/v1/apartment/meterings/modbus/attached", &result);
  DS_NOTICE("Attached Meters", result);

  if (code == 200) {
    try {
      m_attachedMeters.clear();
      m_attachedMeters = ds::fromJson<AttachedMetersReply>(result).data;
    } catch (std::exception& e) {
      DS_WARNING("Could not convert reply", e);
    }
  } else {
    DS_WARNING("Error requesting from server", code, result);
  }

  code = m_client.get("http://127.0.0.1:8088/api/v1/apartment/meterings/modbus/config/supported", &result);
  if (code == 200) {
    DS_NOTICE("Supported Meters", result);
    std::vector<dsModel::ElectricityMeterConfig> meterConfigs;
    try {
      meterConfigs = ds::fromJson<MeterConfigReply>(result).data;
 
      DS_NOTICE("Cycling meters");
      for (auto &meter : m_attachedMeters) {
        DS_DEBUG(meter);
        for (auto &config : meterConfigs) {
          DS_DEBUG(config);
          DS_DEBUG("Comparison ", config.id, meter.configId);
          if (config.id == meter.configId) {
            DS_NOTICE("Reading meter", meter, config);
            readMeter(meter.slaveAddress, config);
          }

        }
      }
     } catch (std::exception& e) {
      DS_WARNING("Could not convert reply", e);
    }
  } else {
    DS_WARNING("Error requesting from server", code, result);
  }
}

void SerialConnection::queryAttachedMeters() {
  std::string result;
  auto code = m_client.get("http://127.0.0.1:8088/api/v1/apartment/meterings/modbus/attached", &result);
  if (code == 200) {
    DS_NOTICE(result);
    try {
      auto serverConfig = ds::fromJson<std::vector<dsModel::ModbusMeter>>(result); 
    } catch (std::exception& e) {
      DS_WARNING("Could not convert reply", e);
    }
  } else {
    DS_WARNING("Error requesting from server", code, result);
  }
}

void SerialConnection::querySerialConnectionConfig() {
  //m_isReading = true;
  std::string result;
  auto code = m_client.get("http://127.0.0.1:8088/api/v1/apartment/meterings/modbus/rtuconfig", &result);
  if (code == 200) {
    DS_NOTICE(result);
    try {
      auto serverConfig = ds::fromJson<SerialPortConfigReply>(result);


      DS_NOTICE("querySerialConnectionConfig", m_config, serverConfig.data);
      if (serverConfig.data != m_config) {
        m_config = serverConfig.data;
        doReconnect();
      }
    } catch (std::exception& e) {
      DS_WARNING("Could not convert reply", e);
    }
  } else {
    DS_WARNING("Error requesting from server", code, result); 
  }
}

void SerialConnection::readMeter(int slaveAddress, dsModel::ElectricityMeterConfig meterConfig) {
  Reading aReading;
  aReading.tariffEnabled = false;

  std::vector<dsModel::ModbusFieldType> failedFields;

  auto dataCountMultiplier = [](dsModel::ModbusDataType modbusDataType) {
    if (modbusDataType == dsModel::ModbusDataType::float32 || modbusDataType == dsModel::ModbusDataType::dateTime || \
      modbusDataType == dsModel::ModbusDataType::uint32 || modbusDataType == dsModel::ModbusDataType::int32) {
      return 2;
    } else if (modbusDataType == dsModel::ModbusDataType::uint64 || modbusDataType == dsModel::ModbusDataType::int64) {   
      return 4;
    } else {
      return 1;
    }
  };

  

  auto setMeasurementField = [&](float value, dsModel::ModbusFieldType modbusFieldType, float scaleValue) {
    DS_NOTICE(value, modbusFieldType, scaleValue);
    auto kWToW = [&value](float scaleValue) {
      DS_DEBUG(value, scaleValue);  
      return value * scaleValue;
    };

    auto kWhToWs = [&value](float scaleValue) {
      DS_DEBUG(value, scaleValue);
      return value * scaleValue * 3600 * 1000;
    };

    switch (modbusFieldType)
    {
    case dsModel::ModbusFieldType::powerImport: {
      DS_DEBUG("power conversion", value, scaleValue);
      float value = kWToW(scaleValue);
      DS_DEBUG("power conversion after ", value);
      if (value > 0) {
        aReading.powerConsumed = value;
      } else {
        aReading.powerProduced = value * -1;
      } 
      return;
    }
    case dsModel::ModbusFieldType::powerExport: aReading.powerProduced = kWToW(scaleValue); return;
    case dsModel::ModbusFieldType::energyImport: aReading.energyConsumed = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyExport: aReading.energyProduced = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyImportTariff1: aReading.t1EnergyConsumed = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyImportTariff2: aReading.t2EnergyConsumed = kWhToWs(scaleValue ); return;
    case dsModel::ModbusFieldType::energyImportTariff3: aReading.t3EnergyConsumed = kWhToWs(scaleValue ); return;
    case dsModel::ModbusFieldType::energyImportTariff4: aReading.t4EnergyConsumed = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyImportTariff5: aReading.t5EnergyConsumed = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyImportTariff6: aReading.t6EnergyConsumed = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyExportTariff1: aReading.t1EnergyProduced = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyExportTariff2: aReading.t2EnergyProduced = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyExportTariff3: aReading.t3EnergyProduced = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyExportTariff4: aReading.t4EnergyProduced = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyExportTariff5: aReading.t5EnergyProduced = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::energyExportTariff6: aReading.t6EnergyProduced = kWhToWs(scaleValue); return;
    case dsModel::ModbusFieldType::tariffEnabled: {
      DS_DEBUG("Tariff", value);
      aReading.tariffEnabled = value > 0; 
      return;
    }
    case dsModel::ModbusFieldType::none: [[fallthrough]];
    case dsModel::ModbusFieldType::model: [[fallthrough]];
    case dsModel::ModbusFieldType::manufacturer: [[fallthrough]];
    case dsModel::ModbusFieldType::serial: [[fallthrough]];
    case dsModel::ModbusFieldType::dateTime: [[fallthrough]];
    default: return;
    }
  };



  DS_NOTICE("Reading meter", slaveAddress, meterConfig.registers);
  ///
  aReading.slaveAddress = slaveAddress;
  aReading.lastChanged = ds::DateTime::clock::now();
  aReading.meteringType = "modbus";
  ////
  DS_NOTICE("Reading meter now", slaveAddress);
  for (auto& aRegister : meterConfig.registers) {
    std::array<uint16_t, 255> data = {};
    /*std::string debug = "";
      for (const auto& value : data) {
        debug += ds::str(" ", value);
      }
      DS_WARNING("ARRAY BEFORE", debug);*/
  
    DS_DEBUG("Reading register", aRegister);
    DS_DEBUG("Reading register params", aRegister.address, aRegister.length * dataCountMultiplier(aRegister.modbusDataType));

    //If it is serial we only query it if we do not know it
    if (aRegister.fieldType == dsModel::ModbusFieldType::serial) {
      for (const auto& attachedMeter: m_attachedMeters) {
        if (attachedMeter.slaveAddress == slaveAddress && attachedMeter.serial != "") {
          aReading.id = attachedMeter.serial;
          break;
        }
      }
      if (std::string(aReading.id) != "") {
        continue;
      }
    }

    modbus_set_slave(ctx, slaveAddress);
    usleep(m_intervalUSec);
    int ret = -1;

    if (aRegister.registerType == dsModel::ModbusRegisterType::input) {
      DS_DEBUG("Reading input registers");
      ret = modbus_read_input_registers(ctx, aRegister.address, 
                aRegister.length * dataCountMultiplier(aRegister.modbusDataType), data.data());
    } else {
      DS_DEBUG("Reading holding registers");
      ret = modbus_read_registers(ctx, aRegister.address, 
                aRegister.length * dataCountMultiplier(aRegister.modbusDataType), data.data());
    }
    
    if (ret == -1) {
      failedFields.push_back(aRegister.fieldType);
      DS_WARNING("Could not read register", strerror(errno), aRegister);
    } else {
      std::string debug = "";
      for (const auto& value : data) {
        debug += ds::str(" ", value);
      }
      DS_DEBUG("ARRAY", debug);
      if (aRegister.modbusDataType == dsModel::ModbusDataType::float32) {
        float value = modbus_get_float_abcd(data.data());
        if (aRegister.fieldType >= dsModel::ModbusFieldType::powerImport && 
            aRegister.fieldType <= dsModel::ModbusFieldType::energyExportTariff6) {
          setMeasurementField(value, aRegister.fieldType, aRegister.scaleValue);
        }
      } else if (aRegister.modbusDataType == dsModel::ModbusDataType::int64) {
        int64_t value = MODBUS_GET_INT64_FROM_INT16(data.data(), 0);
        if (aRegister.fieldType >= dsModel::ModbusFieldType::powerImport && 
            aRegister.fieldType <= dsModel::ModbusFieldType::energyExportTariff6) {
          setMeasurementField(value, aRegister.fieldType, aRegister.scaleValue);
        }
      } else if (aRegister.modbusDataType == dsModel::ModbusDataType::uint64) {
        uint64_t value = MODBUS_GET_INT64_FROM_INT16(data.data(), 0);
        if (aRegister.fieldType >= dsModel::ModbusFieldType::powerImport && 
            aRegister.fieldType <= dsModel::ModbusFieldType::energyExportTariff6) {
          setMeasurementField(value, aRegister.fieldType, aRegister.scaleValue);
        }
      } else if (aRegister.modbusDataType == dsModel::ModbusDataType::uint32) {
        uint32_t value = MODBUS_GET_INT32_FROM_INT16(data.data(), 0);
        if (aRegister.fieldType >= dsModel::ModbusFieldType::powerImport && 
            aRegister.fieldType <= dsModel::ModbusFieldType::energyExportTariff6) {
          setMeasurementField(value, aRegister.fieldType, aRegister.scaleValue);
        } else if (aRegister.fieldType == dsModel::ModbusFieldType::serial) {
          aReading.id = std::to_string(value);
        }
      } else if (aRegister.modbusDataType == dsModel::ModbusDataType::int32) {
        int32_t value = MODBUS_GET_INT32_FROM_INT16(data.data(), 0);
        if (aRegister.fieldType >= dsModel::ModbusFieldType::powerImport && 
            aRegister.fieldType <= dsModel::ModbusFieldType::energyExportTariff6) {
          setMeasurementField(value, aRegister.fieldType, aRegister.scaleValue);
        } 
      } else if (aRegister.modbusDataType == dsModel::ModbusDataType::uint16) {
        uint16_t value = data.data()[0];
        if (aRegister.fieldType >= dsModel::ModbusFieldType::powerImport && 
            aRegister.fieldType <= dsModel::ModbusFieldType::tariffEnabled) {
          setMeasurementField(value, aRegister.fieldType, aRegister.scaleValue);
        } 
      } else if (aRegister.modbusDataType == dsModel::ModbusDataType::ascii) {
        std::string aString;
        for (int i = 0; i < aRegister.length; i++) {
          if((data.at(i) >> 8) != 0) {
            aString.push_back(static_cast<char>(data.at(i) >> 8));
          }
          if ((data.at(i) & 0xff) != 0) {
            aString.push_back(static_cast<char>(data.at(i) & 0xff));
          }
        }
        if (aRegister.fieldType == dsModel::ModbusFieldType::serial) {
          aReading.id = aString;
        }
      }
    }

  }
  auto it = m_previousReading.find(aReading.slaveAddress);
  if (it != m_previousReading.end()) {
    auto& previousReading = *it;
    for (const auto& failedField : failedFields) {
      if (failedField == dsModel::ModbusFieldType::powerImport) {
        aReading.powerConsumed = previousReading.second.powerConsumed;
        aReading.powerProduced = previousReading.second.powerProduced;
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyImport) {
        aReading.energyConsumed = previousReading.second.energyConsumed;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyExport) {
        aReading.energyProduced = previousReading.second.energyProduced;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyExportTariff1) {
        aReading.t1EnergyProduced = previousReading.second.t1EnergyProduced;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyExportTariff2) {
        aReading.t2EnergyProduced = previousReading.second.t2EnergyProduced;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyExportTariff3) {
        aReading.t3EnergyProduced = previousReading.second.t3EnergyProduced;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyExportTariff4) {
        aReading.t4EnergyProduced = previousReading.second.t4EnergyProduced;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyExportTariff5) {
        aReading.t5EnergyProduced = previousReading.second.t5EnergyProduced;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyExportTariff6) {
        aReading.t6EnergyProduced = previousReading.second.t6EnergyProduced;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyImportTariff1) {
        aReading.t1EnergyConsumed = previousReading.second.t1EnergyConsumed;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyImportTariff2) {
        aReading.t2EnergyConsumed = previousReading.second.t2EnergyConsumed;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyImportTariff3) {
        aReading.t3EnergyConsumed = previousReading.second.t3EnergyConsumed;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyImportTariff4) {
        aReading.t4EnergyConsumed = previousReading.second.t4EnergyConsumed;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyImportTariff5) {
        aReading.t5EnergyConsumed = previousReading.second.t5EnergyConsumed;        
        DS_NOTICE("Interpolating field", failedField);
      } else if (failedField == dsModel::ModbusFieldType::energyImportTariff6) {
        aReading.t6EnergyConsumed = previousReading.second.t6EnergyConsumed;        
        DS_NOTICE("Interpolating field", failedField);
      } 

    }
  }
  m_previousReading[aReading.slaveAddress] = aReading;
  DS_NOTICE("Reading", aReading);
  if (std::string(aReading.id) != "") {
    m_senderQueue.send(ds::toJson(aReading));
  }
}

void SerialConnection::doReconnect() {
  DS_NOTICE("doReconnect", m_config);

  std::string baudString = ds::str("", m_config.baudRate);
  std::string stopBitsString = ds::str("", m_config.stopBits);
  std::string dataBitsString = ds::str("", m_config.dataBits);
  std::string parityString = ds::str("", m_config.parity);

  auto config = dss::modbus::loadConfigFile(ds::system());
  if (config.modbusMiscConfig.modbusTCPServer != "") {
    DS_NOTICE("Connecting TCP", config);
    modbus_close(ctx);
    modbus_free(ctx);

    ctx = modbus_new_tcp(config.modbusMiscConfig.modbusTCPServer.c_str(), 
                         config.modbusMiscConfig.port);

    int ret = modbus_connect(ctx);
    if (ret == -1) {
      DS_WARNING("Could not connect TCP", strerror(errno));
    }

    if (config.modbusMiscConfig.timeoutResponseSec != 0) {
      modbus_set_response_timeout(ctx, config.modbusMiscConfig.timeoutResponseSec, 0);
    } else {
      modbus_set_response_timeout(ctx, 5, 0);
    }

    if (config.modbusMiscConfig.timeoutIndicationSec != 0) {
      modbus_set_indication_timeout(ctx, config.modbusMiscConfig.timeoutIndicationSec, 0);
    } else {
      modbus_set_indication_timeout(ctx, 5, 0);
    }

    return;
  }

  if (baudString.length() > 4 && \
      stopBitsString.length() > 4 && \
      dataBitsString.length() > 4 && \
      parityString.length() > 0) {
    
    std::string serialDevice = "/dev/ttymxc2";
    DS_NOTICE("doReconnect", ds::str("", m_config.baudRate));
    int baud = std::atoi(ds::str("", m_config.baudRate).substr(4).c_str());
    DS_NOTICE("doReconnect", ds::str("", m_config.stopBits));
    int stopBits = std::atoi(ds::str("", m_config.stopBits).substr(4).c_str());
    DS_NOTICE("doReconnect", ds::str("", m_config.dataBits));
    int dataBits = std::atoi(ds::str("", m_config.dataBits).substr(4).c_str());
    DS_NOTICE("doReconnect", ds::str("", m_config.parity));
    char parity = std::toupper(ds::str("", m_config.parity).at(0));

    DS_NOTICE("Modbus serial device connecting", serialDevice, baud, parity, dataBits, stopBits);

    modbus_close(ctx);
    modbus_free(ctx);

    struct serial_rs485 rs485conf = {0};

    int fd = open (serialDevice.c_str(), O_RDWR);
    if (fd < 0) {
      DS_ERROR("RS485 could not be configured", serialDevice);
    } else {
      rs485conf.flags |= SER_RS485_ENABLED;
      rs485conf.flags |= SER_RS485_RTS_ON_SEND;
      if (ioctl (fd, TIOCSRS485, &rs485conf) < 0) {
        DS_ERROR("RS485 could not be configured", serialDevice);
      }
    }
    

    ctx = modbus_new_rtu(serialDevice.c_str(), 
                  baud,
                  parity, 
                  stopBits,
                  stopBits);

    int ret = modbus_connect(ctx);
    if (ret == -1) {
      DS_WARNING("Could not connect serial", strerror(errno));
    }

    if (config.modbusMiscConfig.timeoutResponseSec != 0) {
      modbus_set_response_timeout(ctx, config.modbusMiscConfig.timeoutResponseSec, 0);
    } else {
      modbus_set_response_timeout(ctx, 5, 0);
    }

    if (config.modbusMiscConfig.timeoutIndicationSec != 0) {
      modbus_set_indication_timeout(ctx, config.modbusMiscConfig.timeoutIndicationSec, 0);
    } else {
      modbus_set_indication_timeout(ctx, 5, 0);
    }
  }
}

} // namespace dss::modbusreader


DS_REFLECT_STRUCT((dss::modbusreader), SerialPortConfigReply, (data));
DS_REFLECT_STRUCT((dss::modbusreader), AttachedMetersReply, (data));
DS_REFLECT_STRUCT((dss::modbusreader), MeterConfigReply, (data));