/*
    Copyright (c) 2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "actionrequestexecutor.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <memory>

#include "businterface.h"
#include "ds/log.h"
#include "model/apartment.h"
#include "model/modelconst.h"
#include "sceneaccess.h"

static ds::LogChannel dsLogChannel("actionRequestChecker");

DS_REFLECTED_ANONYMOUS_ENUM(
    Action, (increaseValue)(decreaseValue)(setValue)(callScene)(callSceneMin)(saveScene)(undoScene)(undoSceneLast)(
                blink)(increaseOutputChannelValue)(decreaseOutputChannelValue)(stopOutputChannelValue)(pushSensor));

namespace dss::model {

namespace {

template <typename T>
bool authorized(const Apartment& apartment, const T& target, Action action, CallOrigin origin,
                SceneAccessCategory category) {
  switch (action) {
    case Action::increaseValue:
    case Action::decreaseValue:
    case Action::setValue:
    case Action::undoScene:
    case Action::undoSceneLast:
    case Action::blink:
    case Action::increaseOutputChannelValue:
    case Action::decreaseOutputChannelValue:
    case Action::stopOutputChannelValue:
    case Action::pushSensor:
      if (!SceneAccess::checkAccess(apartment, &target, category)) {
        DS_DEBUG(__func__, action, "blocked,");
        return false;
      }
      break;
    case Action::saveScene:
      // TODO no check applied, oversight?
      break;

    case Action::callScene:
    case Action::callSceneMin: DS_WARNING(action, "must be checked with authorizedSceneCall", action); return false;
  }
  return true;
}

template <typename T>
bool authorizedSceneCall(const Apartment& apartment, const T& target, Action action, CallOrigin origin,
                         SceneAccessCategory category, bool force, int sceneNr) {
  switch (action) {
    case Action::callScene:
    case Action::callSceneMin:
      if (!SceneAccess::checkAccess(apartment, &target, category)) {
        DS_DEBUG(__func__, action, "blocked,");
        return false;
      }
      if (origin == CallOrigin::json) {
        if (force == false && (!SceneAccess::checkStates(apartment, &target, sceneNr))) {
          DS_DEBUG(__func__, action, "from web service blocked");
          return false;
        }
      }
      return true;
      break;

    case Action::saveScene:
    case Action::increaseValue:
    case Action::decreaseValue:
    case Action::setValue:
    case Action::undoScene:
    case Action::undoSceneLast:
    case Action::blink:
    case Action::increaseOutputChannelValue:
    case Action::decreaseOutputChannelValue:
    case Action::stopOutputChannelValue:
    case Action::pushSensor:
    default: DS_WARNING("action is not a sceneCall", action); return false;
  }
}

} // namespace

ActionRequestExecutor::ActionRequestExecutor(const Apartment& apartment, ActionRequestInterface& actionRequestInterface)
    : m_apartment(apartment), m_actionRequestInterface(actionRequestInterface) {}

void ActionRequestExecutor::increaseValue(AddressableModelItem& target, Caller&& caller) {
  if (!authorized(m_apartment, target, Action::increaseValue, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.callScene(&target, SceneInc, std::move(caller), false);
} // increaseValue

void ActionRequestExecutor::decreaseValue(AddressableModelItem& target, Caller&& caller) {
  if (!authorized(m_apartment, target, Action::decreaseValue, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.callScene(&target, SceneDec, std::move(caller), false);
} // decreaseValue

void ActionRequestExecutor::setValue(AddressableModelItem& target, Caller&& caller, const uint8_t _value) {
  if (!authorized(m_apartment, target, Action::setValue, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.setValue(&target, _value, std::move(caller));
} // setValue

void ActionRequestExecutor::callScene(AddressableModelItem& target, Caller&& caller, const int _sceneNr,
                                      const bool _force) {
  if (!authorizedSceneCall(m_apartment, target, Action::callScene, caller.origin, caller.category, _sceneNr, _force)) {
    return;
  }
  m_actionRequestInterface.callScene(&target, _sceneNr, std::move(caller), _force);
} // callScene

void ActionRequestExecutor::callSceneMin(AddressableModelItem& target, Caller&& caller, const int _sceneNr) {
  if (!authorizedSceneCall(m_apartment, target, Action::callSceneMin, caller.origin, caller.category, _sceneNr,
                           false)) {
    return;
  }
  m_actionRequestInterface.callSceneMin(&target, _sceneNr, std::move(caller));
} // callSceneMin

void ActionRequestExecutor::saveScene(AddressableModelItem& target, Caller&& caller, const int _sceneNr) {
  if (!authorized(m_apartment, target, Action::saveScene, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.saveScene(&target, _sceneNr, std::move(caller));
} // saveScene

void ActionRequestExecutor::undoScene(AddressableModelItem& target, Caller&& caller, const int _sceneNr) {
  if (!authorized(m_apartment, target, Action::undoScene, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.undoScene(&target, _sceneNr, std::move(caller));
} // undoScene

void ActionRequestExecutor::undoSceneLast(AddressableModelItem& target, Caller&& caller) {
  if (!authorized(m_apartment, target, Action::undoSceneLast, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.undoSceneLast(&target, std::move(caller));
} // undoSceneLast

void ActionRequestExecutor::blink(AddressableModelItem& target, Caller&& caller) {
  if (!authorized(m_apartment, target, Action::blink, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.blink(&target, std::move(caller));
} // blink

void ActionRequestExecutor::increaseOutputChannelValue(AddressableModelItem& target, Caller&& caller,
                                                       const ChannelType _channel) {
  if (!authorized(m_apartment, target, Action::increaseOutputChannelValue, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.increaseOutputChannelValue(&target, _channel, std::move(caller));
} // increaseOutputChannelValue

void ActionRequestExecutor::decreaseOutputChannelValue(AddressableModelItem& target, Caller&& caller,
                                                       const ChannelType _channel) {
  if (!authorized(m_apartment, target, Action::decreaseOutputChannelValue, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.decreaseOutputChannelValue(&target, _channel, std::move(caller));
} // decreaseOutputChannelValue

void ActionRequestExecutor::stopOutputChannelValue(AddressableModelItem& target, Caller&& caller,
                                                   const ChannelType _channel) {
  if (!authorized(m_apartment, target, Action::stopOutputChannelValue, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.stopOutputChannelValue(&target, _channel, std::move(caller));
} // stopOutputChannelValue

void ActionRequestExecutor::pushSensor(AddressableModelItem& target, Caller&& caller, const dsuid_t _sourceID,
                                       SensorType _sensorType, const double _sensorValueFloat) {
  if (!authorized(m_apartment, target, Action::pushSensor, caller.origin, caller.category)) {
    return;
  }
  m_actionRequestInterface.pushSensor(&target, _sourceID, _sensorType, _sensorValueFloat, std::move(caller));
}

} // namespace dss::model
