/*
    Copyright (c) 2010,2011 digitalSTROM.org, Zurich, Switzerland

    Authors: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>
             Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ds-api/array.h>
#include <ds-model/apartment.h>
#include <ds-model/model.h>
#include <ds/conf.h>
#include <ds/log.h>
#include <ds/system.h>

#include <boost/algorithm/string/predicate.hpp>
#include <boost/noncopyable.hpp>
#include <boost/range/algorithm/sort.hpp>
#include <boost/range/algorithm/unique.hpp>
#include <boost/range/algorithm_ext/erase.hpp>
#include <boost/ref.hpp>
#include <boost/uuid/name_generator.hpp>
#include <boost/uuid/nil_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

#include "apartment.h"
#include "businterface.h"
#include "cache/apartment-proxy.h"
#include "ds-model.h"
#include "dss.h"
#include "event.h"
#include "logger.h"
#include "metering/metering.h"
#include "model-features.h"
#include "model/busscanner.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/floor.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/modelevent.h"
#include "model/modelmaintenance.h"
#include "model/set.h"
#include "model/state.h"
#include "model/zone.h"
#include "propertysystem.h"
#include "propertysystem_common_paths.h"

static ds::LogChannel dsLogChannel("modelApartment");

namespace dss {

struct ApartmentImpl : public Apartment, public boost::noncopyable {
  ApartmentImpl(DSS* dss, PropertySystem* propertySystem, BusInterface* busInterface)
      : Apartment(dss, propertySystem, busInterface) {}
  ~ApartmentImpl() = default;

  metering::Metering& getMetering() override { return *m_pMetering; }
  void setMetering(metering::Metering* _value) override { m_pMetering = _value; }

  ModelFeatures* getModelFeatures() override { return m_modelFeatures.get(); }

  metering::Metering* m_pMetering{nullptr}; // not our ownership
  std::unique_ptr<ModelFeatures> m_modelFeatures = makeModelFeatures();
};

std::unique_ptr<Apartment> makeApartment(DSS* dss, PropertySystem* propertySystem, BusInterface* busInterface) {
  return std::make_unique<ApartmentImpl>(dss, propertySystem, busInterface);
};

struct Apartment::Impl : private PropertyListener {
  Apartment& m_apartment;
  PropertySystem* m_propertySystem = m_apartment.getPropertySystem();
  PropertyNodePtr m_remoteConnectivity = m_propertySystem ? m_propertySystem->createProperty(pp_websvc_root) : nullptr;

  Impl(Apartment& apartment) : m_apartment{apartment} {
    if (m_remoteConnectivity) {
      m_remoteConnectivity->addListener(this);
    }
  }

  dsModel::ApartmentId readApartmentId() {
    if (!m_remoteConnectivity) {
      return localApartmentId();
    }
    return m_remoteConnectivity->getProperty("apartmentId")
               ? dsModel::ApartmentId{m_remoteConnectivity->tryChildValue<std::string>("apartmentId")
                                          .value_or(std::string{})}
               : localApartmentId();
  }

  dsModel::ApartmentId localApartmentId() {
    DS_INFO("Remote connectivity is disconnected. Using local apartmentId");
    if (!m_propertySystem) {
      return ds::system().conf().readOr<dsModel::ApartmentId>("apartmentId", {});
    }
    auto apartmentIdGenerator = boost::uuids::name_generator{boost::uuids::nil_generator{}()};
    auto apartmentId = apartmentIdGenerator(m_propertySystem->getStringValue(pp_sysinfo_dsuid));

    return dsModel::ApartmentId{boost::uuids::to_string(apartmentId)};
  }

  void propertyChangedImpl(const PropertyNodePtr& parent, const PropertyNodePtr& node) {
    if (parent == m_remoteConnectivity && node->getName() == "apartmentId") {
      m_apartment.m_changed.emit();
    }
  }

  void propertyChanged(PropertyNodePtr parent, PropertyNodePtr node) override {
    DS_INFO_ENTER(parent, node);
    propertyChangedImpl(parent, node);
  }

  void propertyRemoved(PropertyNodePtr parent, PropertyNodePtr node) override {
    DS_INFO_ENTER(parent, node);
    propertyChangedImpl(parent, node);
  }
  void propertyAdded(PropertyNodePtr parent, PropertyNodePtr node) override {
    DS_INFO_ENTER(parent, node);
    propertyChangedImpl(parent, node);
  }
};

//================================================== Apartment

Apartment::Apartment(DSS* dss, PropertySystem* propertySystem, BusInterface* busInterface)
    : m_dss(dss),
      m_pBusInterface(busInterface),
      m_pModelMaintenance(NULL),
      m_pPropertySystem(propertySystem),
      m_eventSink{},
      m_proxy(makeNullApartmentProxy()),
      m_impl{std::make_unique<Impl>(*this)} {
  // create default floor
  std::shared_ptr<Floor> floor = std::make_shared<Floor>(APARTMENT_FLOOR_ID, this);
  floor->setName("Apartment");
  m_Floors.push_back(floor);

  // create default (broadcast) zone
  std::shared_ptr<Zone> zoneZero = allocateZone(0);
  zoneZero->setIsPresent(true);
  zoneZero->setFloorID(FLOOR_ID_NONE);
  std::shared_ptr<Zone> zoneNewDevices = allocateZone(NEW_DEVICES_ZONE_ID);
  zoneNewDevices->setIsPresent(true);
  zoneNewDevices->setFloorID(FLOOR_ID_NONE);

  if (m_pPropertySystem) {
    m_pPropertyNode = m_pPropertySystem->createProperty("/apartment");
    for (auto& pZone : m_Zones) {
      pZone->publishToPropertyTree();
    }
    for (auto& pFloor : m_Floors) {
      pFloor->publishToPropertyTree();
    }
  }
  DS_INFO_LEAVE(id());
}

Apartment::~Apartment() {
  m_Zones.clear();
  m_Devices.clear();
  m_DSMeters.clear();
  m_States.clear();

  m_pPropertyNode.reset();
  m_pBusInterface = NULL;
  m_pModelMaintenance = NULL;
  m_pPropertySystem = NULL;
} // dtor

dsModel::ApartmentId Apartment::id() const {
  // avoid m_impl->m_apartmentId to prevent races
  return m_impl->readApartmentId();
}

template <class Items, class Id>
static auto ids(const Items& items, const Id& id) {
  auto result = std::vector<decltype(id(items.front()))>{};
  for (const auto& item : items) {
    if (item->isInDsModel()) {
      result.push_back(id(item));
    }
  }
  dsApi::normalizeIdArray(result);
  return result;
}

std::vector<dsModel::ZoneId> Apartment::zoneIds() const {
  return ids(getZones(), [](auto&& x) { return x->id(); });
}

std::vector<dsModel::ClusterId> Apartment::clusterIds() const {
  return ids(getClusters(), [](auto&& x) { return x->id(); });
}

std::vector<dsModel::DSDeviceId> Apartment::deviceIds() const {
  return ids(getDevicesVector(), [](auto&& x) { return x->id(); });
}

std::vector<dsModel::SubmoduleId> Apartment::submoduleIds() const {
  return ids(getDevicesVector(), [](auto&& x) { return x->submoduleId(); });
}

std::vector<dsModel::FunctionBlockId> Apartment::functionBlockIds() const {
  return ids(getDevicesVector(), [](auto&& x) { return x->functionBlockId(); });
}

std::vector<dsModel::FloorId> Apartment::floorIds() const {
  return ids(getFloors(), [](auto&& x) { return x->id(); });
}

std::vector<dsModel::ControllerId> Apartment::controllerIds() const {
  return ids(getDSMeters(), [](auto&& x) { return x->id(); });
}

void Apartment::pushEvent(const std::shared_ptr<Event>& event) {
  if (m_eventSink) {
    m_eventSink->pushEvent(event);
  }
}

void Apartment::addDefaultGroupsToZone(std::shared_ptr<Zone> _zone) {
  std::shared_ptr<Group> grp = std::make_shared<Group>(GroupIDBroadcast, _zone);
  grp->setName("broadcast");
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDYellow, _zone));
  grp->setName("yellow");
  grp->setApplicationType(ApplicationType::lights);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDGray, _zone));
  grp->setName("gray");
  grp->setApplicationType(ApplicationType::blinds);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDHeating, _zone));
  grp->setName("heating");
  grp->setApplicationType(ApplicationType::heating);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDCyan, _zone));
  grp->setName("cyan");
  grp->setApplicationType(ApplicationType::audio);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDViolet, _zone));
  grp->setName("magenta");
  grp->setApplicationType(ApplicationType::video);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDRed, _zone));
  grp->setName("reserved1");
  grp->setApplicationType(ApplicationType::none);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDGreen, _zone));
  grp->setName("reserved2");
  grp->setApplicationType(ApplicationType::none);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDBlack, _zone));
  grp->setName("black");
  grp->setApplicationType(ApplicationType::joker);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDCooling, _zone));
  grp->setName("cooling");
  grp->setApplicationType(ApplicationType::cooling);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDVentilation, _zone));
  grp->setName("ventilation");
  grp->setApplicationType(ApplicationType::ventilation);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDWindow, _zone));
  grp->setName("window");
  grp->setApplicationType(ApplicationType::window);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDRecirculation, _zone));
  grp->setName("recirculation");
  grp->setApplicationType(ApplicationType::recirculation);
  grp->setIsValid(true);
  _zone->addGroup(grp);
  grp.reset(new Group(GroupIDControlTemperature, _zone));
  grp->setName("controltemperature");
  grp->setApplicationType(ApplicationType::temperature);
  grp->setIsValid(true);
  _zone->addGroup(grp);
} // addDefaultGroupsToZone

std::shared_ptr<Device> Apartment::tryGetDeviceByDSID(const dsuid_t& dsid) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::shared_ptr<Device> dev : m_Devices) {
    if (dev->getDSID() == dsid) {
      return dev;
    }
  }
  return std::shared_ptr<Device>();
}

std::shared_ptr<Device> Apartment::getDeviceByDSID(const dsuid_t& dsid) const {
  auto device = tryGetDeviceByDSID(dsid);
  if (!device) {
    throw ItemNotFoundException(dsuid2str(dsid));
  }
  return device;
}

std::shared_ptr<Device> Apartment::device(const dsModel::DSDeviceId& deviceId) const {
  auto device = tryGetDeviceByDSID(str2dsuid(deviceId));
  DS_REQUIRE(device, "Device not found", deviceId);
  return device;
}

std::shared_ptr<Device> Apartment::getDeviceByName(const std::string& _name) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::shared_ptr<Device> dev : m_Devices) {
    if (dev->getName() == _name) {
      return dev;
    }
  }
  throw ItemNotFoundException(_name);
}
std::vector<std::shared_ptr<Device>> Apartment::coupledOEMDevices(const Device& device) {
  if (!device.coupledOEM()) {
    return std::vector<std::shared_ptr<Device>>();
  }

  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::vector<std::shared_ptr<Device>> ret;
  for (auto& dev : m_Devices) {
    if (device.isOemCoupledWith(*dev)) {
      ret.push_back(dev);
    }
  }
  return ret;
}

void Apartment::setName(const std::string& name) {
  if (m_Name == name) {
    return;
  }
  DeviceContainer::setName(name);
  m_changed.emit();
}

Set Apartment::getDevices() const {
  std::vector<DeviceReference> devs;
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::shared_ptr<Device> dev : m_Devices) {
    devs.push_back(DeviceReference(dev, this));
  }

  return Set(devs);
}

std::shared_ptr<Zone> Apartment::getZoneByName(const std::string& _zoneName) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::shared_ptr<Zone> zone : m_Zones) {
    if (zone->getName() == _zoneName) {
      return zone;
    }
  }
  throw ItemNotFoundException(_zoneName);
}

std::weak_ptr<Zone> Apartment::tryGetZone(const int id) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& zone : m_Zones) {
    if (zone->getID() == id) {
      return zone;
    }
  }
  return std::weak_ptr<Zone>();
}

std::shared_ptr<Zone> Apartment::getZone(const int id) const {
  if (auto&& zone = tryGetZone(id).lock()) {
    return std::move(zone);
  }
  throw ItemNotFoundException(intToString(id));
}

std::shared_ptr<Zone> Apartment::zone(const dsModel::ZoneId& id) const {
  if (auto&& zone = tryGetZone(strToInt(id)).lock()) {
    return std::move(zone);
  }
  DS_FAIL_REQUIRE("Zone not found", id);
}

std::vector<std::shared_ptr<Zone>> Apartment::getZones() const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  return m_Zones;
}

int Apartment::getNextFreeZoneId() const {
  for (int zoneId = 1; zoneId < NEW_DEVICES_ZONE_ID; zoneId++) {
    if (!tryGetZone(zoneId).lock()) {
      return zoneId;
    }
  }
  return -1;
}

std::weak_ptr<Floor> Apartment::tryGetFloor(const int id) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& floor : m_Floors) {
    if (floor->getID() == id) {
      return floor;
    }
  }
  return std::weak_ptr<Floor>();
}

std::shared_ptr<Floor> Apartment::getFloor(const int id) const {
  if (auto&& floor = tryGetFloor(id).lock()) {
    return std::move(floor);
  }
  DS_FAIL_REQUIRE("Floor not found", id);
}

std::shared_ptr<Floor> Apartment::floor(const dsModel::FloorId& id) const {
  return getFloor(strToInt(id));
}

std::vector<std::shared_ptr<Floor>> Apartment::getFloors() const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::vector<std::shared_ptr<Floor>> result = m_Floors;
  return result;
}

int Apartment::getNextFreeFloorID() const {
  for (int i = 1; i < INT_MAX; i++) {
    if (!tryGetFloor(i).lock()) {
      return i;
    }
  }
  return FLOOR_ID_NONE;
}

static bool sortFloorsByOrder(const std::shared_ptr<Floor>& f1, const std::shared_ptr<const Floor>& f2) {
  return (f1->getOrder() < f2->getOrder());
}

int Apartment::getNextFloorOrderID() {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  long long id = -1;
  for (size_t i = 0; i < m_Floors.size(); i++) {
    if (m_Floors.at(i)->getOrder() > id) {
      id = m_Floors.at(i)->getOrder();
    }
  }

  if (id < APARTMENT_FLOOR_ID) {
    id = APARTMENT_FLOOR_ID;
  }

  id++;
  if (id >= INT_MAX) { // about to wrap, time to look for holes
    // first sort the floors vector by order id
    std::sort(m_Floors.begin(), m_Floors.end(), sortFloorsByOrder);

    // the array is now sorted by order id, meaning that we set the orders
    // to be equal with the array index and thus eliminate any "jumps" in
    // ordering which might have been caused by previously removed floors
    for (size_t i = 0; i < m_Floors.size(); i++) {
      m_Floors.at(i)->setOrder(i);
    }

    if ((m_Floors.size() >= INT_MAX) || ((m_Floors.size() + 1) >= INT_MAX)) {
      id = -1;
    } else {
      id = (int)m_Floors.size();
    }
  }

  return (int)id;
}

std::shared_ptr<DSMeter> Apartment::getDSMeter(const std::string& _modName) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::shared_ptr<DSMeter> dsMeter : m_DSMeters) {
    if (dsMeter->getName() == _modName) {
      return dsMeter;
    }
  }
  throw ItemNotFoundException(_modName);
} // getDSMeter(name)

std::shared_ptr<DSMeter> Apartment::tryGetDSMeterByDSID(const dsuid_t& _dsid) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::shared_ptr<DSMeter> dsMeter : m_DSMeters) {
    if (dsMeter->getDSID() == _dsid) {
      return dsMeter;
    }
  }
  return std::shared_ptr<DSMeter>();
}

std::shared_ptr<DSMeter> Apartment::getDSMeterByDSID(const dsuid_t& _dsid) const {
  if (auto x = tryGetDSMeterByDSID(_dsid)) {
    return x;
  }
  throw ItemNotFoundException(dsuid2str(_dsid));
}

std::shared_ptr<DSMeter> Apartment::controller(const dsModel::ControllerId& id) const {
  if (auto x = tryGetDSMeterByDSID(str2dsuid(id))) {
    return x;
  }
  DS_FAIL_REQUIRE("Controller not found.", id);
}

std::shared_ptr<DSMeter> Apartment::dsMeterByDeviceDsuid(const dsuid_t& deviceDsuid) const {
  return getDSMeterByDSID(getDeviceByDSID(deviceDsuid)->getDSMeterDSID());
}

std::vector<std::shared_ptr<DSMeter>> Apartment::getDSMeters() const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::vector<std::shared_ptr<DSMeter>> result;
  for (std::shared_ptr<DSMeter> dsMeter : m_DSMeters) {
    // TODO(someday): remove the filter and move this logic to ui - use HasDevices?
    // we should not return vDSMs as they are not currently filtered in ui
    if (dsMeter->getBusMemberType() != BusMember_vDSM) {
      result.push_back(dsMeter);
    }
  }
  return result;
} // getDSMeters

std::vector<std::shared_ptr<DSMeter>> Apartment::getDSMetersAll() const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::vector<std::shared_ptr<DSMeter>> result = m_DSMeters;
  return result;
} // getAllDSMeters

// Group queries
std::shared_ptr<Group> Apartment::getGroup(const std::string& _name) {
  std::shared_ptr<Group> pResult = getZone(0)->getGroup(_name);
  if (pResult != NULL) {
    return pResult;
  }
  throw ItemNotFoundException(_name);
} // getGroup(name)

std::shared_ptr<Group> Apartment::getGroup(const int _id) {
  std::shared_ptr<Group> pResult = getZone(0)->getGroup(_id);
  if (pResult != NULL) {
    return pResult;
  }
  throw ItemNotFoundException(intToString(_id));
} // getGroup(id)

std::weak_ptr<Cluster> Apartment::tryGetCluster(const int id) const {
  if (isCluster(id)) {
    std::shared_ptr<Cluster> pResult = std::dynamic_pointer_cast<Cluster>(getZone(0)->getGroup(id));
    if (pResult != NULL) {
      return pResult;
    }
  }
  return std::weak_ptr<Cluster>();
}

std::shared_ptr<Cluster> Apartment::getCluster(const int id) const {
  if (auto&& cluster = tryGetCluster(id).lock()) {
    return std::move(cluster);
  }
  throw ItemNotFoundException(intToString(id));
} // getCluster(id)

std::shared_ptr<Cluster> Apartment::cluster(const dsModel::ClusterId& id) {
  if (auto&& cluster = tryGetCluster(strToInt(id)).lock()) {
    return std::move(cluster);
  }
  DS_FAIL_REQUIRE("Cluster not found", id);
}

std::vector<std::shared_ptr<Cluster>> Apartment::getClusters() const {
  std::vector<std::shared_ptr<Cluster>> result;
  for (std::shared_ptr<Group> pGroup : getZone(0)->getGroups()) {
    if (isCluster(pGroup->getID())) {
      result.push_back(std::dynamic_pointer_cast<Cluster>(pGroup));
    }
  }
  return result;
}

std::shared_ptr<Cluster> Apartment::getEmptyCluster() {
  // find a group slot with unassigned state machine id
  for (std::shared_ptr<Cluster> pCluster : getClusters()) {
    if (pCluster->getApplicationType() == ApplicationType::none) {
      return pCluster;
    }
  }
  return std::shared_ptr<Cluster>();
}

std::vector<std::shared_ptr<Group>> Apartment::getGlobalApps() {
  std::vector<std::shared_ptr<Group>> result;
  for (std::shared_ptr<Group> pGroup : getZone(0)->getGroups()) {
    if (isGlobalAppGroup(pGroup->getID())) {
      result.push_back(pGroup);
    }
  }
  return result;
}

std::shared_ptr<Device> Apartment::allocateDevice(const dsuid_t dsuid) {
  DS_INFO_ENTER(dsuid);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);

  auto pResult = tryGetDeviceByDSID(dsuid);
  if (pResult == NULL) {
    pResult = makeDevice(dsuid, this);
    pResult->setFirstSeen(DateTime());
    m_Devices.push_back(pResult);
  }

  auto zoneId = pResult->getZoneID();
  DeviceReference devRef(pResult, this);
  getZone(ZONE_ID_BROADCAST)->addDevice(devRef);
  if (auto zone = tryGetZone(zoneId).lock()) {
    zone->addDevice(devRef);
  }
  pResult->publishToPropertyTree();
  return pResult;
}

std::shared_ptr<DSMeter> Apartment::allocateDSMeter(const dsuid_t dsuid) {
  DS_INFO_ENTER(dsuid);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::shared_ptr<DSMeter> dsMeter : m_DSMeters) {
    if (dsMeter->getDSID() == dsuid) {
      return dsMeter;
    }
  }

  std::shared_ptr<DSMeter> pResult = std::make_shared<DSMeter>(dsuid, this);
  m_DSMeters.push_back(pResult);

  return pResult;
}

std::shared_ptr<Zone> Apartment::allocateZone(int zoneId) {
  DS_INFO_ENTER(zoneId);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::shared_ptr<Zone> result;
  for (std::shared_ptr<Zone> zone : m_Zones) {
    if (zone->getID() == zoneId) {
      result = zone;
      break;
    }
  }

  if (result == NULL) {
    result.reset(new Zone(zoneId, this));
    result->publishToPropertyTree();
    addDefaultGroupsToZone(result);
    m_Zones.push_back(result);

    if (zoneId == 0) {
      for (int i = GroupIDAppUserMin; i <= GroupIDAppUserMax; ++i) {
        std::shared_ptr<Cluster> pCluster = std::make_shared<Cluster>(i, boost::ref<Apartment>(*this));
        result->addGroup(pCluster);
      }

      // pre-create apartment ventilation group
      std::shared_ptr<Group> group;
      group.reset(new Group(GroupIDGlobalAppDsVentilation, result));
      group->setName("apartmentVentilation");
      group->setApplicationType(ApplicationType::apartmentVentilation);
      group->setIsValid(
          true); // TODO(soon): this may not be needed for AV (maybe it is only valid when devices are present)
      result->addGroup(group);

      // pre-create awnings group
      group.reset(new Group(GroupIDGlobalAppDsAwnings, result));
      group->setName("awnings");
      group->setApplicationType(ApplicationType::awnings);
      group->setIsValid(true);
      result->addGroup(group);

      // TODO(soon): remove again before R1705
      //        group.reset(new Group(GroupIDGlobalAppDsRecirculation, result));
      //        group->setName("apartmentRecirculation");
      //        group->setApplicationType(ApplicationType::apartmentRecirculation);
      //        group->setIsValid(true);  // TODO(soon): this may not be needed for AV (maybe it is only valid when
      //        devices are present)
      //        result->addGroup(group);
    } else {
      if ((result->getID() != ZONE_ID_BROADCAST) && (result->getID() != NEW_DEVICES_ZONE_ID)) {
        getFloor(0)->addZone(result->getID());
      }
    }
  } else {
    result->publishToPropertyTree();
  }
  return result;
}

void Apartment::removeZone(int zoneId) {
  DS_INFO_ENTER(zoneId);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::vector<std::shared_ptr<Zone>>::iterator ipZone = m_Zones.begin(), e = m_Zones.end(); ipZone != e;
       ++ipZone) {
    std::shared_ptr<Zone> pZone = *ipZone;
    if (pZone->getID() == zoneId) {
      pZone->removeFromPropertyTree();
      if (pZone->getFloorID() != FLOOR_ID_NONE) {
        getFloor(pZone->getFloorID())->removeZone(zoneId);
      }
      m_Zones.erase(ipZone);
      return;
    }
  }
}

std::shared_ptr<Floor> Apartment::allocateFloor(int id) {
  DS_INFO_ENTER(id);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::shared_ptr<Floor> result;

  if (id != FLOOR_ID_NONE) {
    for (std::shared_ptr<Floor> floor : m_Floors) {
      if (floor->getID() == id) {
        result = floor;
        break;
      }
    }
  }

  if (result == NULL) {
    if (id == FLOOR_ID_NONE) {
      id = getNextFreeFloorID();
      if (id == FLOOR_ID_NONE) {
        throw std::runtime_error("Maximum floor limit reached");
      }
    }
    result.reset(new Floor(id, this));
    if (id != APARTMENT_FLOOR_ID) {
      int order = getNextFloorOrderID();
      if (order == -1) {
        throw std::runtime_error("Maximum order id limit reached");
      }
      result->setOrder(order);
    }
    m_Floors.push_back(result);
  }

  result->publishToPropertyTree();
  return result;
}

void Apartment::removeFloor(int id) {
  DS_INFO_ENTER(id);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::vector<std::shared_ptr<Floor>>::iterator ipFloor = m_Floors.begin(), e = m_Floors.end(); ipFloor != e;
       ++ipFloor) {
    std::shared_ptr<Floor> pFloor = *ipFloor;
    if (pFloor->getID() == id) {
      pFloor->removeFromPropertyTree();
      m_Floors.erase(ipFloor);
      return;
    }
  }
}

void Apartment::removeDevice(dsuid_t dsuid) {
  DS_INFO_ENTER(dsuid);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);

  auto pDevice = tryGetDeviceByDSID(dsuid);
  if (!pDevice) {
    DS_NOTICE("not removing unknown device", dsuid);
    return;
  }

  // Remove from zone
  int zoneID = pDevice->getZoneID();
  DeviceReference devRef = DeviceReference(pDevice, this);

  {
    std::shared_ptr<DeviceReference> pDevRef = std::make_shared<DeviceReference>(devRef);
    std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::DeviceEvent, pDevRef);
    mEvent->setProperty("action", "deleted");
    mEvent->setProperty("dsuid", dsuid2str(pDevice->getDSID()));
    mEvent->setProperty("zoneID", intToString(pDevice->getZoneID()));
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(mEvent);
    }
  }

  if (zoneID != 0) {
    getZone(zoneID)->removeDevice(devRef);
  }
  getZone(0)->removeDevice(devRef);

  try {
    // Remove from dsm
    std::shared_ptr<DSMeter> dsMeter = getDSMeterByDSID(pDevice->getDSMeterDSID());
    dsMeter->removeDevice(devRef);
  } catch (ItemNotFoundException& e) {
    Logger::getInstance()->log(std::string("Apartment::removeDevice: Unknown dSM: ") + e.what(), lsWarning);
  }

  pDevice->clearBinaryInputs(); // calls apartment->removeState from inside
  pDevice->clearStates();       // calls apartment->removeState() from inside

  boost::remove_erase(m_Devices, pDevice);
}

void Apartment::removeDSMeter(dsuid_t dsuid) {
  DS_INFO_ENTER(dsuid);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (std::vector<std::shared_ptr<DSMeter>>::iterator ipDSMeter = m_DSMeters.begin(), e = m_DSMeters.end();
       ipDSMeter != e; ++ipDSMeter) {
    std::shared_ptr<DSMeter> pDSMeter = *ipDSMeter;
    if (pDSMeter->getDSID() == dsuid) {
      m_DSMeters.erase(ipDSMeter);
      return;
    }
  }
}

void Apartment::removeInactiveMeters() {
  DS_INFO_ENTER("");
  std::vector<std::shared_ptr<DSMeter>> dsMeters = getDSMeters();

  bool dirtyFlag = false;
  for (std::shared_ptr<DSMeter> dsMeter : dsMeters) {
    if (!dsMeter->isPresent()) {
      removeDSMeter(dsMeter->getDSID());
      dirtyFlag = true;
    }
  }

  if (dirtyFlag) {
    m_pModelMaintenance->addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
  }
}

std::shared_ptr<State> Apartment::allocateState(const eStateType type, const std::string& stateName,
                                                const std::string& scriptId) {
  DS_INFO_ENTER(type, stateName, scriptId);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::shared_ptr<State> result;
  for (auto&& state : m_States) {
    if ((state->getName() == stateName) && (state->getType() == type) && (state->getProviderService() == scriptId)) {
      result = state;
      break;
    }
  }
  if (result == NULL) {
    result = std::make_shared<State>(type, stateName, scriptId);
    m_States.push_back(result);
  }
  return result;
}

std::shared_ptr<State> Apartment::allocateState(const std::string& identifier, const std::string& scriptId,
                                                std::shared_ptr<Device> device, SensorType sensorType,
                                                const std::string& activateCondition,
                                                const std::string& deactivateCondition) {
  DS_INFO_ENTER(identifier, scriptId, device, sensorType, activateCondition, deactivateCondition);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::shared_ptr<State> result;
  auto name = StateSensor::buildName(device, sensorType, identifier);
  for (auto&& state : m_States) {
    if ((state->getName() == name) && (state->getType() == StateType_SensorDevice) &&
        (state->getProviderService() == scriptId)) {
      removeState(state);
      break;
    }
  }
  result =
      std::make_shared<StateSensor>(identifier, scriptId, device, sensorType, activateCondition, deactivateCondition);
  m_States.push_back(result);

  return result;
}

std::shared_ptr<State> Apartment::allocateState(const std::string& identifier, const std::string& scriptId,
                                                std::shared_ptr<Group> group, SensorType sensorType,
                                                const std::string& activateCondition,
                                                const std::string& deactivateCondition) {
  DS_INFO_ENTER(identifier, scriptId, group, sensorType, activateCondition, deactivateCondition);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::shared_ptr<State> result;
  auto name = StateSensor::buildName(group, sensorType, identifier);
  for (auto&& state : m_States) {
    if ((state->getName() == name) && (state->getType() == StateType_SensorZone) &&
        (state->getProviderService() == scriptId)) {
      removeState(state);
      break;
    }
  }
  result =
      std::make_shared<StateSensor>(identifier, scriptId, group, sensorType, activateCondition, deactivateCondition);
  m_States.push_back(result);
  return result;
}

std::shared_ptr<State> Apartment::allocateBinaryInputState(std::shared_ptr<Device> device, int inputIndex) {
  DS_INFO_ENTER(device, inputIndex)
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::shared_ptr<State> result;
  std::string name = BinaryInputState::makeName(device->getDSID(), inputIndex);
  for (auto&& state : m_States) {
    if (state->getName() == name) {
      return state;
    }
  }
  result = std::make_shared<BinaryInputState>(device, inputIndex);
  m_States.push_back(result);
  return result;
}

std::shared_ptr<State> Apartment::allocateDeviceState(std::shared_ptr<Device> device, const std::string& stateName) {
  DS_INFO_ENTER(device, stateName)
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::shared_ptr<State> result;
  std::string name = State::buildName(device, stateName);
  for (auto&& state : m_States) {
    if (state->getName() == name) {
      return state;
    }
  }
  result = std::make_shared<State>(device, stateName);
  m_States.push_back(result);
  return result;
}

std::shared_ptr<State> Apartment::allocateMeterState(std::shared_ptr<DSMeter> meter, int inputIndex) {
  DS_INFO_ENTER(meter, inputIndex)
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::shared_ptr<State> result;
  std::string name = State::buildName(meter, inputIndex);
  for (auto&& state : m_States) {
    if (state->getName() == name) {
      return state;
    }
  }
  result = std::make_shared<State>(meter, inputIndex);
  m_States.push_back(result);
  return result;
}

std::shared_ptr<State> Apartment::allocateGroupState(std::shared_ptr<Group> group, const std::string& name) {
  DS_INFO_ENTER(group, name)
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  std::shared_ptr<State> result;
  for (auto&& state : m_States) {
    if (state->getName() == name) {
      return state;
    }
  }
  result = std::make_shared<State>(group, name);
  m_States.push_back(result);
  return result;
}

std::shared_ptr<State> Apartment::tryGetState(const eStateType type, const std::string& name) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& state : m_States) {
    if ((state->getType() == type) && (state->getName() == name)) {
      return state;
    }
  }
  return std::shared_ptr<State>();
}

std::shared_ptr<State> Apartment::getState(const eStateType type, const std::string& name) const {
  auto state = tryGetState(type, name);
  if (!state) {
    throw ItemNotFoundException(name);
  }
  return state;
}

std::shared_ptr<State> Apartment::tryGetNonScriptState(const std::string& stateName) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& state : m_States) {
    if ((state->getType() != StateType_Script) && (state->getName() == stateName)) {
      return state;
    }
  }
  return {};
}

std::shared_ptr<State> Apartment::getNonScriptState(const std::string& stateName) const {
  if (auto state = tryGetNonScriptState(stateName)) {
    return state;
  }
  throw ItemNotFoundException(stateName);
}

std::shared_ptr<State> Apartment::getState(const eStateType type, const std::string& identifier,
                                           const std::string& stateName) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& state : m_States) {
    if ((state->getType() == type) && (state->getProviderService() == identifier) && (state->getName() == stateName)) {
      return state;
    }
  }
  throw ItemNotFoundException(stateName);
}

std::shared_ptr<State> Apartment::getState(const std::string& identifier, const std::string& name) const {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (auto&& state : m_States) {
    if ((state->getProviderService() == identifier) && (state->getName() == name)) {
      return state;
    }
  }
  DS_FAIL_REQUIRE("State not found", name);
}

// throws (dynamic_cast relies on consistent sensor names)
void Apartment::updateSensorStates(const dsuid_t& dsuid, SensorType sensorType, double value, CallOrigin origin) {
  DS_DEBUG_ENTER(dsuid, sensorType, value, origin);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  // A single sensor can have multiple states, by setting different on/off threshold
  // TODO(someday) SensorState should have a field to identify the input sensor
  std::string prefix("dev." + dsuid2str(dsuid) + ".type" + intToString(static_cast<int>(sensorType)));

  for (auto&& state : m_States) {
    if (boost::starts_with(state->getName(), prefix)) {
      std::dynamic_pointer_cast<StateSensor>(state)->newValue(origin, value);
    }
  }
}

// throws (dynamic_cast relies on consistent sensor names)
void Apartment::updateSensorStates(int zoneId, int groupId, SensorType sensorType, double value, CallOrigin origin) {
  DS_DEBUG_ENTER(zoneId, groupId, sensorType, value, origin);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  // A single sensor can have multiple states, by setting different on/off threshold
  // TODO(someday) SensorState should have a field to identify the input sensor
  std::string prefix("zone.zone" + intToString(zoneId) + ".group" + intToString(groupId) + ".type" +
                     intToString(static_cast<int>(sensorType)));

  for (auto&& state : m_States) {
    if (boost::starts_with(state->getName(), prefix)) {
      std::dynamic_pointer_cast<StateSensor>(state)->newValue(origin, value);
    }
  }
}

void Apartment::removeState(std::shared_ptr<State> state) {
  DS_INFO_ENTER(state);
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  m_States.erase(std::remove(m_States.begin(), m_States.end(), state));
}

std::vector<std::shared_ptr<State>> Apartment::getStates() {
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  return m_States;
}

ActionRequestInterface* Apartment::getActionRequestInterface() {
  if (m_pBusInterface != NULL) {
    return m_pBusInterface->getActionRequestInterface();
  }
  return NULL;
}

DeviceBusInterface* Apartment::getDeviceBusInterface() {
  if (m_pBusInterface != NULL) {
    return m_pBusInterface->getDeviceBusInterface();
  }
  return NULL;
}

std::vector<dsModel::ApartmentApplicationType> Apartment::apartmentApplications() const {
  std::vector<dsModel::ApartmentApplicationType> result;
  boost::recursive_mutex::scoped_lock scoped_lock(m_mutex);
  for (const auto& d : m_Devices) {
    if (auto x = ds::tryToEnum<dsModel::ApartmentApplicationType>(d->applicationType())) {
      result.push_back(*x);
    }
  }
  boost::erase(result, boost::unique<boost::return_found_end>(boost::sort(result)));
  return result;
}

void Apartment::aggregateSensorValue(SensorType _sensorType, const dsuid_t& source, double _value) {
  Logger::getInstance()->log(std::string("Apartment::aggregateSensorValue: ") + std::string("source: ") +
                                 dsuid2str(source) + std::string("sensorValue: ") + doubleToString(_value),
                             lsDebug);

  SensorSourceValue sensorSourceValue = {.source = source, .ts = ds::Clock::now(), .value = _value};
  switch (_sensorType) {
    case SensorType::WindSpeed: m_aggregateWindSpeed.setValue(sensorSourceValue); break;
    case SensorType::GustSpeed: m_aggregateWindGust.setValue(sensorSourceValue); break;
    case SensorType::TemperatureOutdoors: m_aggregateTemperature.setValue(sensorSourceValue); break;
    case SensorType::BrightnessOutdoors: m_aggregateBrightness.setValue(sensorSourceValue); break;
    case SensorType::HumidityOutdoors: m_aggregateHumidity.setValue(sensorSourceValue); break;
    case SensorType::Precipitation: m_aggregatePrecipitation.setValue(sensorSourceValue); break;
    case SensorType::AirPressure: m_aggregateAirPressure.setValue(sensorSourceValue); break;
    case SensorType::WindDirection: [[fallthrough]]; // device sensor - no aggregation
    case SensorType::GustDirection: [[fallthrough]]; // device sensor - no aggregation
    case SensorType::ActivePower: [[fallthrough]];
    case SensorType::OutputCurrent: [[fallthrough]];
    case SensorType::ElectricMeter: [[fallthrough]];
    case SensorType::TemperatureIndoors: [[fallthrough]];
    case SensorType::BrightnessIndoors: [[fallthrough]];
    case SensorType::HumidityIndoors: [[fallthrough]];
    case SensorType::CO2Concentration: [[fallthrough]];
    case SensorType::COConcentration: [[fallthrough]];
    case SensorType::SoundPressureLevel: [[fallthrough]];
    case SensorType::RoomTemperatureSetpoint: [[fallthrough]];
    case SensorType::RoomTemperatureControlVariable: [[fallthrough]];
    case SensorType::Status: [[fallthrough]];
    case SensorType::Reserved1: [[fallthrough]];
    case SensorType::Reserved2: [[fallthrough]];
    case SensorType::OutputCurrent16A: [[fallthrough]];
    case SensorType::ActivePowerVA: [[fallthrough]];
    case SensorType::Temperature: [[fallthrough]];
    case SensorType::Brightness: [[fallthrough]];
    case SensorType::Humidity: [[fallthrough]];
    case SensorType::WaterQuantity: [[fallthrough]];
    case SensorType::WaterFlowRate: [[fallthrough]];
    case SensorType::NotUsed: [[fallthrough]];
    case SensorType::UnknownType: [[fallthrough]];
    case SensorType::SunAzimuth: [[fallthrough]];
    case SensorType::SunElevation: [[fallthrough]];
    default: return; // no aggregation
  }
}

std::optional<SensorSourceValue> Apartment::getAggregatedSensorValue(SensorType _sensorType) {
  switch (_sensorType) {
    case SensorType::WindSpeed: return m_aggregateWindSpeed.get();
    case SensorType::GustSpeed: return m_aggregateWindGust.get();
    case SensorType::TemperatureOutdoors: return m_aggregateTemperature.get();
    case SensorType::BrightnessOutdoors: return m_aggregateBrightness.get();
    case SensorType::HumidityOutdoors: return m_aggregateHumidity.get();
    case SensorType::Precipitation: return m_aggregatePrecipitation.get();
    case SensorType::AirPressure: return m_aggregateAirPressure.get();
    case SensorType::WindDirection: [[fallthrough]]; // device sensor - no aggregation
    case SensorType::GustDirection: [[fallthrough]]; // device sensor - no aggregation
    case SensorType::ActivePower: [[fallthrough]];
    case SensorType::OutputCurrent: [[fallthrough]];
    case SensorType::ElectricMeter: [[fallthrough]];
    case SensorType::TemperatureIndoors: [[fallthrough]];
    case SensorType::BrightnessIndoors: [[fallthrough]];
    case SensorType::HumidityIndoors: [[fallthrough]];
    case SensorType::CO2Concentration: [[fallthrough]];
    case SensorType::COConcentration: [[fallthrough]];
    case SensorType::SoundPressureLevel: [[fallthrough]];
    case SensorType::RoomTemperatureSetpoint: [[fallthrough]];
    case SensorType::RoomTemperatureControlVariable: [[fallthrough]];
    case SensorType::Status: [[fallthrough]];
    case SensorType::Reserved1: [[fallthrough]];
    case SensorType::Reserved2: [[fallthrough]];
    case SensorType::OutputCurrent16A: [[fallthrough]];
    case SensorType::ActivePowerVA: [[fallthrough]];
    case SensorType::Temperature: [[fallthrough]];
    case SensorType::Brightness: [[fallthrough]];
    case SensorType::Humidity: [[fallthrough]];
    case SensorType::WaterQuantity: [[fallthrough]];
    case SensorType::WaterFlowRate: [[fallthrough]];
    case SensorType::NotUsed: [[fallthrough]];
    case SensorType::UnknownType: [[fallthrough]];
    case SensorType::SunAzimuth: [[fallthrough]];
    case SensorType::SunElevation: [[fallthrough]];
    default: return {}; // no aggregation
  }
}

ApartmentSensorStatus_t Apartment::getSensorStatus() const {
  return m_SensorStatus;
}

void Apartment::setTemperature(double _value, DateTime _ts) {
  m_SensorStatus.m_TemperatureValue = _value;
  m_SensorStatus.m_TemperatureTS = _ts;
}

void Apartment::setHumidityValue(double _value, DateTime _ts) {
  m_SensorStatus.m_HumidityValue = _value;
  m_SensorStatus.m_HumidityTS = _ts;
}

void Apartment::setBrightnessValue(double _value, DateTime _ts) {
  m_SensorStatus.m_BrightnessValue = _value;
  m_SensorStatus.m_BrightnessTS = _ts;
}

void Apartment::setWindSpeed(double _value, DateTime _ts) {
  m_SensorStatus.m_WindSpeedValue = _value;
  m_SensorStatus.m_WindSpeedTS = _ts;
}

void Apartment::setWindDirection(double _value, DateTime _ts) {
  m_SensorStatus.m_WindDirectionValue = _value;
  m_SensorStatus.m_WindDirectionTS = _ts;
}

void Apartment::setWindGust(double _value, DateTime _ts) {
  m_SensorStatus.m_WindGustValue = _value;
  m_SensorStatus.m_WindGustTS = _ts;
}

void Apartment::setPrecipitation(double _value, DateTime _ts) {
  m_SensorStatus.m_PrecipitationValue = _value;
  m_SensorStatus.m_PrecipitationTS = _ts;
}

void Apartment::setAirPressure(double _value, DateTime _ts) {
  m_SensorStatus.m_AirPressureValue = _value;
  m_SensorStatus.m_AirPressureTS = _ts;
}

void Apartment::setSunAzimuth(double _value, DateTime _ts) {
  m_SensorStatus.m_SunAzimuthValue = _value;
  m_SensorStatus.m_SunAzimuthTS = _ts;
}

void Apartment::setSunAzimuthLastLogTS(DateTime _ts) {
  m_SensorStatus.m_SunAzimuthLastLogTS = _ts;
}

void Apartment::setSunElevation(double _value, DateTime _ts) {
  m_SensorStatus.m_SunElevationValue = _value;
  m_SensorStatus.m_SunElevationTS = _ts;
}

void Apartment::setSunElevationLastLogTS(DateTime _ts) {
  m_SensorStatus.m_SunElevationLastLogTS = _ts;
}

void Apartment::setWeatherInformation(std::string&& iconId, std::string&& conditionId, std::string&& serviceId,
                                      DateTime ts) {
  m_SensorStatus.m_WeatherIconId = iconId;
  m_SensorStatus.m_WeatherConditionId = conditionId;
  m_SensorStatus.m_WeatherServiceId = serviceId;
  m_SensorStatus.m_WeatherTS = ts;
}

void Apartment::updateWeatherProtectedClusters(const double _sensorValue) {
  for (std::shared_ptr<Cluster> cluster : getClusters()) {
    auto protectionClass = cluster->getProtectionClass();
    if (protectionClass <= wpc_none || wpc_last <= protectionClass) {
      continue;
    }
    cluster->checkWindProtection(_sensorValue);
  }
}

std::pair<std::vector<DeviceLock_t>, std::vector<ZoneLock_t>> Apartment::getClusterLocks() {
  std::vector<DeviceLock_t> lockedDevices;
  std::vector<ZoneLock_t> lockedZones;

  std::vector<std::shared_ptr<Zone>> zones = getZones();
  // loop through all zones
  for (size_t z = 0; z < zones.size(); z++) {
    if (zones.at(z)->getID() == 0) {
      continue;
    }

    Set devices = zones.at(z)->getDevices();

    std::vector<ClassLock_t> classLocks;

    // look for devices of each class in the zone
    for (auto c = static_cast<int>(DeviceClasses_t::DEVICE_CLASS_GE);
         c <= static_cast<int>(DeviceClasses_t::DEVICE_CLASS_WE); c++) {
      if (c == static_cast<int>(DeviceClasses_t::DEVICE_CLASS_SW)) {
        // ignore unconfigured SW devices
        continue;
      }

      std::set<int> lockedScenes;

      Set devclass = devices.getByGroup(c);

      // loop through all devices of a certain group
      for (int d = 0; d < devclass.length(); d++) {
        std::shared_ptr<Device> device = devclass.get(d).getDevice();
        std::vector<int> ls = device->getLockedScenes();
        if (!ls.empty()) {
          DeviceLock_t lockedDevice;
          lockedDevice.dsuid = device->getDSID();
          lockedDevice.lockedScenes = ls;
          lockedDevices.push_back(lockedDevice);
          lockedScenes.insert(ls.begin(), ls.end());
        } else {
          if (device->isInLockedCluster()) {
            DeviceLock_t lockedDevice;
            lockedDevice.dsuid = device->getDSID();
            lockedDevices.push_back(lockedDevice);
          }
        }
      } // devices loop

      if (!lockedScenes.empty()) {
        ClassLock_t cl;
        cl.deviceClass = static_cast<DeviceClasses_t>(c);
        cl.lockedScenes = lockedScenes;
        classLocks.push_back(cl);
      }
    } // device class loop

    if (!classLocks.empty()) {
      ZoneLock_t zl;
      zl.zoneID = zones.at(z)->getID();
      zl.deviceClassLocks = classLocks;
      lockedZones.push_back(zl);
    }
  } // zone loop

  return std::make_pair(lockedDevices, lockedZones);
}

bool Apartment::setDevicesFirstSeen(const DateTime& dateTime) {
  static const DateTime allowedDate = DateTime::parseISO8601("2011-01-01T00:00:00Z");
  if (dateTime < allowedDate) {
    return false;
  }
  for (std::shared_ptr<Device> dev : m_Devices) {
    const DateTime originDate = dev->getFirstSeen();
    if (originDate < allowedDate) {
      dev->setFirstSeen(dateTime);
    }
  }
  m_pModelMaintenance->addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
  return true;
}

void Apartment::setProxy(std::unique_ptr<ApartmentProxy> proxy) {
  m_proxy = std::move(proxy);
}

ApartmentProxy& Apartment::proxy() const {
  return *m_proxy;
}

ItemNotFoundException::ItemNotFoundException(const std::string& name)
    : DSSException(::ds::strArgs("Could not find item. ", ds::hideArg(name), " ", DS_CONTEXT_ARGS)) {}

} // namespace dss
