/*
    Copyright (c) 2015 digitalSTROM.org, Zurich, Switzerland

    Author: Christian Hitz, digitalSTROM AG <christian.hitz@digitalstrom.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ds/locale/ds-locale.h>

#include "apartment.h"
#include "cluster.h"
#include "ds-model.h"
#include "dss.h"
#include "event/event_create.h"
#include "model/modelconst.h"
#include "model/set.h"
#include "model/state.h"
#include "propertysystem.h"

static ds::LogChannel dsLogChannel{"cluster"};

namespace dss {
//============================================= Cluster

Cluster::Cluster(const int _id, Apartment& _apartment)
    : Group(_id, _apartment.getZone(0)),
      m_Location(cd_none),
      m_ProtectionClass(wpc_none),
      m_WindProtectionTimer(std::nullopt),
      m_Floor(0),
      m_ConfigurationLocked(false),
      m_automatic(false) {} // ctor

Cluster::~Cluster() {
  // TODO unpublishPropertyTree
  if (NULL != m_oplock_state) {
    m_pApartment->removeState(m_oplock_state);
  }
}

bool Cluster::isInDsModel() const {
  return applicationType() != dsModel::ClusterApplicationType::none;
}

dsModel::ClusterId Cluster::id() const {
  DS_REQUIRE(isInDsModel(), "Accessing cluster that does not exist in ds model.", *this);
  return dsModel::ClusterId{ds::str(getID())};
}

dsModel::ClusterApplicationType Cluster::applicationType() const {
  switch (getApplicationType()) {
    case dss::ApplicationType::none: break;
    case dss::ApplicationType::lights: return dsModel::ClusterApplicationType::lights;
    case dss::ApplicationType::blinds: {
      if (m_cachedApplicationType == dsModel::ClusterApplicationType::none) {
        calculateCachedApplicationType();
      }
      return m_cachedApplicationType;
    }
    case dss::ApplicationType::awnings: return dsModel::ClusterApplicationType::awnings;
    case dss::ApplicationType::heating: return dsModel::ClusterApplicationType::heating;
    case dss::ApplicationType::audio: return dsModel::ClusterApplicationType::audio;
    case dss::ApplicationType::video: return dsModel::ClusterApplicationType::video;
    case dss::ApplicationType::cooling: return dsModel::ClusterApplicationType::cooling;
    case dss::ApplicationType::ventilation: return dsModel::ClusterApplicationType::ventilation;
    case dss::ApplicationType::recirculation: return dsModel::ClusterApplicationType::recirculation;
    case dss::ApplicationType::security:
    case dss::ApplicationType::access:
    case dss::ApplicationType::joker:
    case dss::ApplicationType::window:
    case dss::ApplicationType::temperature:
    case dss::ApplicationType::apartmentVentilation:
    case dss::ApplicationType::apartmentRecirculation: break;
  }
  return {};
}

void Cluster::calculateCachedApplicationType() const {
  if (getApplicationType() == dss::ApplicationType::blinds) {
    if (hasLegacyAwnings() && !hasShades()) {
      // ignore commands that are adressed to blinds, control awnings/blinds independently
      m_cachedApplicationType = dsModel::ClusterApplicationType::awnings;
    } else {
      m_cachedApplicationType = dsModel::ClusterApplicationType::shades;
    }
  }
}

void Cluster::updateLockedScenes() {
  if (!m_pPropertyNode) {
    return;
  }
  PropertyNodePtr pLockedScenes = m_pPropertyNode->getProperty("lockedScenes");
  if (pLockedScenes) {
    pLockedScenes->getParentNode()->removeChild(pLockedScenes);
  }
  pLockedScenes = m_pPropertyNode->createProperty("lockedScenes");
  for (int scene : m_LockedScenes) {
    PropertyNodePtr gsubnode = pLockedScenes->createProperty("scene" + intToString(scene));
    gsubnode->createProperty("id")->setIntegerValue(scene);
  }
} // updateLockedScenes

void Cluster::updateLockEvents() {
  PropertyNodePtr pOperationLocks = m_pPropertyNode->getProperty("lockEvents");
  if (pOperationLocks) {
    pOperationLocks->getParentNode()->removeChild(pOperationLocks);
  }

  pOperationLocks = m_pPropertyNode->createProperty("lockEvents");
  for (size_t i = 0; i < m_LockEvents.size(); i++) {
    // false is implicit, if something is not in the list then it is not set,
    // basically we only care about locks that are set
    PropertyNodePtr gsubnode = pOperationLocks->createProperty(m_LockEvents.at(i));
    gsubnode->setBooleanValue(true);
  }
} // updateLockEvents

void Cluster::publishToPropertyTree() {
  Group::publishToPropertyTree();
  if (m_pPropertyNode != NULL) {
    m_pPropertyNode->createProperty("CardinalDirection")
        ->linkToProxy(PropertyProxyToString<CardinalDirection_t>(m_Location));
    m_pPropertyNode->createProperty("WindProtectionClass")
        ->linkToProxy(PropertyProxyReference<int, WindProtectionClass_t>(m_ProtectionClass));
    m_pPropertyNode->createProperty("Floor")->linkToProxy(PropertyProxyReference<int>(m_Floor, false));
    m_pPropertyNode->createProperty("ConfigurationLocked")
        ->linkToProxy(PropertyProxyReference<bool>(m_ConfigurationLocked, false));
    m_pPropertyNode->createProperty("Automatic")->linkToProxy(PropertyProxyReference<bool>(m_automatic, false));
    updateLockedScenes();
    updateLockEvents();
  }
} // publishToPropertyTree

void Cluster::setFromSpec(const ClusterSpec_t& spec) {
  Group::setFromSpec(spec);
  setLocation(static_cast<CardinalDirection_t>(spec.location));
  setProtectionClass(static_cast<WindProtectionClass_t>(spec.protectionClass));
  setConfigurationLocked(spec.configurationLocked);
  setLockedScenes(spec.lockedScenes);
}

bool Cluster::isConfigEqual(const ClusterSpec_t& cluster) {
  if (m_LockedScenes.size() != cluster.lockedScenes.size()) {
    return false;
  }
  for (unsigned int i = 0; i < m_LockedScenes.size(); ++i) {
    if (m_LockedScenes[i] != cluster.lockedScenes[i]) {
      return false;
    }
  }
  return (Group::isConfigEqual(cluster) && (m_Location == cluster.location) &&
          (m_ProtectionClass == cluster.protectionClass) && (m_Floor == cluster.floor) &&
          (m_ConfigurationLocked == cluster.configurationLocked));
} // isConfigEqual

void Cluster::reset() {
  setLocation(cd_none);
  setProtectionClass(wpc_none);
  setFloor(0);
  m_LockedScenes.clear();
  setConfigurationLocked(false);
  setAutomatic(false);
  setApplicationType(ApplicationType::none);
  setName("");
  setApplicationConfiguration(0);
  m_LockEvents.clear();
  setLastCalledScene(0);
  updateLockedScenes();
  updateLockEvents();
  if (m_oplock_state) {
      m_oplock_state.reset();
  }
} // reset

bool Cluster::releaseCluster() {
  if (getDevices().isEmpty()) {
    reset();
    return true;
  }
  return false;
}

bool Cluster::isOperationLock() {
  // unknown -> not locked
  return (NULL != m_oplock_state) && (m_oplock_state->getState() == State_Active);
}

void Cluster::checkWindProtection(const double _value) {
  DS_DEBUG_ENTER(_value);
  auto wind = m_pApartment->tryGetNonScriptState("wind.group" + ds::str(getID()));
  bool windActive = wind ? wind->getState() == State_Active : false;

  if (_value > getClassValue(m_ProtectionClass) && !windActive) {
    callScene(CallOrigin::json, SceneAccessCategory::manual, SceneWindActive, "", true);
    setWindProtectionTimer(DateTime());
    DSS::getInstance()->raiseEvent(createClusterCheckWindEvent(getID(), ClusterWindTimeout));
    sleep(1); // prevent flooding dsm queue when wind is called on multiple clusters
  } else if (_value >= (0.90 * getClassValue(m_ProtectionClass)) && windActive) {
    setWindProtectionTimer(DateTime());
  }
}

void Cluster::addLockEvent(std::string stateName) {
  if (std::find(m_LockEvents.begin(), m_LockEvents.end(), stateName) == m_LockEvents.end()) {
    m_LockEvents.push_back(stateName);
    updateLockEvents();
  }
}

void Cluster::removeLockEvent(std::string stateName) {
  m_LockEvents.erase(std::remove(m_LockEvents.begin(), m_LockEvents.end(), stateName), m_LockEvents.end());
  updateLockEvents();
}

void Cluster::setOperationLock(bool _locked, CallOrigin _callOrigin) {
  if (NULL == m_oplock_state) {
    std::string name = "cluster." + intToString(getID()) + ".operation_lock";
    m_oplock_state = m_pApartment->allocateState(StateType_Group, name, "");
    m_oplock_state->setPersistence(true);
    m_oplock_state->setState(CallOrigin::system, State_Unknown);
  }
  m_oplock_state->setState(_callOrigin, _locked ? State_Active : State_Inactive);
}
} // namespace dss
