/*
    Copyright (c) 2009,2011 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/
#pragma once

#include "dssfwd.h"
#include "model/device.h"

DS_REFLECTED_ENUM((dss), InputMode, (button)(sensor)(output))

namespace dss {

Dsuid createNextDSUID();
dsuid_t createMultiDsuid(int deviceCount);

// https://confluence.digitalstrom.org/display/PL/Function+ID
enum class FID100 {
  varButton = 0x20, // value and target of key can be changed
  output = 0x10,    // output present
  reserved1 = 0x08,
  // bits 0..2 encode number of keys present, see FID100keys;
};

enum class FID100keys {
  _1 = 0x00, // CAUTION: 0 means 1 key
  _2 = 0x01,
  _4 = 0x02,
  // values 3..6 reserved
  _0 = 0x07,
};

// NOTE: R100 is broken, e.g. Device::hasInput does not support it
inline auto functionID100(DeviceClasses_t color, FID100keys keys) {
  return static_cast<uint16_t>(static_cast<unsigned int>(color) << 12 | static_cast<unsigned int>(keys));
}
inline uint16_t functionID100(DeviceClasses_t color, FID100 opt1, FID100keys keys) {
  return functionID100(color, keys) | static_cast<uint16_t>(opt1);
}
inline uint16_t functionID100(DeviceClasses_t color, FID100 opt1, FID100 opt2, FID100keys keys) {
  return functionID100(color, opt1, keys) | static_cast<uint16_t>(opt2);
}

// https://confluence.digitalstrom.org/display/PL/Function+ID
enum class FID105 {
  varRampTime = 0x20,    // variable ramp time, can be configure in register RAMPTIMEMAX
  output = 0x10,         // output present
  extraEquipment = 0x08, // device provides extra equipment, sensors, binary inputs
  extraGroups = 0x04,    // group bits 0-15 and 48-63 must be read. device can be member of groups 1-15 and 48-55
  // bits 0..1 encode number of keys present, see FID105keys;
};

enum class FID105keys {
  _0 = 0x0,
  _1 = 0x1,
  _2 = 0x2,
  _4 = 0x3,
};

// 0000              class
//     000000        subclass [ R100(b000), R105(b100), R105v2(b111) ]
//           000000  function module (equipment)

inline uint16_t functionID105(DeviceClasses_t color, FID105keys keys) {
  return static_cast<uint16_t>(static_cast<unsigned int>(color) << 12 | 0b100 << 6 | static_cast<unsigned int>(keys));
}
inline uint16_t functionID105(DeviceClasses_t color, FID105 opt1, FID105keys keys) {
  return functionID105(color, keys) | static_cast<uint16_t>(opt1);
}
inline uint16_t functionID105(DeviceClasses_t color, FID105 opt1, FID105 opt2, FID105keys keys) {
  return functionID105(color, opt1, keys) | static_cast<uint16_t>(opt2);
}

// support groups beyond 64
inline uint16_t functionID105v2(DeviceClasses_t color, FID105keys keys) {
  return static_cast<uint16_t>(static_cast<unsigned int>(color) << 12 | 0b111 << 6 | static_cast<unsigned int>(keys));
}
inline uint16_t functionID105v2(DeviceClasses_t color, FID105 opt1, FID105keys keys) {
  return functionID105v2(color, keys) | static_cast<uint16_t>(opt1);
}
inline uint16_t functionID105v2(DeviceClasses_t color, FID105 opt1, FID105 opt2, FID105keys keys) {
  return functionID105v2(color, opt1, keys) | static_cast<uint16_t>(opt2);
}

// subclass 8: support groups beyond 64, internal temperature control
inline uint16_t functionID105v2Heating(DeviceClasses_t color, FID105keys keys) {
  DS_ASSERT(color == DeviceClasses_t::DEVICE_CLASS_BL);
  return static_cast<uint16_t>(static_cast<unsigned int>(color) << 12 | 0b1000 << 6 | static_cast<unsigned int>(keys));
}

// subclass 9: support groups beyond 64, internal temperature control, active cooling
inline uint16_t functionID105v2Climate(DeviceClasses_t color, FID105keys keys) {
  DS_ASSERT(color == DeviceClasses_t::DEVICE_CLASS_BL);
  return static_cast<uint16_t>(static_cast<unsigned int>(color) << 12 | 0b1001 << 6 | static_cast<unsigned int>(keys));
}

struct DeviceMockConf {
  HardwareType hardwareType;
  std::optional<uint16_t> revisionId; // aka fw version
  std::vector<int> groups;
  std::optional<ButtonInputMode> buttonInputMode;
};

std::shared_ptr<Device> createDevice(Apartment& apartment, const DeviceMockConf& conf);
std::vector<std::shared_ptr<Device>> createSWSKM300(Apartment& apartment, const DeviceMockConf& conf);
std::vector<std::shared_ptr<Device>> createSWUMR200(Apartment& apartment, const DeviceMockConf& conf);
std::vector<std::shared_ptr<Device>> createGRKL300(Apartment& apartment, const DeviceMockConf& conf);

inline void markOnline(dss::Device& dev) {
  dev.setIsPresent(true);
  dev.setIsConnected(true);
}

} // namespace dss
