/*
    Copyright (c) 2012 digitalSTROM.org, Zurich, Switzerland

    Author: Michael Tross, aizo GmbH <michael.tross@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include <ds/log.h>

#include <memory>
#include <regex>

#include "base.h"
#include "dss.h"
#include "event.h"
#include "event/event_create.h"
#include "logger.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/zone.h"
#include "modelmaintenance.h"
#include "propertysystem.h"
#include "state.h"

namespace dss {

  static ds::LogChannel dsLogChannel("modelState");

    //============================================= State
  __DEFINE_LOG_CHANNEL__(State, lsNotice);

  const std::string State::INVALID("invalid");

  State::State(const std::string& name, eStateType type, eState initState, StatePersistency persistency)
    : m_name(name),
      m_persistency(persistency),
      m_callOrigin(CallOrigin::unknown),
      m_originDeviceDSUID(DSUID_NULL),
      m_state(initState),
      m_type(type)
  {
    // common base class; no load/publishToPropertyTree until fully initialized
  }

  State::State(eStateType _type, const std::string& _name, const std::string& _identifier)
  : m_name(_name),
    m_persistency(StatePersistency::volatile_),
    m_callOrigin(CallOrigin::unknown),
    m_originDeviceDSUID(DSUID_NULL),
    m_state(State_Inactive),
    m_type(_type),
    m_serviceName(_identifier)
  {
    // TODO: check if _identifier is valid, e.g. the prop tree does not like
    //       an empty identifier
    load();
    publishToPropertyTree();
  }

  std::string BinaryInputState::makeName(const dsuid_t &dsuid, int inputIndex) {
    return "dev." + dsuid2str(dsuid) + "." + intToString(inputIndex);
  }

  BinaryInputState::BinaryInputState(std::shared_ptr<Device> device, int inputIndex) :
    State(makeName(device->getDSID(), inputIndex),
          StateType_Device,
          State_Invalid,
          StatePersistency::persistent)
  {
    m_providerDev = std::weak_ptr<Device>(device),
    m_providerDevInput = inputIndex;
    load();
    publishToPropertyTree();
  }

  BinaryInputState::~BinaryInputState() = default;

  std::string State::buildName(const std::shared_ptr<Device>& _device,
                               const std::string& stateName) {
    return "dev." + dsuid2str(_device->getDSID()) + "." + stateName;
  }

  State::State(std::shared_ptr<Device> _device, const std::string& stateName) :
      m_name(buildName(_device, stateName)),
      m_persistency(StatePersistency::persistent),
      m_callOrigin(CallOrigin::unknown),
      m_originDeviceDSUID(DSUID_NULL),
      m_state(State_Invalid),
      m_type(StateType_Device),
      m_providerDev(_device),
      m_providerDevInput(-1),
      m_providerDevStateName(stateName)
  {
    load();
    publishToPropertyTree();
  }

  std::string State::buildName(const std::shared_ptr<DSMeter>& _meter, int _inputIndex) {
    return "dsm." + dsuid2str(_meter->getDSID()) + "." + intToString(_inputIndex);
  }

  State::State(std::shared_ptr<DSMeter> _meter, int _inputIndex) :
      m_name(buildName(_meter, _inputIndex)),
      m_persistency(StatePersistency::persistent),
      m_callOrigin(CallOrigin::unknown),
      m_originDeviceDSUID(DSUID_NULL),
      m_state(State_Invalid),
      m_type(StateType_Circuit),
      m_providerDevInput(_inputIndex),
      m_providerDsm(_meter)
  {
    load();
    publishToPropertyTree();
  }

  std::string State::makeGroupName(const Group& group) {
    auto&& logicalName = [&]() -> std::string {
      auto&& applicationType = group.getApplicationType();
      switch (applicationType) {
        case ApplicationType::lights:
          return "light"; // ATTENTION: This is inconsistent with group.getName() == "yellow"
          break;
        case ApplicationType::heating: return "heating"; break;
        case ApplicationType::none:
        case ApplicationType::blinds:
        case ApplicationType::audio:
        case ApplicationType::video:
        case ApplicationType::security:
        case ApplicationType::access:
        case ApplicationType::joker:
        case ApplicationType::cooling:
        case ApplicationType::ventilation:
        case ApplicationType::window:
        case ApplicationType::recirculation:
        case ApplicationType::temperature:
        case ApplicationType::apartmentVentilation:
        case ApplicationType::apartmentRecirculation:
        case ApplicationType::awnings:
        default:
          return "unknown";
          break;
      }
    }();
    return "zone." + intToString(group.getZoneID()) + "." + logicalName;
  }

  State::State(std::shared_ptr<Group> _group, const std::string& name) :
    m_name(name),
    m_persistency(StatePersistency::volatile_),
    m_callOrigin(CallOrigin::unknown),
    m_originDeviceDSUID(DSUID_NULL),
    m_state(State_Unknown),
    m_type(StateType_Group),
    m_providerGroup(_group)
  {
    load();
    publishToPropertyTree();
  }

  State::~State() {
    removeFromPropertyTree();
  }

  void State::publishToPropertyTree() {
    removeFromPropertyTree();
    std::string path;
    switch (m_type) {
      case StateType_Script:
      case StateType_SensorDevice:
      case StateType_SensorZone:
        path = "/usr/addon-states/" + m_serviceName + "/";
        break;
      case StateType_Apartment:
      case StateType_Device:
      case StateType_Service:
      case StateType_Group:
      case StateType_Circuit:
      default:
        path = "/usr/states/";
        break;
    }
    path += m_name;
    if (m_pPropertyNode == NULL && DSS::hasInstance()) {
      if (DSS::getInstance()->getPropertySystem().getProperty(path) != NULL) {
        // avoid publishing duplicate state information when reading out devices multiple times
        // and re-use the existing state node
        m_pPropertyNode = DSS::getInstance()->getPropertySystem().getProperty(path);
        DS_WARNING("StateProperty already exists: " + path);
        return;
      }
      DS_NOTICE("Publishing to the property tree", path);
      m_pPropertyNode = DSS::getInstance()->getPropertySystem().createProperty(path);
      m_pPropertyNode->createProperty("name")
        ->linkToProxy(PropertyProxyReference<std::string>(m_name, false));
      m_pPropertyNode->createProperty("value")
        ->linkToProxy(PropertyProxyReference<int, int>(m_state, false));
      m_pPropertyNode->createProperty("state")
        ->linkToProxy(PropertyProxyMemberFunction<State, std::string, false>(*this, &State::toString));
      if (m_callOrigin != CallOrigin::unknown) {
        m_pPropertyNode->createProperty("callOrigin")
          ->linkToProxy(PropertyProxyReference<int, CallOrigin>(m_callOrigin, false));
      }
      if (m_originDeviceDSUID != DSUID_NULL) {
        m_pPropertyNode->createProperty("originDeviceDSUID")
          ->linkToProxy(PropertyProxyMemberFunction<State, std::string, false>(*this, &State::getOriginDeviceDSUIDString));
      }

      // #5870 - keep compatibility for "presence" and "hibernation"
      if (m_name == "presence" || m_name == "hibernation") {
        m_pPropertyNode->removeChild(m_pPropertyNode->getPropertyByName("value"));
        m_pPropertyNode->createProperty("value")
          ->linkToProxy(PropertyProxyMemberFunction<State, std::string, false>(*this, &State::toString));
      }

      if (std::shared_ptr<Device> providerDev = m_providerDev.lock()) {
        PropertyNodePtr devNode = providerDev->getPropertyNode();
        PropertyNodePtr m_pAliasNode = m_pPropertyNode->createProperty("device/" + dsuid2str(providerDev->getDSID()));
        m_pAliasNode->alias(devNode);
        if (m_providerDevInput != -1) {
          m_pPropertyNode->createProperty("device/inputIndex")
            ->linkToProxy(PropertyProxyReference<int>(m_providerDevInput, false));
        }
        if (!m_providerDevStateName.empty()) {
          m_pPropertyNode->createProperty("device/stateName")
            ->linkToProxy(PropertyProxyReference<std::string>(m_providerDevStateName, false));
        }
      }

      if (std::shared_ptr<DSMeter> providerDsm = m_providerDsm.lock()) {
        PropertyNodePtr devNode = providerDsm->getPropertyNode();
        PropertyNodePtr m_pAliasNode = m_pPropertyNode->createProperty("dSMeter/" + dsuid2str(providerDsm->getDSID()));
        m_pAliasNode->alias(devNode);
        m_pPropertyNode->createProperty("dSMeter/inputIndex")
          ->linkToProxy(PropertyProxyReference<int>(m_providerDevInput, false));
      }

      if (m_serviceName.length() > 0) {
        m_pPropertyNode->createProperty("service/name")
            ->linkToProxy(PropertyProxyReference<std::string>(m_serviceName, false));
      }
    }
  } // publishToPropertyTree

  void State::removeFromPropertyTree() {
    if(m_pPropertyNode != NULL) {
      DS_NOTICE("Removing from the property tree", m_name);
      m_pPropertyNode->unlinkProxy(true);
      PropertyNode *parent = m_pPropertyNode->getParentNode();
      if (parent != NULL) {
        parent->removeChild(m_pPropertyNode);
      }
      m_pPropertyNode.reset();
    }
  } // removeFromPropertyTree

  std::string State::getStorageName() const {
    std::string dirname;
    if (!DSS::hasInstance()) {
      dirname = "/tmp";
    } else {
      dirname = DSS::getInstance()->getSavedPropsDirectory();
    }
    std::string filename = "state." + getName();
    return dirname + "/" + filename;
  }

  void State::save() {
    FILE *fout = fopen(getStorageName().c_str(), "w");
    if (NULL == fout) {
      Logger::getInstance()->log("State " + m_name + " could not be written", lsError);
      return;
    }
    fputc((int) m_state, fout);
    fputc(static_cast<int>(m_callOrigin), fout);
    for (int i = 0; i < DSUID_SIZE; i++) {
      fputc(static_cast<int>(m_originDeviceDSUID.id[i]), fout);
    }
    fclose(fout);
  } // save

  void State::load() {
    FILE *fin = fopen(getStorageName().c_str(), "r");
    if (NULL == fin) {
      return;
    }
    int value = fgetc(fin);
    if (value != EOF) {
      m_state = (eState) value;
    }
    value = fgetc(fin);
    if (value != EOF) {
      m_callOrigin = static_cast<CallOrigin>(value);
    }
    for (int i = 0; i < DSUID_SIZE; i++) {
      value = fgetc(fin);
      if (value != EOF) {
        m_originDeviceDSUID.id[i] = static_cast<char>(value);
      } else {
        m_originDeviceDSUID = DSUID_NULL;
        break;
      }
    }
    fclose(fin);
    return;
  } // load

  bool State::getPersistence() const {
    return m_persistency == StatePersistency::persistent;
  } // getPersistence

  void State::setPersistence(bool _persistent) {
    if (m_persistency == StatePersistency::volatile_ && _persistent) {
      load();
    }
    m_persistency = StatePersistency::persistent;
  } // setPersistence

  bool State::hasPersistentData() {
    std::string fname(getStorageName());
    if (access(fname.c_str(), F_OK) == 0) {
      return true;
    }
    return false;
  } // hasPersistentData

  void State::removePersistentData() const {
    std::string fname(getStorageName());
    if (access(fname.c_str(), F_OK) == 0) {
      unlink(fname.c_str());
    }
  }

  std::string State::toString() const {
    if (!m_values.empty()) {
      if (m_state >= 0 && m_state < (int) m_values.size()) {
        return m_values[m_state];
      } else {
        return INVALID;
      }
    }
    switch(m_state) {
    case State_Unknown:
      return std::string("unknown");
    case State_Active:
      return std::string("active");
    case State_Inactive:
      return std::string("inactive");
    default:
      return INVALID;
    }
  } // toString

  std::optional<int> State::tryValueFromName(const std::string& valueName) {
    if (!m_values.empty()) {
      auto it = std::find(m_values.begin(), m_values.end(), valueName);
      if (it != m_values.end()) {
        return it - m_values.begin();
      }
      return {};
    }

    if (valueName == "active") {
      return State_Active;
    } else if (valueName == "inactive") {
      return State_Inactive;
    } else if (valueName == "unknown") {
      return State_Unknown;
    } else if (valueName == INVALID) {
      return State_Invalid;
    } else {
      return {};
    }
  }

  int State::valueFromName(const std::string& valueName) {
    if (auto&& value = tryValueFromName(valueName)) {
      return *value;
    }
    DS_FAIL_REQUIRE("Unrecorgnized state value", valueName);
  }

  std::optional<int> State::tryGetClusterID() {
    static const std::regex weatherStateRegex(std::string("\\w+\\.group(\\d+)"));
    static const std::regex opLockRegex(std::string("cluster\\.(\\d+)\\..*"));
    if (m_name.empty()) {
      return {};
    }

    std::smatch match{};
    if ((std::regex_match(m_name, match, weatherStateRegex)) ||
        (std::regex_match(m_name, match, opLockRegex))) {
      return strToInt(match[1]);
    }
    return {};
  }

  int State::getState() const {
    return m_state;
  } // getState

  int State::getPrevState() const {
    return m_prevState;
  } // getPrevState

  void State::setState(const CallOrigin _origin, const int value) {
    if (value != m_state) {
      m_prevState = m_state;
      m_state = value;
      m_callOrigin = _origin;
      DS_NOTICE("New value", *this, value, toString());

      if (m_pPropertyNode == nullptr || m_pPropertyNode->getParentNode() == nullptr) {
        DS_WARNING("Property tree entry does not exist, creating");
        publishToPropertyTree();
      }

      if (m_pPropertyNode != NULL && DSS::hasInstance() && (m_callOrigin != CallOrigin::unknown)) {
        m_pPropertyNode->createProperty("callOrigin")
          ->linkToProxy(PropertyProxyReference<int, CallOrigin>(m_callOrigin, false));
      }

      if (m_persistency == StatePersistency::persistent) {
        save();
      }

      if (DSS::hasInstance() && DSS::getInstance()->getState() == ssRunning) {
        DSS::getInstance()->getEventQueue()
          .pushEvent(createStateChangeEvent(shared_from_this(), m_prevState, _origin));
      }
    }
  } // setState

  void State::setState(const CallOrigin _origin, const std::string& valueName) {
    if (auto&& value = tryValueFromName(valueName)) {
      setState(_origin, *value);
    }
  }

  unsigned int State::getValueRangeSize() const {
    return (!m_values.empty()) ? m_values.size() : (State_Unknown + 1);
  }

  void State::setOriginDeviceDSUID(const dsuid_t _dsuid) {
    m_originDeviceDSUID = _dsuid;
    if (m_pPropertyNode != NULL && DSS::hasInstance() &&
        (m_originDeviceDSUID != DSUID_NULL) &&
        m_pPropertyNode->getProperty("originDeviceDSUID") == NULL) {
      m_pPropertyNode->createProperty("originDeviceDSUID")
        ->linkToProxy(PropertyProxyMemberFunction<State, std::string, false>(*this, &State::getOriginDeviceDSUIDString));
    }
  }

  std::ostream& operator<<(std::ostream& ostream, const State& x) {
    const auto& service = x.getProviderService();
    return ostream << service << (service.empty() ? "" : "/") << x.getName();
  }

  void StateSensor::parseCondition(const std::string& _input, eValueComparator& _comp, double& _threshold) {
    _comp = eComp_undef;
    _threshold = 0;
    if (_input.empty()) {
      return;
    }
    std::vector<std::string> tokens = splitString(_input, ';', true);
    if (tokens[0].length() > 0) {
      if (tokens[0][0] == '>') {
        _comp = eComp_higher;
      } else if (tokens[0][0] == '<') {
        _comp = eComp_lower;
      }
    }
    try {
      _threshold = strToDouble(tokens[1]);
    } catch(std::invalid_argument& ex) {
    }
  }

  std::string StateSensor::buildName(const std::shared_ptr<Device>& _device, SensorType _sensorType,
                                     const std::string& _identifier) {
    return "dev." + dsuid2str(_device->getDSID()) + ".type" + intToString(static_cast<int>(_sensorType)) + "."
           + _identifier;
  }

  std::string StateSensor::buildName(const std::shared_ptr<Group>& _group, SensorType _sensorType,
                                     const std::string& _identifier) {
    return "zone.zone" + intToString(_group->getZoneID()) + ".group" + intToString(_group->getID()) + ".type"
           + intToString(static_cast<int>(_sensorType)) + "." + _identifier;
  }

  StateSensor::StateSensor(const std::string& _identifier, const std::string& _scriptId,
      std::shared_ptr<Device> _device, SensorType _sensorType,
      const std::string& _activateCondition, const std::string& _deactivateCondition)
  : State(StateType_SensorDevice,
      buildName(_device, _sensorType, _identifier),
      _scriptId),
    m_activateCondition(_activateCondition),
    m_deactivateCondition(_deactivateCondition)
  {
    parseCondition(m_activateCondition, m_activateComparator, m_activateValue);
    parseCondition(m_deactivateCondition, m_deactivateComparator, m_deactivateValue);
    setProviderDevice(_device);
  }

  StateSensor::StateSensor(const std::string& _identifier, const std::string& _scriptId,
      std::shared_ptr<Group> _group, SensorType _sensorType,
      const std::string& _activateCondition, const std::string& _deactivateCondition)
  : State(StateType_SensorZone,
      buildName(_group, _sensorType, _identifier),
      _scriptId),
    m_activateCondition(_activateCondition),
    m_deactivateCondition(_deactivateCondition)
  {
    parseCondition(m_activateCondition, m_activateComparator, m_activateValue);
    parseCondition(m_deactivateCondition, m_deactivateComparator, m_deactivateValue);
    setProviderGroup(_group);
  }

  void StateSensor::newValue(const CallOrigin _origin, double _value) {
    bool cond0, cond1;

    if (m_activateComparator == eComp_lower) {
      cond1 = _value < m_activateValue;
    } else if (m_activateComparator == eComp_higher) {
      cond1 = _value > m_activateValue;
    } else {
      cond1 = false;
    }
    if (m_deactivateComparator == eComp_lower) {
      cond0 = _value < m_deactivateValue;
    } else if (m_deactivateComparator == eComp_higher) {
      cond0 = _value > m_deactivateValue;
    } else {
      cond0 = false;
    }

    int oldState = getState();
    if (oldState == State_Active) {
      if (cond1 == false && cond0 == true) {
        setState(_origin, State_Inactive);
      } else if (cond1 == true && cond0 == true) {
        setState(_origin, State_Unknown);
      }
    } else {
      if (cond1 == true && cond0 == false) {
        setState(_origin, State_Active);
      } else if (cond1 == true && cond0 == true) {
        setState(_origin, State_Unknown);
      }
    }
  }

  StateSensor::~StateSensor() {
  }

} // namespace dss
