/*
    Copyright (c) 2010,2011,2022 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#include "binaryinputscanner.h"

#include <ds/log.h>

#include "ds485/dsbusinterface.h"
#include "dss.h"
#include "model/apartment.h"
#include "model/device.h"
#include "model/dsmeter.h"
#include "model/set.h"
#include "security/security.h"

static ds::LogChannel dsLogChannel("scannerBinaryInputScanner");

namespace dss {

namespace {
class BinaryInputOrStateDeviceFilter : public IDeviceAction {
private:
  std::vector<std::shared_ptr<Device> > m_devs;

public:
  virtual bool perform(std::shared_ptr<Device> _device) {
    if (_device->isPresent() && (_device->getBinaryInputCount() > 0 || !_device->getStates().empty())) {
      m_devs.push_back(_device);
    }
    return true;
  }
  std::vector<std::shared_ptr<Device> > getDeviceList() { return m_devs; }
};
} // namespace

BinaryInputScanner::BinaryInputScanner(Apartment* _papartment, const std::string& _busConnection)
    : m_pApartment(_papartment), m_busConnection(_busConnection), m_dsmApiHandle(NULL) {
  m_device.reset();
  m_dsm.reset();
}

BinaryInputScanner::~BinaryInputScanner() {
  if (m_dsmApiHandle != NULL) {
    DsmApiClose(m_dsmApiHandle);
    DsmApiCleanup(m_dsmApiHandle);
    m_dsmApiHandle = NULL;
  }
}

uint16_t BinaryInputScanner::getDeviceSensorValue(const dsuid_t& _dsm, dev_t _device, uint8_t _sensorIndex) const {
  if (m_dsmApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint16_t retVal;
  int ret = DeviceSensor_get_value_sync(m_dsmApiHandle, _dsm, _device, _sensorIndex, 30, &retVal);
  DSBusInterface::checkResultCode(ret);
  return retVal;
}

uint8_t BinaryInputScanner::getDeviceConfig(const dsuid_t& _dsm, dev_t _device, uint8_t _configClass,
                                            uint8_t _configIndex) const {
  if (m_dsmApiHandle == NULL) {
    throw std::runtime_error("Invalid libdsm api handle");
  }

  uint8_t retVal;
  int ret = DeviceConfig_get_sync_8(m_dsmApiHandle, _dsm, _device, _configClass, _configIndex, 30, &retVal);
  DSBusInterface::checkResultCode(ret);
  return retVal;
}

void BinaryInputScanner::run() {
  DSS* dss = nullptr;
  if (DSS::hasInstance()) {
    dss = DSS::getInstance();
    dss->getSecurity().loginAsSystemUser("BinaryInputScanner needs system-rights");
  }

  std::vector<std::shared_ptr<Device> > devices;
  std::shared_ptr<Device> device = m_device.lock();
  std::shared_ptr<DSMeter> dsm = m_dsm.lock();

  if (device) {
    // synchronize a single device
    if (device->isPresent() && device->getBinaryInputCount() > 0) {
      devices.push_back(device);
    }
  } else if (dsm) {
    Set D = dsm->getDevices();
    BinaryInputOrStateDeviceFilter filter;
    D.perform(filter);
    devices = filter.getDeviceList();
  } else {
    Set D = m_pApartment->getDevices();
    BinaryInputOrStateDeviceFilter filter;
    D.perform(filter);
    devices = filter.getDeviceList();
  }

  DS_DEBUG("BinaryInputScanner run: dsm=", ds::hideArg(dsm ? ds::str(dsm->getDSID()) : "NULL"),
           "dev=", ds::hideArg(device ? ds::str(device->getDSID()) : "NULL"));

  m_dsmApiHandle = DsmApiInitialize();
  if (!m_dsmApiHandle) {
    throw std::runtime_error("BinaryInputScanner: Unable to init dsmapi handle");
  }

  int result = DsmApiOpen(m_dsmApiHandle, m_busConnection.c_str(), 0);
  if (result < 0) {
    throw std::runtime_error(ds::str("BinaryInputScanner: Unable to open connection to:", m_busConnection));
  }
  if (dss) {
    dss->setDsmApiBlockingCallback(m_dsmApiHandle);
  }

  // send a request to each binary input device to resend its current status
  for (std::shared_ptr<Device> dev : devices) {
    uint16_t value = 0;
    std::shared_ptr<DSMeter> dsm;
    try {
      dsm = m_pApartment->getDSMeterByDSID(dev->getDSMeterDSID());
    } catch (ItemNotFoundException& ex) {
      DS_WARNING("BinaryInputScanner: dsm:", ds::hideArg(dev->getDSMeterDSID()), "is unknown");
      continue;
    }

    while (dsm->isPresent() && (dsm->getState() == DSM_STATE_REGISTRATION)) {
      sleep(5);
    }
    if (!dsm->isPresent()) {
      // ignore non-present dsm's, they will be discovered later with a new event
      DS_WARNING("BinaryInputScanner: dsm:", ds::hideArg(dev->getDSMeterDSID()), "is inactive");
      continue;
    }

    if (!dev->isVdcDevice()) {
      // Powerline Device
      try {
        value = getDeviceSensorValue(dev->getDSMeterDSID(), dev->getShortAddress(), 0x20);
        DS_DEBUG("BinaryInputScanner:", dev, "state = ", value);
        for (int index = 0; index < dev->getBinaryInputCount(); index++) {
          dev->handleBinaryInputEvent(
              index, ((value >> index) & 1) ? BinaryInputStateValue::Active : BinaryInputStateValue::Inactive);
        }
        // avoid powerline bus monopolization
        sleep(5);
      } catch (BusApiError& ex) {
        DS_WARNING("BinaryInputScanner:", dev, "did not respond to input state query");
        for (int index = 0; index < dev->getBinaryInputCount(); index++) {
          dev->handleBinaryInputEvent(index, BinaryInputStateValue::Unknown);
        }
      }

    } else {
      // IP Device
      try {
        VdcHelper::State state = VdcHelper::getState(dsm->getDSID(), dev->getDSID());
        const auto& sInput = state.binaryInputStates;
        DS_DEBUG("BinaryInputScanner:", dev, "state response fields = ", sInput.size());
        for (auto it = sInput.begin(); it != sInput.end(); it++) {
          dev->handleBinaryInputEvent(it->first, static_cast<BinaryInputStateValue>(it->second));
        }
        dev->setStateValues(state.deviceStates);
      } catch (std::runtime_error& e) {
        DS_WARNING("BinaryInputScanner: VdcDevice ", dev, " failure = ", e.what());
        for (int index = 0; index < dev->getBinaryInputCount(); index++) {
          dev->handleBinaryInputEvent(index, BinaryInputStateValue::Unknown);
        }
      }
    }
  }
}

void BinaryInputScanner::setup(std::shared_ptr<DSMeter> _dsMeter, std::shared_ptr<Device> _device) {
  m_dsm = _dsMeter;
  m_device = _device;
}

} // namespace dss
