/*
    Copyright (c) 2010, 2011 digitalSTROM.org, Zurich, Switzerland

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
*/

#include "oemwebquery.h"

#include <ds/log.h>

#include <boost/filesystem.hpp>

#include "dss.h"
#include "model/modelmaintenance.h"
#include "security/security.h"

static ds::LogChannel dsLogChannel("scannerOemWebQuery");

namespace dss {
OEMWebQuery::OEMWebQuery(WebserviceConnection* webserviceConnection, std::shared_ptr<Device> _device,
                         DeviceOEMState_t _oldState, PropertySystem& propertySystem)
    : Task(), m_propertySystem(propertySystem), m_oldOEMState(_oldState), m_connection(webserviceConnection) {
  m_deviceDSUID = _device->getDSID();
  m_EAN = _device->getOemEanAsString();
  m_partNumber = _device->getOemPartNumber();
  m_serialNumber = _device->getOemSerialNumber();
  m_queryConfigLink = _device->getHasActions();
}

OEMWebQuery::OEMWebQueryCallback::OEMWebQueryCallback(dsuid_t _deviceDSUID, DeviceOEMState_t _oldState,
                                                      PropertySystem& propertySystem)
    : m_propertySystem(propertySystem), m_deviceDSUID(_deviceDSUID), m_oldOEMState(_oldState) {}

void OEMWebQuery::OEMWebQueryCallback::result(long code, const std::string& result) {
  std::string productURL;
  std::string productName;
  boost::filesystem::path iconFile;
  std::string defaultName;
  std::string serviceLink;
  std::string infoLink;
  std::string configLink = "/device-configuration-pages";
  DeviceOEMState_t state = m_oldOEMState;
  bool isGetBasicArticleDataResponse = false;

  if (code == 200) {
    DS_DEBUG("OEMWebQueryCallback::result: DSUID:", ds::hideArg(m_deviceDSUID), result);
    struct json_tokener* tok;

    tok = json_tokener_new();
    json_object* json_request = json_tokener_parse_ex(tok, result.c_str(), -1);

    boost::filesystem::path remoteIconPath;
    if (tok->err == json_tokener_success) {
      int resultCode = 99;
      json_object* obj;
      if (json_object_object_get_ex(json_request, "ReturnCode", &obj)) {
        resultCode = json_object_get_int(obj);
      }
      if (resultCode != 0) {
        std::string errorMessage;
        if (json_object_object_get_ex(json_request, "ReturnMessage", &obj)) {
          errorMessage = json_object_get_string(obj);
        }
        DS_ERROR("OEMWebQueryCallback::result: JSON-ERROR: ", errorMessage);
        auto pEvent =
            std::make_shared<ModelEventWithDSID>(ModelEvent::etDeviceOEMDataUpdateProductInfoState, m_deviceDSUID);
        pEvent->addParameter(m_oldOEMState); // switch back from LOADING state
        if (DSS::hasInstance()) {
          DSS::getInstance()->getModelMaintenance().addModelEvent(pEvent);
        }
        json_object_put(json_request);
        json_tokener_free(tok);
        return;
      } else {
        json_object* response;
        if (!json_object_object_get_ex(json_request, "Response", &response)) {
          DS_ERROR("OEMWebQueryCallback::result: no 'Response' object in response");
        } else {
          json_object* article;
          if (json_object_object_get_ex(response, "Article", &article)) {
            isGetBasicArticleDataResponse = true;
            if (json_object_object_get_ex(article, "ServiceLink", &obj)) {
              if (json_object_is_type(obj, json_type_string)) {
                serviceLink = json_object_get_string(obj);
              }
            }
            if (json_object_object_get_ex(article, "Infolink", &obj)) { // mind the "l"
              if (json_object_is_type(obj, json_type_string)) {
                infoLink = json_object_get_string(obj);
              }
            }
            /*if (json_object_object_get_ex(article, "ConfigLink", &obj)) {
              if (json_object_is_type(obj, json_type_string)) {
                configLink = json_object_get_string(obj);
              }
            }*/
          }
          if (json_object_object_get_ex(response, "ArticleName", &obj)) {
            if (json_object_is_type(obj, json_type_string)) {
              productName = json_object_get_string(obj);
            }
          }
          if (json_object_object_get_ex(response, "ArticleIcon", &obj)) {
            if (json_object_is_type(obj, json_type_string)) {
              remoteIconPath = json_object_get_string(obj);
            }
          }
          if (json_object_object_get_ex(response, "ArticleDescriptionForCustomer", &obj)) {
            if (json_object_is_type(obj, json_type_string)) {
              productURL = json_object_get_string(obj);
            }
          }
          if (json_object_object_get_ex(response, "DefaultName", &obj)) {
            if (json_object_is_type(obj, json_type_string)) {
              defaultName = json_object_get_string(obj);
            }
          }
        }
      }
    }
    json_object_put(json_request);
    json_tokener_free(tok);

    if (isGetBasicArticleDataResponse) {
      // done with the first response of GetBasicArticleData
      m_configLink = configLink;
      return;
    }

    state = DEVICE_OEM_VALID;
    iconFile = remoteIconPath.filename();
    if (!remoteIconPath.empty()) {
      HttpClient url;
      long res;
      std::string iconURL = remoteIconPath.string();

      boost::filesystem::path iconBasePath;
      if (DSS::hasInstance()) {
        DSS::getInstance()->getSecurity().loginAsSystemUser("OEMWebQuery needs system-rights");
        iconBasePath = m_propertySystem.getStringValue("/config/subsystems/Apartment/iconBasePath");
      } else {
        return;
      }

      boost::filesystem::path iconPathTemp =
          boost::filesystem::temp_directory_path() / remoteIconPath.filename().replace_extension(".tmp");
      boost::filesystem::path iconPath = iconBasePath / iconFile;
      res = url.downloadFile(iconURL, iconPathTemp.string());
      if (res != 200) {
        DS_WARNING("OEMWebQueryCallback::result: could not download OEM icon from:", iconURL, "; error:", res);
        iconFile.clear();
        try {
          if (boost::filesystem::exists(iconPathTemp)) {
            boost::filesystem::remove(iconPathTemp);
          }
        } catch (boost::filesystem::filesystem_error& e) {
          DS_WARNING("OEMWebQueryCallback::result: cannot delete (incomplete) icon:", e.what());
        }
      } else {
        try {
          if (boost::filesystem::exists(iconPathTemp)) {
            boost::filesystem::copy_file(iconPathTemp, iconPath, boost::filesystem::copy_option::overwrite_if_exists);
            boost::filesystem::remove(iconPathTemp);
          }
        } catch (boost::filesystem::filesystem_error& e) {
          DS_WARNING("OEMWebQueryCallback::result: cannot move icon to target location:", e.what());
        }
      }
    }

    auto pEvent = std::make_shared<ModelEventWithStrings>(ModelEvent::etDeviceOEMDataReady, m_deviceDSUID);
    pEvent->addParameter(state);
    pEvent->addStringParameter(productName);
    pEvent->addStringParameter(iconFile.string());
    pEvent->addStringParameter(productURL);
    pEvent->addStringParameter(defaultName);
    pEvent->addStringParameter(m_configLink);
    if (DSS::hasInstance()) {
      DSS::getInstance()->getModelMaintenance().addModelEvent(pEvent);
    }
  } else {
    DS_WARNING("OEMWebQueryCallback::result: could not download OEM data. Error:", code, "Message:", result);
    // nothing was read out, just keep the old state; if the device
    // has already been read out sometime earlier we do not want to
    // invalidate the data
    auto pEvent =
        std::make_shared<ModelEventWithStrings>(ModelEvent::etDeviceOEMDataUpdateProductInfoState, m_deviceDSUID);
    pEvent->addParameter(m_oldOEMState); // switch back from LOADING state
    if (DSS::hasInstance()) {
      DSS::getInstance()->getModelMaintenance().addModelEvent(pEvent);
    }
  }
}

void OEMWebQuery::run() {
  std::string mac = m_propertySystem.getStringValue("/system/host/interfaces/eth0/mac");
  if (mac.empty()) {
    mac = "00:00:00:00:00:00"; // OEM web service is picky and refuses empty mac address
  }
  std::string country = m_propertySystem.getStringValue("/config/geodata/country");
  std::string locale = m_propertySystem.getStringValue("/system/language/locale");
  std::string language;
  std::string parameters;
  std::shared_ptr<OEMWebQuery::OEMWebQueryCallback> cb =
      std::make_shared<OEMWebQuery::OEMWebQueryCallback>(m_deviceDSUID, m_oldOEMState, m_propertySystem);

  // #15054 - MSHub expects language code without country, and not 'locale' xx_XX code
  if ((locale.length() >= 5) && (locale[2] == '_')) {
    language = locale.substr(0, 2);
    // Fall back to locale region if no geo-location is set
    if (country.empty()) {
      country = locale.substr(3, 2);
    }
  }

  // DSD-807: nummber must be 13 characters
  while (m_EAN.size() < 13) {
    m_EAN = "0" + m_EAN;
  }

  if (m_queryConfigLink) {
    parameters = "ean=" + m_EAN + "&countryCode=" + country + "&languageCode=" + language;
    m_connection->request("public/MasterDataManagement/Article/v1_0/ArticleData/GetBasicArticleData", parameters,
                          HttpMethod::GET, cb, false);
  }
  parameters = "ean=" + m_EAN + "&partNr=" + intToString(m_partNumber) +
               "&oemSerialNumber=" + intToString(m_serialNumber) + "&macAddress=" + mac + "&countryCode=" + country +
               "&languageCode=" + language;
  m_connection->request("public/MasterDataManagement/Article/v1_0/ArticleData/GetArticleData", parameters,
                        HttpMethod::GET, cb, false);
}

} // namespace dss
