/*
    Copyright (c) 2013 digitalSTROM.org, Zurich, Switzerland

    Author: Christian Hitz, aizo AG <christian.hitz@aizo.com>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "sceneaccess.h"

#include <memory>

#include "base.h"
#include "dss.h"
#include "logger.h"
#include "model/addressablemodelitem.h"
#include "model/apartment.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/set.h"
#include "model/state.h"
#include "propertysystem.h"
#include "stdio.h"
#include "string.h"

using namespace std::literals;
static ds::LogChannel dsLogChannel("dssSceneAccess");

namespace dss {

namespace {

bool activeNonScriptState(const Apartment& apartment, const std::string& stateName) {
  if (auto state = apartment.getNonScriptState(stateName); state) { // throws ItemNotFoundException
    return state->getState() == State_Active;
  }
  // NOTE: getNonScriptState throws exception if state does not exist, hence not reached
  return false;
}
} // namespace

bool SceneAccess::checkAccess(const Apartment& apartment, const AddressableModelItem* _pTarget,
                              const SceneAccessCategory _category) {
  // "manual" scene calls are never blocked
  if (_category == SceneAccessCategory::manual) {
    return true;
  }

  {
    /*
     * Fire & Wind: Prohibit automatic actions in group shade
     *
     * Behavior changed 2015/12/10:
     * Fire should not block all groups, but only shades, see #11755
     */
    bool windActive = activeNonScriptState(apartment, "wind");
    bool fireActive = activeNonScriptState(apartment, "fire");

    if (windActive || fireActive) {
      auto protectionMessage = [windActive, fireActive]() -> std::string {
        std::string protectionMessage("Execution blocked: ");
        if (fireActive && windActive) {
          protectionMessage += "fire and wind alarms are active";
        } else if (fireActive) {
          protectionMessage += "fire is active";
        } else if (windActive) {
          protectionMessage += "wind is active";
        }
        return protectionMessage;
      };

      if (const Group* pGroup = dynamic_cast<const Group*>(_pTarget); pGroup != NULL) {
        if (ApplicationType::blinds == pGroup->getApplicationType()) {
          throw SceneAccessException(protectionMessage());
        }
      }

      if (const Device* pDevice = dynamic_cast<const Device*>(_pTarget); pDevice != NULL) {
        if (pDevice->isInGroup(static_cast<int>(DeviceClasses_t::DEVICE_CLASS_GR))) {
          throw SceneAccessException(protectionMessage());
        }
      }
    }
  }

  {
    /*
     * Partial Wind (in user groups): Prevent automatic actions in relevant user group.
     */

    for (const auto& cluster : apartment.getClusters()) {
      if (!cluster->isValid()) {
        continue;
      }
      if (cluster->getApplicationType() != ApplicationType::blinds) {
        continue;
      }

      try {
        // TODO add query function, inverse of Cluster::checkWindProtection
        bool windActive = activeNonScriptState(apartment, ds::str("wind.group", cluster->getID()));
        if (!windActive) {
          continue;
        }
      } catch (ItemNotFoundException& e) {
        continue;
      }

      if (const Group* pGroup = dynamic_cast<const Group*>(_pTarget); pGroup != NULL) {
        // addressing the same cluster with an active wind alarm
        if (pGroup->getID() == cluster->getID()) {
          throw SceneAccessException("Execution blocked: wind is active");
        }
        // addressing group 2 in a single zone (or all zones) with devices
        // that are member of a cluster with an active wind alarm
        if (pGroup->getID() == static_cast<int>(DeviceClasses_t::DEVICE_CLASS_GR)) {
          Set devsInZones = pGroup->getDevices().getByGroup(cluster->getID());
          if (!devsInZones.isEmpty()) {
            throw SceneAccessException("Execution blocked: wind is active");
          }
        }
      }
      if (const Device* pDevice = dynamic_cast<const Device*>(_pTarget); pDevice != NULL) {
        if (pDevice->isInGroup(cluster->getID())) {
          throw SceneAccessException("Execution blocked: wind is active");
        }
      }
    }
  }

  return true;
}

bool SceneAccess::checkStates(const Apartment& apartment, const AddressableModelItem* _pTarget, int _sceneNr) {
  std::string stateName;
  switch (_sceneNr) {
    case SceneWindActive: stateName = "wind"; break;
    case SceneRainActive: stateName = "rain"; break;
    case SceneHailActive: stateName = "hail"; break;
    case SceneFire: stateName = "fire"; break;
    case SceneAlarm: stateName = "alarm"; break;
    case SceneAlarm2: stateName = "alarm2"; break;
    case SceneAlarm3: stateName = "alarm3"; break;
    case SceneAlarm4: stateName = "alarm4"; break;
    default: return true;
  }

  std::shared_ptr<State> s = apartment.getNonScriptState(stateName);
  if (s && (s->getState() == State_Active)) {
    std::string errorMessage("Execution blocked: state " + stateName + " is already active");
    throw SceneAccessException(errorMessage);
  }

  return true;
}

SceneAccessCategory SceneAccess::stringToCategory(const std::string& _categoryString) {
  return ds::toEnum<SceneAccessCategory>(_categoryString);
}

} // namespace dss
