/*
    Copyright (c) 2009 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "structuremanipulator.h"

#include <ds/str.h>

#include <boost/range/algorithm.hpp>
#include <boost/range/algorithm_ext.hpp>
#include <boost/ref.hpp>
#include <regex>
#include <stdexcept>

#include "base.h"
#include "businterface.h"
#include "comm-channel.h"
#include "dss.h"
#include "event.h"
#include "model/apartment.h"
#include "model/buttonmanipulator.h"
#include "model/cluster.h"
#include "model/device.h"
#include "model/devicereference.h"
#include "model/dsmeter.h"
#include "model/group.h"
#include "model/modelconst.h"
#include "model/modelmaintenance.h"
#include "model/set.h"
#include "model/state.h"
#include "model/zone.h"
#include "propertysystem.h"
#include "util.h"
#include "vdc-messages.pb.h"

static ds::LogChannel dsLogChannel("structureManipulator");

using namespace dss::model;

namespace dss {

StructureManipulator::StructureManipulator(Apartment& apartment)
    : m_Apartment(apartment),
      m_Interface(*apartment.getBusInterface()->getStructureModifyingBusInterface()),
      m_QueryInterface(*apartment.getBusInterface()->getStructureQueryBusInterface()) {}

void StructureManipulator::createZone(std::shared_ptr<DSMeter> _dsMeter, std::shared_ptr<Zone> _zone) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_Apartment.getMutex());
  if (!_dsMeter->isPresent()) {
    throw std::runtime_error("Need dsMeter to be present");
  }

  try {
    m_Interface.createZone(_dsMeter->getDSID(), _zone->getID());
  } catch (BusApiError& e) {
    if (e.error == ERROR_ZONE_ALREADY_EXISTS) {
      // this is not error at all

    } else if (e.error == ERROR_NO_FURTHER_ZONES) {
      // WARNING: do not change this string, it is used and translated in
      // the UI
      throw std::runtime_error("The dSM of this device can not handle more rooms");
    } else {
      // do not ignore other errors that might happen
      DS_FAIL_REQUIRE(e.what());
    }
  }

  _zone->addToDSMeter(_dsMeter);
  _zone->setIsPresent(true);
  _zone->setIsConnected(true);
  std::vector<std::shared_ptr<Group> > groupList = _zone->getGroups();
  for (std::shared_ptr<Group> group : groupList) {
    group->setIsConnected(true);
  }
  synchronizeGroups(&m_Apartment, &m_Interface);
} // createZone

void StructureManipulator::checkSensorsOnDeviceRemoval(Zone& _zone, Device& _device) {
  // if the device that is being moved out of the zone was a zone sensor:
  // clear the previous sensor assignment and also check if we can reassign
  // another sensor to the zone
  for (auto&& sensorType : _zone.getAssignedSensorTypes(_device)) {
    resetZoneSensor(_zone, sensorType);
  }
  autoAssignZoneSensors(_zone);
}

void StructureManipulator::addDeviceToZone(std::shared_ptr<Device> _device, std::shared_ptr<Zone> _zone) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_Apartment.getMutex());
  DS_REQUIRE(_device->isPresent(), "Need device to be present.");

  int oldZoneID = _device->getZoneID();
  std::shared_ptr<DSMeter> targetDSMeter = m_Apartment.getDSMeterByDSID(_device->getDSMeterDSID());
  if (!_zone->registeredOnDSMeter(targetDSMeter)) {
    createZone(targetDSMeter, _zone);
  }

  m_Interface.setZoneID(targetDSMeter->getDSID(), _device->getShortAddress(), _zone->getID());
  DeviceReference ref(_device, &m_Apartment);
  _zone->addDevice(ref); // calls device->setZoneID internally

  {
    std::shared_ptr<DeviceReference> pDevRef = std::make_shared<DeviceReference>(ref);
    std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::DeviceEvent, pDevRef);
    mEvent->setProperty("action", "moved");
    mEvent->setProperty("id", intToString(_zone->getID())); // TODO: remove property in next release
    mEvent->setProperty("oldZoneID", intToString(oldZoneID));
    mEvent->setProperty("newZoneID", intToString(_zone->getID()));
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(mEvent);
    }
  }

  // update group presence
  for (int g = 0; g < GroupIDStandardMax; g++) {
    if (_device->isInGroup(g)) {
      std::shared_ptr<Group> group = _zone->getGroup(g);
      if (group) {
        group->setIsPresent(true);
      }
    }
  }

  // check if we can remove the zone from the dsMeter
  if (oldZoneID != 0) {
    DS_INFO("addDeviceToZone: Removing device from old zone", oldZoneID)
    std::shared_ptr<Zone> oldZone = m_Apartment.getZone(oldZoneID);
    oldZone->removeDevice(ref);
    checkSensorsOnDeviceRemoval(*oldZone, *_device);

    Set presentDevicesInZoneOfDSMeter = oldZone->getDevices().getByDSMeter(targetDSMeter).getByPresence(true);
    if (presentDevicesInZoneOfDSMeter.empty()) {
      DS_INFO("addDeviceToZone: Removing zone from meter", targetDSMeter->getDSID());
      try {
        removeZoneOnDSMeter(oldZone, targetDSMeter);
      } catch (std::runtime_error& err) {
        DS_INFO("addDeviceToZone: can't remove zone:", err.what());
      }
    }

    // cleanup group presence
    for (int g = 0; g < GroupIDStandardMax; g++) {
      if (_device->isInGroup(g)) {
        std::shared_ptr<Group> group = oldZone->getGroup(g);
        Set presentDevicesInGroup = oldZone->getDevices().getByGroup(group);
        if (presentDevicesInGroup.empty()) {
          group->setIsPresent(false);
        }
      }
    }
  } else {
    DS_WARNING("addDeviceToZone: No previous zone...");
  }
  // check if newly added device might need to be assigned as sensor
  autoAssignZoneSensors(*_zone);
} // addDeviceToZone

void StructureManipulator::removeZoneOnDSMeter(std::shared_ptr<Zone> _zone, std::shared_ptr<DSMeter> _dsMeter) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_Apartment.getMutex());
  Set presentDevicesInZoneOfDSMeter = _zone->getDevices().getByDSMeter(_dsMeter).getByPresence(true);
  if (!presentDevicesInZoneOfDSMeter.empty()) {
    throw std::runtime_error("cannot delete zone if there are still devices present");
  }

  std::vector<DeviceSpec> inactiveDevices =
      m_QueryInterface.getInactiveDevicesInZone(_dsMeter->getDSID(), _zone->getID());

  for (const DeviceSpec& inactiveDevice : inactiveDevices) {
    std::shared_ptr<Device> dev;
    try {
      dev = m_Apartment.getDeviceByDSID(inactiveDevice.DSID);
    } catch (ItemNotFoundException&) {
      throw std::runtime_error("Inactive device '" + dsuid2str(inactiveDevice.DSID) + "' not known here.");
    }
    if (dev->getLastKnownDSMeterDSID() != _dsMeter->getDSID()) {
      m_Interface.removeDeviceFromDSMeter(_dsMeter->getDSID(), inactiveDevice.ShortAddress);
    } else {
      throw std::runtime_error("Inactive device '" + dsuid2str(inactiveDevice.DSID) +
                               "' only known on this meter, can't remove it.");
    }
  }

  try {
    m_Interface.removeZone(_dsMeter->getDSID(), _zone->getID());
    _zone->removeFromDSMeter(_dsMeter);
    if (!_zone->isRegisteredOnAnyMeter()) {
      _zone->setIsConnected(false);
    }
  } catch (BusApiError& e) {
    DS_WARNING("Can't remove zone", _zone->getID(), "from meter:", _dsMeter->getDSID(), e.what());
  }
} // removeZoneOnDSMeter

void StructureManipulator::wipeZoneOnMeter(dsuid_t _meterdSUID, int _zoneID) {
  // check if the zone consists only of inactive devices and remove them
  // if this is indeed the case, then retry to remove the zone
  if (m_QueryInterface.getDevicesCountInZone(_meterdSUID, _zoneID, true) > 0) {
    throw std::runtime_error("Can not remove zone " + intToString(_zoneID) + ": zone contains active devices");
  }

  std::vector<DeviceSpec> devices = m_QueryInterface.getDevicesInZone(_meterdSUID, _zoneID, false);
  for (size_t i = 0; i < devices.size(); i++) {
    m_Interface.removeDeviceFromDSMeter(_meterdSUID, devices.at(i).ShortAddress);
  }
  m_Interface.removeZone(_meterdSUID, _zoneID);
} // wipeZoneOnMeter

void StructureManipulator::removeZoneOnDSMeters(std::shared_ptr<Zone> _zone) {
  boost::recursive_mutex::scoped_lock scoped_lock(m_Apartment.getMutex());
  try {
    std::vector<std::weak_ptr<const DSMeter> > meters = _zone->getDSMeters();
    for (size_t s = 0; s < meters.size(); s++) {
      if (std::shared_ptr<const DSMeter> meterFromZone = meters.at(s).lock()) {
        std::shared_ptr<dss::DSMeter> meter = m_Apartment.getDSMeterByDSID(meterFromZone->getDSID());
        if (meter != NULL) {
          _zone->removeFromDSMeter(meter);
        }
      }
    }

    // the dSM removes the zone only, if the controller is disabled.
    DSS::getInstance()->assertBlockingAllowed();
    m_Interface.disableZoneHeatingConfig(DSUID_BROADCAST, _zone->getID());

    for (size_t s = 0; s < meters.size(); s++) {
      if (std::shared_ptr<const DSMeter> meterFromZone = meters.at(s).lock()) {
        try {
          m_Interface.removeZone(meterFromZone->getDSID(), _zone->getID());
        } catch (BusApiError& be) {
          if (be.error == ERROR_ZONE_NOT_EMPTY) {
            wipeZoneOnMeter(meterFromZone->getDSID(), _zone->getID());
          }
        }
      }
    }

    if (!_zone->isRegisteredOnAnyMeter()) {
      _zone->setIsConnected(false);
    }
  } catch (BusApiError& e) {
    DS_WARNING("Can't broadcast-remove zone", _zone->getID(), e.what());
  }
} // removeZoneOnDSMeters

void StructureManipulator::removeDeviceFromDSMeter(std::shared_ptr<Device> _device) {
  dsuid_t dsmDsid = _device->getDSMeterDSID();
  dsuid_t devDsid = _device->getDSID();

  if (dsmDsid == DSUID_NULL) {
    dsmDsid = _device->getLastKnownDSMeterDSID();
    // #10453 - unknown last dsmid
    if (dsmDsid == DSUID_NULL) {
      DS_WARNING("removeDevicefromDSMeter: device ", _device->getDSID(), "has invalid dSM ID", dsmDsid);
      return;
    }
  }

  // if the dSM is not connected, then there is no point in trying to remove
  // the device from it, this would only cause a timeout and thus introduce
  // an unnecessary delay in the UI/for the JSON API users
  std::shared_ptr<DSMeter> dsm = m_Apartment.getDSMeterByDSID(_device->getDSMeterDSID());
  if (dsm->isConnected() == false) {
    DS_WARNING("removeDevicefromDSMeter: not removing device ", _device->getDSID(), "from dSM", dsmDsid,
               "this dSM is not connected");
    return;
  }

  m_Interface.removeDeviceFromDSMeters(devDsid);
} // removeDevice

std::vector<std::shared_ptr<Device> > StructureManipulator::removeDevice(std::shared_ptr<Device> _pDevice) {
  std::vector<std::shared_ptr<Device> > result;
  std::shared_ptr<Device> pPartnerDevice;

  if (_pDevice->is2WayMaster()) {
    dsuid_t next;
    dsuid_get_next_dsuid(_pDevice->getDSID(), &next);
    try {
      pPartnerDevice = m_Apartment.getDeviceByDSID(next);
    } catch (ItemNotFoundException& e) {
      DS_INFO("Could not find partner device with dsuid", next);
    }
  }

  if (_pDevice->isMainDevice() && (_pDevice->getPairedDevices() > 0)) {
    bool doSleep = false;
    dsuid_t next;
    dsuid_t current = _pDevice->getDSID();
    for (int p = 0; p < _pDevice->getPairedDevices(); p++) {
      dsuid_get_next_dsuid(current, &next);
      current = next;
      try {
        std::shared_ptr<Device> pPartnerDev;
        pPartnerDev = m_Apartment.getDeviceByDSID(next);
        if (doSleep) {
          usleep(500 * 1000); // 500ms
        }
        removeDeviceFromDSMeter(pPartnerDev);
        checkSensorsOnDeviceRemoval(*m_Apartment.getZone(pPartnerDev->getZoneID()), *pPartnerDev);
        m_Apartment.removeDevice(pPartnerDev->getDSID());
        doSleep = true;

        if (pPartnerDev->isVisible()) {
          result.push_back(pPartnerDev);
        }
      } catch (std::runtime_error& e) {
        DS_INFO("Could not find partner device with dsuid", next);
      }
    }
  }

  try {
    removeDeviceFromDSMeter(_pDevice);
    if (pPartnerDevice != NULL) {
      removeDeviceFromDSMeter(pPartnerDevice);
    }
  } catch (std::runtime_error& e) {
    DS_WARNING("Could not remove device from dSM:", e.what());
  }

  checkSensorsOnDeviceRemoval(*m_Apartment.getZone(_pDevice->getZoneID()), *_pDevice);
  m_Apartment.removeDevice(_pDevice->getDSID());
  result.push_back(_pDevice);

  if (pPartnerDevice != NULL) {
    DS_INFO("Also removing partner device", pPartnerDevice->getDSID());
    checkSensorsOnDeviceRemoval(*m_Apartment.getZone(pPartnerDevice->getZoneID()), *pPartnerDevice);
    m_Apartment.removeDevice(pPartnerDevice->getDSID());
    result.push_back(pPartnerDevice);
  }

  return result;
}

void StructureManipulator::deviceSetName(std::shared_ptr<Device> _pDevice, const std::string& _name) {
  m_Interface.deviceSetName(_pDevice->getDSMeterDSID(), _pDevice->getShortAddress(), _name);

  {
    DeviceReference ref(_pDevice, &m_Apartment);
    std::shared_ptr<DeviceReference> pDevRef = std::make_shared<DeviceReference>(ref);
    std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::DeviceEvent, pDevRef);
    mEvent->setProperty("action", "name");
    mEvent->setProperty("name", _name);
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(mEvent);
    }
  }
} // deviceSetName

void StructureManipulator::meterSetName(std::shared_ptr<DSMeter> _pMeter, const std::string& _name) {
  m_Interface.meterSetName(_pMeter->getDSID(), _name);
} // meterSetName

int StructureManipulator::persistSet(Set& _set, const std::string& _originalSet) {
  // find next empty user-group
  // TODO: As I understand this persist/create group with devices is only for UserGroups? or also ControlGroups?
  int idFound = -1;
  for (int groupID = GroupIDUserGroupStart; groupID <= GroupIDUserGroupEnd; groupID++) {
    try {
      m_Apartment.getGroup(groupID);
    } catch (ItemNotFoundException&) {
      idFound = groupID;
      break;
    }
  }
  if (idFound != -1) {
    return persistSet(_set, _originalSet, idFound);
  }
  return -1;
} // persistSet

int StructureManipulator::persistSet(Set& _set, const std::string& _originalSet, int _groupNumber) {
  std::shared_ptr<Group> pGroup;
  try {
    pGroup = m_Apartment.getGroup(_groupNumber);
  } catch (ItemNotFoundException&) {
    pGroup = std::make_shared<Group>(_groupNumber, m_Apartment.getZone(0));
    m_Apartment.getZone(0)->addGroup(pGroup);
    m_Interface.createGroup(0, _groupNumber, ApplicationType::none, 0, "");
    pGroup->setAssociatedSet(_originalSet);
  }
  DS_INFO("creating new group", _groupNumber);
  DS_ASSERT(pGroup != NULL, "");
  for (int iDevice = 0; iDevice < _set.length(); iDevice++) {
    std::shared_ptr<Device> pDevice = _set.get(iDevice).getDevice();
    pDevice->addToGroup(_groupNumber);
    m_Interface.addToGroup(pDevice->getDSMeterDSID(), _groupNumber, pDevice->getShortAddress());
  }
  return _groupNumber;
} // persistSet

void StructureManipulator::unpersistSet(std::string _setDescription) {
  std::vector<std::shared_ptr<Group> > groups = m_Apartment.getZone(0)->getGroups();
  std::shared_ptr<Group> pGroup;
  for (std::size_t iGroup = 0; iGroup < groups.size(); iGroup++) {
    if (groups[iGroup]->getAssociatedSet() == _setDescription) {
      pGroup = groups[iGroup];
      break;
    }
  }
  if (pGroup != NULL) {
    Set devs = pGroup->getDevices();
    int groupID = pGroup->getID();
    for (int iDevice = 0; iDevice < devs.length(); iDevice++) {
      std::shared_ptr<Device> pDevice = devs.get(iDevice).getDevice();
      pDevice->removeFromGroup(groupID);
      m_Interface.removeFromGroup(pDevice->getDSMeterDSID(), groupID, pDevice->getShortAddress());
    }
    m_Interface.removeGroup(pGroup->getZoneID(), groupID);
    m_Apartment.getZone(0)->removeGroup(pGroup);
  }
} // unpersistSet

void StructureManipulator::createGroup(std::shared_ptr<Zone> _zone, int _groupNumber, ApplicationType applicationType,
                                       const int applicationConfiguration, const std::string& _name) {
  if (isDefaultGroup(_groupNumber) || isGlobalAppDsGroup(_groupNumber)) {
    throw DSSException("Group with id " + intToString(_groupNumber) + " is reserved");
  }

  if (isCluster(_groupNumber)) {
    if (_zone->getID() != 0) {
      throw DSSException("Group with id " + intToString(_groupNumber) + " only allowed in Zone 0");
    }
    try {
      DS_INFO("Configure user group", _groupNumber, " with standard-id ", applicationType);
      std::shared_ptr<Cluster> pCluster = m_Apartment.getCluster(_groupNumber);
      pCluster->setName(_name);
      pCluster->setApplicationType(applicationType);
      m_Interface.createCluster(_groupNumber, applicationType, applicationConfiguration, _name);
    } catch (ItemNotFoundException& e) {
      DS_WARNING("Datamodel-Error creating user group", _groupNumber, e.what());
    } catch (BusApiError& e) {
      DS_WARNING("Bus-Error creating user group", _groupNumber, e.what());
    }
    return;
  }

  if (isZoneUserGroup(_groupNumber)) {
    std::shared_ptr<Group> pGroup;
    try {
      pGroup = _zone->getGroup(_groupNumber);
      throw DSSException("Group id " + intToString(_groupNumber) + " already exists");
    } catch (ItemNotFoundException& e) {
      DS_INFO("Creating user group", _groupNumber, "in zone", _zone->getID());
      pGroup = std::make_shared<Group>(_groupNumber, _zone);
      _zone->addGroup(pGroup);
      m_Interface.createGroup(_zone->getID(), _groupNumber, applicationType, applicationConfiguration, _name);
    }
    return;
  }

  // for groups in User Global Application range, we are locked to zone 0, but we create Groups not clusters
  if (isGlobalAppUserGroup(_groupNumber)) {
    std::shared_ptr<Group> pGroup;
    if (_zone->getID() != 0) {
      throw DSSException("Group with id " + intToString(_groupNumber) + " only allowed in Zone 0");
    }
    try {
      pGroup = _zone->getGroup(_groupNumber);
      throw DSSException("Group id " + intToString(_groupNumber) + " already exists");
    } catch (ItemNotFoundException& e) {
      DS_INFO("Creating user global application group ", _groupNumber);
      pGroup = std::make_shared<Group>(_groupNumber, _zone);
      pGroup->setName(_name);
      pGroup->setApplicationType(applicationType);
      _zone->addGroup(pGroup);
      m_Interface.createGroup(_zone->getID(), _groupNumber, applicationType, applicationConfiguration, _name);
    }
    return;
  }

  throw DSSException("Create group: id " + intToString(_groupNumber) + " too large");
} // createGroup

void StructureManipulator::removeGroup(std::shared_ptr<Zone> _zone, int _groupNumber) {
  if (isDefaultGroup(_groupNumber) || isGlobalAppDsGroup(_groupNumber)) {
    throw DSSException("Group with id " + intToString(_groupNumber) + " is reserved");
  }

  if (isCluster(_groupNumber)) {
    std::shared_ptr<Cluster> pCluster;
    if (_zone->getID() != 0) {
      throw DSSException("Group with id " + intToString(_groupNumber) + " only allowed in Zone 0");
    }
    try {
      DS_INFO("Remove user group ", _groupNumber);
      pCluster = m_Apartment.getCluster(_groupNumber);
    } catch (ItemNotFoundException& e) {
      throw DSSException("Remove group: id " + intToString(_groupNumber) + " does not exist");
    }
    if (pCluster->isConfigurationLocked()) {
      throw DSSException("The group is locked and cannot be modified");
    }
    try {
      removeClusterStates(pCluster);
      pCluster->reset();
      m_Interface.removeCluster(_groupNumber);
    } catch (ItemNotFoundException& e) {
      DS_WARNING("Datamodel-Error removing user group", _groupNumber, e.what());
    } catch (BusApiError& e) {
      DS_WARNING("Bus-Error removing user group", _groupNumber, e.what());
    }
    return;
  }

  if (isZoneUserGroup(_groupNumber)) {
    std::shared_ptr<Group> pGroup;
    try {
      pGroup = _zone->getGroup(_groupNumber);
    } catch (ItemNotFoundException& e) {
      throw DSSException("Group id " + intToString(_groupNumber) + " does not exist");
    }
    try {
      DS_INFO("Removing user group", _groupNumber, "in zone", _zone->getID());
      _zone->removeGroup(pGroup);
      m_Interface.removeGroup(_zone->getID(), _groupNumber);
    } catch (BusApiError& e) {
      DS_WARNING("Bus-Error removing user group", _groupNumber, e.what());
    }
    return;
  }

  // for groups in User Global Application range, we are locked to zone 0, but we remove Groups not clusters
  if (isGlobalAppUserGroup(_groupNumber)) {
    std::shared_ptr<Group> pGroup;
    if (_zone->getID() != 0) {
      throw DSSException("Group with id " + intToString(_groupNumber) + " only allowed in Zone 0");
    }
    try {
      pGroup = _zone->getGroup(_groupNumber);
    } catch (ItemNotFoundException& e) {
      throw DSSException("Group id " + intToString(_groupNumber) + " does not exist");
    }
    try {
      DS_INFO("Removing user global application group", _groupNumber);
      _zone->removeGroup(pGroup);
      m_Interface.removeGroup(_zone->getID(), _groupNumber);
    } catch (BusApiError& e) {
      DS_WARNING("Bus-Error removing user group", _groupNumber, e.what());
    }
    return;
  }

  throw DSSException("Remove group: id " + intToString(_groupNumber) + " not found");
} // removeGroup

void StructureManipulator::removeClusterStates(std::shared_ptr<Cluster> _cluster) {
  DS_REQUIRE(_cluster, "Missing cluster argument");
  auto states = m_Apartment.getStates();
  for (auto& state : states) {
    if (state->tryGetClusterID() == _cluster->getID()) {
      m_Apartment.removeState(state);
    }
  }
}

void StructureManipulator::groupSetName(std::shared_ptr<Group> _group, const std::string& _name) {
  if (isDefaultGroup(_group->getID()) || isGlobalAppDsGroup(_group->getID())) {
    throw DSSException("Group with id " + intToString(_group->getID()) + " is reserved");
  }

  if (isCluster(_group->getID()) || isZoneUserGroup(_group->getID()) || isGlobalAppUserGroup(_group->getID())) {
    _group->setName(_name);
    m_Interface.groupSetName(_group->getZoneID(), _group->getID(), _name);
    return;
  }

  throw DSSException("Rename group: id " + intToString(_group->getID()) + " too large");
} // groupSetName

void StructureManipulator::groupSetApplication(std::shared_ptr<Group> _group, ApplicationType applicationType,
                                               const int applicationConfiguration) {
  if (isDefaultGroup(_group->getID()) || isGlobalAppDsGroup(_group->getID())) {
    // for default groups we allow to change the configuration, but we do not allow changing of application type
    if (_group->getApplicationType() != applicationType) {
      throw DSSException("Group with id " + intToString(_group->getID()) + " cannot be modified");
    }
  }

  if (isCluster(_group->getID())) {
    std::shared_ptr<Cluster> pCluster = m_Apartment.getCluster(_group->getID());
    clusterSetApplication(pCluster, applicationType, applicationConfiguration);
    return;
  }

  if (isValidGroup(_group->getID())) {
    _group->setApplicationType(applicationType);
    _group->setApplicationConfiguration(applicationConfiguration);
    m_Interface.groupSetApplication(_group->getZoneID(), _group->getID(), applicationType, applicationConfiguration);
    return;
  }

  throw DSSException("groupSetApplication group: id " + intToString(_group->getID()) + " not valid");
} // groupSetApplication

void StructureManipulator::sceneSetName(std::shared_ptr<Group> _group, int _sceneNumber, const std::string& nameIn) {
  auto name = fixUtf8(escapeHTML(nameIn));
  _group->setSceneName(_sceneNumber, name);
  m_Interface.sceneSetName(_group->getZoneID(), _group->getID(), _sceneNumber, name);
  if (DSS::hasInstance()) {
    std::shared_ptr<Event> pEvent = std::make_shared<Event>(EventName::SceneNameChanged);
    pEvent->setProperty("zone", intToString(_group->getZoneID()));
    pEvent->setProperty("group", intToString(_group->getID()));
    pEvent->setProperty("sceneId", intToString(_sceneNumber));
    pEvent->setProperty("newSceneName", name);
    DSS::getInstance()->getEventQueue().pushEvent(pEvent);
    DSS::getInstance()->getModelMaintenance().addModelEvent(std::make_shared<ModelEvent>(ModelEvent::etModelDirty));
  }
}

void StructureManipulator::saveScene(IDeviceInterface& deviceInterface, CallOrigin callOrigin, int sceneId,
                                     const std::string& token) {
  if (DSS::hasInstance()) {
    DS_REQUIRE(!CommChannel::getInstance()->isSceneLocked(sceneId),
               "Settings are being updated for selected activity. Please try again later.");
  }
  deviceInterface.saveScene(callOrigin, sceneId, token);
}

void StructureManipulator::deviceAddToGroup(std::shared_ptr<Device> _device, std::shared_ptr<Group> _group) {
  DS_DEBUG_ENTER(_device->id(), _group->getID());
  DS_ASSERT(_device.get() != NULL, "");
  DS_ASSERT(_group.get() != NULL, "");

  std::shared_ptr<Cluster> cluster = std::dynamic_pointer_cast<Cluster>(_group);
  if (cluster && cluster->isConfigurationLocked()) {
    throw DSSException("The group is locked and cannot be modified"); // note: translated in dss-websrc
  }

  m_Interface.addToGroup(_device->getDSMeterDSID(), _group->getID(), _device->getShortAddress());

  // Devices that have a configurable standard group, group bitmask is adjusted by the dSM,
  if ((isDefaultGroup(_group->getID()) || isGlobalAppDsGroup(_group->getID())) &&
      (((_device->getFunctionID() & Fid_105_Mask_VariableStandardGroup) > 0) ||
       ((_device->getFunctionID() & Fid_Subclass_Mask) == Fid_105_Grp_v2))) {
    _device->resetGroups();
  }
  _device->addToGroup(_group->getID());

  // inputs of "simple" devices automatically follow into the new group
  {
    if (isAkmDevice(*_device) && (_device->getBinaryInputCount() == 1)) {
      /* AKM with single input, set active group to last group */
      _device->setDeviceBinaryInputTargetId(0, _group->getID());
      _device->setBinaryInputTargetId(0, _group->getID());
    } else if (_device->getOutputMode() == 0) {
      // device has no output, move the button(=input) into group
      _device->setDeviceButtonActiveGroup(_group->getID());
    }

    if ((_device->getDeviceType() == DEVICE_TYPE_KM) &&
        (_device->getDeviceClass() == DeviceClasses_t::DEVICE_CLASS_BL) && (_device->getDeviceNumber() == 200)) {
      // #18333
      // when moved into temperature control group, set ltmode to 0x41
      // when moved into heating group, reset ltmode to 0x0
      // #dsd-3628: also adjust button usage depending on the grouop
      if (_group->getID() == GroupIDControlTemperature) {
        _device->setDeviceButtonInputMode(ButtonInputMode::HEATING_PUSHBUTTON);
        _device->setButtonUsage(ButtonUsage::manual_unused);
      } else {
        _device->setDeviceButtonInputMode(ButtonInputMode::STANDARD);
        _device->setButtonUsage(ButtonUsage::auto_unused);
      }
    }
  }
  {
    std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::DeviceEvent, _device->reference());
    mEvent->setProperty("action", "groupAdd");
    mEvent->setProperty("id", intToString(_group->getID()));
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(mEvent);
    }
  }
  DS_DEBUG_LEAVE();
} // deviceAddToGroup

void StructureManipulator::deviceRemoveFromGroup(std::shared_ptr<Device> _device, std::shared_ptr<Group> _group,
                                                 bool _adaptInputs) {
  DS_DEBUG_ENTER(_device->id(), _group->getID());
  DS_ASSERT(_device.get() != NULL, "");
  DS_ASSERT(_group.get() != NULL, "");

  std::shared_ptr<Cluster> cluster = std::dynamic_pointer_cast<Cluster>(_group);
  if (cluster && cluster->isConfigurationLocked()) {
    throw DSSException("The group is locked and cannot be modified");
  }

  m_Interface.removeFromGroup(_device->getDSMeterDSID(), _group->getID(), _device->getShortAddress());
  _device->removeFromGroup(_group->getID());

  if (_adaptInputs) {
    if (isAkmDevice(*_device) && (_device->getBinaryInputCount() == 1) &&
        (_group->getID() == _device->getBinaryInput(0)->m_targetGroupId)) {
      /* AKM with single input, active on removed group */
      _device->setDeviceBinaryInputTargetId(0, GroupIDBlack);
      _device->setBinaryInputTargetId(0, GroupIDBlack);
    }

    if (_group->getID() == _device->getButtonActiveGroup()) {
      /* device has button that is active on removed group */
      _device->setDeviceButtonActiveGroup(_device->getActiveGroup());
    }
  }

  {
    std::shared_ptr<Event> mEvent = std::make_shared<Event>(EventName::DeviceEvent, _device->reference());
    mEvent->setProperty("action", "groupRemove");
    mEvent->setProperty("id", intToString(_group->getID()));
    if (DSS::hasInstance()) {
      DSS::getInstance()->getEventQueue().pushEvent(mEvent);
    }
  }
  DS_DEBUG_LEAVE();
} // deviceRemoveFromGroup

void StructureManipulator::deviceRemoveFromGroups(std::shared_ptr<Device> device) {
  std::shared_ptr<Zone> pZone = m_Apartment.getZone(0);
  for (int g = GroupIDAppUserMin; g <= GroupIDAppUserMax; g++) {
    if (!device->isInGroup(g)) {
      continue;
    }
    std::shared_ptr<Group> pGroup = pZone->getGroup(g);
    deviceRemoveFromGroup(device, pGroup);
  }
} // deviceRemoveFromGroups

void StructureManipulator::deviceRemoveFromClusters(std::shared_ptr<Device> device) {
  for (int g = 0; g < device->getGroupsCount(); g++) {
    auto group = device->getGroupByIndex(g);
    if ((group != NULL) && isCluster(group->getID())) {
      deviceRemoveFromGroup(device, group);
    }
  }
}

bool StructureManipulator::setJokerGroup(std::shared_ptr<Device> device, int groupId) {
  // for standard groups force that only one group is active
  for (int g = GroupIDYellow; g <= GroupIDStandardMax; g++) {
    if (g == groupId || GroupIDBlack == g) {
      continue;
    }
    if (device->isInGroup(g)) {
      deviceRemoveFromGroup(device, m_Apartment.getZone(device->getZoneID())->getGroup(g), false);
    }
  }
  // remove from control groups
  for (int g = GroupIDControlGroupMin; g <= GroupIDControlGroupMax; g++) {
    if (device->isInGroup(g)) {
      deviceRemoveFromGroup(device, m_Apartment.getZone(device->getZoneID())->getGroup(g), false);
    }
  }
  // remove also from GA groups
  for (int g = GroupIDGlobalAppMin; g <= GroupIDGlobalAppMax; g++) {
    if (device->isInGroup(g)) {
      deviceRemoveFromGroup(device, m_Apartment.getZone(0)->getGroup(g), false);
    }
  }
  // assign to new "standard" group
  if (groupId >= GroupIDGlobalAppMin) {
    deviceAddToGroup(device, m_Apartment.getZone(0)->getGroup(groupId));
  } else {
    deviceAddToGroup(device, m_Apartment.getZone(device->getZoneID())->getGroup(groupId));
  }

  device->setDeviceJokerGroup(groupId);

  /* check if device is also in a cluster of different application type */
  std::shared_ptr<Zone> pZone = m_Apartment.getZone(0);
  for (int g = GroupIDAppUserMin; g <= GroupIDAppUserMax; g++) {
    if (!device->isInGroup(g)) {
      continue;
    }
    std::shared_ptr<Group> itGroup = pZone->getGroup(g);
    if (static_cast<int>(itGroup->getApplicationType()) == groupId) {
      continue;
    }
    deviceRemoveFromGroup(device, itGroup);
  }

  return true;
}

void StructureManipulator::setDeviceZoneHeatingConfig(const std::shared_ptr<const DSMeter>& meter, Zone& zone) {
  DS_INFO("setDeviceZoneHeatingConfig", meter, zone, zone.getHeatingConfigSpec());
  // disable temperature controller on all meters
  DSS::getInstance()->assertBlockingAllowed();
  m_Interface.disableZoneHeatingConfig(DSUID_BROADCAST, zone.getID());

  // enable temperature controller on selected meter
  if (zone.temperatureConfig().m_mode == HeatingControlMode::PID) {
    m_Interface.setZoneHeatingOperationModes(meter->getDSID(), zone.getID(),
                                             zone.getHeatingControlOperationModeValues());
  } else if (zone.temperatureConfig().m_mode == HeatingControlMode::FIXED) {
    m_Interface.setZoneHeatingOperationModes(meter->getDSID(), zone.getID(), zone.getHeatingFixedOperationModeValues());
  }
  m_Interface.setZoneHeatingConfig(meter->getDSID(), zone.getID(), zone.getHeatingConfigSpec());
}

void StructureManipulator::setZoneHeatingConfig(std::shared_ptr<Zone> zone, const TemperatureConfig& config) {
  zone->setTemperatureConfig(config);
  if (auto&& meter = zone->temperatureController()) {
    setDeviceZoneHeatingConfig(meter, *zone);
  } else {
    DS_WARNING("No temperature controller.", *zone);
  }

  if (auto&& modelMaintenance = zone->tryGetModelMaintenance()) {
    auto pEvent = std::make_shared<ModelEventWithDSID>(ModelEvent::etControllerConfig, DSUID_BROADCAST);
    pEvent->addParameter(zone->getID());
    pEvent->setSingleObjectParameter(std::make_shared<ZoneHeatingConfigSpec_t>(config.spec()));
    modelMaintenance->addModelEvent(pEvent);
  }
}

void StructureManipulator::setZoneHeatingConfigSpec(std::shared_ptr<Zone> zone, const ZoneHeatingConfigSpec_t& spec) {
  auto config = zone->temperatureConfig();
  config.setSpec(spec);
  setZoneHeatingConfig(zone, config);
}

void StructureManipulator::setNextTemperatureController(Zone& zone) {
  DS_CONTEXT(zone);
  // zone is already in DS_CONTEXT
  // rotate zone meters to have next temperature controller candidate in front
  const auto meters = [&] {
    std::vector<std::shared_ptr<const DSMeter> > meters;
    for (const auto& meter : zone.getDSMeters()) {
      if (auto x = meter.lock()) {
        meters.push_back(std::move(x));
      }
    }
    if (auto nextEnd = boost::find<boost::return_next_end>(meters, zone.temperatureController())) {
      boost::rotate(meters, boost::begin(nextEnd));
    }
    return meters;
  }();

  // commit to next capable meter that succeeds
  for (const auto& meter : meters) {
    if (!meter) {
      continue;
    }
    DS_CONTEXT(*meter);
    if (!meter->getCapability_HasTemperatureControl()) {
      DS_DEBUG("!capability");
      continue;
    }
    if (!meter->isPresent()) {
      DS_DEBUG("!present");
      continue;
    }

    try {
      DS_DEBUG("Try select");
      setDeviceZoneHeatingConfig(meter, zone);
      zone.setTemperatureController(meter);
      DS_NOTICE("New temperature controller.");
      return; // commit to new meter for next period
    } catch (const std::exception& e) {
      DS_ERROR("Setting temperature controller failed.", e);
      // try next one
    }
  }
  zone.setTemperatureController(std::shared_ptr<const DSMeter>());
  DS_DEBUG("Nothing to select.");
}

void StructureManipulator::electTemperatureController(Zone& zone) {
  if (zone.getID() == 0) {
    return;
  }
  try {
    const auto& mode = zone.temperatureConfig().m_mode;
    if (mode == HeatingControlMode::OFF) {
      DS_INFO("OK - off", zone);
      return;
    }
    // poll current temperature controller
    auto meter = zone.temperatureController();
    try {
      if (meter) {
        DS_REQUIRE(meter->isPresent(), "Temperatur controller not present.");

        // poll and store state
        ZoneHeatingStateSpec_t state =
            m_Apartment.getBusInterface()->getStructureQueryBusInterface()->getZoneHeatingState(meter->getDSID(),
                                                                                                zone.getID());
        zone.setHeatingControlState(state.State);

        const auto& status = zone.getHeatingStatus();
        const auto& temperatureControllerTs = zone.temperatureControllerTs();

        // Check that temperatur controller sends control value
        if (mode != HeatingControlMode::OFF && temperatureControllerTs.addMinute(3) < DateTime()) {
          const auto& controlValue = status.controlValue;
          DS_REQUIRE(controlValue, "Missing control value.", temperatureControllerTs, *meter);
        }

        DS_INFO("OK", zone, *meter, mode, state, status, temperatureControllerTs);
        return;
      }
      DS_INFO("No controller selected now.", zone);
    } catch (const std::exception& e) {
      DS_ERROR("Temperature controller failed. Select next one.", e, zone);
    }

    setNextTemperatureController(zone);
  } catch (const std::exception& e) {
    DS_ERROR("Failed to update temperature controller.", e);
  }
}

void StructureManipulator::electTemperatureControllers() {
  for (auto&& zone : m_Apartment.getZones()) {
    electTemperatureController(*zone);
  }
}
void StructureManipulator::setZoneSensor(Zone& _zone, SensorType _sensorType, std::shared_ptr<Device> _dev) {
  DS_INFO("SensorAssignment", _zone.getID(), _sensorType, "=>", _dev->getDSID());

  _zone.setSensor(*_dev, _sensorType);
  m_Interface.setZoneSensor(_zone.getID(), _sensorType, _dev->getDSID());
}

void StructureManipulator::resetZoneSensor(Zone& _zone, SensorType _sensorType) {
  DS_INFO("SensorAssignment:", _zone.getID(), _sensorType, "=> none");

  _zone.resetSensor(_sensorType);
  m_Interface.resetZoneSensor(_zone.getID(), _sensorType);
}

void StructureManipulator::autoAssignZoneSensors(Zone& _zone) {
  autoAssignZoneSensorsImpl(_zone);
}

void StructureManipulator::autoAssignZoneSensorsImpl(Zone& _zone) {
  Set devices = _zone.getDevices();
  if (devices.isEmpty()) {
    return;
  }

  auto&& unassigned_sensors = _zone.getUnassignedSensorTypes();

  DS_INFO("SensorAssignment:", _zone.getID(), unassigned_sensors.size(),
          "unassigned sensor types, trying to assign some more... ");

  int assigned = 0;
  // check if our set contains devices that with the matching sensor type
  // and assign the first device that we find automatically: UC 8.1
  for (auto&& sensorType : unassigned_sensors) {
    Set devicesBySensor = devices.getBySensorType(sensorType);
    if (!devicesBySensor.empty()) {
      // select an active sensor
      for (int i = 0; i < devicesBySensor.length(); ++i) {
        if (devicesBySensor.get(i).getDevice()->isPresent()) {
          setZoneSensor(_zone, sensorType, devicesBySensor.get(i).getDevice());
          assigned++;
          break;
        }
      }
      // #13433: removed code that assigned inactive sensors to a zone
    }
  }

  DS_INFO("SensorAssignment:", _zone.getID(), "employed", assigned, "additional sensor as room sensors");
}

void StructureManipulator::synchronizeZoneSensorAssignment(Zone& zone) {
  DS_INFO("SensorAssignment:", zone.getID(), "roll out sensor configuration to dsms");

  {
    // remove sensors that do not belong to the zone
    zone.removeInvalidZoneSensors();

    // erase all unassigned sensors in zone
    try {
      for (auto&& sensorType : zone.getUnassignedSensorTypes()) {
        DS_INFO("SensorAssignment:", zone.getID(), "clear type:", sensorType, "=> none (no available sensor)");
        m_Interface.resetZoneSensor(zone.getID(), sensorType);
      }

      // reassign all assigned sensors in zone
      for (auto&& s : zone.getAssignedSensors()) {
        DS_INFO("SensorAssignment:", zone.getID(), "set type:", s.m_sensorType, "=>", s.m_DSUID);
        m_Interface.setZoneSensor(zone.getID(), s.m_sensorType, s.m_DSUID);
      }
    } catch (std::runtime_error& err) {
      DS_WARNING(zone.getID(), "failed with exception:", err.what());
    }
  }
}

void StructureManipulator::clusterSetName(std::shared_ptr<Cluster> _cluster, const std::string& _name) {
  if (isCluster(_cluster->getID())) {
    _cluster->setName(_name);
    m_Interface.clusterSetName(_cluster->getID(), _name);
    return;
  }

  throw DSSException("Rename cluster: id " + intToString(_cluster->getID()) + " not a cluster");
} // clusterSetName

void StructureManipulator::clusterSetApplication(std::shared_ptr<Cluster> _cluster, ApplicationType applicationType,
                                                 const int applicationConfiguration) {
  if (isCluster(_cluster->getID())) {
    if (_cluster->isConfigurationLocked()) {
      throw DSSException("The group is locked and cannot be modified");
    }
    _cluster->setApplicationType(applicationType);
    _cluster->setApplicationConfiguration(applicationConfiguration);
    m_Interface.clusterSetApplication(_cluster->getID(), applicationType, applicationConfiguration);
    return;
  }

  throw DSSException("clusterSetApplication cluster: id " + intToString(_cluster->getID()) + " not a cluster");
} // clusterSetStandardID

void StructureManipulator::clusterSetConfigurationLock(std::shared_ptr<Cluster> _cluster, bool _locked) {
  if (isCluster(_cluster->getID())) {
    _cluster->setConfigurationLocked(_locked);
    m_Interface.clusterSetConfigurationLock(_cluster->getID(), _locked);
    return;
  }

  throw DSSException("set cluster lock: id " + intToString(_cluster->getID()) + " not a cluster");
} // clusterSetLockedState

void StructureManipulator::setProperty(
    std::shared_ptr<DSMeter> _dsMeter,
    const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& properties) {
  return m_Interface.setProperty(_dsMeter->getDSID(), properties);
}

vdcapi::Message StructureManipulator::getProperty(
    std::shared_ptr<DSMeter> _dsMeter,
    const ::google::protobuf::RepeatedPtrField< ::vdcapi::PropertyElement>& query) {
  return m_Interface.getProperty(_dsMeter->getDSID(), query);
}

} // namespace dss
