/*
 *  Copyright (c) 2015 digitalSTROM.org, Zurich, Switzerland
 *
 *  This file is part of digitalSTROM Server.
 *
 *  digitalSTROM Server is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  digitalSTROM Server is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include <memory>
#include <mutex>
#include <string>
#include <vector>

#include "ds485/specification-fwd.h"
#include "logger.h"
#include "model/modelconst.h"
#include "sqlite3_wrapper.h"

DS_REFLECTED_ENUM((dss), VdcDbAccessLevel,
                  /// Shown everywhere
                  (normal)
                  /// Shown within the UI if the ‘show more events’ option was selected
                  (advanced)
                  /// Not shown within the UI’s at all but might be used by bots and
                  /// other services.
                  /// Meaningless or useless for the user.
                  (service));

namespace dss {

class DSS;

/// Abstract interface
class VdcDb {
public:
  using AccessLevel = VdcDbAccessLevel;
  virtual ~VdcDb();

  struct SpecDesc {
    std::string id;
    std::string tags;
    std::string value;
  };
  typedef SpecDesc Spec;
  typedef std::vector<SpecDesc> Specs;

  virtual Specs getSpec(const std::string &gtin) = 0;

  struct StateDesc {
    std::string id;
    std::vector<std::string> values;
    std::string tags;
  };
  typedef StateDesc State;
  typedef std::vector<State> States;

  virtual std::vector<DeviceStateSpec_t> getStatesLegacy(const std::string &gtin) = 0;
  virtual States getStates(const std::string &gtin) = 0;

  struct EventDesc {
    std::string id;
    std::string categoryId;
    AccessLevel accessLevel;
  };
  typedef EventDesc Event;
  typedef std::vector<EventDesc> Events;

  virtual Events getEvents(const std::string &gtin) = 0;

  enum propertyTypeId { string, integer, numeric, enumeration, boolean };

  struct PropertyDesc {
    std::string id; // technical name
    propertyTypeId typeId;
    std::string typePostfix; // optional postfix, like in: numeric.duration
    std::string defaultValue;
    std::string minValue;
    std::string maxValue;
    std::string resolution;
    std::string siUnit;
    std::string tags;
    bool optional;
    bool readonly;
    std::vector<std::string> values; // enum values
  };
  typedef PropertyDesc Property;
  typedef std::vector<Property> Properties;

  virtual Properties getProperties(const std::string &gtin) = 0;

  struct SensorDesc {
    PropertyDesc prop;
    std::string sensorIndex;
    SensorType dsType;
  };
  typedef SensorDesc Sensor;
  typedef std::vector<Sensor> Sensors;

  virtual Sensors getSensors(const std::string &gtin) = 0;

  struct ActionDesc {
    std::string id;
    std::string categoryId;
    std::string tags;
    std::vector<PropertyDesc> params;
  };
  typedef ActionDesc Action;
  typedef std::vector<Action> Actions;

  virtual Actions getActions(const std::string &gtin) = 0;

  struct StandardActionDesc {
    std::string id;
    std::string action;
    std::vector<std::pair<std::string, std::string>> params;
    // pair <argument name, argument value>
    std::string tags;
  };
  typedef StandardActionDesc DeviceAction;
  typedef std::vector<DeviceAction> StandardActions;

  virtual StandardActions getStandardActions(const std::string &gtin) = 0;

  struct Category {
    std::string id;
    std::string orderId; ///< correspond to category_order_id in category table and to categories_id in events and
                         /// actions tables
    bool alwaysVisible;
  };
  typedef std::vector<Category> Categories;

  virtual Categories getCategories(const std::string &gtin) = 0;

  struct OutputDesc {
    std::string id;
    propertyTypeId typeId;
    std::string typePostfix;
    std::string defaultValue;
    std::string minValue;
    std::string maxValue;
    std::string resolution;
    std::string siUnit;
    std::string outputType;
    std::string outputIndex;
    std::string tags;
    std::string categoryId;
    std::vector<std::pair<std::string, std::string>> options;
  };
  typedef std::vector<OutputDesc> Outputs;

  virtual Outputs getOutputs(const std::string &gtin) = 0;

  virtual bool hasActionInterface(const std::string &gtin) = 0;
};

namespace vdcDb {
std::unique_ptr<VdcDb> make(DSS &dss, SQLite3::Mode mode = SQLite3::Mode::ReadOnly);
void recreate(DSS &dss);
void update(DSS &dss, const std::string &newDb);
} // namespace vdcDb

} // namespace dss
