/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
  #include "config.h"
#endif

#include <digitalSTROM/dsuid.h>

#include "businterface.h"
#include "circuitrequesthandler.h"
#include "ds485types.h"
#include "dss.h"
#include "metering/metering.h"
#include "model/apartment.h"
#include "model/dsmeter.h"
#include "model/modelmaintenance.h"
#include "protobufjson.h"
#include "reregister-devices.h"
#include "structuremanipulator.h"
#include "util.h"
#include "vdc-connection.h"
#include "web/restful.h"

static ds::LogChannel dsLogChannel("jsonApiCircuit");

namespace dss {

  //=========================================== CircuitRequestHandler

  CircuitRequestHandler::CircuitRequestHandler(Apartment& _apartment,
                                               StructureModifyingBusInterface* _pStructureBusInterface,
                                               StructureQueryBusInterface* _pStructureQueryBusInterface)
  : m_Apartment(_apartment),
    m_pStructureBusInterface(_pStructureBusInterface),
    m_pStructureQueryBusInterface(_pStructureQueryBusInterface)
  { }

  WebServerResponse CircuitRequestHandler::jsonHandleRequest(const RestfulRequest& _request, std::shared_ptr<Session> _session, const struct mg_connection* _connection) {
    std::string idString = _request.getParameter("id");
    std::string dsuidStr = _request.getParameter("dsuid");
    if (idString.empty() && dsuidStr.empty()) {
      return JSONWriter::failure("Missing parameter dsuid");
    }

    dsuid_t dsuid;
    try {
      dsuid = dsidOrDsuid2dsuid(idString, dsuidStr);
    } catch(std::runtime_error& e) {
      return JSONWriter::failure(e.what());
    }

    vdcapi::PropertyElement parsedParamsElement;
    const vdcapi::PropertyElement* params = &vdcapi::PropertyElement::default_instance();
    if (const auto& paramsStr = _request.tryGetParameter("params")) {
      parsedParamsElement = ProtobufToJSon::jsonToElement(*paramsStr);
      params = &parsedParamsElement;
    }

    JSONWriter json;
    try {
      std::shared_ptr<DSMeter> dsMeter = m_Apartment.getDSMeterByDSID(dsuid);
      if(_request.getMethod() == "getName") {
        json.add("name", dsMeter->getName());
        return json.successJSON();
      } else if(_request.getMethod() == "setName") {
        std::string name = escapeHTML(_request.getRequiredParameter("newName"));

        DS_NOTICE("Set name.", *dsMeter, name);

        dsMeter->setName(name);
        StructureManipulator manipulator(m_Apartment);
        manipulator.meterSetName(dsMeter, name);
        return json.success();
      } else if (_request.getMethod() == "setMeteringConfiguration") {
        if (!dsMeter->getCapability_HasMetering()) {
          return json.failure("Metering not supported on this device");
        }
        auto spec = m_pStructureQueryBusInterface->getMeteringConfiguration(dsMeter->getDSID());

        if (auto category = ds::tryToEnum<MeteringProducerCategory>(_request.getParameter("category"))) {
          spec.category = *category;
        } else {
          return json.failure("Metering category value invalid");
        }

        if (auto scaleFactor = strToIntDef(_request.getParameter("scaleFactor"), -1); scaleFactor >= 0) {
          spec.scaleFactor = scaleFactor;
        }

        DS_NOTICE("Set metering configuration.", *dsMeter, spec);
        m_pStructureBusInterface->setMeteringConfiguration(dsMeter->getDSID(), spec);
        dsMeter->setCapability_HasMeteringProducerEnabled(spec.category == MeteringProducerCategory::producer);
        return json.success();

      } else if (_request.getMethod() == "getMeteringConfiguration") {
        if (!dsMeter->getCapability_HasMetering()) {
          return json.failure("Metering not supported on this device");
        }

        auto spec = m_pStructureQueryBusInterface->getMeteringConfiguration(dsMeter->getDSID());
        switch (spec.category) {
          case MeteringProducerCategory::consumer: json.add("category", "consumer"); break;
          case MeteringProducerCategory::producer:
            json.add("category", "producer");
            json.add("scaleFactor", spec.scaleFactor);
            break;
          default: return json.failure("Metering configuration invalid");
        }
        return json.successJSON();
      } else if (_request.getMethod() == "getConsumption") {
        if (!dsMeter->getCapability_HasMetering()) {
          return json.failure("Metering not supported on this device");
        }
        const auto& metering = m_Apartment.getMetering();
        json.add("consumption",
                 static_cast<unsigned long long>(metering.getConsumedPower(dsMeter->getDSID()).value_or(0)));
        return json.successJSON();
      } else if (_request.getMethod() == "getEnergyMeterValue") {
        if (!dsMeter->getCapability_HasMetering()) {
          return JSONWriter::failure("Metering not supported on this device");
        }
        json.add("meterValue", static_cast<unsigned long long>(
                                   m_Apartment.getMetering().getLastEnergy(dsMeter->getDSID()).consumption));
        return json.successJSON();
      } else if (_request.getMethod() == "rescan") {
        DS_NOTICE("Rescan.", *dsMeter);

        dsMeter->setIsValid(false);
        return json.success();
      } else if (_request.getMethod() == "learnIn") {
        int timeout = -1;
        _request.getParameter("timeout", timeout);

        DS_NOTICE("Learn in.", *dsMeter, timeout, *params);

        vdcapi::Message res = VdcHelper::callLearningFunction(dsMeter->getDSID(), true, timeout, *params);
        ProtobufToJSon::protoPropertyToJson(res, json);
        return json.successJSON();
      } else if (_request.getMethod() == "learnOut") {
        int timeout = -1;
        _request.getParameter("timeout", timeout);

        DS_NOTICE("Learn out.", *dsMeter, timeout, *params);

        vdcapi::Message res = VdcHelper::callLearningFunction(dsMeter->getDSID(), false, timeout, *params);
        ProtobufToJSon::protoPropertyToJson(res, json);
        return json.successJSON();
      } else if (_request.getMethod() == "firmwareCheck") {
        DS_NOTICE("Check firmware.", *dsMeter, *params);

        vdcapi::Message res = VdcHelper::callFirmwareFunction(dsMeter->getDSID(), true, false, *params);
        ProtobufToJSon::protoPropertyToJson(res, json);
        return json.successJSON();
      } else if (_request.getMethod() == "firmwareUpdate") {
        bool clearSettings = false;
        _request.getParameter("clearsettings", clearSettings);

        DS_NOTICE("Firmware update.", *dsMeter, clearSettings, *params);

        vdcapi::Message res = VdcHelper::callFirmwareFunction(dsMeter->getDSID(), false, clearSettings, *params);
        ProtobufToJSon::protoPropertyToJson(res, json);
        return json.successJSON();
      } else if (_request.getMethod() == "storeAccessToken") {
        const auto& authData = _request.getRequiredParameter("authData");
        std::string authScope = _request.getParameter("authScope");

        DS_NOTICE("Store access token.", *dsMeter, authData, authScope, *params);

        vdcapi::Message res = VdcHelper::callAuthenticate(dsMeter->getDSID(), authData, authScope, *params);
        ProtobufToJSon::protoPropertyToJson(res, json);
        return json.successJSON();
      } else if (_request.getMethod() == "identify") {
        DS_NOTICE("Identify.", *dsMeter);

        vdcapi::Message res = VdcHelper::callIdentify(dsMeter->getDSID());
        ProtobufToJSon::protoPropertyToJson(res, json);
        return json.successJSON();
      } else if (_request.getMethod() == "invokeMethod") {
        std::string method;
        _request.getParameter("method", method);
        vdcapi::Message res =
            VdcConnection::genericRequest(dsMeter->getDSID(), dsMeter->getDSID(), method, params->elements());
        ProtobufToJSon::protoPropertyToJson(res, json);
        return json.successJSON();
      } else if (_request.getMethod() == "reregisterDevices") {
        DS_NOTICE("Reregistering devices.", *dsMeter);

        auto&& dss = *DSS::getInstance();
        auto dsa = makeDsaHttpClient();
        auto& backendVdcs = dss.backendVdcs();
        reregisterDevices(*dsMeter, *dsa, backendVdcs);

        return json.successJSON();
      } else if (_request.getMethod() == "removeEmptyRooms") {
        try {
          m_pStructureBusInterface->removeInactiveDevicesAndRooms(dsMeter->getDSID());
        } catch(std::runtime_error& e) {
          return JSONWriter::failure(e.what());
        }
        
        return json.successJSON();
      } else {
        throw std::runtime_error("Unhandled function");
      }
    } catch(ItemNotFoundException&) {
      return JSONWriter::failure("Could not find dSMeter with given dSUID");
    }
  } // handleRequest

} // namespace dss
