
#include "deviceinterfacerequesthandler.h"

#include <ds/catch/catch.h>
#include <ds/log.h>

#include "businterface-mock.h"
#include "model/apartment-mock.h"
#include "model/deviceinterface-mock.h"
#include "web/restful.h"

static ds::LogChannel dsLogChannel("webHandlerDeviceInterfaceRequestHandlerTest");

using namespace dss;
using namespace dss::web::handler;
using namespace trompeloeil;
using trompeloeil::_;
using namespace std::literals;


namespace {

static const char* TAGS = "[web][handler][devicerequesthandler]";

TEST_CASE("isDeviceRpc", TAGS) {
  CHECK(isDeviceInterfaceCall(RestfulRequest{"foo/blink", ""s}) == true);
  CHECK(isDeviceInterfaceCall(RestfulRequest{"foo/bogus", ""s}) == false);
  // not exposed in the original ajax interface
  CHECK(isDeviceInterfaceCall(RestfulRequest{"foo/undoSceneLast", ""s}) == false);
}

TEST_CASE("rpcCalls", TAGS) {
  auto deviceInterfaceMock = std::make_shared<IDeviceInterfaceMock>();
  auto apartmentMock = ApartmentMock{};

  BusInterfaceMock busInterfaceMock{};
  apartmentMock.setBusInterface(&busInterfaceMock);

  auto deviceInterfaceRequestHandler = DeviceInterfaceRequestHandler(apartmentMock);

  SECTION("turnOff") {
    REQUIRE_CALL(*deviceInterfaceMock, turnOff(_, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(RestfulRequest{"foo/turnOff", ""s},
                                                                             deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("turnOn") {
    REQUIRE_CALL(*deviceInterfaceMock, turnOn(_, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(RestfulRequest{"foo/turnOn", ""s},
                                                                             deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("increaseValue") {
    REQUIRE_CALL(*deviceInterfaceMock, increaseValue(_, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(RestfulRequest{"foo/increaseValue", ""s},
                                                                             deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("decreaseValue") {
    REQUIRE_CALL(*deviceInterfaceMock, decreaseValue(_, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(RestfulRequest{"foo/decreaseValue", ""s},
                                                                             deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("setValue") {
    REQUIRE_CALL(*deviceInterfaceMock, setValue(_, _, 5, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(RestfulRequest{"foo/setValue", "value=5"s},
                                                                             deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("callScene") {
    REQUIRE_CALL(*deviceInterfaceMock, callScene(_, _, 5, _, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(
        RestfulRequest{"foo/callScene", "sceneNumber=5"s}, deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("callSceneMin") {
    REQUIRE_CALL(*deviceInterfaceMock, callSceneMin(_, _, 5, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(
        RestfulRequest{"foo/callSceneMin", "sceneNumber=5"s}, deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("saveScene") {
    REQUIRE_CALL(*deviceInterfaceMock, saveScene(_, 5, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(
        RestfulRequest{"foo/saveScene", "sceneNumber=5"s}, deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("undoScene") {
    REQUIRE_CALL(*deviceInterfaceMock, undoScene(_, _, 5, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(
        RestfulRequest{"foo/undoScene", "sceneNumber=5"s}, deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("getConsumption") {
    REQUIRE_CALL(*deviceInterfaceMock, getPowerConsumption()).RETURN(50);
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(RestfulRequest{"foo/getConsumption", ""s},
                                                                             deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
    auto consumptionReply = ds::fromJson<ConsumptionReply>(reply.result);
    REQUIRE(consumptionReply.consumption == 50);
  }
  SECTION("nextScene") {
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(RestfulRequest{"foo/nextScene", ""s},
                                                                             deviceInterfaceMock, {});
    DS_NOTICE(buffer);
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(!reply.ok);
    REQUIRE(reply.message == "Operation allowed only on Ventilation groups."s);
  }
  SECTION("previousScene") {
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(RestfulRequest{"foo/previousScene", ""s},
                                                                             deviceInterfaceMock, {});
    DS_NOTICE(buffer);
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(!reply.ok);
    REQUIRE(reply.message == "Operation allowed only on Ventilation groups."s);
  }
  SECTION("blink") {
    REQUIRE_CALL(*deviceInterfaceMock, blink(_, _, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(RestfulRequest{"foo/blink", ""s},
                                                                             deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("increaseOutputChannelValue") {
    REQUIRE_CALL(*deviceInterfaceMock, increaseOutputChannelValue(_, _, _, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(
        RestfulRequest{"foo/increaseOutputChannelValue", "channel=brightness"s}, deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("decreaseOutputChannelValue") {
    REQUIRE_CALL(*deviceInterfaceMock, decreaseOutputChannelValue(_, _, _, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(
        RestfulRequest{"foo/decreaseOutputChannelValue", "channel=brightness"s}, deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("stopOutputChannelValue") {
    REQUIRE_CALL(*deviceInterfaceMock, stopOutputChannelValue(_, _, _, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(
        RestfulRequest{"foo/stopOutputChannelValue", "channel=brightness"s}, deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
  SECTION("pushSensorValue") {
    REQUIRE_CALL(*deviceInterfaceMock, pushSensor(_, _, _, SensorType::RoomTemperatureSetpoint, 25.5, _));
    auto buffer = deviceInterfaceRequestHandler.handleDeviceInterfaceRequest(
        RestfulRequest{"foo/pushSensorValue", "sensorType=50&sensorValue=25.5"s}, deviceInterfaceMock, {});
    auto reply = ds::fromJson<dss::web::JSONResponse>(buffer);
    REQUIRE(reply.ok);
  }
}

} // namespace
