/*
    Copyright (c) 2010 digitalSTROM.org, Zurich, Switzerland

    Author: Patrick Staehlin, futureLAB AG <pstaehlin@futurelab.ch>

    This file is part of digitalSTROM Server.

    digitalSTROM Server is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    digitalSTROM Server is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#define BOOST_TEST_NO_MAIN
#include <boost/test/unit_test.hpp>

#include "base.h"
#include "businterface-mock.h"
#include "model/apartment-mock.h"
#include "model/deviceinterface-mock.h"
#include "model/deviceinterface.h"
#include "session.h"
#include "web/handler/deviceinterfacerequesthandler.h"
#include "web/restful.h"
#include "webfixture.h"

using namespace dss;

BOOST_AUTO_TEST_SUITE(WebDeviceInterface)

  class DeviceInterfaceDummy : public IDeviceInterface {
  public:

    DeviceInterfaceDummy()
    : m_NumberOfCalls(0)
    {
    }
    virtual void turnOn(const CallOrigin _origin, const SceneAccessCategory _category) {
      functionCalled("turnOn");
    }
    virtual void turnOff(const CallOrigin _origin, const SceneAccessCategory _category) {
      functionCalled("turnOff");
    }
    virtual void increaseValue(const CallOrigin _origin, const SceneAccessCategory _category)  {
      functionCalled("increaseValue");
    }
    virtual void decreaseValue(const CallOrigin _origin, const SceneAccessCategory _category)  {
      functionCalled("decreaseValue");
    }
    virtual void setValue(const CallOrigin _origin, const SceneAccessCategory _category, const uint8_t _value, const std::string _token)  {
      functionCalled("setValue(" + intToString(_value) + ")");
    }
    virtual void callScene(const CallOrigin _origin, const SceneAccessCategory _category, const int _sceneNr, const std::string _token, const bool _force)  {
      functionCalled("callScene(" + intToString(_sceneNr) + ")");
    }
    virtual void callSceneMin(const CallOrigin _origin, const SceneAccessCategory _category, const int _sceneNr, const std::string _token) {
      functionCalled("callSceneMin(" + intToString(_sceneNr) + ")");
    }
    virtual void saveScene(const CallOrigin _origin, const int _sceneNr, const std::string _token)  {
      functionCalled("saveScene(" + intToString(_sceneNr) + ")");
    }
    virtual void undoScene(const CallOrigin _origin, const SceneAccessCategory _category, const int _sceneNr, const std::string _token)  {
      functionCalled("undoScene(" + intToString(_sceneNr) + ")");
    }
    virtual void undoSceneLast(const CallOrigin _origin, const SceneAccessCategory _category, const std::string _token)  {
      functionCalled("undoSceneLast");
    }
    virtual unsigned long getPowerConsumption()  {
      functionCalled("getConsumption");
      return 50;
    }
    virtual void nextScene(const CallOrigin _origin, const SceneAccessCategory _category)  {
      functionCalled("nextScene");
    }
    virtual void previousScene(const CallOrigin _origin, const SceneAccessCategory _category)  {
      functionCalled("previousScene");
    }
    virtual void blink(const CallOrigin _origin, const SceneAccessCategory _category, const std::string _token)  {
      functionCalled("blink");
    }
    virtual void increaseOutputChannelValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                            const ChannelType _channel, const std::string _token) {
      functionCalled("increaseOutputChannelValue(" + intToString(static_cast<int>(_channel)) + ")");
    }
    virtual void decreaseOutputChannelValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                            const ChannelType _channel, const std::string _token) {
      functionCalled("decreaseOutputChannelValue(" + intToString(static_cast<int>(_channel)) + ")");
    }
    virtual void stopOutputChannelValue(const CallOrigin _origin, const SceneAccessCategory _category,
                                        const ChannelType _channel, const std::string _token) {
      functionCalled("stopOutputChannelValue(" + intToString(static_cast<int>(_channel)) + ")");
    }
    virtual void pushSensor(const CallOrigin _origin, const SceneAccessCategory _category, dsuid_t _sourceID, SensorType _sensorType, double _sensorValueFloat, const std::string _token) {
      (void)doubleToSensorValue(_sensorType, _sensorValueFloat);
      functionCalled("pushSensorValue");
    }

    int getNumberOfCalls() const { return m_NumberOfCalls; }
    const std::string& getLastFunction() const { return m_LastFunction; }

  private:
    void functionCalled(const std::string& _name) {
      m_NumberOfCalls++;
      m_LastFunction = _name;
    }
  private:
    int m_NumberOfCalls;
    std::string m_LastFunction;
  }; // IDeviceInterface


class DeviceInterfaceRequestHandlerValid : public DeviceInterfaceRequestHandler {
public:
  using DeviceInterfaceRequestHandler::DeviceInterfaceRequestHandler;
};

class Fixture : public WebFixture {
public:

  void testFunction(const std::string& _functionName) {
    testFunctionDo(_functionName, _functionName, "");
  }

  void testFunctionOK(const std::string& _functionName, const std::string& _params) {
    testFunctionDo(_functionName, _functionName, _params);
  }

  void testFunctionThrow(const std::string& _functionName, const std::string& _params) {
    testFunctionThrow(_functionName, _functionName, _params);
  }

  void testFunction(const std::string& _functionName, const std::string& _paramName, const std::string& _paramValue) {
    std::string params = urlEncode(_paramName) + "=" + urlEncode(_paramValue);
    testFunctionDo(_functionName, _functionName + "(" + _paramValue + ")", params);
  }

  void testFunction(const std::string& _functionName,
                    const std::string& _paramName,
                    const std::string& _paramValue,
                    const std::string& _resultingFunctionName) {
    std::string params = urlEncode(_paramName) + "=" + urlEncode(_paramValue);
    testFunctionDo(_functionName, _resultingFunctionName, params);
  }
private:
  void testFunctionDo(const std::string& _functionName,
                      const std::string& _functionNameWithParams,
                      const std::string& _params) {
    std::shared_ptr<DeviceInterfaceDummy> dummy = std::make_shared<DeviceInterfaceDummy>();
    std::shared_ptr<Session> dummySession = std::make_shared<Session>("dummy");
    RestfulRequest req("bla/" + _functionName, _params);
    std::string temp;
    BOOST_CHECK_NO_THROW(temp =
      m_RequestHandler.handleDeviceInterfaceRequest(req, dummy, dummySession));
    WebServerResponse response = temp;
    BOOST_CHECK_EQUAL(dummy->getNumberOfCalls(), 1);
    BOOST_CHECK_EQUAL(dummy->getLastFunction(), _functionNameWithParams);
    testOkIs(response, true);
  }
  void testFunctionThrow(const std::string& _functionName,
                         const std::string& _functionNameWithParams,
                         const std::string& _params) {
    std::shared_ptr<DeviceInterfaceDummy> dummy = std::make_shared<DeviceInterfaceDummy>();
    std::shared_ptr<Session> dummySession = std::make_shared<Session>("dummy");
    RestfulRequest req("bla/" + _functionName, _params);
    std::string temp;
    BOOST_CHECK_THROW(temp =
      m_RequestHandler.handleDeviceInterfaceRequest(req, dummy, dummySession), DSSException);
  }

  BusInterfaceMock m_busInterface{};
  std::unique_ptr<Apartment> pApartment = makeApartment(nullptr, nullptr, &m_busInterface);
  Apartment& m_apartment = *pApartment;
  DeviceInterfaceRequestHandlerValid m_RequestHandler{m_apartment};
};

BOOST_FIXTURE_TEST_CASE(testTurnOff, Fixture) {
  testFunction("turnOff");
}

BOOST_FIXTURE_TEST_CASE(testTurnOn, Fixture) {
  testFunction("turnOn");
}

BOOST_FIXTURE_TEST_CASE(testIncreaseValue, Fixture) {
  testFunction("increaseValue");
}

BOOST_FIXTURE_TEST_CASE(testDecreaseValue, Fixture) {
  testFunction("decreaseValue");
}

BOOST_FIXTURE_TEST_CASE(testSetValue, Fixture) {
  testFunction("setValue", "value", "5");
}

BOOST_FIXTURE_TEST_CASE(testCallScene, Fixture) {
  testFunction("callScene", "sceneNumber", "5");
}

BOOST_FIXTURE_TEST_CASE(testSaveScene, Fixture) {
  testFunction("saveScene", "sceneNumber", "5");
}

BOOST_FIXTURE_TEST_CASE(testUndoScene, Fixture) {
  testFunction("undoScene", "sceneNumber", "5");
}

BOOST_FIXTURE_TEST_CASE(testGetPowerConsumption, Fixture) {
  testFunction("getConsumption");
}

BOOST_FIXTURE_TEST_CASE(testBlink, Fixture) {
  testFunction("blink");
}

BOOST_FIXTURE_TEST_CASE(testPushSensorIDRoomTemperatureSetpoint, Fixture) {
  testFunctionOK("pushSensorValue", "sensorType=50&sensorValue=25.5");
}

BOOST_AUTO_TEST_SUITE_END()
